/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.action.impl;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import magma.agent.action.IAction;
import magma.agent.action.IActionPerformer;
import magma.agent.action.IEffector;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.ICompositeJointConfiguration;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.meta.agent.ISensorConfiguration;
import magma.agent.meta.agent.IUniversalJointConfiguration;

/**
 * This class generates server messages from Effector objects.
 */
public class Action implements IAction
{
	/** handle for logging */
	protected static final Logger logger = Logger.getLogger("");

	protected final HashMap<String, IEffector> effectors;

	private final SayEffector say;

	private final BeamEffector beam;

	private Map<String, IEffector> actionEffectors;

	private IActionPerformer actionPerformer;

	private float maxGain;

	/**
	 * Initializes all effectors that occur in the simulation
	 * 
	 * @param actionPerformer the component that can send the actions
	 */
	public Action(IActionPerformer actionPerformer)
	{
		this.actionPerformer = actionPerformer;
		effectors = new HashMap<String, IEffector>();
		beam = new BeamEffector();
		say = new SayEffector();
		maxGain = 0;
	}

	/**
	 * Initializes the effectors
	 * @param metaModel the robot meta model
	 */
	@Override
	public void init(IAgentMetaModel metaModel)
	{
		initJointEffectors(metaModel);
		actionEffectors = new HashMap<String, IEffector>(effectors);
		actionEffectors.put(say.getName(), say);
	}

	/**
	 * Initializes all HingeJoint- and UniversalJoint-Effectors based on the
	 * given meta-model
	 * 
	 * @param metaModel - the meta model of the agent
	 */
	protected void initJointEffectors(IAgentMetaModel metaModel)
	{
		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			ISensorConfiguration jointConfig = config.getJointConfiguration();

			// create Joint Effectors
			if (jointConfig instanceof IHingeJointConfiguration) {
				String effectorName = ((IHingeJointConfiguration) jointConfig)
						.getEffectorName();
				effectors.put(effectorName, new HingeEffector(effectorName));

			} else if (jointConfig instanceof IUniversalJointConfiguration) {
				IUniversalJointConfiguration conf = (IUniversalJointConfiguration) jointConfig;
				HingeEffector effector1 = new HingeEffector(conf
						.getHj1Configuration().getEffectorName());
				HingeEffector effector2 = new HingeEffector(conf
						.getHj2Configuration().getEffectorName());
				UniversalEffector ue = new UniversalEffector(conf.getName(),
						effector1, effector2);
				effectors.put(conf.getName(), ue);

			} else if (jointConfig instanceof ICompositeJointConfiguration) {
				ICompositeJointConfiguration conf = (ICompositeJointConfiguration) jointConfig;
				IHingeJointConfiguration[] hjConfigurations = conf
						.getHjConfigurations();
				HingeEffector[] he = new HingeEffector[hjConfigurations.length];
				for (int i = 0; i < hjConfigurations.length; i++) {
					he[i] = new HingeEffector(hjConfigurations[i].getEffectorName());
				}
				CompositeEffector ce = new CompositeEffector(conf.getName(), he);
				effectors.put(conf.getName(), ce);

			} else {
				// Further Joints go here
			}
		}
	}

	@Override
	public void setEffectorValues(String name, float... values)
	{
		effectors.get(name).setEffectorValues(maxGain, values);
	}

	@Override
	public void sendBeamString(float x, float y, float rotation)
	{
		beam.setEffectorValues(0, x, y, rotation);
		send(Collections.singletonMap(beam.getName(), (IEffector) beam));
		beam.resetAfterAction();
	}

	@Override
	public void sayMessage(String message)
	{
		say.setMessage(message);
	}

	@Override
	public void sendMotorMessage()
	{
		send(actionEffectors);

		for (IEffector effector : actionEffectors.values()) {
			effector.resetAfterAction();
		}
	}

	/**
	 * @param effectors
	 */
	protected void send(Map<String, IEffector> effectors)
	{
		actionPerformer.performAction(effectors);
	}

	@Override
	public void setMaxGain(float motorGain)
	{
		maxGain = motorGain;
	}

	@Override
	public float getMaxGain()
	{
		return maxGain;
	}
}
