/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel;

import magma.util.geometry.Pose6D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * 
 * @author dorer
 */
public interface IBodyModel
{
	/**
	 * @param name the name of the sensor to return
	 * @return the first sensor in a body part that is found having passed name
	 */
	ISensor getSensor(String name);

	/**
	 * @param name name of the joint to retrieve
	 * @return the first joint specified by its name found in a body part, null
	 *         if not existing in this model
	 */
	IJoint getJoint(String name);

	/**
	 * @return the torso of this body model
	 */
	IBodyPart getTorso();

	/**
	 * @param name the name of the body part to return
	 * @return the first body part found with passed name, null if not existing
	 */
	IBodyPart getBodyPart(String name);

	/**
	 * Center of Mass
	 * @return
	 */
	Vector3D getCenterOfMass();

	/**
	 * Center of Mass relative to the ground
	 * 
	 * TODO: Should be renamed to "Center of Stability" or something, since
	 * "center of Gravity" would resoult in the same as "Center of Mass" as all
	 * Body Parts have the same density
	 * 
	 * @return
	 */
	Vector3D getCenterOfGravity();

	/**
	 * 4 Vectors that define the stability hull<br>
	 * <br>
	 * 0: Left foot, left tiptoe <br>
	 * 1: Left foot, left heel <br>
	 * 2: Right foot, right tiptoe <br>
	 * 3: Right foot, right heel
	 * 
	 * @return Stability Hull
	 */
	public Vector3D[] getStabilityHull();

	/**
	 * @param name the name of the hinge joint to change axis speed
	 * @param value the position this joint should reach (in degrees?)
	 */
	void setHingeJointPosition(String name, double value);

	/**
	 * Moves all joints towards their zero position. This method is usually
	 * called to bring the pose as close as possible to the initial pose while
	 * beaming.
	 */
	void performInitialPose();

	/**
	 * Reset all movements (set all joint speeds to 0).
	 */
	void resetAllMovements();

	/**
	 * Adjusts the desired speed of the passed joint to reach current position +
	 * delta
	 * @param name the name of the hinge joint to change axis speed
	 * @param delta the delta angle (deg) to add to the current position
	 */
	void adjustHingeJointPosition(String name, double delta);

	/**
	 * Get the position of the corners of the given BodyPart relative to the
	 * torso center.
	 * 
	 * e.g. for a cube this will return eight vectors to the corners of it
	 * 
	 * <pre>
	 *     6___________4
	 *    /|          /|
	 *   / |         / |
	 *  7___________5  |
	 *  |  |        |  |
	 *  |  3________|__1     z       
	 *  | /         | /      |   x   
	 *  |/          |/       | /     
	 *  2___________0	       |/_____y
	 * </pre>
	 * 
	 * TODO: change the method (BodyPart.getCorners()) that the sequence makes
	 * more sense
	 * 
	 * @param part The BodyPart
	 * @return corners of the given part relative to the robots torso center
	 */
	Vector3D[] getCorners(IBodyPart part);

	/**
	 * Convenient method for
	 * {@link #moveBodyPartToPose(String, Vector3D, Vector3D)}.
	 * 
	 * @param targetBodyName the name of the body part that should be moved
	 * @param targetPose the target pose
	 * @return success
	 */
	public boolean moveBodyPartToPose(String targetBodyName, Pose6D targetPose);

	/**
	 * Move the targetBodyPart to the specified target pose relative to the torso
	 * using the underlying agent IK-solver.
	 * 
	 * @param targetBodyName the name of the body part that should be moved
	 * @param targetPosition the target position
	 * @param targetAngles the target angles
	 * @return success
	 */
	public boolean moveBodyPartToPose(String targetBodyName,
			Vector3D targetPosition, Vector3D targetAngles);
}