/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel;

/**
 * SimSpark "UniversalJoint" sensor interface.
 * 
 * @author Stefan Glaser, Klaus Dorer
 */
public interface IUniversalJoint extends IJoint
{
	/**
	 * @return the minimal angle that can be reached by joint1 (in degrees)
	 */
	float getMinAxis1();

	/**
	 * @return the maximal angle that can be reached by joint1 (in degrees)
	 */
	float getMaxAxis1();

	/**
	 * @return the minimal angle that can be reached by joint2 (in degrees)
	 */
	float getMinAxis2();

	/**
	 * @return the maximal angle that can be reached by joint2 (in degrees)
	 */
	float getMaxAxis2();

	/**
	 * Retrieve joint axis1 angle
	 * 
	 * @return first angle of the joint
	 */
	float getAxis1();

	/**
	 * Retrieve joint axis2 angle
	 * 
	 * @return second angle of the joint
	 */
	float getAxis2();

	/**
	 * Let the universal joint perform a movement along axis1 with the given
	 * speed in the current cycle.
	 * 
	 * @param speed to move the joint axis1 (in deg/cycle)
	 */
	void performAxis1Speed(float speed);

	/**
	 * Let the universal joint perform a movement along axis2 with the given
	 * speed in the current cycle.
	 * 
	 * @param speed to move the joint axis2 (in deg/cycle)
	 */
	void performAxis2Speed(float speed);

	/**
	 * Let the hinge joint perform a movement along axis1 in direction of the
	 * given target position (at maximum speed)
	 * 
	 * @param position - target position of axis1 movement
	 */
	void performAxis1Position(float position);

	/**
	 * Let the hinge joint perform a movement along axis2 in direction of the
	 * given target position (at maximum speed)
	 * 
	 * @param position - target position of axis2 movement
	 */
	void performAxis2Position(float position);

	/**
	 * Let the hinge joint perform a movement along axis1 in direction of the
	 * given target position with the given speed as maximum speed
	 * 
	 * @param position - target position of axis1 movement
	 * @param maxSpeed - maximum allowed speed for movement (in deg/cycle)
	 */
	void performAxis1Position(float position, float maxSpeed);

	/**
	 * Let the hinge joint perform a movement along axis2 in direction of the
	 * given target position with the given speed as maximum speed
	 * 
	 * @param position - target position of axis2 movement
	 * @param maxSpeed - maximum allowed speed for movement (in deg/cycle)
	 */
	void performAxis2Position(float position, float maxSpeed);

	/**
	 * Returns the next axis1 speed to the joint.<br>
	 * This method is called by the coordinator, during reflecting the future
	 * actions out of the agent model.
	 * 
	 * @return next axis1 speed to perform (in deg/cycle)
	 */
	float getNextAxis1Speed();

	/**
	 * Returns the next axis2 speed to the joint.<br>
	 * This method is called by the coordinator, during reflecting the future
	 * actions out of the agent model.
	 * 
	 * @return next axis2 speed to perform (in deg/cycle)
	 */
	float getNextAxis2Speed();
}