/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.io.Serializable;
import java.util.List;

import magma.agent.IMagmaConstants;
import magma.agent.action.IAction;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.agentmodel.ik.IAgentIKSolver;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.perception.IPerception;
import magma.util.geometry.Geometry;
import magma.util.observer.IObserver;
import magma.util.observer.IPublishSubscribe;
import magma.util.observer.Subject;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Implementation of the AgentModel. Used to represent all the information the
 * agent has about itself.
 * 
 * @author Stefan Glaser, Ingo Schindler
 */

public class AgentModel implements IAgentModel, Serializable
{
	private static final long serialVersionUID = 1L;

	/** observers of agent model */
	protected final transient IPublishSubscribe<IAgentModel> observer;

	/** action processing of the agent */
	protected final transient IAction action;

	// TODO: THink about how to remove the following (robot specific constants)
	// from the agent model

	private double MAX_X_KICK;

	private double MAX_Y_KICK;

	private double MAX_X_BOTH;

	private double MAX_Y_BOTH;

	/** the body part model of this robot as sensed by last perception */
	protected BodyModel bodyModelSensed;

	/** the body part model of this robot as we expect it after next sense */
	protected BodyModel bodyModelExpected;

	/** the body part model of this robot to define the future */
	protected BodyModel bodyModelFuture;

	/** the name of the camera body part containing the camera */
	protected final String nameOfCameraBodyPart;

	/** the static pivot point used in static movement definitions */
	protected final Vector3D staticPivotPoint;

	/**
	 * initializes all known Sensors like: UniversalJoints, HingeJoints,
	 * ForceResistances and GyroRates
	 */

	public AgentModel(IAction action, IAgentMetaModel metaModel,
			IAgentIKSolver ikSolver, double[] kickArea)
	{
		this.action = action;
		bodyModelSensed = new BodyModel(metaModel, ikSolver);
		bodyModelExpected = bodyModelSensed;
		bodyModelFuture = new BodyModel(bodyModelSensed);
		nameOfCameraBodyPart = metaModel.getNameOfCameraBodyPart();
		staticPivotPoint = metaModel.getStaticPivotPoint();
		observer = new Subject<IAgentModel>();
		MAX_X_KICK = kickArea[0];
		MAX_Y_KICK = kickArea[1];
		MAX_X_BOTH = kickArea[2];
		MAX_Y_BOTH = kickArea[3];
	}

	@Override
	public boolean update(IPerception perception)
	{
		// update all Sensors
		bodyModelSensed.updateFromPerception(perception);
		// checkForDiscrepancies();
		bodyModelExpected = bodyModelFuture;
		bodyModelExpected.updateFromPerception(perception);

		if (IMagmaConstants.USE_FUTURE_BODY_MODEL) {
			bodyModelFuture = new BodyModel(bodyModelExpected);
		} else {
			bodyModelFuture = new BodyModel(bodyModelSensed);
		}

		// now inform observers about changes
		observer.onStateChange(this);
		return perception.containsMotion();
	}

	@Override
	public void attach(IObserver<IAgentModel> newObserver)
	{
		observer.attach(newObserver);
	}

	@Override
	public boolean detach(IObserver<IAgentModel> oldObserver)
	{
		return observer.detach(oldObserver);
	}

	@Override
	public UniversalJoint getUJ(String name)
	{
		return (UniversalJoint) bodyModelSensed.getJoint(name);
	}

	@Override
	public HingeJoint getHJ(String name)
	{
		return (HingeJoint) bodyModelSensed.getJoint(name);
	}

	@Override
	public HingeJoint getHJExpected(String name)
	{
		return (HingeJoint) bodyModelExpected.getJoint(name);
	}

	@Override
	public IHingeJoint getWriteableHJ(String name)
	{
		// TODO: kdo here we can now switch to a separate instance
		return (IHingeJoint) bodyModelFuture.getSensor(name);
	}

	@Override
	public IBodyPart getBodyPart(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return bodyModelSensed.getBodyPart(name);
	}

	@Override
	public IBodyPart getTorso()
	{
		return bodyModelSensed.getTorso();
	}

	@Override
	public Vector3D getCenterOfMass()
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return bodyModelSensed.getCenterOfMass();
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof AgentModel)) {
			return false;
		}
		AgentModel other = (AgentModel) o;
		return bodyModelSensed.equals(other.bodyModelSensed);
	}

	@Override
	public Vector3D[] getBothLegKickRectangle()
	{
		// points are relative to player coordinates
		double[][] points = { { 0, MAX_Y_BOTH }, { MAX_X_BOTH, MAX_Y_BOTH },
				{ MAX_X_BOTH, -MAX_Y_BOTH }, { 0, -MAX_Y_BOTH } };
		return Geometry.getPolygon(points);
	}

	@Override
	public Vector3D[] getLeftLegKickRectangle()
	{
		// points are relative to player coordinates
		double[][] points = { { 0, MAX_Y_KICK }, { MAX_X_KICK, MAX_Y_KICK },
				{ MAX_X_KICK, 0 }, { 0, 0 } };
		return Geometry.getPolygon(points);
	}

	/**
	 * Calculate the rectangle within the ball is kickable by the right leg
	 * 
	 * @return Calculated rectangle
	 */
	@Override
	public Vector3D[] getRightLegKickRectangle()
	{
		// points are relative to player coordinates
		double[][] points = { { 0, 0 }, { MAX_X_KICK, 0 },
				{ MAX_X_KICK, -MAX_Y_KICK }, { 0, -MAX_Y_KICK } };
		return Geometry.getPolygon(points);
	}

	@Override
	public int hashCode()
	{
		return bodyModelSensed.hashCode();
	}

	@Override
	public void shutOff(boolean off)
	{
		bodyModelFuture.shutOff(off);
	}

	@Override
	public void beamToPosition(Vector3D position)
	{
		bodyModelFuture.beamToPosition(position);
	}

	@Override
	public void sayMessage(String message)
	{
		bodyModelFuture.sayMessage(message);
	}

	@Override
	public ForceResistance getForceResistance(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return (ForceResistance) bodyModelSensed.getSensor(name);
	}

	@Override
	public GyroRate getGyroRate(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return (GyroRate) bodyModelSensed.getSensor(name);
	}

	@Override
	public Accelerometer getAccelerometer(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return (Accelerometer) bodyModelSensed.getSensor(name);
	}

	@Override
	public String toString()
	{
		return bodyModelSensed.toString();
	}

	@Override
	public void reflectTargetStateToAction()
	{
		bodyModelFuture.generateTargetStateToAction(action);
	}

	@Override
	public void updateJointsSpeed(IBodyModel source)
	{
		bodyModelFuture.updateJointsSpeed(source);
	}

	@Override
	public IBodyModel getFutureBodyModel()
	{
		return bodyModelFuture;
	}

	@Override
	public IBodyPart getBodyPartContainingCamera()
	{
		if (nameOfCameraBodyPart != null) {
			return getBodyPart(nameOfCameraBodyPart);
		}

		return null;
	}

	@Override
	public Vector3D getStaticPivotPoint()
	{
		if (staticPivotPoint == null) {
			return Vector3D.ZERO;
		}

		return staticPivotPoint;
	}

	@Override
	public List<IHingeJoint> getAllHingeJoints()
	{
		return bodyModelSensed.getAllHingeJoints();
	}

	@Override
	public double getMaxMotorTemperature()
	{
		List<IHingeJoint> joints = getAllHingeJoints();
		double max = 0;
		for (IHingeJoint joint : joints) {
			float temperature = joint.getMotor().getTemperature();
			if (temperature > max) {
				max = temperature;
			}
		}
		return max;
	}

	@Override
	public double getAverageMotorTemperature()
	{
		List<IHingeJoint> joints = getAllHingeJoints();
		if (joints.isEmpty()) {
			return 0;
		}

		double sum = 0;
		for (IHingeJoint joint : joints) {
			sum += joint.getMotor().getTemperature();
		}
		return sum / joints.size();
	}

	@Override
	public double getMaxCoilTemperature()
	{
		List<IHingeJoint> joints = getAllHingeJoints();
		double max = 0;
		for (IHingeJoint joint : joints) {
			float temperatureCoil = joint.getMotor()
					.getCalculatedTemperatureCoil();
			if (temperatureCoil > max) {
				max = temperatureCoil;
			}
		}
		return max;
	}

	@Override
	public double getAverageCoilTemperature()
	{
		List<IHingeJoint> joints = getAllHingeJoints();
		if (joints.isEmpty()) {
			return 0;
		}

		double sum = 0;
		for (IHingeJoint joint : joints) {
			sum += joint.getMotor().getCalculatedTemperatureCoil();
		}
		return sum / joints.size();
	}

	@Override
	public void setStiffness(float stiffness)
	{
		List<IHingeJoint> hinges = bodyModelFuture.getAllHingeJoints();
		for (IHingeJoint hinge : hinges) {
			hinge.getMotor().setStiffness(stiffness);
		}
	}
}
