/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel.impl;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.agentmodel.ik.IAgentIKSolver;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.perception.IPerception;
import magma.robots.nao.INaoConstants;
import magma.util.FuzzyCompare;
import magma.util.geometry.Pose3D;
import magma.util.geometry.Pose6D;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Represents the body with its body parts and joints
 * @author dorer
 */
class BodyModel implements Serializable, IBodyModel
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** link to the parent body part */
	private final IBodyPart torso;

	/** the next say message to the server */
	protected String sayMessage;

	/** the next beam position */
	protected Vector3D beamPosition;

	/** indicator if beam action has been performed and should be reset */
	private transient boolean beamPerformed;

	private Draw draw;

	/** a cache for all sensors including joints for faster access */
	private Map<String, ISensor> sensorCache;

	private boolean centerOfMassValid;

	private Vector3D centerOfMass;

	private transient final IAgentIKSolver ikSolver;

	private boolean shutOff;

	/**
	 * Constructor to create the body model from a passed meta model
	 * @param metaModel the meta model containing body part information
	 */
	BodyModel(IAgentMetaModel metaModel, IAgentIKSolver ikSolver)
	{
		this.ikSolver = ikSolver;
		Map<String, IBodyPart> bodyParts = createBodyParts(metaModel);
		torso = connectBodyParts(metaModel, bodyParts);
		sensorCache = collectAllSensors();
		sayMessage = null;
		beamPosition = null;
		beamPerformed = false;
		centerOfMassValid = false;
		// draw = new Draw();

	}

	/**
	 * Copy constructor
	 * @param source the source body model to copy
	 */
	BodyModel(BodyModel source)
	{
		this.ikSolver = source.ikSolver;
		torso = new BodyPart((BodyPart) source.torso, null);
		sayMessage = source.sayMessage;
		if (source.beamPosition == null) {
			beamPosition = null;
		} else {
			beamPosition = source.beamPosition;
		}
		sensorCache = collectAllSensors();
	}

	private Map<String, IBodyPart> createBodyParts(IAgentMetaModel metaModel)
	{
		Map<String, IBodyPart> bodyParts = new HashMap<String, IBodyPart>();

		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			BodyPart newPart = new BodyPart(config);
			bodyParts.put(config.getName(), newPart);
		}
		return bodyParts;
	}

	/**
	 * @param metaModel
	 * @param bodyParts
	 */
	private BodyPart connectBodyParts(IAgentMetaModel metaModel,
			Map<String, IBodyPart> bodyParts)
	{
		BodyPart result = null;
		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			BodyPart parent = (BodyPart) bodyParts.get(config.getParent());
			BodyPart currentPart = (BodyPart) bodyParts.get(config.getName());
			currentPart.setParent(parent);
			if (parent == null) {
				// we have what we call torso
				result = currentPart;
			}
		}
		return result;
	}

	/**
	 * @return
	 */
	private Map<String, ISensor> collectAllSensors()
	{
		Map<String, ISensor> result = new HashMap<String, ISensor>(50);
		torso.collectSensors(result);
		return result;
	}

	/**
	 * Creates the action content for sending to the server
	 */
	void generateTargetStateToAction(IAction action)
	{
		generateShutoffAction(action);
		generateBeamAction(action);

		Map<String, float[]> actions = new HashMap<String, float[]>();
		generateJointActions(actions);
		for (String key : actions.keySet()) {
			action.setEffectorValues(key, actions.get(key));
		}

		generateSayAction(action);
	}

	/**
	 * Informs the action component to create action commands from this body
	 * model
	 * @param action the action component that creates the action protocol
	 */
	void generateJointActions(Map<String, float[]> actions)
	{
		torso.generateJointActions(actions);
	}

	/**
	 * Called to forward the beam action.
	 */
	void generateShutoffAction(IAction action)
	{
		if (shutOff) {
			action.setMaxGain(0.0f);
		}
	}

	/**
	 * Called to forward the beam action.
	 */
	void generateBeamAction(IAction action)
	{
		if (beamPosition != null) {
			action.sendBeamString((float) beamPosition.getX(),
					(float) beamPosition.getY(), (float) beamPosition.getZ());
			beamPerformed = true;
		}
	}

	void shutOff(boolean off)
	{
		shutOff = off;
	}

	void beamToPosition(Vector3D position)
	{
		beamPosition = position;
	}

	/**
	 * Called to forward the say message.
	 */
	void generateSayAction(IAction action)
	{
		if (sayMessage != null) {
			action.sayMessage(sayMessage);
		}
	}

	void sayMessage(String message)
	{
		sayMessage = message;
	}

	/**
	 * Updates the joint values in the body model from perception
	 * @param perception the new perception we made
	 */
	void updateFromPerception(IPerception perception)
	{
		// reset actions
		sayMessage = null;
		if (beamPerformed) {
			beamPosition = null;
			beamPerformed = false;
		}

		torso.updateFromPerception(perception);
		centerOfMassValid = false;
		if (draw != null)
			draw.repaint();
	}

	/**
	 * Updates the joint values in the body model if not connected to perception
	 */
	void updateNoPerception()
	{
		// reset actions
		sayMessage = null;
		if (beamPerformed) {
			beamPosition = null;
			beamPerformed = false;
		}

		torso.updateNoPerception();
		centerOfMassValid = false;
		if (draw != null)
			draw.repaint();
	}

	/**
	 * Called to copy back joint speed values from the passed model to this
	 * model. The passed model is usually a copy that has been created earlier
	 * from this object so that the joint model matches.
	 * @param source the body model to take the values from
	 */
	public void updateJointsSpeed(IBodyModel source)
	{
		torso.updateJointsSpeed(source.getTorso());
	}

	@Override
	public ISensor getSensor(String name)
	{
		return sensorCache.get(name);
		// return torso.getSensorDeep(name);
	}

	@Override
	public IJoint getJoint(String name)
	{
		return (IJoint) getSensor(name);
	}

	@Override
	public IBodyPart getTorso()
	{
		return torso;
	}

	@Override
	public IBodyPart getBodyPart(String name)
	{
		return torso.getBodyPart(name);
	}

	@Override
	public Vector3D getCenterOfMass()
	{
		if (!centerOfMassValid) {
			centerOfMass = torso.getCenterOfMass();
		}
		return centerOfMass;
	}

	public Vector3D getCenterOfGravityOld()
	{
		// // m should be the global orientation but is not available any more
		// // (was gyro orientation before)
		// Matrix3d m = new Matrix3d();
		// m.setIdentity();
		//
		// Vector3d v = MatrixUtil.mulMatrixVector(m, getCenterOfMass());
		//
		// Matrix3d m3 = new Matrix3d();
		// m3.rotZ(Math.atan2(m.m01, m.m11));
		//
		// return MatrixUtil.mulMatrixVector(m3, v);
		return getCenterOfMass();
	}

	@Override
	public Vector3D getCenterOfGravity()
	{
		// TODO: Remove dependency to Nao robot model
		Vector3D acc = ((Accelerometer) torso
				.getSensorDeep(INaoConstants.TorsoAccelerometer)).getAcceleration();

		double distToGroundLeft = getBodyPart(INaoConstants.LFoot).getPosition()
				.getNorm();
		double distToGroundRight = getBodyPart(INaoConstants.RFoot).getPosition()
				.getNorm();

		double distToExpectedGround;
		if (FuzzyCompare.gt(distToGroundLeft, distToGroundRight, 4)) {
			distToExpectedGround = distToGroundLeft;
		} else if (FuzzyCompare.lt(distToGroundLeft, distToGroundRight, 4)) {
			distToExpectedGround = distToGroundRight;
		} else {
			distToExpectedGround = (distToGroundRight + distToGroundLeft) / 2;
		}

		if (FuzzyCompare.eq(0, acc.getNorm(), 0.00001)) {
			return getCenterOfMass();
		}

		acc = acc.negate();

		acc = acc.normalize();

		acc = acc.scalarMultiply(distToExpectedGround);

		acc = acc.add(getCenterOfMass());

		double x = Double.isNaN(acc.getX()) ? 0 : acc.getX();
		double y = Double.isNaN(acc.getY()) ? 0 : acc.getY();
		double z = Double.isNaN(acc.getZ()) ? 0 : acc.getZ();

		return new Vector3D(x, y, z);
	}

	@Override
	public Vector3D[] getStabilityHull()
	{
		// TODO: Remove dependency to Nao robot model
		IBodyPart lfoot = torso.getBodyPart(INaoConstants.LFoot);
		Pose3D lPose = lfoot.getPose();
		Vector3D lPos = lPose.getPosition();
		Rotation lOrientation = lPose.getOrientation();
		Vector3D lHalfSize = lfoot.getGeometry().scalarMultiply(0.5);
		IBodyPart rfoot = torso.getBodyPart(INaoConstants.RFoot);
		Pose3D rPose = rfoot.getPose();
		Vector3D rPos = rPose.getPosition();
		Rotation rOrientation = rPose.getOrientation();
		int index = 0;

		Vector3D leftFrontCorner = new Vector3D(-lHalfSize.getX(),
				lHalfSize.getY(), -lHalfSize.getZ());
		Vector3D rightFrontCorner = new Vector3D(lHalfSize.getX(),
				lHalfSize.getY(), -lHalfSize.getZ());
		Vector3D rightBackCorner = new Vector3D(lHalfSize.getX(),
				-lHalfSize.getY(), -lHalfSize.getZ());
		Vector3D leftBackCorner = new Vector3D(-lHalfSize.getX(),
				-lHalfSize.getY(), -lHalfSize.getZ());

		Vector3D[] positions = new Vector3D[6];

		positions[index] = lPos.add(lOrientation.applyTo(leftFrontCorner));
		index++;
		if (lPos.getY() > rPos.getY()) {
			positions[index] = lPos.add(lOrientation.applyTo(rightFrontCorner));
			index++;
		} else {
			// if (rPos.getY() > lPos.getY()) {
			positions[index] = rPos.add(rOrientation.applyTo(leftFrontCorner));
			index++;
		}
		positions[index] = rPos.add(rOrientation.applyTo(rightFrontCorner));
		index++;
		positions[index] = rPos.add(rOrientation.applyTo(rightBackCorner));
		index++;
		if (lPos.getY() > rPos.getY()) {
			positions[index] = rPos.add(rOrientation.applyTo(leftBackCorner));
			index++;
		} else {
			// if (rPos.getY() > lPos.getY()) {
			positions[index] = lPos.add(lOrientation.applyTo(rightBackCorner));
			index++;
		}
		positions[index] = lPos.add(lOrientation.applyTo(leftBackCorner));

		// Vector3D[] lFoot = getCorners(torso.getBodyPart(INaoConstants.LFoot));
		// Vector3D[] rFoot = getCorners(torso.getBodyPart(INaoConstants.RFoot));
		//
		// positions[0] = lFoot[0];
		// positions[1] = rFoot[1];
		//
		// positions[2] = rFoot[2];
		// positions[3] = lFoot[3];

		return positions;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof BodyModel)) {
			return false;
		}
		BodyModel other = (BodyModel) o;
		return torso.equals(other.torso);
	}

	@Override
	public void setHingeJointPosition(String name, double value)
	{
		IJoint joint = getJoint(name);
		((HingeJoint) joint).performAxisPosition(value);
	}

	@Override
	public void adjustHingeJointPosition(String name, double delta)
	{
		IJoint joint = getJoint(name);
		((HingeJoint) joint).adjustAxisPosition(delta);
	}

	@Override
	public void performInitialPose()
	{
		LinkedList<IBodyPart> bodyPartsToInitialize = new LinkedList<IBodyPart>();
		Collection<IBodyPart> children;
		IBodyPart currentBodyPart = torso;

		while (currentBodyPart != null) {
			if (currentBodyPart.getJoint() != null) {
				currentBodyPart.getJoint().performInitialPosition();
			}

			// Add all children to list of body part to initialize and poll the
			// next body part to initialize
			children = currentBodyPart.getChildren();
			if (children != null) {
				bodyPartsToInitialize.addAll(children);
			}
			currentBodyPart = bodyPartsToInitialize.pollFirst();
		}
	}

	@Override
	public void resetAllMovements()
	{
		LinkedList<IBodyPart> bodyPartsToInitialize = new LinkedList<IBodyPart>();
		Collection<IBodyPart> children;
		IBodyPart currentBodyPart = torso;

		while (currentBodyPart != null) {
			if (currentBodyPart.getJoint() != null) {
				currentBodyPart.getJoint().resetMovement();
			}

			// Add all children to list of body part to initialize and poll the
			// next body part to initialize
			children = currentBodyPart.getChildren();
			if (children != null) {
				bodyPartsToInitialize.addAll(children);
			}
			currentBodyPart = bodyPartsToInitialize.pollFirst();
		}
	}

	@Override
	public Vector3D[] getCorners(IBodyPart part)
	{
		Pose3D pose = part.getPose();

		Vector3D[] edges = part.getCorners();
		Vector3D[] edgesReturn = new Vector3D[edges.length];

		for (int i = 0; i < edges.length; i++) {
			edgesReturn[i] = pose.getPosition().add(
					pose.getOrientation().applyTo(edges[i]));
		}

		return edgesReturn;
	}

	@Override
	public boolean moveBodyPartToPose(String targetBodyName, Pose6D targetPose)
	{
		return moveBodyPartToPose(targetBodyName, targetPose.getPosition(),
				targetPose.getAngles());
	}

	@Override
	public boolean moveBodyPartToPose(String targetBodyName,
			Vector3D targetPosition, Vector3D targetAngles)
	{
		IBodyPart targetBody = getBodyPart(targetBodyName);

		if (targetBody != null) {
			return ikSolver.solve(targetBody, targetPosition, targetAngles);
		}

		return false;
	}

	private class Draw extends JPanel
	{
		private static final long serialVersionUID = -3570338876522740810L;

		final ArrayList<Vector3D> rForceList = new ArrayList<Vector3D>(250);

		ArrayList<Vector3D> lForceList;

		@Override
		protected void paintComponent(Graphics g)
		{
			super.paintComponent(g);

			Graphics2D g2d = (Graphics2D) g;
			g2d.transform(getTransform());

			// ///////////////////////////////////////////////////////////////////
			// Coordinate-System
			// ///////////////////////////////////////////////////////////////////
			g2d.setStroke(new BasicStroke(1));
			g2d.setColor(Color.GREEN);
			drawLine(g2d, -10, 0, 10, 0);
			drawLine(g2d, 0, -10, 0, 10);
			// ///////////////////////////////////////////////////////////////////

			// ///////////////////////////////////////////////////////////////////
			// Stability Hull
			// ///////////////////////////////////////////////////////////////////
			Vector3D[] stab = getStabilityHull();

			g2d.setStroke(new BasicStroke(2));
			g2d.setColor(Color.black);

			// left left
			drawLine(g2d, stab[0].getX(), stab[0].getY(), stab[1].getX(),
					stab[1].getY());
			drawString(g2d,
					String.format("%f : %f", stab[0].getX(), stab[0].getY()),
					stab[0].getX() - 0.1, stab[0].getY() + 0.01);

			// left right
			drawLine(g2d, stab[1].getX(), stab[1].getY(), stab[2].getX(),
					stab[2].getY());
			drawString(g2d,
					String.format("%f : %f", stab[1].getX(), stab[1].getY()),
					stab[1].getX() - 0.1, stab[1].getY() - 0.02);

			// right right
			drawLine(g2d, stab[2].getX(), stab[2].getY(), stab[3].getX(),
					stab[3].getY());
			drawString(g2d,
					String.format("%f : %f", stab[2].getX(), stab[2].getY()),
					stab[2].getX() + 0.001, stab[2].getY() + 0.01);

			// right left
			drawLine(g2d, stab[3].getX(), stab[3].getY(), stab[0].getX(),
					stab[0].getY());
			drawString(g2d,
					String.format("%f : %f", stab[3].getX(), stab[3].getY()),
					stab[3].getX() + 0.001, stab[3].getY() - 0.02);
			// ///////////////////////////////////////////////////////////////////

			// ///////////////////////////////////////////////////////////////////
			// Foot + FootForce
			// ///////////////////////////////////////////////////////////////////
			g2d.setStroke(new BasicStroke(1));
			g2d.setColor(Color.BLACK);
			drawForce(g2d, getBodyPart(INaoConstants.RFoot));

			g2d.setColor(Color.BLACK);
			g2d.setStroke(new BasicStroke(1));
			drawForce(g2d, getBodyPart(INaoConstants.LFoot));
			// ------------------------------------------------------------
			// ///////////////////////////////////////////////////////////////////

			// CoM
			g2d.setStroke(new BasicStroke(2));
			g2d.setColor(Color.BLUE);
			Vector3D com = getCenterOfMass();
			drawPoint(g2d, com.getX(), com.getY());
			drawString(
					g2d,
					String.format("CoM: %f : %f : %f", com.getX(), com.getY(),
							com.getZ()), 2);

			// CoG
			g2d.setColor(Color.RED);
			Vector3D cog = getCenterOfGravity();
			drawPoint(g2d, cog.getX(), cog.getY());
			drawString(
					g2d,
					String.format("CoG: %f : %f : %f", cog.getX(), cog.getY(),
							cog.getZ()), 3);

			g2d.setColor(Color.GREEN);
			Vector3D acc = ((Accelerometer) getSensor(INaoConstants.TorsoAccelerometer))
					.getAcceleration();

			drawString(
					g2d,
					String.format("ACC: %f : %f : %f", acc.getX(), acc.getY(),
							acc.getZ()), 4);

		}

		/**
		 * @param g2d
		 * @param foot
		 */
		private void drawForce(Graphics2D g2d, IBodyPart foot)
		{

			ForceResistance footForce = ((ForceResistance) foot.getSensorDeep(foot
					.getName() + "Force"));

			Pose3D footPose = foot.getPose();

			drawFeet(g2d, foot);

			if (footForce.getForce().getNorm() > ((torso.getWholeMass() * 10) / 2)) {
				g2d.setColor(Color.RED);
			} else {
				g2d.setColor(Color.ORANGE);
			}

			g2d.setStroke(new BasicStroke(5));

			Vector3D force = footForce.getForceOrigin();
			ArrayList<Vector3D> forceList;
			int line = 0;
			if (foot.getName().startsWith("R")) {
				forceList = rForceList;
			} else {
				forceList = lForceList;
				line = 1;
			}

			drawString(g2d, String.format(foot.getName() + "Force: %f", footForce
					.getForce().getNorm()), line);

			forceList.add(force);

			if (forceList.size() == 26)
				forceList.remove(0);

			int i = 255;
			for (Vector3D v : forceList) {
				i -= 6;
				g2d.setColor(new Color(i, i, i));

				Vector3D forcePos = footPose.getPosition().add(
						footPose.getOrientation().applyTo(v));

				drawPoint(g2d, forcePos.getX(), forcePos.getY());
			}
		}

		/**
		 * @param g2d
		 * @param foot
		 */
		private void drawFeet(Graphics2D g2d, IBodyPart foot)
		{

			drawString(g2d, foot.getName().substring(0, 1), foot.getPosition()
					.getX(), foot.getPosition().getY());

			Vector3D[] edge = getCorners(foot);

			for (int i = 0; i < 3; i++)
				drawLine(g2d, edge[i].getX(), edge[i].getY(), edge[i + 1].getX(),
						edge[i + 1].getY());

			drawLine(g2d, edge[3].getX(), edge[3].getY(), edge[0].getX(),
					edge[0].getY());

		}

		private void drawLine(Graphics2D g, double x1, double y1, double x2,
				double y2)
		{
			int ix1 = (int) (x1 * 1000);

			int iy1 = (int) (y1 * 1000);

			int ix2 = (int) (x2 * 1000);

			int iy2 = (int) (y2 * 1000);

			g.drawLine(ix1, iy1, ix2, iy2);

		}

		private void drawPoint(Graphics2D g, double x, double y)
		{
			int ix = (int) (x * 1000);

			int iy = (int) (y * 1000);

			g.drawOval(ix - 2, iy - 2, 4, 4);

		}

		private AffineTransform getTransform()
		{
			AffineTransform transform = new AffineTransform();

			double deltaX = (double) this.getWidth() / 2;

			double deltaY = (double) this.getHeight() / 2;

			transform.translate(deltaX, deltaY);

			transform.scale(1, -1);

			return transform;
		}

		private void drawString(Graphics2D g, String str, double x, double y)
		{

			AffineTransform transform = new AffineTransform();

			int ix = (int) (x * 1000);

			int iy = (int) (y * 1000);

			transform.translate(ix, iy);
			transform.scale(1, -1);

			g.transform(transform);

			g.drawString(str, 0, 0);

			transform.scale(1, 1);
			transform.translate(-ix * 2, 0);
			g.transform(transform);
		}

		private void drawString(Graphics2D g, String str, int line)
		{

			AffineTransform transform = new AffineTransform();

			transform.translate(-this.getWidth() / 2 + 10, this.getHeight() / 2
					- 20 - 15 * line);
			transform.scale(1, -1);

			g.transform(transform);

			g.drawString(str, 0, 0);

			transform.scale(1, 1);
			transform.translate(this.getWidth() - 20, 0);
			g.transform(transform);
		}
	}

	/**
	 * @return
	 */
	public List<IHingeJoint> getAllHingeJoints()
	{
		List<IHingeJoint> result = new ArrayList<IHingeJoint>();
		torso.getAllHingeJoints(result);
		return result;
	}
}
