/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import magma.agent.agentmodel.ICompositeJoint;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.meta.agent.ICompositeJointConfiguration;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.meta.agent.IJointToMotorMapper;
import magma.agent.perception.ICompositeJointPerceptor;
import magma.agent.perception.IHingeJointPerceptor;
import magma.agent.perception.IPerception;
import magma.agent.perception.impl.HingeJointPerceptor;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;

/**
 * SimSpark "UniversalJoint" sensor implementation
 * 
 * @author Stefan Glaser, Klaus Dorer
 * 
 */
public class CompositeJoint extends Joint implements ICompositeJoint
{
	private static final long serialVersionUID = 3124679237337000038L;

	/** first hinge joint */
	private HingeJoint[] joints;

	private final transient IJointToMotorMapper mapper;

	/**
	 * Instantiates a new UniversalJoint sensor
	 * 
	 * @param jointConfig - the joint meta model
	 * @param effector1
	 * @param effector2
	 */
	public CompositeJoint(ICompositeJointConfiguration jointConfig)
	{
		super(jointConfig.getName(), jointConfig.getPerceptorName());

		IHingeJointConfiguration[] configs = jointConfig.getHjConfigurations();
		joints = new HingeJoint[configs.length];
		for (int i = 0; i < configs.length; i++) {
			IHingeJointConfiguration config = configs[i];
			this.joints[i] = new HingeJoint(config);
		}
		this.mapper = jointConfig.getJointToMotorMapper();
	}

	/**
	 * Copy constructor
	 * @param source to copy from
	 */
	private CompositeJoint(CompositeJoint source)
	{
		super(source);

		joints = new HingeJoint[source.joints.length];
		for (int i = 0; i < joints.length; i++) {
			this.joints[i] = new HingeJoint(source.joints[i]);
		}
		this.mapper = source.mapper;
	}

	@Override
	public float getAxis(int i)
	{
		return joints[i].getAxis();
	}

	@Override
	public float getMinAxis(int i)
	{
		return joints[i].getMin();
	}

	@Override
	public float getMaxAxis(int i)
	{
		return joints[i].getMax();
	}

	@Override
	public float getNextAxisSpeed(int i)
	{
		return joints[i].getNextAxisSpeed();
	}

	@Override
	public void performAxisSpeed(int i, float speed)
	{
		joints[i].performAxisSpeed(speed);
	}

	@Override
	public void performAxisPosition(int i, float position)
	{
		joints[i].performAxisPosition(position);
	}

	@Override
	public void performAxisPosition(int i, float position, float maxSpeed)
	{
		joints[i].performAxisPosition(position, maxSpeed);
	}

	@Override
	public void performInitialPosition()
	{
		for (int i = 0; i < joints.length; i++) {
			joints[i].performInitialPosition();
		}
	}

	@Override
	public void resetMovement()
	{
		for (int i = 0; i < joints.length; i++) {
			joints[i].resetMovement();
		}
	}

	@Override
	public Rotation getRotation()
	{
		Rotation result = joints[0].getRotation();
		for (int i = 1; i < joints.length; i++) {
			result = joints[i].getRotation().applyTo(result);
		}
		return result;
	}

	@Override
	public ISensor copy()
	{
		return new CompositeJoint(this);
	}

	@Override
	public void generateJointAction(Map<String, float[]> actions)
	{
		double[] jointAngles = new double[joints.length];
		for (int i = 0; i < joints.length; i++) {
			jointAngles[i] = joints[i].getAxis();
		}

		double[] motorAngles = jointAngles;
		if (mapper != null) {
			motorAngles = mapper.jointToMotorAngle(jointAngles);
		}

		float[] values = new float[joints.length * 3];
		for (int i = 0; i < joints.length; i++) {
			int j = i * 3;
			values[j] = joints[i].getMotor().getNextSpeed((float) motorAngles[i]);
			values[j + 1] = (float) motorAngles[i];
			values[j + 2] = joints[i].getMotor().getGain();
		}

		// if (getName().equals(ISweatyRealConstants.NeckYawCombined)) {
		// System.out.println(jointAngles[0] + "; " + motorAngles[0]);
		// }

		actions.put(getName(), values);
	}

	@Override
	public void updateFromPerception(IPerception perception)
	{
		ICompositeJointPerceptor ujPerceptor = perception
				.getCompositeJointPerceptor(getPerceptorName());

		if (ujPerceptor == null) {
			return;
		}

		IHingeJointPerceptor[] jointPerceptors = ujPerceptor.getJoints();
		double[] motorAngles = new double[joints.length];
		for (int i = 0; i < joints.length; i++) {
			motorAngles[i] = jointPerceptors[i].getAxis();
		}

		// map motor to joint angles
		double[] jointAngles = motorAngles;
		if (mapper != null) {
			jointAngles = mapper.motorToJointAngle(motorAngles);
		}

		// if (getName().equalsIgnoreCase(ISweatyRealConstants.LHipCombined)) {
		// System.out.println("Left Hip: " + Arrays.toString(jointAngles));
		// }

		for (int i = 0; i < joints.length; i++) {
			// apply the possible change of the motor mapper
			((HingeJointPerceptor) jointPerceptors[i])
					.setAxis((float) jointAngles[i]);
			joints[i].updateFromPerception(jointPerceptors[i]);
		}
	}

	@Override
	public void updateNoPerception()
	{
		for (int i = 0; i < joints.length; i++) {
			joints[i].updateNoPerception();
		}
	}

	@Override
	public void updateJointPositionFromJoint(IJoint joint)
	{
		CompositeJoint cj = (CompositeJoint) joint;
		for (int i = 0; i < joints.length; i++) {
			joints[i].updateJointPositionFromJoint(cj.joints[i]);
		}
	}

	@Override
	public List<IJoint> getAllSubJoints()
	{
		List<IJoint> result = new ArrayList<IJoint>();
		for (IHingeJoint joint : joints) {
			result.add(joint);
		}
		return result;
	}

	@Override
	public void updateSensors(Map<String, ISensor> flatSensors,
			Map<String, ISensor> structuredSensors)
	{
		for (int i = 0; i < joints.length; i++) {
			flatSensors.put(joints[i].getName(), joints[i]);
		}
		structuredSensors.put(getName(), this);
	}
}
