/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import magma.agent.agentmodel.IForceResistance;
import magma.agent.agentmodel.ISensor;
import magma.agent.perception.IForceResistancePerceptor;
import magma.agent.perception.IPerception;
import magma.util.FuzzyCompare;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Implementation of the SimSpark "ForceRestistance" sensor
 * 
 * @author Klaus Dorer
 */
public class ForceResistance extends Sensor implements IForceResistance
{
	private static final long serialVersionUID = 7431711977473627245L;

	/** point where the force acts */
	private Vector3D forceOrigin;

	/** the force itself */
	private Vector3D force;

	/**
	 * Instantiates a new ForceResistance sensor
	 * 
	 * @param name Sensor name
	 */
	public ForceResistance(String name, String perceptorName)
	{
		super(name, perceptorName);
		forceOrigin = Vector3D.ZERO;
		force = Vector3D.ZERO;
	}

	/**
	 * Copy constructor
	 * @param source the object to copy from
	 */
	private ForceResistance(ForceResistance source)
	{
		super(source.name, source.perceptorName);
		forceOrigin = source.forceOrigin;
		force = source.force;
	}

	@Override
	public Vector3D getForceOrigin()
	{
		return forceOrigin;
	}

	/**
	 * Set the force origin
	 * 
	 * @param forceOrigin Force origin point
	 */
	void setForceOrigin(Vector3D forceOrigin)
	{
		this.forceOrigin = forceOrigin;
	}

	@Override
	public Vector3D getForce()
	{
		return force;
	}

	/**
	 * Set the force vector
	 * 
	 * @param force Force vector
	 */
	void setForce(Vector3D force)
	{
		this.force = force;
	}

	/**
	 * Updates this ForceResistances from perception
	 * @param perception the result from server message parsing
	 */
	@Override
	public void updateFromPerception(IPerception perception)
	{
		IForceResistancePerceptor frPerceptor = perception
				.getForceResistancePerceptor(getPerceptorName());

		if (frPerceptor == null) {
			setForce(Vector3D.ZERO);
			return;
		}

		// Fetch new Force Values
		setForce(frPerceptor.getForce());
		setForceOrigin(frPerceptor.getForceOrigin());
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof ForceResistance)) {
			return false;
		}
		ForceResistance other = (ForceResistance) o;
		if (!super.equals(other)) {
			return false;
		}
		if (!FuzzyCompare.eq(force, other.force, 0.00001)) {
			return false;
		}
		return FuzzyCompare.eq(forceOrigin, other.forceOrigin, 0.00001f);
	}

	@Override
	public ISensor copy()
	{
		return new ForceResistance(this);
	}
}
