/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import magma.agent.IMagmaConstants;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.IMotor;
import magma.agent.agentmodel.ISensor;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.perception.IHingeJointPerceptor;
import magma.agent.perception.IPerception;
import magma.util.FuzzyCompare;
import magma.util.ValueUtil;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * SimSpark "HingeJoint" sensor implementation
 * 
 * @author Stefan Glaser, Klaus Dorer
 * 
 */
public class HingeJoint extends Joint implements IHingeJoint
{
	private static final long serialVersionUID = -9142499983257109635L;

	// from sensors
	/** current angle of the joint */
	private float ax;

	// self calculated
	/** the original angle of the source body model */
	private float originalAx;

	/** the last speed value of the joint */
	private float lastSpeed;

	/** a vector describing the rotation axis of the joint */
	private final Vector3D jointAxis;

	/** minimum axis position */
	private final float min_ax;

	/** maximum axis position */
	private final float max_ax;

	/** maximum axis speed */
	private final float maxSpeed;

	/** the maximal amount of speed change (in degrees per cycle) */
	private float maxAcceleration;

	/** the rotation to this joint */
	private transient Rotation rotation;

	/** -1 if this joint is turning negated with respect to our standard model */
	private final float negateRotation;

	/** the motor by which this joint is operated */
	private Motor motor;

	/** the acceleration in deg/cycle//cycle */
	private float acceleration;

	/**
	 * Instantiates a new HingeJoint sensor
	 * 
	 * @param name Sensor name
	 * @param effectorName Effector name linked to this sensor
	 * @param min Minimum sensor value
	 * @param max Maximum sensor value
	 */
	HingeJoint(String name, String perceptorName, String effectorName,
			Vector3D jointAxis, float min, float max, float max_speed,
			float max_acc, float gain)
	{
		super(name, perceptorName);
		// TODO: in indirect cases the motor max speed is not the joint max speed
		motor = new Motor(effectorName, max_speed, gain);
		this.ax = 0;
		this.originalAx = 0;
		this.lastSpeed = 0;
		this.jointAxis = jointAxis;
		this.min_ax = min;
		this.max_ax = max;
		this.maxSpeed = max_speed;
		this.maxAcceleration = max_acc;
		acceleration = 0;

		// HACK: to get Webots and 3Dsim to work
		double sum = jointAxis.getX() + jointAxis.getY() + jointAxis.getZ();
		if (sum < -0.9 && sum > -1.1) {
			negateRotation = -1;
		} else {
			negateRotation = 1;
		}
	}

	public HingeJoint(IHingeJointConfiguration jointConfig)
	{
		this(jointConfig.getName(), jointConfig.getPerceptorName(), jointConfig
				.getEffectorName(), jointConfig.getJointAxis(), jointConfig
				.getMinAngle(), jointConfig.getMaxAngle(), jointConfig
				.getMaxSpeed(), jointConfig.getMaxAcceleration(), jointConfig
				.getGain());
	}

	/**
	 * Copy constructor
	 * @param source the object to copy from
	 */
	HingeJoint(HingeJoint source)
	{
		super(source);
		this.motor = new Motor(source.motor);
		this.ax = source.ax;
		// we do not copy the original value but use current as new original
		this.originalAx = ax;
		motor.setPerceivedAngle(ax);
		this.lastSpeed = source.lastSpeed;
		this.acceleration = source.acceleration;
		// we do not expect jointAxis to change. Anyhow, as soon as immutable
		// Vector3ds are used, we need no copy of jointAxis
		this.jointAxis = source.jointAxis;
		this.min_ax = source.min_ax;
		this.max_ax = source.max_ax;
		this.maxSpeed = source.maxSpeed;
		this.maxAcceleration = source.maxAcceleration;
		this.negateRotation = source.negateRotation;
	}

	@Override
	public float getAxis()
	{
		return ax;
	}

	@Override
	public float getMin()
	{
		return min_ax;
	}

	@Override
	public float getMax()
	{
		return max_ax;
	}

	/**
	 * Called to act by setting an angular speed
	 * @param speed the angular speed (in degrees per cycle)
	 */
	@Override
	public void performAxisSpeed(float speed)
	{
		acceleration = speed - lastSpeed;
		if (Math.abs(acceleration) > maxAcceleration) {
			acceleration = ValueUtil.limitAbs(acceleration, maxAcceleration);
			speed = lastSpeed + acceleration;
		}

		float usedSpeed = ValueUtil.limitAbs(speed, maxSpeed);
		float newAx = originalAx + usedSpeed;
		ax = ValueUtil.limitValue(newAx, min_ax, max_ax);

		// Reset rotation matrix
		rotation = null;
	}

	@Override
	public void adjustAxisPosition(double delta)
	{
		performAxisPosition(ax + delta, maxSpeed);
	}

	@Override
	public float performAxisPosition(double position)
	{
		return performAxisPosition(position, maxSpeed);
	}

	@Override
	public float performAxisPosition(double position, float max_speed)
	{
		// float axisSpeed = (float) (position - originalAx) / 2.0f;
		float axisSpeed;
		if (IMagmaConstants.USE_FUTURE_BODY_MODEL) {
			axisSpeed = (float) (position - originalAx);
		} else {
			axisSpeed = (float) (position - originalAx) / (maxSpeed / 2);
		}

		performAxisSpeed(ValueUtil.limitAbs(axisSpeed, max_speed));
		return axisSpeed;
	}

	@Override
	public void performInitialPosition()
	{
		performAxisPosition(0);
	}

	@Override
	public void resetMovement()
	{
		performAxisSpeed(0);
	}

	@Override
	public float getNextAxisSpeed()
	{
		return ax - originalAx;
	}

	@Override
	public Rotation getRotation()
	{
		if (rotation == null) {
			rotation = new Rotation(jointAxis, Math.toRadians(ax));
		}

		return rotation;
	}

	@Override
	public Vector3D getJointAxis()
	{
		return jointAxis;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof HingeJoint)) {
			return false;
		}
		HingeJoint other = (HingeJoint) o;
		if (!super.equals(other)) {
			return false;
		}
		if (!FuzzyCompare.eq(ax, other.ax, 0.00001f)) {
			return false;
		}
		if (!FuzzyCompare.eq(min_ax, other.min_ax, 0.00001f)) {
			return false;
		}
		return FuzzyCompare.eq(max_ax, other.max_ax, 0.00001f);
	}

	@Override
	public ISensor copy()
	{
		return new HingeJoint(this);
	}

	@Override
	public void generateJointAction(Map<String, float[]> actions)
	{
		lastSpeed = ax - originalAx;
		motor.generateMotorAction(actions, ax);
	}

	@Override
	public void updateFromPerception(IPerception perception)
	{
		IHingeJointPerceptor hingeJointPerceptor = perception
				.getHingeJointPerceptor(getPerceptorName());
		if (hingeJointPerceptor != null) {
			updateFromPerception(hingeJointPerceptor);
		} else if (perception.containsMotion()) {
			// TODO: we should have received data from this sensor
		}
	}

	/**
	 * @param hingeJointPerceptor
	 */
	void updateFromPerception(IHingeJointPerceptor hingeJointPerceptor)
	{
		motor.updateFromPerception(hingeJointPerceptor);

		float axis = negateRotation * hingeJointPerceptor.getAxis();
		float delta = ax - originalAx;
		originalAx = axis;
		ax = originalAx + delta;

		// Reset rotation matrix
		rotation = null;
	}

	@Override
	public void updateNoPerception()
	{
		// in the case where we are connected to the percpetion this is done in
		// the copy constructor
		originalAx = ax;
		super.updateNoPerception();
	}

	@Override
	public void updateJointPositionFromJoint(IJoint joint)
	{
		ax = ((HingeJoint) joint).ax;

		// Reset rotation matrix
		rotation = null;
	}

	/**
	 * @return the max_speed
	 */
	@Override
	public float getMaxSpeed()
	{
		return maxSpeed;
	}

	@Override
	public IMotor getMotor()
	{
		return motor;
	}

	/**
	 * @return the max_speed
	 */
	@Override
	public float getAcceleration()
	{
		return acceleration;
	}

	@Override
	public List<IJoint> getAllSubJoints()
	{
		List<IJoint> result = new ArrayList<IJoint>();
		result.add(this);
		return result;
	}

}
