/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.io.Serializable;
import java.util.Map;

import magma.agent.agentmodel.IMotor;
import magma.agent.perception.IHingeJointPerceptor;
import magma.util.ValueUtil;

/**
 * Motor attached to a joint implementation
 * 
 * @author Stefan Glaser, Klaus Dorer
 * 
 */
public class Motor implements Serializable, IMotor
{
	private static final long serialVersionUID = -9142499983257109635L;

	private String name;

	private float speed;

	private float load;

	private float voltage;

	private float temperature;

	private float calculatedTemperature;

	private float calculatedTemperatureCoil;

	// self calculated
	/** the sensor reading of the angle of the motor */
	private float perceivedAngle;

	/** maximum angular speed */
	private final float max_speed;

	/** the servo error code */
	private byte error;

	/** The stiffness of the motor */
	private float stiffness;

	/** the maximum gain of the motor when active */
	private float maxGain;

	/**
	 * Instantiates a new Motor
	 * 
	 * @param name Effector name
	 * @param effector the effector this joint is linked to
	 * @param min Minimum sensor value
	 * @param max Maximum sensor value
	 */
	Motor(String name, float max_speed, float maxGain)
	{
		this.name = name;
		this.perceivedAngle = 0;
		this.max_speed = max_speed;
		this.maxGain = maxGain;
		this.stiffness = 1;
	}

	/**
	 * Copy constructor
	 * @param source the object to copy from
	 */
	Motor(Motor source)
	{
		name = source.name;
		// TODO: check this ???
		// we do not copy the original value but use current as new original
		this.perceivedAngle = source.perceivedAngle;
		this.max_speed = source.max_speed;
		this.speed = source.speed;
		this.load = source.load;
		this.voltage = source.voltage;
		this.temperature = source.temperature;
		this.calculatedTemperature = source.calculatedTemperature;
		this.calculatedTemperatureCoil = source.calculatedTemperatureCoil;
		this.error = source.error;
		this.maxGain = source.maxGain;
		this.stiffness = source.stiffness;
	}

	public IMotor copy()
	{
		return new Motor(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getPerceivedAngle()
	 */
	@Override
	public float getPerceivedAngle()
	{
		return perceivedAngle;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#setPerceivedAngle(float)
	 */
	@Override
	public void setPerceivedAngle(float angle)
	{
		perceivedAngle = angle;
	}

	public void generateMotorAction(Map<String, float[]> actions, float angle)
	{
		float nextAxisSpeed = getNextSpeed(angle);
		float[] values = { nextAxisSpeed, angle, stiffness * maxGain };
		actions.put(name, values);
	}

	/**
	 * @param angle the new desired angle
	 * @return the speed to get to the desired angle
	 */
	@Override
	public float getNextSpeed(float angle)
	{
		return (angle - perceivedAngle) / 2.0f;
	}

	/**
	 * @param hingeJointPerceptor
	 */
	void updateFromPerception(IHingeJointPerceptor hingeJointPerceptor)
	{
		perceivedAngle = hingeJointPerceptor.getAxis();
		speed = hingeJointPerceptor.getSpeed();
		load = hingeJointPerceptor.getLoad();
		voltage = hingeJointPerceptor.getVoltage();
		temperature = hingeJointPerceptor.getTemperature();
		calculatedTemperature = hingeJointPerceptor.getCalculatedTemperature();
		calculatedTemperatureCoil = hingeJointPerceptor
				.getCalculatedTemperatureCoil();
		error = hingeJointPerceptor.getError();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getMaxSpeed()
	 */
	@Override
	public float getMaxSpeed()
	{
		return max_speed;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getSpeed()
	 */
	@Override
	public float getSpeed()
	{
		return speed;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getLoad()
	 */
	@Override
	public float getLoad()
	{
		return load;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getVoltage()
	 */
	@Override
	public float getVoltage()
	{
		return voltage;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getTemperature()
	 */
	@Override
	public float getTemperature()
	{
		return temperature;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getCalculatedTemperature()
	 */
	@Override
	public float getCalculatedTemperature()
	{
		return calculatedTemperature;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getCalculatedTemperatureCoil()
	 */
	@Override
	public float getCalculatedTemperatureCoil()
	{
		return calculatedTemperatureCoil;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getError()
	 */
	@Override
	public byte getError()
	{
		return error;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#setStiffness(float)
	 */
	@Override
	public void setStiffness(float stiffness)
	{
		this.stiffness = ValueUtil.limitValue(stiffness, 0, 1);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getStiffness()
	 */
	@Override
	public float getStiffness()
	{
		return stiffness;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see magma.agent.agentmodel.impl.IMotor#getGain()
	 */
	@Override
	public float getGain()
	{
		return maxGain;
	}
}
