/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.agentmodel.IUniversalJoint;
import magma.agent.meta.agent.IJointToMotorMapper;
import magma.agent.meta.agent.IUniversalJointConfiguration;
import magma.agent.perception.IPerception;
import magma.agent.perception.IUniversalJointPerceptor;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;

/**
 * SimSpark "UniversalJoint" sensor implementation
 * 
 * @author Stefan Glaser, Klaus Dorer
 * 
 */
public class UniversalJoint extends Joint implements IUniversalJoint
{
	private static final long serialVersionUID = 3124679237337000038L;

	/** first hinge joint */
	private HingeJoint joint1;

	/** second hinge joint */
	private HingeJoint joint2;

	private final transient IJointToMotorMapper mapper;

	/**
	 * Instantiates a new UniversalJoint sensor
	 * 
	 * @param jointConfig - the joint meta model
	 * @param effector1
	 * @param effector2
	 */
	public UniversalJoint(IUniversalJointConfiguration jointConfig)
	{
		super(jointConfig.getName(), jointConfig.getPerceptorName());
		this.joint1 = new HingeJoint(jointConfig.getHj1Configuration());
		this.joint2 = new HingeJoint(jointConfig.getHj2Configuration());
		this.mapper = jointConfig.getJointToMotorMapper();
	}

	/**
	 * Copy constructor
	 * @param source to copy from
	 */
	private UniversalJoint(UniversalJoint source)
	{
		super(source);

		this.joint1 = new HingeJoint(source.joint1);
		this.joint2 = new HingeJoint(source.joint2);
		this.mapper = source.mapper;
	}

	@Override
	public float getAxis1()
	{
		return joint1.getAxis();
	}

	@Override
	public float getAxis2()
	{
		return joint2.getAxis();
	}

	@Override
	public float getMinAxis1()
	{
		return joint1.getMin();
	}

	@Override
	public float getMinAxis2()
	{
		return joint2.getMin();
	}

	@Override
	public float getMaxAxis1()
	{
		return joint1.getMax();
	}

	@Override
	public float getMaxAxis2()
	{
		return joint2.getMax();
	}

	@Override
	public float getNextAxis1Speed()
	{
		return joint1.getNextAxisSpeed();
	}

	@Override
	public float getNextAxis2Speed()
	{
		return joint2.getNextAxisSpeed();
	}

	@Override
	public void performAxis1Speed(float speed)
	{
		joint1.performAxisSpeed(speed);
	}

	@Override
	public void performAxis2Speed(float speed)
	{
		joint2.performAxisSpeed(speed);
	}

	@Override
	public void performAxis1Position(float position)
	{
		joint1.performAxisPosition(position);
	}

	@Override
	public void performAxis2Position(float position)
	{
		joint2.performAxisPosition(position);
	}

	@Override
	public void performAxis1Position(float position, float maxSpeed)
	{
		joint1.performAxisPosition(position, maxSpeed);
	}

	@Override
	public void performAxis2Position(float position, float maxSpeed)
	{
		joint2.performAxisPosition(position, maxSpeed);
	}

	@Override
	public void performInitialPosition()
	{
		performAxis1Position(0);
		performAxis2Position(0);
	}

	@Override
	public void resetMovement()
	{
		performAxis1Speed(0);
		performAxis2Speed(0);
	}

	@Override
	public Rotation getRotation()
	{
		return joint2.getRotation().applyTo(joint1.getRotation());
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof UniversalJoint)) {
			return false;
		}
		UniversalJoint other = (UniversalJoint) o;
		if (!super.equals(other)) {
			return false;
		}
		if (!joint1.equals(other.joint1)) {
			return false;
		}
		return joint2.equals(other.joint2);
	}

	@Override
	public ISensor copy()
	{
		return new UniversalJoint(this);
	}

	@Override
	public void generateJointAction(Map<String, float[]> actions)
	{
		double[] jointAngles = { joint1.getAxis(), joint2.getAxis() };
		double[] motorAngles = jointAngles;
		if (mapper != null) {
			motorAngles = mapper.jointToMotorAngle(jointAngles);
		}

		float angle1 = (float) motorAngles[0];
		float speed1 = joint1.getMotor().getNextSpeed(angle1);
		float gain1 = joint1.getMotor().getGain();
		float angle2 = (float) motorAngles[1];
		float speed2 = joint2.getMotor().getNextSpeed(angle2);
		float gain2 = joint2.getMotor().getGain();
		float[] values = { speed1, angle1, gain1, speed2, angle2, gain2 };
		actions.put(getName(), values);
	}

	@Override
	public void updateFromPerception(IPerception perception)
	{
		IUniversalJointPerceptor ujPerceptor = perception
				.getUniversalJointPerceptor(getPerceptorName());

		if (ujPerceptor == null) {
			return;
		}

		// TODO: put motor to joint mapping here

		joint1.updateFromPerception(ujPerceptor.getJoint1());
		joint2.updateFromPerception(ujPerceptor.getJoint2());
	}

	@Override
	public void updateJointPositionFromJoint(IJoint joint)
	{
		joint1.updateJointPositionFromJoint(((UniversalJoint) joint).joint1);
		joint2.updateJointPositionFromJoint(((UniversalJoint) joint).joint2);
	}

	@Override
	public int hashCode()
	{
		return name.hashCode() + joint1.hashCode() * joint2.hashCode();
	}

	@Override
	public List<IJoint> getAllSubJoints()
	{
		List<IJoint> result = new ArrayList<IJoint>();
		result.add(joint1);
		result.add(joint2);
		return result;
	}

	@Override
	public void updateSensors(Map<String, ISensor> flatSensors,
			Map<String, ISensor> structuredSensors)
	{
		flatSensors.put(joint1.getName(), joint1);
		flatSensors.put(joint2.getName(), joint2);
		structuredSensors.put(getName(), this);
	}
}
