/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior;

/**
 * Interface to access all behaviors of the agent
 */
public interface IBehavior
{
	/**
	 * Reinitializes the Behavior-StateMachine (if present)
	 */
	public void init();

	/**
	 * Called to perform the behavior
	 */
	public void perform();

	/**
	 * Retrieve the behavior name
	 * 
	 * @return Behavior name
	 */
	public String getName();

	/**
	 * Check if this behavior is finished performing
	 * 
	 * @return True if finished, false if not
	 */
	public boolean isFinished();

	/**
	 * Called when to stop this behavior without caring for stopping it softly
	 */
	public void abort();

	/**
	 * Decide if it is possible to switch to this behavior from the current
	 * behavior.
	 * 
	 * @param actualBehavior the currently/actually performed behavior
	 * @return This behavior if it is possible to switch to this behavior,
	 *         otherwise the current behavior
	 */
	public IBehavior switchFrom(IBehavior actualBehavior);

	/**
	 * Called to notify this behavior that is no longer performed and replaced by
	 * the new behavior.
	 * 
	 * @param newBehavior the new behavior which will be performed after this
	 *        behavior
	 */
	public void onLeavingBehavior(IBehavior newBehavior);

	/**
	 * Called if we decided again for that behavior
	 */
	public void stayIn();

	/**
	 * @return the intensity with which to perform the behavior.
	 */
	float getIntensity();

	/**
	 * @param intensity the intensity with which to perform the behavior. Values
	 *        are free to use, but are typically in the range of [0..1]
	 */
	void setIntensity(float intensity);

	/**
	 * Retrieve the number of times this behavior has been performed/initialized
	 * 
	 * @return Number of performs
	 */
	int getPerforms();

	/**
	 * This method will only return sensible values for those behaviors that have
	 * implemented a corresponding estimation what successful means.
	 * @return the number of times this behavior has been performed successfully
	 */
	int getSuccessfulPerforms();

	/**
	 * This method will only return sensible values for those behaviors that have
	 * implemented a corresponding estimation what successful means.
	 * @return the number of times this behavior has been performed
	 *         unsuccessfully
	 */
	int getUnsuccessfulPerforms();

	/**
	 * A call does <b><u>not</u></b> have side effects. But a call before
	 * decideNextBasicBehavior will result in the old current behavior in case of
	 * Complex behaviors.
	 * @return the underlying basic behavior that will be performed.
	 */
	IBehavior getRootBehavior();

	/**
	 * @return the number of consecutive performs without interruption
	 */
	int getConsecutivePerforms();

	/**
	 * Reloads this behaviors. Only works for function behaviors.
	 */
	void reload();
}
