/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior;

import magma.util.geometry.Angle;
import magma.util.geometry.IPose2D;

public interface IKick extends IBehavior
{
	/**
	 * Set the intended global kick direction.
	 * 
	 * @param intendedKickDirection - the intended kick direction in the global
	 *        system
	 */
	public void setIntendedKickDirection(Angle intendedKickDirection);

	/**
	 * Retrieve the foot with which this kick kicks.
	 * 
	 * @return the kicking foot
	 */
	public SupportFoot getKickingFoot();

	/**
	 * Retrieve the pose relative to the ball and intended kick direction, to
	 * which we should navigate in order to be able to perform this kick.
	 * 
	 * @return the target pose relative to the ball
	 */
	public IPose2D getRelativeRunToPose();

	/**
	 * Check if the ball is kickable with this kick into the intended kick
	 * direction set via {@link #setIntendedKickDirection(Angle)}.<br>
	 * This method works like a kind of measure how well we are currently able to
	 * perform this kick in a certain direction. A negative value indicates that
	 * the kick will certainly fail and should not be performed. Values grater
	 * than zero represent the deviation from the current to the optimal kick
	 * situation. With old kicks, this deviation is represented by the deviation
	 * from the kick direction. New kicks return a size measure of the last step
	 * that has to be performed before the kick. This way, old and new kicks can
	 * be prioritized by the same system. If some wants to mix old and new kicks,
	 * he has to ensure a proper weighting.
	 * 
	 * @return -1 if the ball is not kickable<br>
	 *         >= 0 the deviation of the optimal situation
	 */
	public float isBallKickable();

	/**
	 * @return the maximum kick distance this kick is able to kick
	 */
	public double getMaxKickDistance();

	/**
	 * @param kickPower the kickPower to set (factor 2 means half long)
	 */
	void setKickPower(float kickPower);
}
