/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.Behavior;
import magma.agent.thoughtmodel.IThoughtModel;

/**
 * Base class for all behaviors that use a couple of other behaviors. <br>
 * A complex behavior is typically used as a decision behavior, instrumenting
 * other behaviors in order to fulfill a more complex and abstract task. For
 * this purpose, a complex behavior supports switching to a list of alternative
 * behaviors in each cycle, which are provided and parameterized by the
 * {@link #decideNextBasicBehaviors()} method.<br>
 * <br>
 * Example: A simple example is the RunToPosition complex behavior, which is
 * instrumenting the underlying walk behavior in order to walk a path to the
 * intended position.
 * 
 * @author Klaus Dorer, Stefan Glaser
 */
public abstract class ComplexBehavior extends Behavior
{
	/** */
	private static final long serialVersionUID = 1L;

	/** the behaviors that are available */
	protected transient final Map<String, IBehavior> behaviors;

	/** the behavior that was last selected to perform */
	private IBehavior currentBehavior;

	/** the behavior to which we initialize */
	private transient final IBehavior defaultBehavior;

	/**
	 * Default constructor for complex behaviors
	 * 
	 * @param name name of behavior
	 * @param thoughtModel Reference to the thought model object
	 * @param behaviors map of all available behaviors
	 * @param defaultBehavior the default behavior to start with
	 */
	public ComplexBehavior(String name, IThoughtModel thoughtModel,
			Map<String, IBehavior> behaviors, IBehavior defaultBehavior)
	{
		super(name, thoughtModel);
		this.behaviors = behaviors;
		this.defaultBehavior = defaultBehavior;
		currentBehavior = defaultBehavior;
	}

	/**
	 * Default constructor for complex behaviors
	 * 
	 * @param name name of behavior
	 * @param thoughtModel Reference to the thought model object
	 * @param behaviors map of all available behaviors
	 */
	public ComplexBehavior(String name, IThoughtModel thoughtModel,
			Map<String, IBehavior> behaviors)
	{
		this(name, thoughtModel, behaviors, behaviors
				.get(IBehaviorConstants.NONE));
	}

	@Override
	public void perform()
	{
		List<IBehavior> possibleBehaviors = decideNextBasicBehaviors();
		IBehavior newBehavior = currentBehavior;
		IBehavior toExecute;

		for (int i = 0; i < possibleBehaviors.size(); i++) {
			toExecute = possibleBehaviors.get(i);

			if (toExecute == currentBehavior) {
				// stop try to switch if we reach a behavior that is in execution
				newBehavior = currentBehavior;
				break;
			} else {
				// we prefer to switch to another behavior, so try to do so
				newBehavior = toExecute.switchFrom(currentBehavior);
				logger.log(Level.INFO, "current: {0} new: {1}", new Object[] {
						newBehavior.getName(), toExecute.getName() });

				if (newBehavior == toExecute) {
					break;
				}
			}
		}

		if (currentBehavior == newBehavior) {
			currentBehavior.stayIn();
		} else {
			currentBehavior = newBehavior;
		}

		currentBehavior.perform();
	}

	/**
	 * Decide for a list of possible next behaviors, sorted from the most to the
	 * least preferred behavior.
	 * 
	 * @return next basic behaviors
	 */
	protected abstract List<IBehavior> decideNextBasicBehaviors();

	@Override
	public IBehavior switchFrom(IBehavior actualBehavior)
	{
		// decide for the next behaviors which would be performed in this complex
		// behavior
		List<IBehavior> possibleBehaviors = decideNextBasicBehaviors();
		IBehavior toExecute;

		for (int i = 0; i < possibleBehaviors.size(); i++) {
			toExecute = possibleBehaviors.get(i);

			// if this complex behavior decides to perform a behavior that is
			// actually in execution, directly switch to this behavior
			if (isBehaviorInExecution(toExecute, actualBehavior)) {
				if (actualBehavior != toExecute) {
					// notify the actual behavior that it is replaced with our
					// decision
					actualBehavior.onLeavingBehavior(toExecute);
				}

				// initialize this behavior
				currentBehavior = toExecute;
				return this;
			}

			// if we decided for a behavior, which is not actually performed
			// already, ask our decision if it is capable to switch from the actual
			// behavior
			IBehavior newB = toExecute.switchFrom(actualBehavior);

			// if the behavior was able to switch, return this behavior, otherwise
			// stick to the actual behavior
			if (newB == toExecute) {
				return this;
			}
		}

		return actualBehavior;
	}

	@Override
	public void onLeavingBehavior(IBehavior newBehavior)
	{
		if (this == newBehavior) {
			// just to make sure to not leave the new behavior
			return;
		}

		// only forward leaving notification to the current behavior if it is not
		// the new behavior
		if (currentBehavior != null && currentBehavior != newBehavior) {
			currentBehavior.onLeavingBehavior(newBehavior);
		}

		super.onLeavingBehavior(newBehavior);
	}

	@Override
	public void abort()
	{
		getCurrentBehavior().abort();
		super.abort();
	}

	/**
	 * @return the currentBehavior
	 */
	public IBehavior getCurrentBehavior()
	{
		return currentBehavior;
	}

	@Override
	public boolean isFinished()
	{
		return currentBehavior.isFinished();
	}

	@Override
	public String toString()
	{
		return super.toString() + " performing " + currentBehavior.toString();
	}

	@Override
	public void init()
	{
		super.init();
		// we have to initialize the current behavior. Otherwise on the next
		// invocation of this complex behavior the old currentBehavior will be
		// asked if it is finished which does not make sense.
		if (behaviors != null) {
			currentBehavior = defaultBehavior;
		}
	}

	/**
	 * @return the root of the currentBehavior
	 */
	@Override
	public IBehavior getRootBehavior()
	{
		return currentBehavior.getRootBehavior();
	}

	/**
	 * Checks if the testee behavior is in execution by the referenceBehavior.
	 * The check is performed recursively along the chain of currently performed
	 * behaviors in the referenceBehavior.
	 * 
	 * @param testee the behavior we are looking for
	 * @param referenceBehavior the reference behavior in which we search for the
	 *        testee
	 * @return true if the testee behavior is currently in execution by the
	 *         referenceBehavior, false otherwise
	 */
	public static boolean isBehaviorInExecution(IBehavior testee,
			IBehavior referenceBehavior)
	{
		IBehavior reference = referenceBehavior;

		// traverse down the behavior chain and search for the testee behavior
		while (reference != null && reference instanceof ComplexBehavior) {
			if (testee == reference) {
				return true;
			}

			// fetch the next child behavior
			reference = ((ComplexBehavior) reference).getCurrentBehavior();
		}

		// if we reached a basic behavior, make the final check
		return reference == testee;
	}

	/**
	 * Retrieve the chain of behavior that are currently performed, starting with
	 * the given referenceBehavior.
	 * 
	 * @param referenceBehavior the top level behavior to expand
	 * @return the list of currently performed behaviors of the referenceBehavior
	 */
	public static List<IBehavior> getCurrentBehaviorchain(
			IBehavior referenceBehavior)
	{
		List<IBehavior> chain = new ArrayList<IBehavior>();
		IBehavior reference = referenceBehavior;

		// traverse down the behavior chain and search for the testee behavior
		while (reference != null && reference instanceof ComplexBehavior) {
			chain.add(reference);

			// fetch the next child behavior
			reference = ((ComplexBehavior) reference).getCurrentBehavior();
		}

		// if we reached a basic behavior, add it and return the chain
		if (reference != null) {
			chain.add(reference);
		}

		return chain;
	}
}
