/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.Map;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Behavior to get in a suitable scoring position, behind the ball, in line with
 * opponent's goal and ball
 * 
 * @author Mathias Ehret
 */
public class GetInScorePosition extends SingleComplexBehavior
{
	/**  */
	private static final long serialVersionUID = 1L;

	/**  */
	static final float BEHIND_BALL_DISTANCE = 0.15f;

	private static final float BALL_KEEP_AWAY_DISTANCE = 0.5f;

	/** the minimal speed for running */
	private float minSpeed;

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param behaviors List of applicable behaviors
	 */
	public GetInScorePosition(IThoughtModel thoughtModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.GET_IN_SCORE_POSITION, thoughtModel, behaviors);
		minSpeed = 15;
	}

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param behaviors List of applicable behaviors
	 */
	public GetInScorePosition(String name, IThoughtModel thoughtModel,
			Map<String, IBehavior> behaviors, float minSpeed)
	{
		super(name, thoughtModel, behaviors);
		this.minSpeed = minSpeed;
	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		RunToPosition run = (RunToPosition) behaviors
				.get(IBehaviorConstants.RUN_TO_POSITION);

		Pose2D pose = getPosition();

		// if (worldModel.getThisPlayer().getDistanceToXYZ(posVector) > 0.4) {
		run.setPosition(pose, minSpeed, 50, 100, false);
		// } else {
		// // when close we run directly to ball
		// run.setPosition(worldModel.getBall().getPosition(), angle);
		// }

		return run;
	}

	Pose2D getPosition()
	{
		IMoveableObject ball = worldModel.getBall();
		Vector3D ballPosition = ball.getPosition();
		IThisPlayer thisPlayer = worldModel.getThisPlayer();

		Vector3D otherGoalPosition = worldModel.getOtherGoalPosition();
		Vector3D shootVector = otherGoalPosition.subtract(ballPosition);
		// FEATURE: uncomment these lines to use the dynamic shoot direction
		// provided by the worldModel
		// Vector3D shootVector = new Vector3D(worldModel
		// .getPreferredKickDirection().radians(), 0);

		Vector3D weBall = ballPosition.subtract(thisPlayer.getPosition());
		Vector3D scorePosition = ballPosition.add(-BEHIND_BALL_DISTANCE,
				shootVector.normalize());

		// we run to some point that is opposite of ball-goal direction
		Vector3D shoot2D = new Vector3D(shootVector.getX(), shootVector.getY(), 0);
		Vector3D meBall2D = new Vector3D(weBall.getX(), weBall.getY(), 0);
		Angle delta1 = Angle.rad(Vector3D.angle(shoot2D, meBall2D));
		double delta = Math.abs(delta1.degrees());

		final double safetyDistance = 1.0;
		Vector3D goal = worldModel.getOwnGoalPosition();
		Vector3D goalLeft = goal.add(new Vector3D(0, -worldModel.goalHalfWith()
				- safetyDistance, 0));
		Vector3D goalRight = goal.add(new Vector3D(0, worldModel.goalHalfWith()
				+ safetyDistance, 0));

		Angle playerBall = thisPlayer.getDirectionTo(ball);
		Angle playerGoalLeft = thisPlayer.getDirectionTo(goalLeft);
		Angle playerGoalRight = thisPlayer.getDirectionTo(goalRight);

		// System.out.print("pgl: " + playerGoalLeft + " - \tpb: " + playerBall
		// + " - \tpgr: " + playerGoalRight);

		if (playerBall.isRightOf(playerGoalLeft)
				&& playerBall.isLeftOf(playerGoalRight)) {
			if (delta > 20) {
				// if (!directlyToBall()) {
				double distanceFromBall = Geometry.getLinearFuzzyValue(20, 80,
						true, delta) * BALL_KEEP_AWAY_DISTANCE + BEHIND_BALL_DISTANCE;

				scorePosition = ballPosition.add(-distanceFromBall,
						shootVector.normalize());
				// }
			}
		} else {
			shootVector = ballPosition.subtract(thisPlayer.getPosition())
					.normalize();
			// System.out.print(" - \tpos: " + scorePosition);
			// System.out.print(" - \talpha: "
			// + Angle.rad(shootVector.getAlpha()).degrees());
			scorePosition = ballPosition.add(-BEHIND_BALL_DISTANCE, shootVector);
		}
		// System.out.println();
		return new Pose2D(scorePosition.getX(), scorePosition.getY(),
				Angle.rad(shootVector.getAlpha()));
	}
}
