/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.IWalk;
import magma.agent.behavior.complex.WalkEstimator.WalkMode;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IPositionManager;
import magma.agent.worldmodel.IThisPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Implements a behavior which is able to walk/run to a specified target
 * position
 */
public class RunToPosition extends SingleComplexBehavior
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** the minimum speed with which to walk forward */
	private double minSpeedForward;

	/** the minimum speed with which to walk forward */
	private double maxSpeedForward;

	/** the minimum speed with which to walk sideward */
	private double minSpeedSideward;

	private transient WalkEstimator walkEstimator;

	/** how much we deviate from destination facing before just turning (deg) */
	// private double angularPrecision;

	/** distance at which we consider angular precision (m) */
	// private double fastestToRunDistance;

	/** distance at which we start turning to desired direction (m) */
	private double turnToDesiredDirdistance;

	/** true if movement to home position */
	private boolean passiveMovement;

	/** class which calculates optimal path */
	// private OptimalPathCalculation optPathCalc;

	/**
	 * Constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param behaviors List of applicable behaviors
	 */
	public RunToPosition(IThoughtModel thoughtModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.RUN_TO_POSITION, thoughtModel, behaviors);
		minSpeedForward = 0;
		maxSpeedForward = 100;
		minSpeedSideward = 0;
		walkEstimator = new WalkEstimator();
		// angularPrecision = 60;
		// fastestToRunDistance = 1.5;
		turnToDesiredDirdistance = 0.2;
		// optPathCalc = new OptimalPathCalculation(OptPathVisualizer.VIS_ON);
		// String parameter = System.getenv().get("VARIATION_PARAMETER");
		// if (parameter != null) {
		// fastestToRunDistance = Double.parseDouble(parameter);
		// System.out.println("parameter: " + fastestToRunDistance);
		// }
	}

	/**
	 * 
	 */
	@Override
	public IBehavior decideNextBasicBehavior()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IPositionManager poseManager = thisPlayer.getPositionManager();
		Pose2D targetPose = poseManager.getFinalPosition();
		Vector3D finalPosition = targetPose.getPosition();
		Angle absDirAtPosition = targetPose.getAngle();

		Pose2D nextPose = avoidCollision(targetPose);

		// clear desired list from any previous collisions avoidances
		poseManager.clear();
		poseManager.addDesiredPosition(0, targetPose);
		double distanceToFinal = thisPlayer.getDistanceToXY(finalPosition);
		double distanceToNextPose = distanceToFinal;
		if (nextPose != targetPose) {
			// player is avoiding an obstacle so add new position to the list
			poseManager.addDesiredPosition(0, nextPose);
			distanceToNextPose = thisPlayer
					.getDistanceToXY(nextPose.getPosition());
		}

		// dir is the relative direction in which to face
		// when far we want to face to the destination position
		Vector3D nextPosition = nextPose.getPosition();
		Angle nextPositionAngle = thisPlayer.getBodyDirectionTo(nextPosition);
		// if (distanceToFinal < TURN_TO_DESIRED_DIR_DISTANCE) {
		// // when close to destination we want to face into desired direction
		// if (desiredPos == targetPositionOrientation) {
		// // but only if not avoiding obstacles
		// dirToGo = relDirAtPosition;
		// }
		// }

		double nextAngleDegree = nextPositionAngle.degrees();

		WalkMode fastestWalkMode = walkEstimator.getFastestWalkMode(thisPlayer,
				nextPose);

		// System.out.println("WalkMode: " + fastestWalkMode);

		switch (fastestWalkMode) {
		case BACKWARD:
			if (nextAngleDegree < 0) {
				nextAngleDegree = 180 + nextAngleDegree;
			} else {
				nextAngleDegree = nextAngleDegree - 180;
			}
			break;
		case LEFTSIDE:
			nextAngleDegree = nextAngleDegree - 90;
			break;
		case RIGHTSIDE:
			nextAngleDegree = 90 + nextAngleDegree;
			break;
		case DIAGONALLEFT:
			nextAngleDegree = nextAngleDegree - 45;
			break;
		case DIAGONALRIGHT:
			nextAngleDegree = 45 + nextAngleDegree;
			break;
		default:
			// forward
			break;
		}

		// if (Math.abs(nextAngleDegree) > angularPrecision
		// && distanceToNextPose > fastestToRunDistance) {
		// // in case of big distances and high rotation we just turn
		// walk.walk(0, 0, nextAngleDegree);
		// return walk;
		// }

		if (distanceToFinal < turnToDesiredDirdistance) {
			// when close to destination we want to face into desired direction
			// but only if not avoiding obstacles
			if (nextPose == targetPose) {
				Angle relDirAtPosition = absDirAtPosition.subtract(thisPlayer
						.getHorizontalAngle());
				nextAngleDegree = relDirAtPosition.degrees();
			}
		}

		double speedXAtGoal = 0.05;
		Vector3D speedThere = new Vector3D(speedXAtGoal, 0, 0);
		if (nextPose != targetPose) {
			if (distanceToNextPose <= 0.8) {
				// if we run to an intermediate position and if we are almost
				// there we do not turn
				nextAngleDegree = 0;
				// System.out.println("no turn: " + distanceToDesired);
			}
			// double speedX = Geometry.getLinearFuzzyValue(0.2, 0.5, true,
			// distanceToFinal) * (1 - speedXAtGoal) + speedXAtGoal;
			double speedX = 1.0;
			if (distanceToFinal < 1.0) {
				// System.out.println("dist: " + distanceToFinal);
				speedX = 0.3;
			}
			speedThere = new Vector3D(speedX, 0, 0);
		}

		// we rotate the position by the rotation we want to perform to avoid
		// that a sidewise movement is done where turning would have sufficed
		Vector3D myPosition = thisPlayer.getPosition();
		Vector3D virtualPosition = nextPosition;

		// Vector3D virtualPosition = position.subtract(myPosition);
		// virtualPosition = Geometry.createZRotation(
		// (float) Math.toRadians(-dirDegree)).applyTo(virtualPosition);
		// virtualPosition = virtualPosition.add(myPosition);

		// desired position in forward or backward direction
		IWalk walk = (IWalk) behaviors.get(IBehaviorConstants.WALK);
		Pose2D walkTo = new Pose2D(virtualPosition.subtract(myPosition),
				Angle.deg(nextAngleDegree));

		walk.globalWalk(walkTo, speedThere, minSpeedForward, minSpeedSideward,
				maxSpeedForward, !passiveMovement);

		return walk;

		// Stefan Grossmann Version with optimal path calculation
		// speed is position changes of the last 3 cycles -> 1 cycle(20ms) * 50 =
		// distance in m/s
		// Pose2D walkTo = new Pose2D(virtualPosition/*.subtract(myPosition)*/,
		// nextPose.angle);
		// double startSpeed =
		// Math.sqrt(Math.pow(thisPlayer.getSpeed().getX(),2)+Math.pow(thisPlayer.getSpeed().getY(),2))/4*50;
		// boolean newSit = optPathCalc.newSituation(new
		// Pose2D(thisPlayer.getPosition(), thisPlayer.getHorizontalAngle()),
		// startSpeed, walkTo, speedThere, null);
		// IBehavior myWalk;
		// myWalk = optPathCalc.walkIt(behaviors);
		//
		// return myWalk;
	}

	/**
	 * Calculates if obstacles are in our way
	 * @param goalPos the destination position to run to
	 * @return old position (goalPos) if there is nothing in way, a new position
	 *         to avoid collisions otherwise
	 */
	Pose2D avoidCollision(Pose2D goalPos)
	{
		Pose2D result = goalPos;
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		double goalPosDistance = thisPlayer
				.getDistanceToXY(goalPos.getPosition());

		List<IVisibleObject> obstacles = thoughtModel.getObstacles();
		for (IVisibleObject obstacle : obstacles) {

			// check if the obstacle intersects our trajectory
			Vector3D obstaclePos = obstacle.getPosition();
			double distanceToMe = thisPlayer.getDistanceToXY(obstaclePos);
			if (distanceToMe > goalPosDistance) {
				// farer than goal position
				continue;
			}

			// calculate distance to line me -> goalPos
			Vector3D mePos2D = new Vector3D(thisPlayer.getPosition().getX(),
					thisPlayer.getPosition().getY(), 0);
			Vector3D goalPos2D = new Vector3D(goalPos.getPosition().getX(),
					goalPos.getPosition().getY(), 0);
			Vector3D meGoalPos2D = goalPos2D.subtract(mePos2D);

			Vector3D obstaclePos2D = new Vector3D(obstaclePos.getX(),
					obstaclePos.getY(), 0);
			Vector3D meObstaclePos2D = obstaclePos2D.subtract(mePos2D);

			// check if destination position and obstacle position lie in opposite
			// direction
			Angle angle = Angle.rad(Vector3D.angle(meGoalPos2D, meObstaclePos2D));
			logger.log(Level.FINEST, "avoidCollision angle: {0}",
					new Object[] { angle });

			if (Math.abs(angle.degrees()) > 90) {
				// obstacle is not in goal direction but opposite direction
				continue;
			}

			// get a normalized orthogonal (in 2 D space)
			if (meGoalPos2D.getNorm() < 0.01) {
				// avoid an exception on normalization
				continue;
			}
			meGoalPos2D = meGoalPos2D.normalize();
			Vector3D norm = new Vector3D(-meGoalPos2D.getY(), meGoalPos2D.getX(),
					0);
			double distanceToLine = Vector3D.dotProduct(meObstaclePos2D, norm);
			distanceToLine = Math.abs(distanceToLine);

			double criticalDistance = obstacle.getCollisionDistance();
			if (obstacle == worldModel.getBall()) {
				// if we move to home position, we try to avoid to getting too close
				// to the ball.
				if (passiveMovement) {
					criticalDistance = 1.2;
				}
			}

			double delta = criticalDistance - distanceToLine;
			if (delta <= 0) {
				continue;
			}

			// we have to run around the obstacle
			// calculate base for intermediate positions
			// double factor = distanceToMe / (3.0 * criticalDistance);
			double factor = Geometry.getLinearFuzzyValue(0.0, 0.2, true,
					distanceToMe - delta);
			// if (factor > 1.0) {
			// factor = 1.0;
			// }
			Vector3D base = mePos2D.add(factor, meObstaclePos2D);

			Vector3D intermediate;
			// criticalDistance = (Geometry.getLinearFuzzyValue(0, 60, false,
			// Math.abs(angle.degrees())) + 0.2)
			// * criticalDistance;
			Vector3D intermediate1 = base.add(criticalDistance, norm);
			Vector3D intermediate2 = base.subtract(criticalDistance, norm);

			// decide for the shortest path
			double distance1 = thisPlayer.getDistanceToXY(intermediate1);
			double distance2 = thisPlayer.getDistanceToXY(intermediate2);
			if (distance1 < distance2) {
				intermediate = intermediate1;
			} else {
				intermediate = intermediate2;
			}

			// calculate desired angle at intermediate point
			double desiredAngle = goalPos2D.subtract(intermediate).getAlpha();
			result = new Pose2D(intermediate, Angle.rad(desiredAngle));

			logger.log(Level.FINE,
					"avoiding obstacle {0} at (x/y): {1}/{2} going to {3}",
					new Object[] { obstacle.getName(), obstaclePos2D.getX(),
							obstaclePos2D.getY(), result });

			// System.out.println("obstacle: " + obstacle.getName()
			// + " want to go to " + goalPos);

			// TODO: do not stop in this case but calculate a path taking all
			// players into account using AStar search
			break;
		}
		return result;
	}

	/**
	 * Set new target position
	 * 
	 * @param newPosition New target Position, null, if current position should
	 *        be used
	 * @param rotation Target rotation (radians)
	 * @param minSpeedForward the minimum speed for forward direction
	 * @param minSpeedSideward the minimum speed for sideward direction
	 * @param passiveMovement true if movement to home position
	 */
	public void setPosition(Pose2D newPose, double minSpeedForward,
			double minSpeedSideward, double maxSpeedForward,
			boolean passiveMovement)
	{
		this.minSpeedForward = minSpeedForward;
		this.minSpeedSideward = minSpeedSideward;
		this.maxSpeedForward = maxSpeedForward;
		this.passiveMovement = passiveMovement;
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		if (newPose == null) {
			newPose = new Pose2D(thisPlayer.getPosition(),
					thisPlayer.getHorizontalAngle());
		} else {
			newPose = new Pose2D(avoidTeammate(newPose.getPosition()),
					newPose.getAngle());
		}
		IPositionManager posManager = thisPlayer.getPositionManager();
		posManager.setDesiredPosition(newPose, false);
		Vector3D current = thisPlayer.getPosition();

		logger.log(Level.FINE,
				"setDesiredPosition (x/y/angle): {0}/{1}/{2} current: {3}/{4}/{5}",
				new Object[] { newPose.getX(), newPose.getY(),
						newPose.getAngle().degrees(), current.getX(), current.getY(),
						thisPlayer.getHorizontalAngle() });
	}

	/**
	 * We calculate whether teammates stop us from moving to a position
	 * @param newPosition
	 * @return
	 */
	Vector3D avoidTeammate(Vector3D newPosition)
	{
		List<IPlayer> allTeammates = thoughtModel.getTeammatesAtBall();
		Vector3D ownPosition = worldModel.getThisPlayer().getPosition();
		Vector3D ownPosNewPos = newPosition.subtract(ownPosition);
		double norm = ownPosNewPos.getNorm();

		for (IPlayer player : allTeammates) {
			if (player.getDistanceToXY(newPosition) < 0.5) {
				logger.log(Level.INFO,
						"avoidTeammate norm: {0} newPosition: {1},{2}", new Object[] {
								norm, newPosition.getX(), newPosition.getY() });
				// If we are near enough already we can just stay in place
				if (norm < 1) {
					return ownPosition;
				}

				return ownPosNewPos.normalize().scalarMultiply(norm - 1);
			}
		}

		// If no teammates are near to the new position just go there
		return newPosition;
	}
}
