/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package magma.agent.behavior.complex;

import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Calculates an estimation for how long it is to walk somewhere
 * @author kdorer
 */
public class WalkEstimator
{
	public enum WalkMode {
		FORWARD, BACKWARD, LEFTSIDE, RIGHTSIDE, DIAGONALLEFT, DIAGONALRIGHT
	};

	// the speed we can achieve running straight forward (m/s)
	private static final float FORWARD_SPEED = 0.85f;

	// the speed we can achieve running straight backward (m/s)
	private static final float BACKWARD_SPEED = 0.60f;

	// the speed we can achieve running forward and sideward (m/s)
	private static final float DIAGONAL_SPEED = 0.65f;

	// the speed we can achieve running sideward (m/s)
	private static final float SIDEWARD_SPEED = 0.45f;

	// the speed we can achieve turning (deg/s)
	private static final float TURNING_SPEED = 180f;

	// the speed lookup indexed by WALKMODE
	private static final float speedDefault[] = { FORWARD_SPEED, BACKWARD_SPEED,
			SIDEWARD_SPEED, SIDEWARD_SPEED, DIAGONAL_SPEED, DIAGONAL_SPEED,
			TURNING_SPEED };

	private final float speed[];

	private final float turnSpeed;

	// the relative body direction lookup in degrees indexed by WALKMODE
	private final float walkDir[] = { 0, 180, 90, -90, 45, -45 };

	public WalkEstimator()
	{
		speed = speedDefault;
		turnSpeed = speed[6];
	}

	public WalkEstimator(float speed[])
	{
		this.speed = speed;
		turnSpeed = speed[6];
	}

	public WalkMode getFastestWalkMode(IThisPlayer thisPlayer, Pose2D pose)
	{
		float bestTime = Float.POSITIVE_INFINITY;
		WalkMode bestMode = null;
		for (WalkMode mode : WalkMode.values()) {

			// if (mode == WalkMode.DIAGONALLEFT || mode == WalkMode.DIAGONALRIGHT)
			// {
			// // switch off diagonal
			// continue;
			// }

			float time = estimateWalkTime(thisPlayer, pose, mode);
			if (time < bestTime) {
				bestTime = time;
				bestMode = mode;
			}
		}
		// System.out.println(bestMode + ": " + bestTime);

		return bestMode;
	}

	/**
	 * Estimates the time (in s) it takes to get to the specifed position turned
	 * into the specified direction.
	 * @param worldModel
	 * @param position the position to get to
	 * @param absFacingDirectionAtPosition the direction to turn to at position
	 * @param mode how to get there
	 * @return the time (in s)
	 */
	public float estimateWalkTime(IThisPlayer thisPlayer, Pose2D pose,
			WalkMode mode)
	{
		float walkTime = 0;

		Vector3D position = pose.getPosition();
		// time for turning into desired walk mode direction
		Angle dirToGo = thisPlayer.getBodyDirectionTo(position);
		dirToGo = dirToGo.subtract(Angle.deg(walkDir[mode.ordinal()]));
		walkTime += Math.abs(dirToGo.degrees()) / turnSpeed;

		// time for getting there
		double distanceToFinal = thisPlayer.getDistanceToXY(position);
		walkTime += distanceToFinal / speed[mode.ordinal()];

		// time for turning into final direction
		Angle absDirToPos = thisPlayer.getDirectionTo(position);
		Angle delta = absDirToPos.subtract(pose.getAngle());
		delta = delta.subtract(Angle.deg(walkDir[mode.ordinal()]));
		walkTime += Math.abs(delta.degrees()) / turnSpeed;

		return walkTime;
	}
}
