/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.motor;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.util.geometry.Geometry;
import magma.util.logging.CSVFormatter;
import magma.util.parameter.IParameterList;
import magma.util.parameter.ParameterMap;

/**
 * Implements a static motor movement behavior, movement data is read from a
 * file
 * 
 * @author Klaus Dorer
 */
public class AdjustableMotorFileBehavior extends MotorBehavior
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** values of each joint angles (in degrees) for each step */
	private transient ArrayList<HashMap<String, Float>> steps;

	private boolean startModus = true;

	private float currentStep;

	private float stepSizes[];

	private float stepSizeStart;

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param name Behavior name
	 * @param thoughtModel Reference to the thought model object
	 * @param params Behavior parameters
	 * @param filepath Path to movement data file
	 * @param slowDown Slow down factor
	 * @param startSlowDown Startup slowdown factor
	 * @throws IOException If the movement data file could not be successfully
	 *         opened and parsed
	 */
	public AdjustableMotorFileBehavior(String name, IThoughtModel thoughtModel,
			ParameterMap params, String filepath, int slowDown, int startSlowDown)
			throws IOException
	{
		super(name, thoughtModel);

		ReadCSV valuesFromFile = new ReadCSV(filepath);
		this.steps = valuesFromFile.getFormattedCSV();

		this.currentStep = 0.0f;
		this.stepSizeStart = 1.0f / startSlowDown;
		this.finished = false;
		if (LOG) {
			initLogger(Level.INFO);
		}
		IParameterList param = params.get(name);
		if (param != null) {
			stepSizes = param.getParameterValues();
		} else {
			float[] speeds = new float[steps.size()];
			Arrays.fill(speeds, 1.0f / slowDown);
			stepSizes = speeds;
		}
	}

	@Override
	public void init()
	{
		super.init();
		currentStep = 0.0f;
		startModus = true;
	}

	@Override
	public void perform()
	{
		logger.log(Level.FINEST, "{0}: {1}", new Object[] { getName(),
				currentStep });
		finished = false;

		IBodyModel body = agentModel.getFutureBodyModel();
		if (startModus) {
			performSlowDown(body);

			currentStep += stepSizeStart;
			if (currentStep >= 1.0f) {
				startModus = false;
				currentStep = 1.0f;
			}
		} else {
			performSlowDown(body);

			currentStep += stepSizes[(int) currentStep];
			if (currentStep >= steps.size()) {
				finished = true;
				currentStep = 0.0f;
			}
		}
		pastPerform(body, (int) currentStep);
	}

	/**
	 * Called after the motor file was translated into joint speeds. The default
	 * implementation does nothing
	 * @param body the body model on which this was done
	 * @param realStep the step within the behavior file
	 */
	protected void pastPerform(IBodyModel body, int realStep)
	{
	}

	private void performSlowDown(IBodyModel body)
	{
		IHingeJoint hingeJoint;
		float currentAngle;
		float delta;
		float speed;

		StringBuilder buffer = new StringBuilder();
		Map<String, Float> desiredAngles1 = steps.get((int) currentStep);
		int index2 = (int) currentStep + 1;
		if (index2 >= steps.size()) {
			index2 = (int) currentStep;
		}
		Map<String, Float> desiredAngles2 = steps.get(index2);

		// loop through all joints affected
		for (String angleName : desiredAngles1.keySet()) {

			// calculate speed
			float desiredAngle1 = desiredAngles1.get(angleName).floatValue();
			float desiredAngle2 = desiredAngles2.get(angleName).floatValue();
			hingeJoint = (IHingeJoint) body.getJoint(angleName);

			float desiredAngle = (float) Geometry.linearInterpolation(
					(int) currentStep, desiredAngle1, (int) currentStep + 1,
					desiredAngle2, currentStep);

			if (hingeJoint == null) {
				logger.log(Level.SEVERE, "unknown hinge joint: {0}", angleName);
			} else {
				currentAngle = hingeJoint.getAxis();
				delta = (desiredAngle - currentAngle);
				// use this line if calculating on sensed model
				// speed = delta / 2.0f;
				// use this line if calculating on expected model
				speed = delta;

				// speed /= speedFactor;

				if (speed > -0.02 && speed < 0.02) {
					speed = 0.0f;
				}

				if (speed > hingeJoint.getMaxSpeed()) {
					speed = hingeJoint.getMaxSpeed();
				}

				if (speed < -hingeJoint.getMaxSpeed()) {
					speed = -hingeJoint.getMaxSpeed();
				}

				if (LOG) {
					buffer.append(desiredAngle).append(";");
					buffer.append(currentAngle).append(";");
					buffer.append(speed).append(";");
				}
				// set axis speed
				hingeJoint.performAxisSpeed(speed);
			}
		}
		if (LOG) {
			Logger.getLogger(getName()).log(Level.INFO, buffer.toString());
		}
	}

	/**
	 * Initializes logging functionality
	 * 
	 * @param level Logging level
	 */
	private void initLogger(Level level)
	{
		try {
			Logger loggerCSV = Logger.getLogger(getName());
			FileHandler handler = new FileHandler("log/" + getName() + ".csv");
			handler.setFormatter(new CSVFormatter());
			loggerCSV.addHandler(handler);
			loggerCSV.setLevel(level);

			// write heading
			StringBuilder buffer = new StringBuilder();
			Map<String, Float> namesMap = steps.get(0);

			for (String stepName : namesMap.keySet()) {
				buffer.append(stepName).append(";");
				buffer.append("angleAsIs").append(";");
				buffer.append("speed").append(";");
			}

			loggerCSV.log(Level.INFO, buffer.toString());
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
