/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.motor;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import magma.util.function.FunctionLoader;
import magma.util.function.IFunction;
import magma.util.parameter.Parameter;
import magma.util.parameter.ParameterList;

/**
 * @author Klaus Dorer, Stefan Glaser
 */
public class FunctionBehaviorParameter extends ParameterList implements
		Serializable
{
	private static final String USED_BEHAVIOR_PATH = "config/behaviors/nao/FunctionBehaviors/UsedBehaviors/";

	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** map containing an entry for all joints involved in the behavior */
	private final Map<String, IFunction> joints;

	/** the version of the file format we use */
	private int version;

	/** the name of the corresponding behavior */
	private String name;

	/** the number of cycles the behavior takes */
	private float period;

	public FunctionBehaviorParameter(Parameter[] params, String name,
			int version, float duration, Map<String, IFunction> jointFunctions)
	{
		super(params);

		this.name = name;
		this.version = version;
		this.period = duration;
		this.joints = jointFunctions;
	}

	/**
	 * Writes a function motor file
	 * @param filename the name to use as filename
	 */
	@Override
	public void writeToFile(String filename)
	{
		try {
			writeBehaviorFile(new File(filename), version, name, period, joints);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void writeToFile()
	{
		writeToFile(USED_BEHAVIOR_PATH + name + ".txt");
	}

	/**
	 * @return a map of functions for each joint that is involved in the behavior
	 */
	public Map<String, IFunction> getJoints()
	{
		return joints;
	}

	/**
	 * @return the file version
	 */
	public int getVersion()
	{
		return version;
	}

	/**
	 * @return the behavior name
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @return the period
	 */
	public float getPeriod()
	{
		return period;
	}

	// /**
	// * Converts the passed function parameters into a chromosom for learning
	// *
	// * @param param The function parameters to convert
	// * @return A chromosom suitable for genetic learning
	// */
	// public float[] toChromosom()
	// {
	// int count = 0;
	// Map<String, IFunction> joints = getJoints();
	// String[] keys = getSplineJointNamesToLearn();
	// // Hack: magic 6, fix by using a list taking real number spline
	// function
	// // params and converting it after the loop to an array
	// int size = keys.length * 12;
	// float[] result = new float[size];
	// for (String key : keys) {
	// IFunction currentFunction = joints.get(key);
	// // we only learn the spline functions
	// SupportPointFunction func = (SupportPointFunction) currentFunction;
	// double[] xValues = func.getXValues();
	// double[] yValues = func.getYValues();
	// // we do not learn 1st and last spline values
	// for (int i = 1; i < xValues.length - 1; i++) {
	// result[count++] = (float) xValues[i];
	// result[count++] = (float) yValues[i];
	// }
	// }
	//
	// return result;
	//
	// return null;
	// }
	//
	// /**
	// * Converts the passed chromosom parameters into a function parameter for
	// * execution
	// *
	// * @param chromosom the chromosom to convert
	// * @return a function parameter that can be used for behavior creation
	// */
	// public void fromChromosom(float[] chromosom)
	// {
	// Map<String, IFunction> joints = getJoints();
	//
	// int count = 0;
	// String[] keys = getSplineJointNamesToLearn();
	// for (String key : keys) {
	// IFunction currentFunction = joints.get(key);
	// SupportPointFunction func = (SupportPointFunction) currentFunction;
	// double[] xValues = func.getXValues();
	// double[] yValues = func.getYValues();
	// // we do not learn 1st and last spline values
	// for (int i = 1; i < xValues.length - 1; i++) {
	// if (learnXValues) {
	// xValues[i] = chromosom[count++];
	// } else {
	// count++;
	// }
	// yValues[i] = chromosom[count++];
	// }
	// func.initialize();
	// }
	// }
	//
	// /**
	// * @return the names of the joints using splines that are learned
	// */
	// protected String[] getSplineJointNamesToLearn()
	// {
	// Parameter[] params = getParameters();
	// String[] result = new String[params.length];
	// int i = 0;
	// for (Parameter param : params) {
	// result[i++] = param.getName();
	// }
	// return result;
	// }

	/**
	 * Create a new {@link FunctionBehaviorParameter} instance from the given
	 * behavior file.
	 * 
	 * @param file - the file to read
	 * @return a new {@link FunctionBehaviorParameter} instance
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public static FunctionBehaviorParameter readBehaviorFile(Parameter[] params,
			String fileName) throws FileNotFoundException, IOException
	{
		ClassLoader cl = FunctionBehaviorParameter.class.getClassLoader();
		InputStream is = cl.getResourceAsStream(fileName);
		BufferedReader in2 = new BufferedReader(new InputStreamReader(is));

		return readBehaviorFile(params, in2);
	}

	/**
	 * Create a new {@link FunctionBehaviorParameter} instance from the given
	 * behavior file <b>without throwing Exceptions</b>.
	 * 
	 * @param file - the file to read
	 * @return a new {@link FunctionBehaviorParameter} instance
	 */
	public static FunctionBehaviorParameter readBehaviorFileNoEx(
			Parameter[] params, File file)
	{
		try {
			return readBehaviorFile(params, new BufferedReader(
					new FileReader(file)));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * Create a new {@link FunctionBehaviorParameter} instance from the given
	 * behavior file.
	 * 
	 * @param file - the file to read
	 * @return a new {@link FunctionBehaviorParameter} instance
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public static FunctionBehaviorParameter readBehaviorFile(Parameter[] params,
			File file) throws FileNotFoundException, IOException
	{
		return readBehaviorFile(params, new BufferedReader(new FileReader(file)));
	}

	/**
	 * Create a new {@link FunctionBehaviorParameter} instance from the given
	 * input reader.
	 * 
	 * @param in - an buffered reader instance
	 * @return a new {@link FunctionBehaviorParameter} instance
	 * @throws IOException
	 */
	public static FunctionBehaviorParameter readBehaviorFile(Parameter[] params,
			BufferedReader in) throws IOException
	{
		int version;
		String name;
		float duation;
		Map<String, IFunction> jointFunctions = new HashMap<String, IFunction>();

		try {
			String line;
			String[] csvLineContents;

			line = in.readLine();
			csvLineContents = line.split(",");

			// read the date from the first line
			name = csvLineContents[0];
			version = Integer.valueOf(csvLineContents[1]);
			duation = Float.valueOf(csvLineContents[2]);

			// read all data for each joint
			while ((line = in.readLine()) != null) {
				int sepIndex = line.indexOf(",");
				String jointName = line.substring(0, sepIndex);
				String functionString = line.substring(sepIndex + 1, line.length());
				IFunction jointFunction = FunctionLoader.INSTANCE.fromCSVString(
						functionString, version);

				jointFunctions.put(jointName, jointFunction);
			}
		} catch (IOException e) {
			in.close();
			throw e;
		}

		in.close();

		return new FunctionBehaviorParameter(params, name, version, duation,
				jointFunctions);
	}

	/**
	 * Write a function-behavior-file with the given
	 * {@link FunctionBehaviorParameter}.
	 * 
	 * @param file - the file to write to
	 * @param parameter - the parameter which should be written to the file
	 * @throws IOException
	 */
	public static void writeBehaviorFile(File file,
			FunctionBehaviorParameter parameter) throws IOException
	{
		writeBehaviorFile(file, parameter.version, parameter.name,
				parameter.period, parameter.joints);
	}

	/**
	 * Write a function-behavior-file with the given information.
	 * 
	 * @param file - the file to write to
	 * @param version - the format version
	 * @param behaviorName - the name of the behavior
	 * @param duration - the duration of the behavior
	 * @param jointFunctions - the joint functions of this behaviors
	 * @throws IOException
	 */
	public static void writeBehaviorFile(File file, int version,
			String behaviorName, float duration,
			Map<String, IFunction> jointFunctions) throws IOException
	{
		PrintWriter out = new PrintWriter(new FileWriter(file));
		out.println(behaviorName + "," + version + "," + duration + ",1,1");

		// write all data for each joint
		IFunction jointFunction;
		for (String key : jointFunctions.keySet()) {
			jointFunction = jointFunctions.get(key);
			out.println(key + ","
					+ FunctionLoader.INSTANCE.toCSVString(jointFunction, version));
		}

		out.close();
	}
}
