/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.motor;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Reads a CSV file of the format: (BOF) Time,Pose,Effector1,Effector2
 * TimeValue,PoseId1,Effector1_Value,Effector2_Value
 * TimeValue,PoseId2,Effector1_Value,Effector2_Value ... (EOF)
 * 
 * @author Srinivasa Ragavan
 */
public class ReadCSV
{
	/**
	 * The formattedCSV holds the whole CSV file in the format: ArrayList of each
	 * line in CSV. Each line is held as a HashMap of Effectors as 'key' and
	 * Target_Angle(in Radians) as 'value'. formattedCSV.size() will give the
	 * number of steps for the motion.
	 */
	private ArrayList<HashMap<String, Float>> formattedCSV = null;

	private final ArrayList<String[]> csvContents = new ArrayList<String[]>();

	/** Number of records in each line of CSV */
	private int noOfRecords = 0;

	private String filePath = null;

	private String[] csvHeader = null;

	/**
	 * Constructor
	 * 
	 * @param pathOfFile Path to the CSV file
	 */
	public ReadCSV(String pathOfFile)
	{
		super();
		this.filePath = pathOfFile;
	}

	/**
	 * @return ArrayList<HashMap<String, String>> formattedCSV
	 * @throws IOException If the file could not be successfully opened and
	 *         parsed
	 */
	public ArrayList<HashMap<String, Float>> getFormattedCSV()
			throws IOException
	{
		readCsvFile();
		formatCsv();

		return formattedCSV;
	}

	/**
	 * Reads the CSV file and populates the contents in an ArrayList of Strings.
	 * @throws IOException If the file could not be successfully opened and
	 *         parsed
	 */
	private void readCsvFile() throws IOException
	{
		ClassLoader cl = this.getClass().getClassLoader();
		InputStream is = cl.getResourceAsStream(filePath);

		if (is == null) {
			Logger.getLogger("").log(Level.SEVERE, "File can not be found: {0}",
					filePath);
			return;
		}

		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		String line;

		while ((line = br.readLine()) != null) {
			String[] csvLineContents = line.split(",");
			noOfRecords = csvLineContents.length;
			csvContents.add(csvLineContents);
		}

		br.close();
	}

	/**
	 * Formats the CSV and populates formattedCSV
	 * 
	 * @see magma.agent.behavior.motor.ReadCSV#getFormattedCSV()
	 */
	private void formatCsv()
	{
		String effectorName;
		String effectorValue;
		HashMap<String, Float> poseNameAndValue;
		formattedCSV = new ArrayList<HashMap<String, Float>>();

		for (int i = 1; i < csvContents.size(); i++) {
			poseNameAndValue = new HashMap<String, Float>();

			for (int arrayLen = 2; arrayLen < noOfRecords; arrayLen++) {
				effectorName = csvContents.get(0)[arrayLen];

				effectorValue = csvContents.get(i)[arrayLen];
				double valueD = Double.parseDouble(effectorValue);

				// TODO: this is robot specific!!
				if ("sweaty".equals(csvContents.get(0)[1])) {
					mapToSweaty(poseNameAndValue, effectorName, valueD);
				} else {
					mapToNao(poseNameAndValue, effectorName, valueD);
				}
			}

			formattedCSV.add(poseNameAndValue);
		}
	}

	/**
	 * @param poseNameAndValue
	 * @param effectorName
	 * @param valueD
	 * @return
	 */
	private void mapToNao(HashMap<String, Float> poseNameAndValue,
			String effectorName, double valueD)
	{
		effectorName = effectorName.replace("Ankle", "Foot");
		effectorName = effectorName.replace("Head", "Neck");
		effectorName = effectorName.replace("Elbow", "Arm");
		effectorName = effectorName.replace("ShoulderRoll", "ShoulderYaw");

		valueD = -Math.toDegrees(valueD);

		// Negate some joints
		if (effectorName.contains("Roll") || effectorName.contains("ShoulderYaw")
				|| effectorName.contains("YawPitch")) {
			valueD = -valueD;
		}

		Float value = new Float(valueD);
		poseNameAndValue.put(effectorName, value);
	}

	/**
	 * @param poseNameAndValue
	 * @param effectorName
	 * @param valueD
	 * @return
	 */
	private void mapToSweaty(HashMap<String, Float> poseNameAndValue,
			String effectorName, double valueD)
	{
		valueD = Math.toDegrees(valueD);
		Float value = new Float(valueD);
		poseNameAndValue.put(effectorName, value);
	}

	/**
	 * Reads the Header of the CSV File into an ArrayList of Strings.
	 * @throws IOException If the file could not be successfully opened and
	 *         parsed
	 */
	private void getCsvHeader() throws IOException
	{

		ClassLoader cl = this.getClass().getClassLoader();
		InputStream is = cl.getResourceAsStream(filePath);

		if (is == null) {
			Logger.getLogger("").log(Level.SEVERE, "File not found 1: {0}",
					filePath);
			System.out.println("File not found 1: " + filePath);
			return;
		}

		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		String line;

		if ((line = br.readLine()) != null) {
			csvHeader = line.split(",");
		} else {
			System.out.println("File empty: " + filePath);
		}

		br.close();

	}

	public String getBehaviorValuesName() throws IOException
	{
		getCsvHeader();
		return csvHeader[0];

	}

	public int getIntSlowDown() throws IOException
	{

		return Integer.parseInt(csvHeader[csvHeader.length - 2]);

	}

	public int getIntStartSlowDown() throws IOException
	{

		return Integer.parseInt(csvHeader[csvHeader.length - 1]);

	}

	public String getFunctionBehaviorName() throws IOException
	{

		getCsvHeader();
		return csvHeader[0];

	}

	public float getFloatSlowDown() throws IOException
	{

		return Float.parseFloat(csvHeader[csvHeader.length - 2]);

	}

	public float getFloatStartSlowDown() throws IOException
	{

		return Float.parseFloat(csvHeader[csvHeader.length - 1]);

	}

	/**
	 * Returns an array of filenames from the resource path specified. Works for
	 * jar files.
	 * @param the path to return the files, ending with "/", no "/" at start
	 * @return the names of each file in the specified URI not including path
	 *         information.
	 * @throws URISyntaxException if the path is not a valid URI
	 * @throws IOException if reading the folder causes problems
	 */
	public static String[] getResourceListing(String path)
			throws URISyntaxException, IOException
	{
		ClassLoader cl = ReadCSV.class.getClassLoader();
		URL pathURL = cl.getResource(path);
		if (pathURL != null && pathURL.getProtocol().equals("file")) {
			// reading from a file
			return new File(pathURL.toURI()).list();
		}

		if (pathURL == null) {
			// when reading from a jar file we use the classloaders path to get
			// hold of the jar
			String className = ReadCSV.class.getName().replace(".", "/")
					+ ".class";
			pathURL = cl.getResource(className);
		}

		if (pathURL.getProtocol().equals("jar")) {
			String jarPath = pathURL.getPath().substring(5,
					pathURL.getPath().indexOf("!"));
			JarFile jar = new JarFile(URLDecoder.decode(jarPath, "UTF-8"));

			Enumeration<JarEntry> entries = jar.entries();
			Set<String> result = new HashSet<String>();
			while (entries.hasMoreElements()) {
				// read through all files of the jar file
				String name = entries.nextElement().getName();
				if (name.startsWith(path)) {
					// but only take the ones in the specified path
					String entry = name.substring(path.length());
					int slashPosition = entry.indexOf("/");
					if (slashPosition >= 0) {
						// only the name in case of a subdirectory
						entry = entry.substring(0, slashPosition);
					}
					result.add(entry);
				}
			}
			jar.close();
			return result.toArray(new String[result.size()]);
		}

		throw new UnsupportedOperationException("Cannot list files for URL "
				+ pathURL);
	}
}
