/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.behavior.movement;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.motor.FunctionBehaviorParameter;
import magma.util.function.FunctionLoader;
import magma.util.function.IFunction;
import magma.util.function.PiecewiseLinearFunction;
import magma.util.function.SupportPoint;

/**
 * 
 * @author dorer
 */
public class Movement implements Serializable
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** the current phase we are in for performing */
	private int currentPhaseIndex;

	/** the phases of this behavior */
	private List<MovementPhase> phases;

	/** the name of this movement */
	private final String name;

	public Movement(String name)
	{
		this.name = name;
		currentPhaseIndex = 0;
		phases = new ArrayList<MovementPhase>();
	}

	/**
	 * Adds a phase to the list of phases
	 * @param phase the phase to add
	 */
	public void add(MovementPhase phase)
	{
		phases.add(phase);
	}

	public void init()
	{
		currentPhaseIndex = 0;
		for (MovementPhase current : phases) {
			current.init();
		}
	}

	/**
	 * @param agent reference to the agentmodel
	 * @return false if this movement is over
	 */
	public boolean perform(IAgentModel agent)
	{
		if (isFinished()) {
			return false;
		}

		MovementPhase currentPhase = getCurrentPhase();

		// System.out.println("phase: " + currentPhaseIndex + " name: "
		// + currentPhase.getName() + " cycle: "
		// + currentPhase.getCyclesPerformed() + //
		// " hip: " + agent.getHJ(INaoConstants.LHipPitch).getAxis() + //
		// " knee: " + agent.getHJ(INaoConstants.LKneePitch).getAxis() + //
		// " foot: " + agent.getHJ(INaoConstants.LFootPitch).getAxis() + //
		// " rhipRoll: " + agent.getHJ(INaoConstants.RHipRoll).getAxis() + //
		// " rfootRoll: " + agent.getHJ(INaoConstants.RFootRoll).getAxis());

		boolean result = currentPhase.perform(agent);
		if (result == false) {
			currentPhaseIndex++;
			if (isFinished()) {
				return false;
			}
		}
		return true;
	}

	public MovementPhase getCurrentPhase()
	{
		return phases.get(currentPhaseIndex);
	}

	/**
	 * @return true if this movement is over
	 */
	public boolean isFinished()
	{
		return currentPhaseIndex >= phases.size();
	}

	/**
	 * goes to the next phase no matter if this phase is finished
	 * @return
	 */
	public boolean proceed()
	{
		getCurrentPhase().init();
		currentPhaseIndex++;
		if (isFinished()) {
			return false;
		}
		return true;
	}

	public Movement getLeftVersion()
	{
		String newName = name.replace("Right", "Left");
		Movement result = new Movement(newName);
		for (MovementPhase current : phases) {
			MovementPhase newPhase = current.getLeftVersion();
			result.add(newPhase);
		}
		return result;
	}

	@Override
	public String toString()
	{
		return name + ": " + (isFinished() ? "finished" : getCurrentPhase());
	}

	public FunctionBehaviorParameter toFunctionBehaviorParameter()
	{
		float duration = 0;
		Map<String, ArrayList<SupportPoint>> rawJointPoints = new HashMap<String, ArrayList<SupportPoint>>();

		// get the raw data from the phases
		for (MovementPhase phase : phases) {
			duration += phase.getCycles();

			Map<String, SupportPoint> jointPoints = phase.getJointSupportPoints();
			for (Map.Entry<String, SupportPoint> entry : jointPoints.entrySet()) {
				ArrayList<SupportPoint> points = rawJointPoints.get(entry.getKey());
				if (points == null) {
					points = new ArrayList<SupportPoint>();
					points.add(new SupportPoint(0, 0));
				}

				points.add(entry.getValue());
				rawJointPoints.put(entry.getKey(), points);
			}
		}

		// convert to functions
		Map<String, IFunction> jointFunctions = new HashMap<String, IFunction>();
		for (Map.Entry<String, ArrayList<SupportPoint>> entry : rawJointPoints
				.entrySet()) {
			jointFunctions.put(entry.getKey(),
					new PiecewiseLinearFunction(entry.getValue(), 1));
		}

		return new FunctionBehaviorParameter(null, name,
				FunctionLoader.VERSION_2, duration, jointFunctions);
	}
}
