/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.supportPoint;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.agentmodel.IHingeJoint;
import magma.agent.behavior.motor.MotorBehavior;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.util.function.FunctionLoader;
import magma.util.function.IFunction;
import magma.util.logging.CSVFormatter;

/**
 * Behavior representing sinoid repeated motor commands
 * 
 * @author Klaus Dorer
 */
public abstract class BasicMotorBehavior extends MotorBehavior
{
	/**  */
	private static final long serialVersionUID = 1L;

	private static final String BEHAVIOR_FILE_VERSION = "1";

	/** name of a file that was not saved */
	public static final String UNNAMED_FILE_TXT = "unnamed.txt";

	/** constants for state change events */
	static enum BehaviorEvent {
		INIT, NEXT_STEP
	}

	private transient Logger motorLogger;

	private final float slowDownFactor;

	private final float startSlowDownFactor;

	protected float currentStep;

	/**
	 * true if progress on the currentStep is controlled from outside the
	 * behavior
	 */
	protected boolean fixCurrentStep;

	/** if necessary this can be passed by dependency injection */
	transient BMBStateFactory factory;

	/** the current state (finite state machine) */
	protected transient BMBState state;

	/** motor manager for each motor involved in the behavior */
	protected transient final Map<String, MotorManager> motorManagers;

	/** the number of values after which the motor function repeats */
	protected float period;

	/** true if the behavior should log the actual function to memory */
	private boolean softLogging;

	/** path to the file that has the underlying behavior definition */
	protected String filePath;

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param name Behavior name
	 * @param thoughtModel Reference to the thought model object
	 * @param filepath Path to movement data file
	 * @param slowDown Slow down factor
	 * @param startSlowDown Startup slowdown factor
	 */
	public BasicMotorBehavior(String name, IThoughtModel thoughtModel,
			String filepath, float slowDown, float startSlowDown)
	{
		super(name, thoughtModel);
		this.currentStep = 0.0f;
		this.slowDownFactor = slowDown;
		this.startSlowDownFactor = startSlowDown;
		this.motorLogger = null;
		this.motorManagers = new HashMap<String, MotorManager>();
		this.factory = new BMBStateFactory();
		this.state = (BMBState) factory.getInitialState();
		this.filePath = filepath;
		this.fixCurrentStep = false;
	}

	/**
	 * Copy constructor
	 * 
	 * @param source Source object to copy from
	 */
	public BasicMotorBehavior(BasicMotorBehavior source)
	{
		super(source);

		motorLogger = null;
		slowDownFactor = source.slowDownFactor;
		startSlowDownFactor = source.startSlowDownFactor;
		currentStep = source.currentStep;
		state = source.state;
		motorManagers = new HashMap<String, MotorManager>();

		for (String managerName : source.motorManagers.keySet()) {
			MotorManager sourceManager = source.motorManagers.get(managerName);
			MotorManager copyManager = new MotorManager(sourceManager);
			motorManagers.put(managerName, copyManager);
		}

		period = source.period;
		filePath = source.filePath;
		fixCurrentStep = false;
	}

	@Override
	public void init()
	{
		super.init();
		setCurrentStep(0.0f);
		state = (BMBState) state.tiggerEvent(BehaviorEvent.INIT, this);
		if (softLogging) {
			for (MotorManager manager : motorManagers.values()) {
				manager.resetSoftLog();
			}
		}
	}

	@Override
	public void perform()
	{
		for (String joint : motorManagers.keySet()) {
			agentModel
					.getWriteableHJ(joint)
					.getMotor()
					.setStiffness(
							motorManagers.get(joint).getMotorFunction().getStiffness());
		}

		state.perform(this);
		state = (BMBState) state.tiggerEvent(BehaviorEvent.NEXT_STEP, this);
	}

	protected boolean willFinish()
	{
		return currentStep >= period;
	}

	void performSlowDown()
	{
		IHingeJoint hingeJoint;
		float currentAngle;
		float desiredAngle;
		float delta;

		StringBuilder buffer = new StringBuilder();

		// loop through all joints affected
		for (MotorManager manager : motorManagers.values()) {

			// calculate speed
			desiredAngle = manager.getDesiredAngle(currentStep);
			hingeJoint = manager.getHingeJoint();

			if (hingeJoint == null) {
				logger.log(Level.SEVERE, "unknown hinge joint: {0}", name);
			} else {
				currentAngle = hingeJoint.getAxis();
				delta = desiredAngle - currentAngle;
				float speed = state.calculateSpeed(this, delta);
				if (speed > -0.02 && speed < 0.02) {
					speed = 0.0f;
				}

				if (softLogging) {
					manager.setLogPoint(currentStep, currentAngle);
				}

				if (LOG) {
					buffer.append(desiredAngle).append(";");
					buffer.append(currentAngle).append(";");
					buffer.append(speed).append(";");
				}
				// set axis speed
				hingeJoint.performAxisSpeed(speed);

			}
			if (LOG) {
				if (motorLogger == null) {
					initLogger(Level.INFO);
				}
				motorLogger.log(Level.INFO, buffer.toString());
			}
		}
	}

	/**
	 * Initializes logging functionality
	 * 
	 * @param level Logging level
	 */
	private void initLogger(Level level)
	{
		try {
			motorLogger = Logger.getLogger(getName());
			FileHandler handler = new FileHandler("log/" + getName() + ".csv");
			handler.setFormatter(new CSVFormatter());
			motorLogger.addHandler(handler);
			motorLogger.setLevel(level);

			// write heading
			StringBuilder buffer = new StringBuilder();
			for (MotorManager manager : motorManagers.values()) {
				buffer.append(manager.getName()).append(";");
				buffer.append("angleAsIs").append(";");
				buffer.append("speed").append(";");
			}
			motorLogger.log(Level.INFO, buffer.toString());

		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @return the list of motor managers for this behavior
	 */
	public Map<String, MotorManager> getMotorManagers()
	{
		return motorManagers;
	}

	/**
	 * @return the list of motor managers for this behavior
	 */
	public Set<String> getMotorManagerNames()
	{
		return motorManagers.keySet();
	}

	/**
	 * Creates a copy of this instance
	 * @return a copy of this instance
	 */
	public abstract BasicMotorBehavior copy();

	/**
	 * Save internal movement data to file
	 */
	public void saveToFile()
	{
		PrintWriter out = null;

		try {
			File file = new File(getAbsoluteFilepath());
			out = new PrintWriter(new FileWriter(file));
			out.println(BEHAVIOR_FILE_VERSION + "," + name + "," + period);

			// write all data for each joint
			for (MotorManager manager : motorManagers.values()) {
				String line = manager.toCSVString();
				out.println(line);
			}

		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (out != null) {
				out.close();
			}
		}
	}

	/**
	 * Adds a new joint to the joints of the selected behavior
	 * @param jointName the name of the joint
	 * @param functionType the type (spline, sin, linear, ...)
	 * @param forceAdd True if the addition process should be forced if the joint
	 *        already exists
	 * @return false if the joint is already part of this behavior
	 */
	public MotorManager addNewMotorManager(String jointName,
			String functionType, boolean forceAdd)
	{
		if (motorManagers.containsKey(jointName)) {
			if (!forceAdd) {
				return null;
			}
			motorManagers.remove(jointName);
		}

		// create function
		IFunction function = FunctionLoader.INSTANCE.createFunction(functionType,
				null, 1);

		// create motor manager
		MotorManager manager = new MotorManager(agentModel, jointName, function);

		// add it
		motorManagers.put(jointName, manager);

		return manager;
	}

	/**
	 * Re-Add a motor manager
	 * 
	 * @param manager Reference to motor manager object
	 */
	public void reAddMotorManager(MotorManager manager)
	{
		motorManagers.put(manager.getName(), manager);
	}

	/**
	 * Remove a motor manager
	 * 
	 * @param manager Name of the motor manager object to remove
	 * @return Reference to removed motor manager object
	 */
	public MotorManager removeMotorManager(String manager)
	{
		return motorManagers.remove(manager);
	}

	/**
	 * @param softLogging the softLogging to set
	 */
	public void setSoftLogging(boolean softLogging)
	{
		this.softLogging = softLogging;
	}

	/**
	 * @return the slowDownFactor
	 */
	public float getSlowDownFactor()
	{
		return slowDownFactor;
	}

	/**
	 * @return the currentStep
	 */
	public float getCurrentStep()
	{
		return currentStep;
	}

	/**
	 * @return the period
	 */
	public float getPeriod()
	{
		return period;
	}

	/**
	 * @return the startSlowDownFactor
	 */
	public float getStartSlowDownFactor()
	{
		return startSlowDownFactor;
	}

	/**
	 * @param step the current x coordinate of the motor file
	 */
	void setCurrentStep(float step)
	{
		if (!fixCurrentStep) {
			currentStep = step;
		}
	}

	/**
	 * @param step the current x coordinate of the motor file
	 * @param fix true to fix current step
	 */
	public void fixCurrentStep(float step, boolean fix)
	{
		fixCurrentStep = fix;
		currentStep = step;
	}

	/**
	 * Retrieve the absolute path to the movement data file
	 * 
	 * @return Absolute path
	 */
	public String getAbsoluteFilepath()
	{
		File file = new File(filePath);

		if (!file.isAbsolute()) {
			// workaround for being able to save files specified with relative
			// path
			file = new File(file.getAbsolutePath().replaceFirst("behaviors",
					"config/behaviors"));
		}

		return file.getAbsolutePath();
	}

	/**
	 * @return the filePath
	 */
	public String getFilePath()
	{
		return filePath;
	}

	/**
	 * @param filePath the filePath to set
	 */
	public void setFilePath(String filePath)
	{
		this.filePath = filePath;
	}
}
