/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.supportPoint;

import java.io.IOException;
import java.util.Map;

import magma.agent.behavior.motor.FunctionBehaviorParameter;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.util.function.IFunction;
import magma.util.parameter.ParameterMap;

/**
 * Behavior representing motor commands based on a mathematical function
 * 
 * @author Klaus Dorer
 */
public class FunctionBehavior extends BasicMotorBehavior
{
	/**  */
	private static final long serialVersionUID = 1L;

	/**
	 * Default constructor for motor behaviors
	 * @param name identifier for this behavior
	 * @param thoughtModel Reference to the thought model object
	 * @param filename the name of a file containing the parameters for this
	 *        behavior
	 * @param slowDown factor of the speed with which the behavior is performed.
	 *        2 means half the speed.
	 * @param startSlowDown the speed to approach the first value
	 */
	public FunctionBehavior(String name, IThoughtModel thoughtModel,
			String filename, float slowDown, float startSlowDown)
	{
		super(name, thoughtModel, filename, slowDown, startSlowDown);
		loadFromFile();
	}

	private void loadFromFile()
	{
		FunctionBehaviorParameter parameters = null;
		try {
			parameters = FunctionBehaviorParameter
					.readBehaviorFile(null, filePath);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(-1);
		}
		extractMotorManagers(parameters);
		period = parameters.getPeriod();
	}

	/**
	 * Default constructor for motor behaviors
	 * @param name identifier for this behavior
	 * @param coordinator handle for action sending to server
	 * @param thoughtModel Reference to the thought model object
	 * @param params parameters to control the behavior
	 * @param slowDown factor of the speed with which the behavior is performed.
	 *        2 means half the speed.
	 * @param startSlowDown the speed to approach the first value
	 */
	public FunctionBehavior(String name, IThoughtModel thoughtModel,
			ParameterMap params, float slowDown, float startSlowDown)
	{
		super(name, thoughtModel, BasicMotorBehavior.UNNAMED_FILE_TXT, slowDown,
				startSlowDown);
		FunctionBehaviorParameter param = (FunctionBehaviorParameter) params
				.get(name);
		extractMotorManagers(param);
		period = param.getPeriod();
	}

	/**
	 * Copy constructor
	 * 
	 * @param source Source object to copy state from
	 */
	private FunctionBehavior(FunctionBehavior source)
	{
		super(source);
	}

	/**
	 * Extract motor managers from behavior parameters
	 * 
	 * @param parameters Function behavior parameters
	 */
	private void extractMotorManagers(FunctionBehaviorParameter parameters)
	{
		Map<String, IFunction> joints = parameters.getJoints();
		motorManagers.clear();

		for (String jointName : joints.keySet()) {
			MotorManager manager = new MotorManager(agentModel, jointName,
					joints.get(jointName));
			motorManagers.put(jointName, manager);
		}
	}

	@Override
	public BasicMotorBehavior copy()
	{
		return new FunctionBehavior(this);
	}

	@Override
	public void reload()
	{
		super.reload();

		if (!filePath.equals(BasicMotorBehavior.UNNAMED_FILE_TXT)) {
			loadFromFile();
		}
	}
}
