/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.belief.impl;

import java.awt.geom.Line2D;

import magma.agent.belief.IBeliefConstants;
import magma.agent.thoughtmodel.IThoughtModel;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Belief to decide if the Ball could be keepable for the goal keeping behavior
 * @author Rajit Shahi
 * 
 */
public class BallIsKeepable extends Belief
{
	/** the height of the player */
	private final float playerHeight;

	/** the number of cycles we predict ball position into future */
	private final int goalPredictionTime;

	/**
	 * Constructor
	 * 
	 * @param agentModel Reference to the thought model object
	 */
	public BallIsKeepable(IThoughtModel thoughtModel, float playerHeight,
			int goalPredictionTime)
	{
		super(IBeliefConstants.BALL_IS_KEEPABLE, thoughtModel);
		this.playerHeight = playerHeight;
		this.goalPredictionTime = goalPredictionTime;
	}

	@Override
	public float getTruthValue()
	{
		Vector3D[] futureBallPositions = worldModel.getBall().getFuturePositions(
				goalPredictionTime);
		Vector3D futureBallPosition = futureBallPositions[futureBallPositions.length - 1];
		Vector3D ballPosition = worldModel.getBall().getPosition();

		Line2D ballLine = new Line2D.Double(futureBallPosition.getX(),
				futureBallPosition.getY(), ballPosition.getX(), ballPosition.getY());

		Vector3D playerPosition = worldModel.getThisPlayer().getPosition();
		double playerAngle = worldModel.getThisPlayer().getHorizontalAngle()
				.radians();

		Vector3D playerDirection = new Vector3D(playerAngle, 0);
		playerDirection.normalize().scalarMultiply(playerHeight);

		Vector3D right = new Vector3D(playerDirection.getAlpha() - (Math.PI / 2),
				0).normalize();
		Vector3D left = new Vector3D(playerDirection.getAlpha() + (Math.PI / 2),
				0).normalize();

		Vector3D leftLess = left.scalarMultiply(playerHeight * 0.1).add(
				playerPosition);
		Vector3D rightLess = right.scalarMultiply(playerHeight * 0.1).add(
				playerPosition);
		Vector3D leftMost = left.scalarMultiply(playerHeight).add(playerPosition);
		Vector3D rightMost = right.scalarMultiply(playerHeight).add(
				playerPosition);

		Line2D keepableDiameter = new Line2D.Double(rightMost.getX(),
				rightMost.getY(), leftMost.getX(), leftMost.getY());

		Line2D unkeepableDiameter = new Line2D.Double(rightLess.getX(),
				rightLess.getY(), leftLess.getX(), leftLess.getY());

		if (unkeepableDiameter.intersectsLine(ballLine)) {
			return 0;
		}

		if (keepableDiameter.intersectsLine(ballLine)) {
			return 1;
		}

		return 0;
	}

}
