/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.belief.impl;

import java.util.logging.Level;

import magma.agent.belief.IBeliefConstants;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Checks if we are properly aligned with ball and kick direction
 * 
 * @author Klaus Dorer
 */
public class FacingAdvantageousKickDirection extends Belief
{
	/**
	 * Default constructor
	 * 
	 * @param agentModel Reference to the thought model object
	 */
	public FacingAdvantageousKickDirection(IThoughtModel thoughtModel)
	{
		super(IBeliefConstants.FACING_ADV_KICK_DIRECTION, thoughtModel);
	}

	/**
	 * Constructor
	 * 
	 * @param name Belief name
	 * @param agentModel Reference to the thought model object
	 */
	FacingAdvantageousKickDirection(String name, IThoughtModel thoughtModel)
	{
		super(name, thoughtModel);
	}

	@Override
	public float getTruthValue()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IMoveableObject ball = worldModel.getBall();
		Vector3D ballPos = ball.getPosition();
		Vector3D kickTo = getDesiredPosition();

		double ownDistanceToBall = thisPlayer.getDistanceToXY(ballPos);
		double kickToDistance = thisPlayer.getDistanceToXY(kickTo);
		Angle ballAngle1 = thisPlayer.getBodyDirectionTo(ballPos);
		Angle kickToAngle1 = thisPlayer.getBodyDirectionTo(kickTo);

		double ballAngle = ballAngle1.degrees();
		double kickToAngle = kickToAngle1.degrees();

		// LOG
		logger.log(
				Level.FINE,
				"Kick: ownDistanceToBall: {0} BallAngle: {1} kickToAngle: {2} "
						+ "last seen: {3} kickToDistance: {4} ",
				new Object[] { ownDistanceToBall, ballAngle, kickToAngle,
						ball.getLastSeenTime(), kickToDistance });

		if (shouldWeKick(kickToDistance, ballAngle, kickToAngle)) {
			return 1.0f;
		}

		return 0.0f;
	}

	/**
	 * Get desired position
	 * 
	 * @return Desired position
	 */
	Vector3D getDesiredPosition()
	{
		return thoughtModel.getPointToKickBall();
	}

	/**
	 * Decide if the agent should kick
	 * 
	 * @param kickToDistance Distance to kick target
	 * @param ballAngle Relative angle to ball
	 * @param kickToAngle Relative angle of the direction we want to kick in
	 * 
	 * @return True if agent should kick
	 */
	boolean shouldWeKick(double kickToDistance, double ballAngle,
			double kickToAngle)
	{
		double ownGoalDistance = worldModel.getThisPlayer().getDistanceToXY(
				worldModel.getOwnGoalPosition());

		if (ownGoalDistance < 3 && Math.abs(ballAngle) < 30
				&& Math.abs(kickToAngle) < 80) {
			return true;
		}
		if ((ownGoalDistance < 9 && Math.abs(ballAngle) < 30 && Math
				.abs(kickToAngle) < 60)) {
			return true;
		}

		return Math.abs(ballAngle) < 30 && Math.abs(kickToAngle) < 45;
	}
}
