/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.belief.impl;

import java.util.logging.Level;

import magma.agent.belief.IBeliefConstants;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Checks if we are properly aligned with ball and kick direction
 * 
 * @author Klaus Dorer
 */
public class FacingKickDirection extends Belief
{
	/**
	 * Default constructor
	 * 
	 * @param agentModel Reference to the thought model object
	 */
	public FacingKickDirection(IThoughtModel thoughtModel)
	{
		super(IBeliefConstants.FACING_KICK_DIRECTION, thoughtModel);
	}

	@Override
	public float getTruthValue()
	{
		// TODO: this belief is still specific for kicking towards the goal

		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IMoveableObject ball = worldModel.getBall();
		Vector3D ballPos = ball.getPosition();
		Vector3D ownPos = thisPlayer.getPosition();

		double ownDistanceToBall = thisPlayer.getDistanceToXY(ballPos);
		double goalDistance = thisPlayer.getDistanceToXY(worldModel
				.getOtherGoalPosition());
		double ownGoalDistance = thisPlayer.getDistanceToXY(worldModel
				.getOwnGoalPosition());
		Angle ballAngle1 = thisPlayer.getBodyDirectionTo(ballPos);
		Angle goalAngle1 = thisPlayer.getBodyDirectionTo(worldModel
				.getOtherGoalPosition());

		double ballAngle = ballAngle1.degrees();
		double goalAngle = goalAngle1.degrees();

		// Calculate the position where the straight line through the player and
		// the ball will cross the end of the field

		double m;
		double b;

		// Calculate the straight line to the opponent goal
		double x = worldModel.getOtherGoalPosition().getX();
		m = (ownPos.getY() - ballPos.getY()) / (ownPos.getX() - ballPos.getX());
		b = ownPos.getY() - m * ballPos.getX();

		// Calculate the Y point of the goal line where the ball will cross the
		// line if we perform a straight kick
		double y = m * x + b;

		// LOG
		logger.log(
				Level.FINE,
				"Kick: Distance: {0} BallAngle: {1} goalAngle: {2} "
						+ "ownGoalDistance: {3} last seen: {4} goalDistance: {5} y: {6}",
				new Object[] { ownDistanceToBall, ballAngle, goalAngle,
						ownGoalDistance, ball.getLastSeenTime(), goalDistance, y });

		if (shouldWeKick(goalDistance, ownGoalDistance, ballAngle, goalAngle, y)) {
			return 1.0f;
		}

		return 0.0f;
	}

	private boolean shouldWeKick(double goalDistance, double ownGoalDistance,
			double ballAngle, double goalAngle, double y)
	{
		if (ownGoalDistance < 3 && Math.abs(ballAngle) < 30
				&& Math.abs(goalAngle) < 80) {
			return true;
		}
		if ((ownGoalDistance < 9 && Math.abs(ballAngle) < 30 && Math
				.abs(goalAngle) < 60)) {
			return true;
		}

		if (goalDistance < 4.5) {
			return y < worldModel.goalHalfWith() && y > -worldModel.goalHalfWith()
					&& Math.abs(ballAngle) < 15;
		}

		return Math.abs(ballAngle) < 30 && Math.abs(goalAngle) < 45;
	}
}
