/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.belief.impl;

import magma.agent.belief.IBeliefConstants;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Geometry;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

public class ShouldWeDribble extends Belief
{
	/** length of the dribble area */
	private static final double MAX_X_DRIBBLE = 0.35;

	/** width of the dribble area */
	private static final double MAX_Y_DRIBBLE = 0.10;

	/** assumed maximum distance the ball would move while dribbling */
	private static final double MAX_DRIBBLE_KICK_DISTANCE = 3;

	private transient final Vector3D[] dribbleablePoly;

	/**
	 * @param agentModel Reference to the thought model object
	 */
	public ShouldWeDribble(IThoughtModel thoughtModel)
	{
		super(IBeliefConstants.SHOULD_WE_DRIBBLE, thoughtModel);
		dribbleablePoly = Geometry.getPolygon(new double[][] {
				{ 0, MAX_Y_DRIBBLE }, { MAX_X_DRIBBLE, MAX_Y_DRIBBLE },
				{ MAX_X_DRIBBLE, -MAX_Y_DRIBBLE }, { 0, -MAX_Y_DRIBBLE } });
	}

	@Override
	public float getTruthValue()
	{
		return shouldWeDribble();
	}

	/**
	 * Checks if we are in a position to dribble an as well in the need to. It
	 * prevents bad situations with respect to our goal and the field border.
	 * 
	 * @return the position to dribble to, or null otherwise
	 */
	private float shouldWeDribble()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		Vector3D ballPos = worldModel.getBall().getPosition();

		if (!thisPlayer.isInsidePolygonXY(ballPos, dribbleablePoly)) {
			return 0.0f;
		}
		IPlayer closestOpponent = thoughtModel.getOpponentAtBall();

		if (closestOpponent == null) {
			return 0.0f;
		}
		double opponentDistance = closestOpponent.getDistanceToXY(ballPos);

		if (opponentDistance > 1.2) {
			// no dribbling if opponent is far
			return 0.0f;
		}

		// Since there is an opponent near the ball and we are in
		// dribbleable distance, we should maybe dribble
		double viewDirection = thisPlayer.getHorizontalAngle().degrees();

		// 1. Check if we would run towards our goal
		if (Math.abs(viewDirection) > 140
				|| (ballPos.getX() < 0 && Math.abs(viewDirection) > 70)) {
			return 0.0f;
		}

		// 2. Check if we would run outside the field
		Vector3D resultingBallPos = ballPos.add(new Vector3D(
				MAX_DRIBBLE_KICK_DISTANCE, new Vector3D(Math
						.toRadians(viewDirection), 0)));

		if (resultingBallPos.getX() > worldModel.fieldHalfLength()
				|| resultingBallPos.getX() < -worldModel.fieldHalfLength()
				|| resultingBallPos.getY() > worldModel.fieldHalfWith()
				|| resultingBallPos.getY() < -worldModel.fieldHalfWith()) {
			return 0.0f;
		}
		return 1.0f;
	}

}
