/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package magma.agent.communication.base;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import magma.agent.action.IEffector;
import magma.agent.communication.IChannelManager;
import magma.agent.communication.IInputChannel;
import magma.agent.communication.IOutputChannel;
import magma.agent.perception.IPerceptor;

/**
 * 
 * @author kdorer
 */
public class ChannelManager implements IChannelManager
{
	/** the container for incoming perceptions */
	private List<Map<String, IPerceptor>> perceptors;

	/** the single main channel */
	private IInputChannel mainChannel;

	/** channels for receiving sensor information */
	protected List<IInputChannel> inChannels;

	/** channels for sending sensor information */
	protected List<IOutputChannel> outChannels;

	private ChannelManagerStatus status;

	/**
	 * @param perception
	 */
	public ChannelManager()
	{
		perceptors = new ArrayList<Map<String, IPerceptor>>();
		this.inChannels = new ArrayList<IInputChannel>();
		this.outChannels = new ArrayList<IOutputChannel>();
		status = ChannelManagerStatus.CREATED;
	}

	@Override
	public void addPerceptors(Map<String, IPerceptor> newPerceptors)
	{
		if (newPerceptors == null) {
			return;
		}

		synchronized (perceptors) {
			perceptors.add(newPerceptors);
			// if (perceptors.size() > 1) {
			// System.out
			// .println("Have multiple perceptors: " + perceptors.size());
			// }
		}

		// outside the synchronized block we trigger the updates
		synchronized (this) {
			notify();
		}
	}

	@Override
	public Map<String, IPerceptor> getNextPerceptorMap()
	{
		synchronized (perceptors) {
			if (perceptors.isEmpty()) {
				return null;
			}
			return perceptors.remove(0);
		}
	}

	@Override
	public void performAction(Map<String, IEffector> effectors)
	{
		for (IOutputChannel outChannel : outChannels) {
			outChannel.sendMessage(effectors);
		}
	}

	@Override
	public void addInputChannel(IInputChannel channel, boolean isMainChannel)
	{
		inChannels.add(channel);
		if (isMainChannel) {
			mainChannel = channel;
		}
	}

	@Override
	public void addOutputChannel(IOutputChannel channel)
	{
		outChannels.add(channel);
	}

	@Override
	public boolean start()
	{
		// TODO: we have to decide what to do if some channels do not work
		boolean result = true;
		for (IInputChannel in : inChannels) {
			boolean connected = in.startChannel();
			if (!connected && in == mainChannel) {
				stop();
				return false;
			}
			result &= connected;
		}
		for (IOutputChannel out : outChannels) {
			result &= out.startChannel();
		}
		status = ChannelManagerStatus.STARTED;
		return result;
	}

	@Override
	public void stop()
	{
		for (IInputChannel in : inChannels) {
			in.stopChannel();
		}
		for (IOutputChannel out : outChannels) {
			out.stopChannel();
		}
		synchronized (this) {
			notify();
		}

		status = ChannelManagerStatus.STOPPED;
	}

	@Override
	public boolean isConnected()
	{
		if (mainChannel == null) {
			return false;
		}
		return mainChannel.isConnected();
	}

	@Override
	public void lostConnection(IInputChannel channel)
	{
		if (mainChannel == channel) {
			// we have to inform that a main connection is lost
			status = ChannelManagerStatus.LOST_MAIN_CONNECTION;
			synchronized (this) {
				notify();
			}
			stop();
			status = ChannelManagerStatus.LOST_MAIN_CONNECTION;
		}
	}

	@Override
	public ChannelManagerStatus getStatus()
	{
		return status;
	}

	@Override
	public String toString()
	{
		if (mainChannel != null) {
			return mainChannel.getConnectionState().toString();
		}
		return status.toString();
	}
}
