/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package magma.agent.communication.base;

import java.util.Arrays;

import magma.agent.communication.IChannelManager;
import magma.agent.communication.IChannelState.ConnectionState;
import magma.agent.communication.IChannelState.DataState;
import magma.agent.communication.IInputChannel;
import magma.util.connection.ConnectionException;
import magma.util.connection.IServerConnection;
import magma.util.observer.IObserver;

/**
 * 
 * @author kdorer
 */
public abstract class ConnectionChannel extends ChannelBase implements
		IInputChannel, IObserver<byte[]>
{
	/** the connection to the server */
	protected final IServerConnection connection;

	/** the number of received messages */
	private int numberOfReceivedMessages;

	public ConnectionChannel(IChannelManager manager,
			IServerConnection connection)
	{
		super(manager);
		this.connection = connection;
		numberOfReceivedMessages = 0;
	}

	protected void sendMessage(byte[] message)
	{
		try {
			connection.sendMessage(message);
			// System.out.println("sending: " + Arrays.toString(message));
			// System.out.println("sending: " + new String(message));

		} catch (ConnectionException e) {
			System.out.println(e);
			e.printStackTrace();
			state.setLastErrorMessage(e.getMessage());
		}
	}

	/**
	 * Starts the connection to the server, will only return after disconnection
	 * Uses default IP and port
	 * 
	 * @return true, if connection was successful and the agent stopped receiving
	 *         messages, false if connection was refused or the server shut down
	 */
	@Override
	public boolean startChannel()
	{
		if (connection == null) {
			return false;
		}
		if (connection.isConnected()) {
			return true;
		}
		connection.attach(this);
		state.setConnectionState(ConnectionState.TRYING_TO_CONNECT);
		try {
			connection.establishConnection();
			state.setConnectionState(ConnectionState.CONNECTED);
		} catch (ConnectionException e) {
			state.setConnectionState(ConnectionState.CONNECTION_FAILED);
			state.setLastErrorMessage(e.getMessage());
			System.out.println(e);
			return false;
		}

		boolean success = startReceiveLoop();

		return success;
	}

	/**
	 * Starts to receive data from the connection. The default implementation
	 * calls the connections startReceiveLoop method and may block depending on
	 * the connection.
	 * @return
	 */
	protected boolean startReceiveLoop()
	{
		try {
			connection.startReceiveLoop();
			return true;
		} catch (ConnectionException e) {
			state.setLastErrorMessage(e.getMessage());
			state.setConnectionState(ConnectionState.CONNECTION_FAILED);
			return false;
		}
	}

	@Override
	public void stopChannel()
	{
		if (connection == null || !connection.isConnected()) {
			return;
		}
		connection.stopReceiveLoop();
		state.setConnectionState(ConnectionState.NOT_CONNECTED);
	}

	@Override
	public void update(byte[] message)
	{
		try {
			if (numberOfReceivedMessages == 0) {
				onFirstMessage(message);
			}
			onEachMessage(message);
			numberOfReceivedMessages++;

		} catch (ConnectionException e) {
			e.printStackTrace();
			state.setConnectionState(ConnectionState.DISCONNECTED);

		} catch (RuntimeException e) {
			System.out.println("Error parsing message: "
					+ Arrays.toString(message));
			// System.out.println(new String(message));
			state.setDataState(DataState.INVALID_DATA);
		}
	}

	/**
	 * Notification when the first message was received on this channel, before
	 * processing it
	 * @param message
	 */
	protected void onFirstMessage(byte[] message) throws ConnectionException
	{
	}

	/**
	 * Notification for each message was received on this channel
	 * @param message
	 */
	protected void onEachMessage(byte[] message) throws ConnectionException
	{
	}

	/**
	 * @return the lastErrorMessage
	 */
	public String getLastErrorMessage()
	{
		return getLastErrorMessage();
	}

	/**
	 * @return the numberOfReceivedMessages
	 */
	public int getNumberOfReceivedMessages()
	{
		return numberOfReceivedMessages;
	}
}
