/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.decision.simple;

import java.io.Serializable;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.belief.IBelief;
import magma.agent.belief.IBeliefConstants;
import magma.agent.decision.IDecisionMaker;
import magma.agent.thoughtmodel.IThoughtModel;

/**
 * Base class for all decision makers
 * @author Klaus Dorer
 */
public abstract class DecisionMakerBase implements IDecisionMaker, Serializable
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** handle for logging */
	protected static final Logger logger = Logger.getLogger("");

	/** the behavior that is currently selected */
	protected IBehavior currentBehavior;

	/** the believes that are available */
	protected transient final Map<String, IBelief> believes;

	/** the behaviors that are available */
	protected transient final Map<String, IBehavior> behaviors;

	/** link to the thought model */
	protected final transient IThoughtModel thoughtModel;

	public IThoughtModel getThoughtModel()
	{
		return thoughtModel;
	}

	/** the number of decisions taken */
	protected int numberOfDecisions;

	/** the behavior that is planned to switch to */
	private IBehavior desiredBehavior;

	/**
	 * Constructor
	 * 
	 * @param believes List of believes
	 * @param behaviors List of applicable behaviors
	 */
	public DecisionMakerBase(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors, IThoughtModel thoughtModel)
	{
		this.believes = believes;
		this.behaviors = behaviors;
		this.thoughtModel = thoughtModel;
		numberOfDecisions = 0;
		if (behaviors != null) {
			currentBehavior = behaviors.get(IBehaviorConstants.NONE);
		}
		desiredBehavior = currentBehavior;
	}

	/**
	 * Retrieve a specific belief
	 * 
	 * @param name Belief name
	 * @return The specified believe, null if no matching belief was found
	 */
	@Override
	public IBelief getBelieve(String name)
	{
		return believes.get(name);
	}

	/**
	 * Retrieve a list of all beliefs
	 * 
	 * @return Belief list
	 */
	public Map<String, IBelief> getBelieves()
	{
		return believes;
	}

	/**
	 * Retrieve a specific behavior
	 * 
	 * @param name Behavior name
	 * @return The specified behavior, null if no matching behavior was found
	 */
	@Override
	public IBehavior getBehavior(String name)
	{
		return behaviors.get(name);
	}

	/**
	 * Retrieve a list of all behaviors
	 * 
	 * @return Behavior list
	 */
	public Map<String, IBehavior> getBehaviors()
	{
		return behaviors;
	}

	/**
	 * Retrieve the number of times decide() was called
	 * 
	 * @return Call count
	 */
	public int getNumberOfDecisions()
	{
		return numberOfDecisions;
	}

	@Override
	public boolean decide()
	{
		numberOfDecisions++;

		if (isPaused()) {
			desiredBehavior = getPausedBehavior();

		} else {
			if (numberOfDecisions > 10) {
				desiredBehavior = decideSoccer();

			} else {
				desiredBehavior = behaviors.get(IBehaviorConstants.NONE);
			}
		}
		if (desiredBehavior != currentBehavior) {
			String oldName = currentBehavior.toString();
			currentBehavior = desiredBehavior.switchFrom(currentBehavior);
			logger.log(Level.INFO, "old: {0} wanted: {1} new: {2}",
					new Object[] { oldName, desiredBehavior.toString(),
							currentBehavior.toString() });
		} else {
			currentBehavior.stayIn();
		}
		logger.log(Level.INFO, "going to perform: {0}",
				new Object[] { currentBehavior });
		currentBehavior.perform();
		logger.log(Level.INFO, "performed: {0}", new Object[] { currentBehavior });
		// System.out.println(currentBehavior.getName());
		return true;
	}

	protected boolean isPaused()
	{
		return believes.get(IBeliefConstants.PAUSING).getTruthValue() > 0.5f
				|| believes.get(IBeliefConstants.SOFT_PAUSING).getTruthValue() > 0.5f;
	}

	/**
	 * @return
	 */
	protected IBehavior getPausedBehavior()
	{
		return behaviors.get(IBehaviorConstants.NONE);
	}

	/**
	 * Decide which behavior to use next
	 * 
	 * @return Next behavior
	 */
	public abstract IBehavior decideSoccer();

	@Override
	public IBehavior getCurrentBehavior()
	{
		// System.out.println(currentBehavior);
		return currentBehavior;
	}

	/**
	 * @return the desiredBehavior
	 */
	@Override
	public IBehavior getDesiredBehavior()
	{
		return desiredBehavior;
	}
}
