/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.decision.simple;

import java.util.Map;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.belief.IBelief;
import magma.agent.thoughtmodel.IThoughtModel;

/**
 * Decision maker used to be trigger from external source
 * 
 * @author Klaus Dorer
 */
public class KeyboardDecisionMaker extends SoccerDecisionMaker
{
	/**  */
	private static final long serialVersionUID = 1L;

	public static enum ActionEnum {
		StandStill, Forward, Backward, LeftSlide, RightSlide, LeftTurn, RightTurn, //
		FastForward, FastBackward, FastLeftSlide, FastRightSlide, FastLeftTurn, FastRightTurn, //
		Kick
	}

	/** we just want to beam once */
	private boolean haveBeamed;

	/** current action to perform */
	private ActionEnum currentAction;

	/**
	 * Constructor
	 * 
	 * @param believes List of believes
	 * @param behaviors List of applicable behaviors
	 * @param serverVersion Server version number
	 */
	public KeyboardDecisionMaker(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors, IThoughtModel thoughtModel,
			int serverVersion)
	{
		super(believes, behaviors, thoughtModel, serverVersion);
		haveBeamed = true;
		currentAction = ActionEnum.StandStill;
	}

	@Override
	protected IBehavior beamHome()
	{
		if (haveBeamed) {
			return null;
		}
		haveBeamed = true;
		return super.beamHome();
	}

	/**
	 * Called to decide if we should wait for game start or not. Default
	 * implementation waits.
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	@Override
	protected IBehavior waitForGameStart()
	{
		return null;
	}

	/**
	 * Called to decide if we should kick the ball
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	@Override
	protected IBehavior doKickBehavior()
	{
		return null;
	}

	/**
	 * Called to decide if movement is necessary
	 * @return walk behavior if no team mate is closer to ball
	 */
	@Override
	protected IBehavior movement()
	{
		IBehavior behavior;
		switch (getCurrentAction()) {
		case Forward:
			behavior = behaviors.get(IBehaviorConstants.WALKFORWARD);
			behavior.setIntensity(0.5f);
			break;
		case Backward:
			behavior = behaviors.get(IBehaviorConstants.WALKFORWARD);
			behavior.setIntensity(-0.5f);
			break;
		case LeftSlide:
			behavior = behaviors.get(IBehaviorConstants.WALKSIDEWARD);
			behavior.setIntensity(0.5f);
			break;
		case RightSlide:
			behavior = behaviors.get(IBehaviorConstants.WALKSIDEWARD);
			behavior.setIntensity(-0.5f);
			break;
		case LeftTurn:
			behavior = behaviors.get(IBehaviorConstants.WALKTURNING);
			behavior.setIntensity(0.3f);
			break;
		case RightTurn:
			behavior = behaviors.get(IBehaviorConstants.WALKTURNING);
			behavior.setIntensity(-0.3f);
			break;
		case FastForward:
			behavior = behaviors.get(IBehaviorConstants.WALKFORWARD);
			behavior.setIntensity(1.0f);
			break;
		case FastBackward:
			behavior = behaviors.get(IBehaviorConstants.WALKFORWARD);
			behavior.setIntensity(-1.0f);
			break;
		case FastLeftSlide:
			behavior = behaviors.get(IBehaviorConstants.WALKSIDEWARD);
			behavior.setIntensity(1.0f);
			break;
		case FastRightSlide:
			behavior = behaviors.get(IBehaviorConstants.WALKSIDEWARD);
			behavior.setIntensity(-1.0f);
			break;
		case FastLeftTurn:
			behavior = behaviors.get(IBehaviorConstants.WALKTURNING);
			behavior.setIntensity(0.6f);
			break;
		case FastRightTurn:
			behavior = behaviors.get(IBehaviorConstants.WALKTURNING);
			behavior.setIntensity(-0.6f);
			break;
		case Kick:
			behavior = behaviors.get(IBehaviorConstants.SHOOT_STRAIGHT);
			break;
		default:
			behavior = behaviors.get(IBehaviorConstants.WALKFORWARD);
			behavior.setIntensity(0.0f);
		}

		return behavior;
	}

	/**
	 * @param currentAction the currentAction to set
	 */
	public void setCurrentAction(ActionEnum currentAction)
	{
		this.currentAction = currentAction;
	}

	/**
	 * @return the currentAction
	 */
	public ActionEnum getCurrentAction()
	{
		return currentAction;
	}
}
