/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.decision.simple;

import java.util.Map;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.complex.RunToPosition;
import magma.agent.belief.IBelief;
import magma.agent.belief.IBeliefConstants;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Simple decision maker for primitive field player
 * 
 * @author Klaus Dorer
 */
public class SoccerDecisionMaker extends DecisionMakerBase
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** The version of the soccer server used */
	protected int serverVersion;

	/**
	 * Default constructor
	 * @param believes map of believes we can decide on
	 * @param behaviors map of behaviors we can decide for
	 * @param serverVersion the version of the soccer server used
	 */
	public SoccerDecisionMaker(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors, IThoughtModel thoughtModel,
			int serverVersion)
	{
		super(believes, behaviors, thoughtModel);
		this.serverVersion = serverVersion;
	}

	@Override
	public IBehavior decideSoccer()
	{
		performFocusBall();
		performSay();

		IBehavior toExecute;

		// temperature cool down
		toExecute = temperatureCoolDown();
		if (toExecute != null) {
			return toExecute;
		}

		// dance or cry
		toExecute = gameEndedAction();
		if (toExecute != null) {
			return toExecute;
		}

		// beam
		toExecute = beamHome();
		if (toExecute != null) {
			return toExecute;
		}

		toExecute = getUp();
		if (toExecute != null) {
			return toExecute;
		}

		// get ready
		toExecute = getReady();
		if (toExecute != null) {
			return toExecute;
		}

		// wait till game starts
		toExecute = waitForGameStart();
		if (toExecute != null) {
			return toExecute;
		}

		// wait till opponent has restarted the game
		toExecute = waitForOpponentActions();
		if (toExecute != null) {
			return toExecute;
		}

		// own kickoff action
		toExecute = ownKickOffActions();
		if (toExecute != null) {
			return toExecute;
		}

		// search ball
		toExecute = doSearchBall();
		if (toExecute != null) {
			return toExecute;
		}

		// kick
		toExecute = doKickBehavior();
		if (toExecute != null) {
			return toExecute;
		}

		// walk
		toExecute = movement();
		if (toExecute != null) {
			return toExecute;
		}

		// stop any running motors and do nothing
		return behaviors.get(IBehaviorConstants.NONE);
	}

	/**
	 * Tries to keep the ball in the visible area.
	 */
	protected void performFocusBall()
	{
		behaviors.get(IBehaviorConstants.FOCUS_BALL).perform();
	}

	/**
	 * Say something.
	 */
	protected void performSay()
	{
		behaviors.get(IBehaviorConstants.SAY_POSITIONS_BEHAVIOR).perform();
	}

	/**
	 * Checks if the temperature of the motors is too high and the robot should
	 * cool down.
	 * 
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior temperatureCoolDown()
	{
		IBelief tooHot = believes.get(IBeliefConstants.TEMPERATURE_HOT);
		if (tooHot.getTruthValue() > 0.7) {
			return behaviors.get(IBehaviorConstants.SHUT_OFF);
			// return behaviors.get(IBehaviorConstants.MOVE_ZERO);
		}
		if (tooHot.getTruthValue() > 0.4) {
			return behaviors.get(IBehaviorConstants.MOVE_ZERO);
		}
		return null;
	}

	/**
	 * Called to decide if we should beam us to our home position
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior beamHome()
	{
		IBelief beamTime = believes.get(IBeliefConstants.BEAM_TIME);
		if (beamTime.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.BEAM_HOME);
		}
		return null;
	}

	protected IBehavior gameEndedAction()
	{
		IBelief gameRunning = believes.get(IBeliefConstants.GAME_RUNNING);
		if (gameRunning.getTruthValue() < 0.5) {
			if (believes.get(IBeliefConstants.WIN).getTruthValue() > 0.5) {
				return behaviors.get(IBehaviorConstants.WIN);
			}
			if (believes.get(IBeliefConstants.LOOSE).getTruthValue() > 0.5) {
				return behaviors.get(IBehaviorConstants.LOOSE);
			}

		}

		return null;
	}

	/**
	 * Called to decide if we should wait for game start or not. Default
	 * implementation waits.
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior waitForGameStart()
	{
		IBelief gameRunning = believes.get(IBeliefConstants.GAME_RUNNING);
		if (gameRunning.getTruthValue() < 0.5) {
			return behaviors.get(IBehaviorConstants.NONE);
		}

		return null;
	}

	/**
	 * Called to decide if we should wait for opponent actions or not. Default
	 * implementation waits only on OpponentKickOff.
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior waitForOpponentActions()
	{
		IBelief opponentKickOff = believes
				.get(IBeliefConstants.OPPONENT_KICK_OFF);
		if (opponentKickOff.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.NONE);
		}

		return null;
	}

	/**
	 * Called to decide if we should stand up from laying on ground
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior getUp()
	{
		// get up back
		IBelief onBack = believes.get(IBeliefConstants.LAY_ON_BACK);
		if (onBack.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.GET_UP_BACK);
		}

		// get up front
		IBelief onFront = believes.get(IBeliefConstants.LAY_ON_FRONT);
		if (onFront.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.GET_UP_FRONT);
		}

		// get up side
		IBelief onSide = believes.get(IBeliefConstants.LAY_ON_SIDE);
		if (onSide.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.MOVE_ARM_TO_FALL_BACK);
		}

		return null;
	}

	/**
	 * Called to decide if we should get into ready position
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior getReady()
	{
		IBelief soccerPosition = believes.get(IBeliefConstants.SOCCER_POSITION);
		if (soccerPosition.getTruthValue() < 0.5) {
			return behaviors.get(IBehaviorConstants.GET_READY);
		}
		return null;
	}

	/**
	 * @return
	 */
	protected IBehavior ownKickOffActions()
	{
		IBelief ownKickoff = believes.get(IBeliefConstants.OWN_KICK_OFF);
		if (ownKickoff.getTruthValue() < 0.5) {
			return null;
		}

		IBelief teammateBetter = believes
				.get(IBeliefConstants.TEAMMATE_CLOSER_TO_BALL);
		if (teammateBetter.getTruthValue() < 0.5) {
			return behaviors.get(IBehaviorConstants.ATTACK_KICKOFF);
		}
		return null;
	}

	protected IBehavior doSearchBall()
	{
		IBelief seeBall = believes.get(IBeliefConstants.SEE_BALL);
		if (seeBall.getTruthValue() < 0.5) {
			return behaviors.get(IBehaviorConstants.SEARCH_BALL);
		}
		return null;
	}

	/**
	 * Called to decide if we should kick the ball
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior doKickBehavior()
	{
		IBelief ballKickable = believes.get(IBeliefConstants.BALL_KICKABLE);
		if (ballKickable.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.SHOOT_TO_GOAL);
		}
		return null;
	}

	/**
	 * Called to decide if movement is necessary
	 * @return walk behavior if no team mate is closer to ball
	 */
	protected IBehavior movement()
	{
		// IBelief teammateCloser =
		// believes.get(IBelief.TEAMMATE_CLOSER_TO_BALL);
		IBelief teammateBetter = believes
				.get(IBeliefConstants.TEAMMATE_CLOSER_TO_BALL);
		if (teammateBetter.getTruthValue() < 0.5) {
			// we want to go to the ball
			RunToPosition run = (RunToPosition) behaviors
					.get(IBehaviorConstants.RUN_TO_POSITION);
			Vector3D position = thoughtModel.getWorldModel().getBall()
					.getPosition();
			Vector3D destination = new Vector3D(position.getX() - 0.18,
					position.getY(), position.getZ());
			run.setPosition(new Pose2D(destination, Angle.ZERO), 30, 0, 100, false);
			return run;
		}

		return behaviors.get(IBehaviorConstants.MOVE_WITH_BALL);
	}

	/**
	 * @return the serverVersion
	 */
	public int getServerVersion()
	{
		return serverVersion;
	}
}
