/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.general;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;

import magma.agent.action.IAction;
import magma.agent.action.IActionPerformer;
import magma.agent.action.impl.Action;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.ik.IAgentIKSolver;
import magma.agent.agentmodel.ik.impl.JacobianTransposeAgentIKSolver;
import magma.agent.agentmodel.impl.AgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.BeamBehavior;
import magma.agent.behavior.basic.BeamHomeBehavior;
import magma.agent.behavior.basic.NoneBehavior;
import magma.agent.behavior.basic.SayPositionsBehavior;
import magma.agent.behavior.complex.GoalieStrikePositioning;
import magma.agent.behavior.complex.RemoteCommandBehavior;
import magma.agent.behavior.complex.RemoteControllBehavior;
import magma.agent.behavior.motor.MotorFileBehavior;
import magma.agent.behavior.motor.ReadCSV;
import magma.agent.behavior.supportPoint.FunctionBehavior;
import magma.agent.belief.IBelief;
import magma.agent.belief.IBeliefConstants;
import magma.agent.belief.impl.BallEndsUpInOwnGoal;
import magma.agent.belief.impl.BallInCriticalArea;
import magma.agent.belief.impl.BallInPenaltyArea;
import magma.agent.belief.impl.BallKickable;
import magma.agent.belief.impl.BeamTime;
import magma.agent.belief.impl.FacingAdvantageousKickDirection;
import magma.agent.belief.impl.FacingGoalKickDirection;
import magma.agent.belief.impl.FacingKickDirection;
import magma.agent.belief.impl.GameRunning;
import magma.agent.belief.impl.GameStateBelief;
import magma.agent.belief.impl.GoalieShouldKick;
import magma.agent.belief.impl.LayOnBack;
import magma.agent.belief.impl.LayOnFront;
import magma.agent.belief.impl.LayOnSide;
import magma.agent.belief.impl.Lose;
import magma.agent.belief.impl.OpponentNearBall;
import magma.agent.belief.impl.OpponentWillKick;
import magma.agent.belief.impl.PlayDefense;
import magma.agent.belief.impl.PlayOffense;
import magma.agent.belief.impl.SeeBall;
import magma.agent.belief.impl.ShouldWeDribble;
import magma.agent.belief.impl.SimpleFlagBelief;
import magma.agent.belief.impl.TeammateCloserToBall;
import magma.agent.belief.impl.TemperatureHot;
import magma.agent.belief.impl.Win;
import magma.agent.communication.IChannelManager;
import magma.agent.communication.base.ChannelParameters;
import magma.agent.decision.IDecisionMaker;
import magma.agent.decision.simple.DecisionMakerBase;
import magma.agent.decision.simple.DoNothingDecisionMaker;
import magma.agent.decision.simple.GoalieDecisionMaker;
import magma.agent.decision.simple.GoaliePenaltyDecisionMaker;
import magma.agent.decision.simple.PenaltyDecisionMaker;
import magma.agent.decision.simple.RunChallengeDecisionMaker;
import magma.agent.decision.simple.SoccerDecisionMaker;
import magma.agent.decision.simple.TrainingDecisionMaker;
import magma.agent.flags.Flag;
import magma.agent.flags.IFlagModel;
import magma.agent.flags.impl.FlagModel;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.server.IRCServerMetaModel;
import magma.agent.meta.server.impl.RCServerMetaModelV62;
import magma.agent.meta.server.impl.RCServerMetaModelV63;
import magma.agent.meta.server.impl.RCServerMetaModelV64;
import magma.agent.meta.server.impl.RCServerMetaModelV66;
import magma.agent.perception.IPerception;
import magma.agent.perception.impl.Perception;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.thoughtmodel.impl.ThoughtModel;
import magma.agent.thoughtmodel.strategy.IRoleManager;
import magma.agent.thoughtmodel.strategy.ITeamStrategy;
import magma.agent.thoughtmodel.strategy.impl.IRoleConstants;
import magma.agent.thoughtmodel.strategy.impl.RoleManager;
import magma.agent.thoughtmodel.strategy.impl.strategies.AcceptorStrategy;
import magma.agent.worldmodel.GameState;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.impl.GlobalMap;
import magma.agent.worldmodel.localizer.ILocalizer;
import magma.agent.worldmodel.localizer.impl.CompositeLocalizer;
import magma.agent.worldmodel.localizer.impl.LocalizerFieldNormal;
import magma.agent.worldmodel.localizer.impl.LocalizerGyro;
import magma.robots.nao.behavior.movement.KickFastBehaviorParameter;
import magma.robots.nao.behavior.movement.KickStraightOneLegBehaviorParameter;
import magma.util.parameter.ParameterMap;

/**
 * Factory class for the components of our RoboCup player
 * 
 * @author Klaus Dorer
 */
public abstract class ComponentFactory
{
	/**
	 * Create a agent meta model
	 * 
	 * @return the agent meta model
	 */
	public abstract IAgentMetaModel createAgentMetaModel();

	/**
	 * @return the channel manager including the channels
	 */
	public abstract IChannelManager createChannelManager(ChannelParameters info);

	/**
	 * Create a agent IK-solver.
	 * 
	 * @return the agent IK-solver
	 */
	public IAgentIKSolver createAgentIKSolver()
	{
		return new JacobianTransposeAgentIKSolver();
	}

	/**
	 * Create a new rc-server meta model based on serverVersion.
	 * 
	 * @param serverVerion - the version of the meta model
	 * @return the rc-server meta model
	 */
	public IRCServerMetaModel createRCServerMetaModel(int serverVerion)
	{
		if (serverVerion < 62) {
			serverVerion = 62;
		}

		switch (serverVerion) {
		case 62:
			return RCServerMetaModelV62.INSTANCE;
		case 63:
			return RCServerMetaModelV63.INSTANCE;
		case 64:
		case 65:
			return RCServerMetaModelV64.INSTANCE;
		case 66:
		default:
			return RCServerMetaModelV66.INSTANCE;
		}
	}

	/**
	 * Create a Perception object
	 * 
	 * @return New Perception
	 */
	public IPerception createPerception(IAgentMetaModel metaModel)
	{
		return new Perception();
	}

	/**
	 * Create an new Action
	 * 
	 * @param actionPerformer Server connection
	 * @param scene Scene name
	 * @param teamname Team name
	 * @param playerNumber Player number
	 * @return New Action object
	 */
	public IAction createAction(IActionPerformer actionPerformer,
			IAgentMetaModel metaModel)
	{
		IAction action = new Action(actionPerformer);
		action.init(metaModel);
		return action;
	}

	/**
	 * Create a new AgentModel
	 * 
	 * @return New AgentModel object
	 */
	public IAgentModel createAgentModel(IAction action, IAgentMetaModel metaModel)
	{
		return new AgentModel(action, metaModel, createAgentIKSolver(),
				new double[] { 0.20, 0.10, 0.18, 0.03 });
	}

	/**
	 * Create a new GlobalMap
	 * 
	 * @param agentModel Reference to the AgentModel
	 * @param teamname Team name
	 * @param playerNumber Player number
	 * @param serverVersion Currently used server version
	 * @return New GlobalMap object
	 */
	public IGlobalMap createWorldModel(IAgentModel agentModel,
			IRCServerMetaModel metaModel, String teamname, int playerNumber)
	{
		return new GlobalMap(agentModel, createLocalizer(), metaModel, teamname,
				playerNumber);
	}

	public IThoughtModel createThoughtModel(IAgentModel agentModel,
			IGlobalMap worldModel, IFlagModel flags)
	{
		ThoughtModel thoughtModel = new ThoughtModel(agentModel, worldModel,
				flags);

		// strategy = new DefaultStrategy(worldModel);
		ITeamStrategy strategy = new AcceptorStrategy(worldModel);
		IRoleManager roleManager = new RoleManager(worldModel, strategy);
		thoughtModel.setRoleManager(roleManager, IRoleConstants.DUMMY_ROLE);

		return thoughtModel;
	}

	/**
	 * Create a new Localizer
	 * 
	 * @return New Localizer
	 */
	protected ILocalizer createLocalizer()
	{
		ILocalizer[] localizer = new ILocalizer[2];

		// Note: the order of localizer is important! I should start with the
		// localizer needing the highest number of visible reference points and
		// end with the localizer, which needs the least amount of visible
		// reference points.
		localizer[0] = new LocalizerFieldNormal();
		localizer[1] = new LocalizerGyro();

		return new CompositeLocalizer(localizer);
	}

	/**
	 * Create a new model holding flags.
	 * 
	 * @return New flags model.
	 */
	public IFlagModel createFlags()
	{
		return new FlagModel();
	}

	/**
	 * Create a new DecisionMaker
	 * 
	 * @param believes Agent believes
	 * @param behaviors Agent behaviors
	 * @param playerNumber Player number
	 * @param serverVersion Server version
	 * @param decisionMakerID 0 (default) if we decide based on player number, 1
	 *        for goalie, 2 for field player, 3 for simple decision maker, 4 for
	 *        Penalty decision maker, 5 for do nothing decision maker, 6 for
	 *        training decision maker, 7 for goalie penalty decision maker, 8 for
	 *        walk demo decision maker, 9 for TDM2, 10 for run challenge decision
	 *        maker
	 * @param learningParam used for parameterization learning
	 * @return new DecisionMaker object
	 */
	public IDecisionMaker createDecisionMaker(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors, IThoughtModel thoughtModel,
			int playerNumber, int serverVersion, int decisionMakerID,
			ParameterMap learningParam)
	{
		DecisionMakerBase result;

		switch (decisionMakerID) {
		case 1:
			result = new GoalieDecisionMaker(believes, behaviors, thoughtModel,
					serverVersion);
			break;

		case 2:
			result = new SoccerDecisionMaker(believes, behaviors, thoughtModel,
					serverVersion);
			break;

		case 4:
			result = new PenaltyDecisionMaker(believes, behaviors, thoughtModel,
					serverVersion);
			break;

		case 5:
			result = new DoNothingDecisionMaker(believes, behaviors, thoughtModel);
			break;

		case 6:
			result = new TrainingDecisionMaker(believes, behaviors, thoughtModel,
					serverVersion);
			break;

		case 7:
			result = new GoaliePenaltyDecisionMaker(believes, behaviors,
					thoughtModel, serverVersion);
			break;

		case 10:
			result = new RunChallengeDecisionMaker(believes, behaviors,
					thoughtModel, serverVersion);
			break;

		default:
			if (playerNumber == 1) {
				result = new GoalieDecisionMaker(believes, behaviors, thoughtModel,
						serverVersion);
			} else {
				result = new SoccerDecisionMaker(believes, behaviors, thoughtModel,
						serverVersion);
			}
		}

		return result;

	}

	/**
	 * Creates all believe objects
	 * 
	 * @param thoughtModel Thought model
	 * @return A map with all believes that may be used
	 */
	public Map<String, IBelief> createBeliefs(IThoughtModel thoughtModel)
	{
		Map<String, IBelief> beliefs = new HashMap<String, IBelief>();

		// General beliefs
		IBelief belief = new GameRunning(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new BeamTime(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new Win(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new Lose(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new PlayOffense(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new PlayDefense(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new TemperatureHot(thoughtModel, 90, 110);
		beliefs.put(belief.getName(), belief);

		// Player beliefs
		belief = new TeammateCloserToBall(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new OpponentNearBall(thoughtModel);
		beliefs.put(belief.getName(), belief);

		// Kick beliefs
		belief = new BallKickable(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new FacingKickDirection(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new FacingAdvantageousKickDirection(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new FacingGoalKickDirection(thoughtModel);
		beliefs.put(belief.getName(), belief);

		// Laying beliefs
		belief = new LayOnBack(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new LayOnFront(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new LayOnSide(thoughtModel);
		beliefs.put(belief.getName(), belief);

		// GameState beliefs
		belief = new GameStateBelief(IBeliefConstants.OWN_CORNER_KICK,
				thoughtModel, GameState.OWN_CORNER_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_FREE_KICK,
				thoughtModel, GameState.OWN_FREE_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_GOAL_KICK,
				thoughtModel, GameState.OWN_GOAL_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_KICK_IN, thoughtModel,
				GameState.OWN_KICK_IN);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_KICK_OFF, thoughtModel,
				GameState.OWN_KICK_OFF);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_CORNER_KICK,
				thoughtModel, GameState.OPPONENT_CORNER_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_FREE_KICK,
				thoughtModel, GameState.OPPONENT_FREE_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_GOAL_KICK,
				thoughtModel, GameState.OPPONENT_GOAL_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_KICK_IN,
				thoughtModel, GameState.OPPONENT_KICK_IN);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_KICK_OFF,
				thoughtModel, GameState.OPPONENT_KICK_OFF);
		beliefs.put(belief.getName(), belief);

		// Flag beliefs
		belief = new SimpleFlagBelief(IBeliefConstants.PAUSING, thoughtModel,
				Flag.PAUSE);
		beliefs.put(belief.getName(), belief);

		belief = new SimpleFlagBelief(IBeliefConstants.SOFT_PAUSING,
				thoughtModel, Flag.SOFT_PAUSE);
		beliefs.put(belief.getName(), belief);

		belief = new SimpleFlagBelief(IBeliefConstants.PENALTY_SHOOT_OUT,
				thoughtModel, Flag.PENALTY);
		beliefs.put(belief.getName(), belief);

		// Goalie beliefs
		belief = new BallInCriticalArea(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new OpponentWillKick(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new BallInPenaltyArea(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new BallEndsUpInOwnGoal(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new ShouldWeDribble(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new GoalieShouldKick(thoughtModel);
		beliefs.put(belief.getName(), belief);

		belief = new SeeBall(thoughtModel);
		beliefs.put(belief.getName(), belief);

		// Robot specific beliefs
		createSpecificBeliefs(thoughtModel, beliefs);

		return beliefs;
	}

	/**
	 * Creates all specific belief objects
	 * 
	 * @param thoughtModel - the thought model
	 * @param beliefs - a map for storing the created specific beliefs
	 */
	protected abstract void createSpecificBeliefs(IThoughtModel thoughtModel,
			Map<String, IBelief> beliefs);

	/**
	 * Create all behavior objects that are used during the game
	 * 
	 * @param thouhgtModel Reference to the ThoughtModel
	 * @param coordinator Reference to the Coordinator
	 * @param params Behavior parameterization, null if default should be used
	 * @return A map of available behaviors
	 */
	public Map<String, IBehavior> createBehaviors(IThoughtModel thoughtModel,
			ParameterMap params)
	{
		Map<String, IBehavior> behaviors = new HashMap<String, IBehavior>();

		IBehavior behavior = new NoneBehavior();
		behaviors.put(behavior.getName(), behavior);

		behavior = new BeamHomeBehavior(IBehaviorConstants.BEAM_HOME,
				thoughtModel, 0.26f);
		behaviors.put(behavior.getName(), behavior);

		behavior = new BeamBehavior(IBehaviorConstants.BEAM_TO_POSITION,
				thoughtModel, 0.26f);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SayPositionsBehavior(
				IBehaviorConstants.SAY_POSITIONS_BEHAVIOR, thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		readFileBehaviors(getBehaviorFilesBasePath()
				+ "BehaviorValues/UsedBehaviors/", false, thoughtModel, params,
				behaviors);

		readFileBehaviors(getBehaviorFilesBasePath()
				+ "FunctionBehaviors/UsedBehaviors/", true, thoughtModel, params,
				behaviors);

		// the new behavior by Klaus & Julian
		behavior = new GoalieStrikePositioning(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new RemoteControllBehavior(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new RemoteCommandBehavior(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		createSpecificBehaviors(thoughtModel, params, behaviors);

		return behaviors;
	}

	/**
	 * Return the path to the behavior files base folder. This method is used to
	 * read dynamically all used behaviors of the given path.<br>
	 * The following folder structure is expected:<br>
	 * <i>&lt;path&gt; -> BehaviorValues -> UsedBehaviors</i><br>
	 * for CSV behavior files and:<br>
	 * <i>&lt;path&gt; -> FunctionBehaviors -> UsedBehaviors</i><br>
	 * for function behavior files.
	 * 
	 * @return the path to the behavior files base folder
	 */
	protected abstract String getBehaviorFilesBasePath();

	/**
	 * Create all specific behavior objects.
	 * 
	 * @param thoughtModel - the thought model
	 * @param params - behavior parameterization, null if default should be used
	 * @param behaviors - a map for storing the created specific behaviors
	 */
	protected abstract void createSpecificBehaviors(IThoughtModel thoughtModel,
			ParameterMap params, Map<String, IBehavior> behaviors);

	/**
	 * Reads either CSV- or function behavior files, creates corresponding
	 * objects and adds them to the map of behaviors.
	 * 
	 * @param agentModel - the agent model
	 * @param worldModel - the world model
	 * @param params - behavior parameterization, null if default should be used
	 * @param behaviors - a map for storing the created behaviors
	 * @throws IOException
	 */
	protected void readFileBehaviors(String pathname, boolean functionBehavior,
			IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors)
	{
		try {
			String[] files = ReadCSV.getResourceListing(pathname);

			for (String fileName : files) {
				// to avoid .svn file
				if (fileName.endsWith(".txt")) {
					String filePath = pathname + fileName;
					ReadCSV csvHeaderReader = new ReadCSV(filePath);

					if (functionBehavior) {
						String name = csvHeaderReader.getFunctionBehaviorName();
						behaviors.put(name, new FunctionBehavior(name, thoughtModel,
								filePath, csvHeaderReader.getFloatSlowDown(),
								csvHeaderReader.getFloatStartSlowDown()));
					} else {
						String name = csvHeaderReader.getBehaviorValuesName();
						behaviors.put(name, new MotorFileBehavior(name, thoughtModel,
								params, filePath, csvHeaderReader.getIntSlowDown(),
								csvHeaderReader.getIntStartSlowDown()));
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("Rule file not found" + pathname);
		} catch (URISyntaxException e) {
			e.printStackTrace();
			System.out.println("URI exception when reading: " + pathname);
		}
	}

	/**
	 * Creates behavior or decision maker parameters. fromExtern parameters
	 * overwrite default parameters.
	 * @param fromExtern parametrization from an external source like learning
	 * @return a parameter map that has to contain all ParameterLists needed for
	 *         default behaviors
	 */
	public ParameterMap createParameters(ParameterMap fromExtern)
	{
		ParameterMap result = new ParameterMap();
		result.put(IBehaviorConstants.DYNAMIC_KICK_LEFT,
				KickStraightOneLegBehaviorParameter.getInstance());
		result.put(IBehaviorConstants.DYNAMIC_KICK_RIGHT,
				KickStraightOneLegBehaviorParameter.getInstance());

		result.put(IBehaviorConstants.SHOOT_STRAIGHT,
				KickFastBehaviorParameter.getInstance());

		// specific parameters have higher priority than general and are put after
		// them in order to allow overwriting
		result.putAll(createSpecificParameters());

		// external parameters (e.g. from learning) have higher priority and are
		// therefore overwriting parameters defined here
		result.putAll(fromExtern);
		return result;
	}

	/**
	 * Overwrite this to create robot specific parametrization
	 * @param fromExtern
	 * @return
	 */
	protected abstract ParameterMap createSpecificParameters();

}
