/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.general;

import java.util.logging.Level;

import magma.agent.communication.base.ChannelParameters;
import magma.util.parameter.ParameterMap;

/**
 * Stores all necessary plaser parameters (Server connection, decision maker
 * etc.)#
 * 
 * @author Klaus Dorer
 */
public class PlayerParameters
{
	private final ChannelParameters channelParams;

	private final ParameterMap parameterMap;

	private final Level level;

	private final int serverVersion;

	private final int decisionMakerID;

	final ComponentFactory factory;

	/**
	 * Instantiates and initializes a new PlayerParameters object
	 * 
	 * @param teamname Team name
	 * @param playerNumber Player number
	 * @param host Server host
	 * @param port Server port
	 * @param behaviorParams Behavior parameters
	 * @param level Logging level
	 * @param serverVersion Server version
	 * @param factory Reference to the component factory
	 * @param decisionMakerID Index of the decision maker
	 */
	public PlayerParameters(String teamname, byte teamID, int playerNumber,
			String host, int port, ParameterMap behaviorParams, Level level,
			int serverVersion, ComponentFactory factory, int decisionMakerID)
	{
		this(new ChannelParameters(teamname, teamID, playerNumber, host, port),
				behaviorParams, level, serverVersion, factory, decisionMakerID);
	}

	public PlayerParameters(ChannelParameters channelParams,
			ParameterMap behaviorParams, Level level, int serverVersion,
			ComponentFactory factory, int decisionMakerID)
	{
		this.channelParams = channelParams;
		if (behaviorParams == null) {
			behaviorParams = new ParameterMap();
		}
		this.parameterMap = behaviorParams;
		this.level = level;
		this.serverVersion = serverVersion;
		this.factory = factory;
		this.decisionMakerID = decisionMakerID;
	}

	/**
	 * @return the parameters that are used to setup channels
	 */
	public ChannelParameters getChannelParams()
	{
		return channelParams;
	}

	/**
	 * Retrieve the team name
	 * 
	 * @return Team name
	 */
	public String getTeamname()
	{
		return channelParams.getTeamname();
	}

	/**
	 * @return
	 */
	public byte getTeamID()
	{
		return channelParams.getTeamID();
	}

	/**
	 * Retrieve the player number
	 * 
	 * @return Player number
	 */
	public int getPlayerNumber()
	{
		return channelParams.getPlayerNumber();
	}

	/**
	 * Retrieve the server host
	 * 
	 * @return Server host string
	 */
	public String getHost()
	{
		return channelParams.getHost();
	}

	/**
	 * Retrieve the server port
	 * 
	 * @return Server port
	 */
	public int getPort()
	{
		return channelParams.getPort();
	}

	/**
	 * Retrieve the behavior parameters
	 * 
	 * @return Behavior parameters
	 */
	public ParameterMap getParameterMap()
	{
		return parameterMap;
	}

	/**
	 * Retrieve the logging level
	 * 
	 * @return Logging level
	 */
	public Level getLevel()
	{
		return level;
	}

	/**
	 * Retrieve the server version
	 * 
	 * @return Server version
	 */
	public int getServerVersion()
	{
		return serverVersion;
	}

	/**
	 * Retrieve the reference to the component factory
	 * 
	 * @return Component factory reference
	 */
	public ComponentFactory getComponentFactory()
	{
		return factory;
	}

	/**
	 * Retrieve the decision maker index
	 * 
	 * @return Decision maker index
	 */
	public int getDecisionMakerID()
	{
		return decisionMakerID;
	}
}