/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.meta.agent;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

public interface IAgentMetaModel extends Serializable
{
	/**
	 * Retrieve the name of this meta model.
	 * 
	 * @return the name of this model
	 */
	String getName();

	/**
	 * Returns the corresponding action scene string, to this meta model.
	 * 
	 * @return the scene-string
	 */
	public String getSceneString();

	/**
	 * Returns the static pivot-point used as replacement of the CoM in the
	 * balancing engine related movements. This pivot-point is usually close to
	 * the initial CoM location or somewhere in the pelvis of the robot.
	 * 
	 * @return the static pivot-point
	 */
	public Vector3D getStaticPivotPoint();

	/**
	 * Returns a list of body part configurations, specifying the different body
	 * parts of the agent and their related sensors/effectors.
	 * 
	 * @return the list of body part configurations
	 */
	public List<IBodyPartConfiguration> getBodyPartConfiguration();

	/**
	 * Returns the name of the body part containing the camera. This method is
	 * used to dynamically relate the vision to the system of the root-body.
	 * 
	 * @return the name of the body part containing the camera
	 */
	public String getNameOfCameraBodyPart();

	/**
	 * Returns an unmodifiable collection of available joints of the robot
	 * described by this meta model.
	 * 
	 * @return an unmodifiable collection of available joints of the robot model
	 */
	public Collection<ISensorConfiguration> getAvailableJoints();

	/**
	 * Returns a list of child body part configurations to the given body part
	 * configuration. This method is useful when browsing the robot model.
	 * 
	 * @param bodyPart the parent body part configuration
	 * @return a list of child body part configurations
	 */
	public List<IBodyPartConfiguration> getChildBodyConfigurations(
			IBodyPartConfiguration bodyPart);

	/**
	 * Returns the root body configuration. The root body configuration is the
	 * only body part configuration which has no parent body part.
	 * 
	 * @return the root body configuration
	 */
	public IBodyPartConfiguration getRootBodyConfiguration();

	/**
	 * Returns a string array containing the names of all available joints of the
	 * robot described by this meta model.
	 * 
	 * @return an array of names to available joints of the robot model
	 */
	public String[] getAvailableJointNames();

	/**
	 * @return the names of available effectors
	 */
	public String[] getAvailableEffectorNames();
}
