/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.meta.agent;

import java.io.Serializable;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

public interface IBodyPartConfiguration extends Serializable
{
	/**
	 * Returns the name of the body part.
	 * 
	 * @return the name of the body part
	 */
	public String getName();

	/**
	 * Returns the name of the parent body part.
	 * 
	 * @return the name of the parent body part - or null if there is no parent
	 *         body part.
	 */
	public String getParent();

	/**
	 * Returns a vector that describes the position of the body part relative to
	 * it's parent.
	 * 
	 * @return the body translation (meter) - or null if there is no parent body
	 *         part.
	 */
	public Vector3D getTranslation();

	/**
	 * Retruns the mass of the body part.
	 * 
	 * @return the body mass (kg)
	 */
	public float getMass();

	/**
	 * Returns a vector, containing the geometry of the body part.
	 * 
	 * @return the body geometry (meter)
	 */
	public Vector3D getGeometry();

	/**
	 * Returns the JointConfiguration object to that body part.
	 * 
	 * @return the joint configuration - or null if there is no joint attached
	 */
	public ISensorConfiguration getJointConfiguration();

	/**
	 * Returns the ancher of the joint, relative to the body part.
	 * 
	 * @return the ancher of the joint - or null if there is no joint attached
	 */
	public Vector3D getJointAnchor();

	/**
	 * Returns the configuration of the GyroRate perceptor inside the body part.
	 * 
	 * @return the configuration of the gyro-rate perceptor - or null if the body
	 *         contains no such perceptor
	 */
	public ISensorConfiguration getGyroRateConfiguration();

	/**
	 * Returns the configuration of the Accelerometer inside the body part.
	 * 
	 * @return the configuration of the accelerometer - or null if the body
	 *         contains no such perceptor
	 */
	public ISensorConfiguration getAccelerometerConfiguration();

	/**
	 * Returns the configuration of the ForceResistance percaptor inside the body
	 * part.
	 * 
	 * @return the configuration of the force-resistance perceptor - or null if
	 *         the body contains no such perceptor
	 */
	public ISensorConfiguration getForceResistanceConfiguration();

	/**
	 * Returns the configuration of the CompassConfig percaptor inside the body
	 * part.
	 * 
	 * @return the configuration of the force-resistance perceptor - or null if
	 *         the body contains no such perceptor
	 */
	public ISensorConfiguration getCompassConfig();
}
