/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.meta.agent.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.ICompositeJointConfiguration;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.meta.agent.ISensorConfiguration;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Agent meta model class
 * 
 * @author Stefan Glaser
 */
public class AgentMetaModel implements IAgentMetaModel
{
	private static final long serialVersionUID = 3424907664438065954L;

	protected final String modelName;

	protected final String sceneString;

	protected List<IBodyPartConfiguration> bodyPartConfig;

	protected final String bodyPartContainingCamera;

	protected Collection<ISensorConfiguration> jointConfigs;

	protected String[] jointNames;

	protected String[] effectorNames;

	protected final Vector3D staticPivotPoint;

	/**
	 * Constructor.
	 * 
	 * @param modelName - the name of this model
	 * @param sceneString - the scene string
	 * @param staticPivotPoint - the static pivot-point related to this robot
	 * @param bodyPartConfig - a list of body parts configurations
	 * @param bodyPartContainingCamera - the name of the body part containing the
	 *        camera
	 */
	public AgentMetaModel(String modelName, String sceneString,
			Vector3D staticPivotPoint, String bodyPartContainingCamera)
	{
		this.modelName = modelName;
		this.sceneString = sceneString;
		this.staticPivotPoint = staticPivotPoint;
		this.bodyPartContainingCamera = bodyPartContainingCamera;
	}

	@Override
	public String getName()
	{
		return modelName;
	}

	@Override
	public String getSceneString()
	{
		return sceneString;
	}

	@Override
	public Vector3D getStaticPivotPoint()
	{
		return staticPivotPoint;
	}

	@Override
	public List<IBodyPartConfiguration> getBodyPartConfiguration()
	{
		return bodyPartConfig;
	}

	public void setBodyPartConfiguration(List<IBodyPartConfiguration> config)
	{
		bodyPartConfig = config;
	}

	@Override
	public String getNameOfCameraBodyPart()
	{
		return bodyPartContainingCamera;
	}

	@Override
	public List<IBodyPartConfiguration> getChildBodyConfigurations(
			IBodyPartConfiguration bodyPart)
	{
		List<IBodyPartConfiguration> childBodies = new ArrayList<IBodyPartConfiguration>();

		if (bodyPart != null) {
			String parentName = bodyPart.getName();

			for (int i = 0; i < bodyPartConfig.size(); i++) {
				if (parentName.equals(bodyPartConfig.get(i).getParent())) {
					childBodies.add(bodyPartConfig.get(i));
				}
			}
		}

		return childBodies;
	}

	@Override
	public IBodyPartConfiguration getRootBodyConfiguration()
	{
		for (int i = 0; i < bodyPartConfig.size(); i++) {
			if (bodyPartConfig.get(i).getParent() == null) {
				return bodyPartConfig.get(i);
			}
		}

		return null;
	}

	@Override
	public Collection<ISensorConfiguration> getAvailableJoints()
	{
		if (jointConfigs == null) {
			ArrayList<ISensorConfiguration> jointConfigsList = new ArrayList<ISensorConfiguration>();

			// Extract joints
			ISensorConfiguration jointConfig;
			for (IBodyPartConfiguration config : bodyPartConfig) {
				jointConfig = config.getJointConfiguration();
				if (jointConfig != null) {
					if (jointConfig instanceof ICompositeJointConfiguration) {
						for (ISensorConfiguration conf : ((ICompositeJointConfiguration) jointConfig)
								.getHjConfigurations()) {
							jointConfigsList.add(conf);
						}
					} else {
						jointConfigsList.add(jointConfig);
					}
				}
			}
			jointConfigs = Collections.unmodifiableCollection(jointConfigsList);

			jointNames = new String[jointConfigsList.size()];
			effectorNames = new String[jointConfigsList.size()];
			for (int i = 0; i < jointConfigsList.size(); i++) {
				ISensorConfiguration config = jointConfigsList.get(i);
				jointNames[i] = config.getName();
				effectorNames[i] = ((IHingeJointConfiguration) config)
						.getEffectorName();
			}
		}

		return jointConfigs;
	}

	@Override
	public String[] getAvailableJointNames()
	{
		if (jointNames == null) {
			getAvailableJoints();
		}

		return jointNames;
	}

	@Override
	public String[] getAvailableEffectorNames()
	{
		if (effectorNames == null) {
			getAvailableJoints();
		}

		return effectorNames;
	}
}
