/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.meta.server.impl;

import magma.agent.meta.server.IHomePosition;
import magma.agent.meta.server.IRCServerMetaModel;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

public class RelativeHomePosition implements IHomePosition
{
	private float[] weights;

	private Vector3D offset;

	/**
	 * Specifies a relative home position. The relative home position is
	 * calculated by an x and y weight plus an offset. The weights are used as
	 * multipliers to the half field length and width of the RCServerMetaModel in
	 * use. The range of the weights is within -1 <= weight[0] <= 0 for the first
	 * weight (x-direction) and -1 <= weight[1] <= 1 for the second
	 * (y-direction). <br>
	 * <br>
	 * Examples: (weight[0] = x and weight[1] = y)<br>
	 * - [x->0, y->0] results in the field center.<br>
	 * - [x->-1, y->0] results in the center below the goal.<br>
	 * - [x->-0.5, y->1] results in the middle of our half on the upper field
	 * border line.<br>
	 * 
	 * <br>
	 * <br>
	 * The first two values of the offset is added to the via weights calculated
	 * position to the final position. The third value is directly interpreted as
	 * final turn angle.
	 * 
	 * @param weights
	 * @param offset
	 */
	public RelativeHomePosition(float[] weights, Vector3D offset)
	{
		this.weights = weights;
		this.offset = offset;

		if (this.weights[0] < -1) {
			this.weights[0] = -1;
		} else if (this.weights[0] > 0) {
			this.weights[0] = 0;
		}

		if (this.weights[1] < -1) {
			this.weights[1] = -1;
		} else if (this.weights[1] > 1) {
			this.weights[1] = 1;
		}
	}

	public float getXWeight()
	{
		return weights[0];
	}

	public float getYWeight()
	{
		return weights[1];
	}

	public Vector3D getOffset()
	{
		return offset;
	}

	@Override
	public Vector3D getHomePosition(IRCServerMetaModel metaModel)
	{
		// Use some initial value, that in an error case not all agents wants to
		// be beamed at Point (0, 0).
		double fieldLength = 10.5;
		double fieldWidht = 7;

		if (metaModel != null) {
			fieldLength = metaModel.getFieldDimensions()[0] / 2;
			fieldWidht = metaModel.getFieldDimensions()[1] / 2;
		}

		double x = fieldLength * weights[0] + offset.getX();
		double y = fieldWidht * weights[1] + offset.getY();

		return new Vector3D(x, y, offset.getZ());
	}

	@Override
	public Vector3D getHomePosition(float fieldHalfLength, float fieldHalfWidth)
	{
		double x = fieldHalfLength * weights[0] + offset.getX();
		double y = fieldHalfWidth * weights[1] + offset.getY();

		return new Vector3D(x, y, offset.getZ());
	}
}