/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.perception;

import java.util.List;
import java.util.Map;

/**
 * The perception represents all input an agent can get from the outside.
 * 
 * @author Simon Raffeiner
 * 
 */
public interface IPerception
{
	/** object name constants */
	String BALL = "B";

	String GOAL_LEFT_LEFTPOST = "G1L";

	String GOAL_LEFT_RIGHTPOST = "G2L";

	String GOAL_RIGHT_LEFTPOST = "G1R";

	String GOAL_RIGHT_RIGHTPOST = "G2R";

	/**
	 * Get a specific Universal Joint Perceptor
	 * 
	 * @param name Perceptor name
	 * @return perceptor
	 */
	ICompositeJointPerceptor getCompositeJointPerceptor(String name);

	/**
	 * Get a specific Universal Joint Perceptor
	 * 
	 * @param name Perceptor name
	 * @return perceptor
	 */
	IUniversalJointPerceptor getUniversalJointPerceptor(String name);

	/**
	 * Get a specific Hinge Joint Perceptor
	 * 
	 * @param name Perceptor name
	 * @return perceptor
	 */
	IHingeJointPerceptor getHingeJointPerceptor(String name);

	/**
	 * Get a specific Accelerometer Perceptor
	 * 
	 * @param name Perceptor name
	 * @return Acceleration perceptor
	 */
	IAccelerometerPerceptor getAccelerationPerceptor(String name);

	/**
	 * Get Agent State
	 * 
	 * @return AgentState perceptor
	 */
	IAgentStatePerceptor getAgentState();

	/**
	 * Get a specific Gyro Perceptor
	 * 
	 * @param name Perceptor name
	 * @return Gyro perceptor
	 */
	IGyroPerceptor getGyroRatePerceptor(String name);

	/**
	 * Get a specific Compass Perceptor
	 * 
	 * @param name Perceptor name
	 * @return Compass perceptor
	 */
	ICompassPerceptor getCompassPerceptor(String name);

	/**
	 * Get a specific Force Resistance Perceptor
	 * 
	 * @param name Perceptor name
	 * @return perceptor
	 */
	IForceResistancePerceptor getForceResistancePerceptor(String name);

	/**
	 * Get a specific Visible Object
	 * 
	 * @param name Perceptor name
	 * @return perceptor
	 */
	IVisibleObjectPerceptor getVisibleObject(String name);

	/**
	 * Get the global pose perceptor
	 * 
	 * @param name Perceptor name
	 * @return perceptor
	 */
	IGlobalPosePerceptor getGlobalPose();

	/**
	 * Get the gloabl time perceptor
	 * 
	 * @return perceptor
	 */
	ITimerPerceptor getTime();

	/**
	 * Get the Game State perceptor
	 * 
	 * @return perceptor
	 */
	IGameStatePerceptor getGameState();

	/**
	 * Get the Flags perceptors.
	 * 
	 * @return a list of flag perceptors
	 */
	List<IFlagPerceptor> getFlags();

	/**
	 * Get the Hear perceptor
	 * 
	 * @return perceptor
	 */
	List<IHearPerceptor> getHearPerceptors();

	/**
	 * Get remote command
	 * 
	 * @return perceptor
	 */
	ICommandPerceptor getComandPerceptor();

	/**
	 * Get camera tilt
	 * 
	 * @return perceptor
	 */
	ICameraTiltPerceptor getCameraTiltPerceptor();

	/**
	 * Get a list of all visible players
	 * 
	 * @return perceptor
	 */
	List<IPlayerPos> getVisiblePlayers();

	/**
	 * Get a list of all visible lines
	 * 
	 * @return perceptor
	 */
	List<ILinePerceptor> getVisibleLines();

	/**
	 * @return true, if the current perception contains vision information
	 */
	boolean containsVision();

	/**
	 * @return true, if the current perception contains vision information
	 */
	boolean containsMotion();

	/**
	 * @param perceptor
	 */
	void updatePerceptors(Map<String, IPerceptor> perceptor);
}
