/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.perception.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import magma.agent.perception.IAccelerometerPerceptor;
import magma.agent.perception.IAgentStatePerceptor;
import magma.agent.perception.ICameraTiltPerceptor;
import magma.agent.perception.ICommandPerceptor;
import magma.agent.perception.ICompassPerceptor;
import magma.agent.perception.ICompositeJointPerceptor;
import magma.agent.perception.IFlagPerceptor;
import magma.agent.perception.IForceResistancePerceptor;
import magma.agent.perception.IGameStatePerceptor;
import magma.agent.perception.IGlobalPosePerceptor;
import magma.agent.perception.IGyroPerceptor;
import magma.agent.perception.IHearPerceptor;
import magma.agent.perception.IHingeJointPerceptor;
import magma.agent.perception.ILinePerceptor;
import magma.agent.perception.IPerception;
import magma.agent.perception.IPerceptor;
import magma.agent.perception.IPlayerPos;
import magma.agent.perception.ITimerPerceptor;
import magma.agent.perception.IUniversalJointPerceptor;
import magma.agent.perception.IVisibleObjectPerceptor;

/**
 * Represents all data the agent is able to perceive from its environment.
 * Should be updated in every simulation cycle.
 * 
 * @author Klaus Dorer, Simon Raffeiner, Stefan Glaser
 */
public class Perception implements IPerception
{
	/** handle for logging */
	protected static final Logger logger = Logger.getLogger("");

	// Map for all named perceptors
	protected Map<String, IPerceptor> perceptors;

	// Flag - if the current perception contains vision information
	private boolean containsVision;

	// Flag - if the current perception contains motor information
	private boolean containsMotion;

	/**
	 * 
	 */
	public Perception()
	{
		perceptors = new HashMap<String, IPerceptor>();
	}

	@Override
	public ICompositeJointPerceptor getCompositeJointPerceptor(String name)
	{
		return (ICompositeJointPerceptor) perceptors.get(name);
	}

	@Override
	public IUniversalJointPerceptor getUniversalJointPerceptor(String name)
	{
		return (IUniversalJointPerceptor) perceptors.get(name);
	}

	@Override
	public IHingeJointPerceptor getHingeJointPerceptor(String name)
	{
		return (IHingeJointPerceptor) perceptors.get(name);
	}

	@Override
	public IAccelerometerPerceptor getAccelerationPerceptor(String name)
	{
		return (IAccelerometerPerceptor) perceptors.get(name);
	}

	@Override
	public ICompassPerceptor getCompassPerceptor(String name)
	{
		return (ICompassPerceptor) perceptors.get(name);
	}

	@Override
	public IAgentStatePerceptor getAgentState()
	{
		return (IAgentStatePerceptor) perceptors.get("AgentState");
	}

	@Override
	public IGyroPerceptor getGyroRatePerceptor(String name)
	{
		return (IGyroPerceptor) perceptors.get(name);
	}

	@Override
	public IForceResistancePerceptor getForceResistancePerceptor(String name)
	{
		return (IForceResistancePerceptor) perceptors.get(name);
	}

	@Override
	public IVisibleObjectPerceptor getVisibleObject(String name)
	{
		return (IVisibleObjectPerceptor) perceptors.get(name);
	}

	@Override
	public IGlobalPosePerceptor getGlobalPose()
	{
		return (IGlobalPosePerceptor) perceptors.get("globalPose");
	}

	@Override
	public ITimerPerceptor getTime()
	{
		return (ITimerPerceptor) perceptors.get("time");
	}

	@Override
	public IGameStatePerceptor getGameState()
	{
		return (IGameStatePerceptor) perceptors.get("GS");
	}

	@Override
	public List<IFlagPerceptor> getFlags()
	{
		List<IFlagPerceptor> result = new ArrayList<IFlagPerceptor>();
		for (IPerceptor perceptor : perceptors.values()) {
			if (perceptor instanceof IFlagPerceptor) {
				result.add((IFlagPerceptor) perceptor);
			}
		}
		return result;
	}

	@Override
	public List<IHearPerceptor> getHearPerceptors()
	{
		List<IHearPerceptor> result = new ArrayList<IHearPerceptor>();
		for (IPerceptor perceptor : perceptors.values()) {
			if (perceptor instanceof IHearPerceptor) {
				result.add((IHearPerceptor) perceptor);
			}
		}
		return result;
	}

	@Override
	public ICommandPerceptor getComandPerceptor()
	{
		return (ICommandPerceptor) perceptors.get("command");
	}

	@Override
	public ICameraTiltPerceptor getCameraTiltPerceptor()
	{
		return (ICameraTiltPerceptor) perceptors.get("cameraTilt");
	}

	@Override
	public List<IPlayerPos> getVisiblePlayers()
	{
		List<IPlayerPos> result = new ArrayList<IPlayerPos>();
		for (IPerceptor perceptor : perceptors.values()) {
			if (perceptor instanceof IPlayerPos) {
				result.add((IPlayerPos) perceptor);
			}
		}
		return result;
	}

	@Override
	public List<ILinePerceptor> getVisibleLines()
	{
		List<ILinePerceptor> result = new ArrayList<ILinePerceptor>();
		for (IPerceptor perceptor : perceptors.values()) {
			if (perceptor instanceof ILinePerceptor) {
				result.add((ILinePerceptor) perceptor);
			}
		}
		return result;
	}

	@Override
	public boolean containsVision()
	{
		return containsVision;
	}

	@Override
	public boolean containsMotion()
	{
		return containsMotion;
	}

	@Override
	public void updatePerceptors(Map<String, IPerceptor> perceptors)
	{
		if (perceptors == null || perceptors.isEmpty()) {
			// nothing to do, might happen at disconnection
			return;
		}

		this.perceptors = perceptors;

		// Clear list of visible objects
		containsVision = false;
		containsMotion = false;

		// Process
		for (IPerceptor perceptor : perceptors.values()) {
			processInputPerceptor(perceptor);
		}
	}

	private void processInputPerceptor(IPerceptor perceptor)
	{
		// Handle sensor perceptors
		if (perceptor instanceof IHingeJointPerceptor
				|| perceptor instanceof IUniversalJointPerceptor
				|| perceptor instanceof ICompositeJointPerceptor) {
			containsMotion = true;
		}

		if (perceptor instanceof IVisibleObjectPerceptor) {
			containsVision = true;
		}
	}
}