/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.perception.impl;

import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.IMagmaConstants;
import magma.agent.perception.IPerceptor;
import magma.agent.perception.IServerMessageParser;
import magma.agent.perception.PerceptorConversionException;
import magma.common.spark.TeamColor;
import magma.util.symboltreeparser.IllegalSymbolInputException;
import magma.util.symboltreeparser.SymbolNode;
import magma.util.symboltreeparser.SymbolTreeParser;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Parses incoming server messages, converts them into single Perceptor messages
 * and delivers them to all observers
 * 
 * @author Simon Raffeiner
 * 
 */
public class ServerMessageParser implements IServerMessageParser
{
	// handle for logging
	protected static final Logger logger = Logger.getLogger("");

	// Internal tree parser
	private final SymbolTreeParser treeparser;

	/**
	 * Default constructor
	 */
	public ServerMessageParser()
	{
		treeparser = new SymbolTreeParser();
	}

	/**
	 * Parse a string into a list of Perceptor objects
	 * 
	 * @param message Message string
	 * @return List of parsed Perceptor objects
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	@Override
	public Map<String, IPerceptor> parseMessage(byte[] message)
			throws PerceptorConversionException
	{
		SymbolNode root;

		// Generate the symbol tree
		try {
			String msg = new String(message, 0, message.length, "UTF-8");
			// System.out.println(msg);
			root = treeparser.parse(msg);

		} catch (IllegalSymbolInputException e) {
			throw new PerceptorConversionException("Parsing error: "
					+ e.getMessage());
		} catch (UnsupportedEncodingException e) {
			throw new PerceptorConversionException(
					"Parsing error: unsupported encoding " + e.getMessage());
		}

		Map<String, IPerceptor> map = new HashMap<String, IPerceptor>();

		// Parse all top-level nodes into messages
		for (int i = 0; i < root.children.size(); i++) {
			Object node = root.children.get(i);

			/*
			 * The tree may contain top-level leaves, ATM they are not specified
			 * and therefore not parsed
			 */
			if (node instanceof SymbolNode) {
				parseNode((SymbolNode) node, map);
			}
		}

		return map;
	}

	/**
	 * Parse a symbol tree node into a Perceptor object (if possible)
	 * 
	 * @param node Symbol tree node
	 * @return Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private void parseNode(SymbolNode node, Map<String, IPerceptor> map)
			throws PerceptorConversionException
	{
		/* Check message type */
		Object child = node.children.get(0);

		if (child == null || !(child instanceof String))
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());

		String type = (String) child;
		IPerceptor perceptor = null;
		if (type.equals("HJ")) {
			perceptor = parseHingeJoint(node);
		} else if (type.equals("time"))
			perceptor = parseTime(node);
		else if (type.equals("GS"))
			perceptor = parseGameState(node);
		else if (type.equals("GYR"))
			perceptor = parseGyro(node);
		else if (type.equals("ACC"))
			perceptor = parseAccelerometer(node);
		else if (type.equals("See"))
			parseVision(node, map);
		else if (type.equals("TCH"))
			perceptor = parseTouch(node);
		else if (type.equals("AgentState"))
			perceptor = parseAgentState(node);
		else if (type.equals("hear"))
			perceptor = parseHear(node);
		else if (type.equals("FRP"))
			perceptor = parseForceResistance(node);
		else if (type.equals("UJ"))
			perceptor = parseUniversalJoint(node);
		else {
			/* Unknown node? */
			logger.log(Level.FINEST, "Got an unknown node: " + node);
		}
		if (perceptor != null) {
			map.put(perceptor.getName(), perceptor);
		}
	}

	/**
	 * Parse a symbol tree node into a Gyro Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Gyro Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private GyroPerceptor parseGyro(SymbolNode node)
			throws PerceptorConversionException
	{
		if (!(node.children.get(1) instanceof SymbolNode)
				|| !(node.children.get(2) instanceof SymbolNode))
			throw new PerceptorConversionException("Malformed Message: "
					+ node.toString());

		try {
			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children.get(1);
			SymbolNode rotationNode = (SymbolNode) node.children.get(2);

			if (!nameNode.children.get(0).equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());

			if (!rotationNode.children.get(0).equals("rt"))
				throw new PerceptorConversionException("rotation expected: "
						+ node.toString());

			return new GyroPerceptor((String) nameNode.children.get(1) + "Gyro",
					Float.valueOf((String) rotationNode.children.get(1)),
					Float.valueOf((String) rotationNode.children.get(2)),
					Float.valueOf((String) rotationNode.children.get(3)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Gyro perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Gyro perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private AccelerometerPerceptor parseAccelerometer(SymbolNode node)
			throws PerceptorConversionException
	{
		if (!(node.children.get(1) instanceof SymbolNode)
				|| !(node.children.get(2) instanceof SymbolNode))
			throw new PerceptorConversionException("Malformed Message: "
					+ node.toString());

		try {
			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children.get(1);
			SymbolNode accelerationNode = (SymbolNode) node.children.get(2);

			if (!nameNode.children.get(0).equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());

			if (!accelerationNode.children.get(0).equals("a"))
				throw new PerceptorConversionException("rotation expected: "
						+ node.toString());

			return new AccelerometerPerceptor((String) nameNode.children.get(1)
					+ "Accel", Float.valueOf((String) accelerationNode.children
					.get(1)), Float.valueOf((String) accelerationNode.children
					.get(2)), Float.valueOf((String) accelerationNode.children
					.get(3)));

		} catch (NumberFormatException e) {
			// seems that the server sometimes sends NAN, in which case we ignore
			// the reading for accelerometer
			logger.log(Level.INFO,
					"Parsing accelerometer problem: " + node.toString() + "{0}", e);
			return null;

		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Hinge Joint Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Hinge Joint Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private HingeJointPerceptor parseHingeJoint(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// if (!(node.children.get(1) instanceof SymbolNode)
			// || !(node.children.get(2) instanceof SymbolNode))
			// throw new PerceptorConversionException("Malformed Message: "
			// + node.toString());

			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children.get(1);
			SymbolNode rotationNode = (SymbolNode) node.children.get(2);

			// if (!((String) nameNode.children.get(0)).equals("n"))
			// throw new PerceptorConversionException("Malformed Message: "
			// + node.toString() + ": name expected");
			//
			// if (!((String) rotationNode.children.get(0)).equals("ax"))
			// throw new PerceptorConversionException("Malformed Message: "
			// + node.toString() + ": axis expected");

			return new HingeJointPerceptor((String) nameNode.children.get(1),
					Float.valueOf((String) rotationNode.children.get(1)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Universal Joint Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Universal Joint Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private UniversalJointPerceptor parseUniversalJoint(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// Sanity checks
			if (!(node.children.get(1) instanceof SymbolNode)
					|| !(node.children.get(2) instanceof SymbolNode)
					| !(node.children.get(3) instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());

			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children.get(1);
			SymbolNode axis1Node = (SymbolNode) node.children.get(2);
			SymbolNode axis2Node = (SymbolNode) node.children.get(3);

			if (!nameNode.children.get(0).equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());
			if (!axis1Node.children.get(0).equals("ax1"))
				throw new PerceptorConversionException("axis expected: "
						+ node.toString());
			if (!axis2Node.children.get(0).equals("ax2"))
				throw new PerceptorConversionException("axis expected: "
						+ node.toString());

			return new UniversalJointPerceptor((String) nameNode.children.get(1),
					Float.valueOf((String) axis1Node.children.get(1)),
					Float.valueOf((String) axis2Node.children.get(1)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Force Resistance Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Force Resistance Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private ForceResistancePerceptor parseForceResistance(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// Sanity checks
			if (!(node.children.get(1) instanceof SymbolNode)
					|| !(node.children.get(2) instanceof SymbolNode)
					| !(node.children.get(3) instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());

			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children.get(1);
			SymbolNode originNode = (SymbolNode) node.children.get(2);
			SymbolNode forceNode = (SymbolNode) node.children.get(3);

			if (!nameNode.children.get(0).equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());
			if (!originNode.children.get(0).equals("c"))
				throw new PerceptorConversionException("origin expected: "
						+ node.toString());
			if (!forceNode.children.get(0).equals("f"))
				throw new PerceptorConversionException("force expected: "
						+ node.toString());

			return new ForceResistancePerceptor((String) nameNode.children.get(1),
					Float.valueOf((String) originNode.children.get(1)),
					Float.valueOf((String) originNode.children.get(2)),
					Float.valueOf((String) originNode.children.get(3)),
					Float.valueOf((String) forceNode.children.get(1)),
					Float.valueOf((String) forceNode.children.get(2)),
					Float.valueOf((String) forceNode.children.get(3)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Touch Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Touch Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private TouchPerceptor parseTouch(SymbolNode node)
			throws PerceptorConversionException
	{
		if (node.children.size() != 5 || !node.children.get(1).equals("n")
				|| !node.children.get(3).equals("val"))
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());

		try {
			String name = (String) node.children.get(2);

			boolean value = false;
			if (node.children.get(4).equals("1"))
				value = true;

			return new TouchPerceptor(name, value);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree vision node into a list of Perceptor objects
	 * 
	 * @param node Symbol tree vision node
	 * @return List of Perceptor objects
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private void parseVision(SymbolNode node, Map<String, IPerceptor> map)
			throws PerceptorConversionException
	{
		// Sanity checks
		if (node.children.size() == 0)
			throw new PerceptorConversionException("Malformed Node: "
					+ node.toString());

		try {
			// Parse visible objects
			int lineID = 0;
			for (int i = 1; i < node.children.size(); i++) {
				Object subitem = node.children.get(i);

				if (subitem instanceof SymbolNode) {
					SymbolNode subnode = (SymbolNode) subitem;

					// Player object?
					if (!(subnode.children.get(0) instanceof String))
						throw new PerceptorConversionException(
								"Malformed Node, empty name: " + node.toString());
					else if (subnode.children.get(0).equals("P"))
						parsePlayer(subnode, map);

					else if (subnode.children.get(0).equals("L"))
						parseLine(subnode, map, lineID++);

					else
						parseVisibleObject(subnode, map);
				}
			}

		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * @param subnode
	 * @return
	 */
	private void parseLine(SymbolNode node, Map<String, IPerceptor> map,
			int lineID) throws PerceptorConversionException
	{
		try {
			assert node.children.size() == 3 : "Malformed node";
			assert node.children.get(1) instanceof SymbolNode : "Malformed node: "
					+ node.toString();
			assert node.children.get(2) instanceof SymbolNode : "Malformed node: "
					+ node.toString();

			String name = (String) node.children.get(0);
			Vector3D pol1 = parsePol((SymbolNode) node.children.get(1));
			Vector3D pol2 = parsePol((SymbolNode) node.children.get(2));

			LinePerceptor perceptor = new LinePerceptor(name + lineID, pol1, pol2,
					true);
			map.put(perceptor.getName(), perceptor);

		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Player object
	 * 
	 * @param node Symbol tree node
	 * @return Player object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private void parsePlayer(SymbolNode node, Map<String, IPerceptor> map)
			throws PerceptorConversionException
	{
		try {
			/* Parse content */
			String teamName = IMagmaConstants.UNKNOWN_PLAYER_TEAMNAME;
			int id = IMagmaConstants.UNKNOWN_PLAYER_NUMBER;
			Vector3D pol = Vector3D.ZERO;
			String bodyPartIdentifier = "";
			Map<String, Vector3D> bodyPartMap = new HashMap<String, Vector3D>();

			for (int i = 1; i < node.children.size(); i++) {
				if (!(node.children.get(i) instanceof SymbolNode)) {
					throw new PerceptorConversionException("Malformed node: "
							+ node.toString());
				}

				SymbolNode param = (SymbolNode) node.children.get(i);

				if (param.children.get(0).equals("team")) {
					teamName = (String) param.children.get(1);
				} else if (param.children.get(0).equals("id")) {
					id = Integer.parseInt((String) param.children.get(1));
				} else {
					// in case of seeing parts of opponent, we have to look into them
					if (!param.children.get(0).equals("pol")) {
						bodyPartIdentifier = (String) param.children.get(0);
						param = (SymbolNode) param.children.get(1);
					}

					if (param.children.get(0).equals("pol")) {
						pol = parsePol(param);
						bodyPartMap.put(bodyPartIdentifier, pol);
					}
				}
			}

			String name = "P" + teamName + id;
			if (id == IMagmaConstants.UNKNOWN_PLAYER_NUMBER) {
				name = "P" + teamName + pol.toString();
			}
			PlayerPos perceptor = new PlayerPos(name, pol, id, teamName,
					bodyPartMap, true);

			map.put(perceptor.getName(), perceptor);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Player object
	 * 
	 * @param node Symbol tree node
	 * @return Player object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private void parseVisibleObject(SymbolNode node, Map<String, IPerceptor> map)
			throws PerceptorConversionException
	{
		try {
			assert node.children.size() == 2 : "Malformed node";
			assert node.children.get(1) instanceof SymbolNode : "Malformed node: "
					+ node.toString();

			String name = (String) node.children.get(0);
			Vector3D pol = parsePol((SymbolNode) node.children.get(1));

			VisibleObjectPerceptor perceptor = new VisibleObjectPerceptor(name,
					pol, true, 1.0);
			map.put(perceptor.getName(), perceptor);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse object coordinates
	 * 
	 * @param node SymbolNode to parse
	 * @return Coordinates
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private Vector3D parsePol(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			if (node.children.size() != 4)
				throw new PerceptorConversionException("Malformed node: "
						+ node.toString());
			if (!node.children.get(0).equals("pol"))
				throw new PerceptorConversionException("Expecting a pol object: "
						+ node.toString());

			float val1 = Float.parseFloat((String) node.children.get(1));
			float val2 = Float.parseFloat((String) node.children.get(2));
			float val3 = Float.parseFloat((String) node.children.get(3));

			return new Vector3D(val1, new Vector3D(Math.toRadians(val2),
					Math.toRadians(val3)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());

		} catch (NumberFormatException e) {
			// seems that the server sometimes sends NAN, in which case we ignore
			// the reading for position
			logger.log(Level.INFO, "Parsing position problem: " + node.toString()
					+ "{0}", e);
		}
		return null;
	}

	/**
	 * Parse a symbol tree node into a Game State Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Game State Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private GameStatePerceptor parseGameState(SymbolNode node)
			throws PerceptorConversionException
	{
		float time = 0.0f;
		String playmode = "", team = "";
		int unum = 0;
		int scoreLeft = 0, scoreRight = 0;

		SymbolNode child = null;

		// Evaluate content
		for (int i = 1; i < node.children.size(); i++) {
			try {
				child = (SymbolNode) node.children.get(i);

				String type = (String) child.children.get(0);

				// Check sub-node type
				if (type.equals("t")) {
					time = Float.parseFloat((String) child.children.get(1));
				} else if (type.equals("pm")) {
					playmode = (String) child.children.get(1);
				} else if (type.equals("unum")) {
					unum = Integer.parseInt((String) child.children.get(1));
				} else if (type.equals("team")) {
					team = (String) child.children.get(1);
				} else if (type.equals("sl")) {
					scoreLeft = Integer.parseInt((String) child.children.get(1));
				} else if (type.equals("sr")) {
					scoreRight = Integer.parseInt((String) child.children.get(1));
				} else
					throw new PerceptorConversionException(
							"Malformed GameState node, unknown sub-node: "
									+ child.toString());
			} catch (Exception e) {
				if (child != null)
					throw new PerceptorConversionException(
							"Malformed GameState node, conversion error: "
									+ child.toString());

				throw new PerceptorConversionException(
						"Malformed GameState node: child was null!");
			}
		}
		// we do not have sufficient information for team color, so team color
		// information is missing in Simspark!
		TeamColor color = TeamColor.Unknown;
		return new GameStatePerceptor(time, playmode, team, color, unum,
				scoreLeft, scoreRight);
	}

	/**
	 * Parse a symbol tree node into an Agent State Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Agent State Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private AgentStatePerceptor parseAgentState(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// Sanity checks
			if (!(node.children.get(1) instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());
			if (!(node.children.get(2) instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());

			/* Check content */
			SymbolNode temperatureNode = (SymbolNode) node.children.get(1);
			SymbolNode batteryNode = (SymbolNode) node.children.get(2);

			if (!temperatureNode.children.get(0).equals("temp"))
				throw new PerceptorConversionException("temperature expected: "
						+ node.toString());
			if (!batteryNode.children.get(0).equals("battery"))
				throw new PerceptorConversionException("battery level expected: "
						+ node.toString());

			return new AgentStatePerceptor(
					Integer.parseInt((String) temperatureNode.children.get(1)),
					Integer.parseInt((String) batteryNode.children.get(1)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Hear Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Hear Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private HearPerceptor parseHear(SymbolNode node)
			throws PerceptorConversionException
	{
		float time;
		String message = "";
		String target;
		String team = "";
		int messageStartID = 3;

		logger.log(Level.FINEST, "hear: ({0})", new Object[] { node });

		// Sanity checks
		if (node.children.size() < 4)
			throw new PerceptorConversionException("Malformed hear node: "
					+ node.toString());

		try {
			// version 0.6.6 or earlier protocol
			time = Float.parseFloat((String) node.children.get(1));
			target = (String) node.children.get(2);
		} catch (NumberFormatException e) {
			// BAD: it would really be helpful to get protocol version information
			// from outside
			// version 0.6.7 protocol
			team = (String) node.children.get(1);
			time = Float.parseFloat((String) node.children.get(2));
			target = (String) node.children.get(3);
			messageStartID = 4;

		} catch (Exception e) {
			throw new PerceptorConversionException(
					"Malformed hear node, conversion error: " + node.toString());
		}

		// Concatenate following nodes
		for (int i = messageStartID; i < node.children.size(); i++) {
			if (i > messageStartID)
				message += " " + node.children.get(i);
			else
				message += node.children.get(i);
		}

		return new HearPerceptor(time, team, target, message);
	}

	/**
	 * Parse a symbol tree node into a Time Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Time Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private TimePerceptor parseTime(SymbolNode node)
			throws PerceptorConversionException
	{
		// Sanity check
		if (node.children.size() != 2
				|| !(node.children.get(1) instanceof SymbolNode))
			throw new PerceptorConversionException("Malformed time node: "
					+ node.toString());

		// Sub-node sanity check
		SymbolNode timeNode = (SymbolNode) node.children.get(1);
		if (timeNode.children.size() != 2
				|| timeNode.children.get(1).equals("now"))
			throw new PerceptorConversionException("Malformed time sub-node: "
					+ timeNode.toString());

		try {
			// Sanity checks
			float time = Float.parseFloat((String) timeNode.children.get(1));

			return new TimePerceptor(time);
		} catch (Exception e) {
			throw new PerceptorConversionException(
					"Malformed time node, conversion error: " + node.toString());
		}
	}
}
