/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel;

import java.util.List;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.flags.IFlagModel;
import magma.agent.perception.IPerception;
import magma.agent.thoughtmodel.strategy.IRole;
import magma.agent.thoughtmodel.strategy.IRoleManager;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.util.geometry.Angle;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Represents all information related to thoughts of the agent.<br>
 * While the {@link IGlobalMap} and {@link IAgentModel} represent the measured
 * states based on sensor information, the IThoughtModel is mend to represent a
 * higher level of information abstraction. Furthermore, it is the central model
 * to store all information that comes from inside the agent and cannot be
 * measured.<br>
 * To give a simple example: <i>IFOCalculator</i>:<br>
 * The location information to each player is a measured value and contained in
 * the {@link IGlobalMap}. The question which player is closest to the ball can
 * be calculated based on this information but represents a thought which is not
 * directly related to a sensor value. This "thought" (which player is closest
 * to the ball) can be integrated and buffered in the IThoughtModel, which
 * improves reusability of common calculations and therefore reduces runtime.
 * 
 * @author Stefan Glaser
 */
public interface IThoughtModel
{
	/**
	 * Retrieve the agent model.
	 * 
	 * @return the agent model
	 */
	IAgentModel getAgentModel();

	/**
	 * Retrieve the world model.
	 * 
	 * @return the world model
	 */
	IGlobalMap getWorldModel();

	/**
	 * Retrieve the flags.
	 * 
	 * @return the flags
	 */
	IFlagModel getFlags();

	/**
	 * Update models based on new perception information.
	 * 
	 * @param perception The actual perception or <code>null</code> if only the
	 *        internal state of the IThoughtModel should be refreshed
	 */
	boolean update(IPerception perception);

	/**
	 * Get a list of points around ball in order of priority to kick to.
	 * 
	 * @return List of points
	 */
	Vector3D getPointToKickBall();

	/**
	 * @return the preferred kick direction
	 */
	Angle getPreferredKickDirection();

	/**
	 * @return the kick direction profile
	 */
	float[][] getKickDirectionProfile();

	/**
	 * Method gives a list of Polygons around this player and the ball. Is
	 * @deprecated as it is used only by the Debugger tool.
	 * 
	 * @return List of polygons
	 */
	@Deprecated
	List<Vector3D[]> getInfluencePolys();

	/**
	 * @return the roll manager
	 */
	IRoleManager getRoleManager();

	/**
	 * @return the current role of the agent
	 */
	IRole getRole();

	/**
	 * @return the opponent that is closest to ball
	 */
	IPlayer getOpponentAtBall();

	/**
	 * Get the team-mate which is closest to the ball
	 * 
	 * @return the team-mate that is closest to the ball, null if we see no
	 *         team-mates
	 */
	IPlayer getTeammateAtBall();

	/**
	 * Get the player of the own team (including myself) which is closest to the
	 * ball
	 * 
	 * @return the own player that is closest to the ball, ourself if we see no
	 *         team-mates
	 */
	IPlayer getClosestOwnPlayerAtBall();

	/**
	 * Get a list of all players, sorted by distance from the own position
	 * 
	 * @return a list of all players sorted by the distance to me
	 */
	List<IPlayer> getPlayersAtMeList();

	/**
	 * Get a list of all opponents, sorted by distance from the own position
	 * 
	 * @return a list of all opponents sorted by the distance to me
	 */
	List<IPlayer> getOpponentsAtMeList();

	/**
	 * Get a list of all opponents, sorted by distance from the ball position
	 * 
	 * @return a list of all opponents sorted by the distance to ball
	 */
	List<IPlayer> getOpponentsAtBallList();

	/**
	 * Get a list of all team-mates, sorted by distance from the ball position
	 * 
	 * @return list of team-mates
	 */
	List<IPlayer> getTeammatesAtBall();

	/**
	 * Get a list of all players, sorted by distance from the ball position
	 * 
	 * @return List of players
	 */
	List<IPlayer> getPlayersAtBallList();

	/**
	 * @return the obstacles to avoid
	 */
	List<IVisibleObject> getObstacles();

	/**
	 * @param delta how much closer has the opponent to be than own player
	 * @return true if one of our own team (not including myself!) is closer to
	 *         ball than an opponent player
	 */
	boolean ownTeammateHasBall(double delta);

	/**
	 * @return a command sent by remote control, null if none
	 */
	String getRemoteCommand();

	String getCameraTilt();
}
