/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package magma.agent.thoughtmodel.impl;

import java.util.Comparator;

import magma.agent.IMagmaConstants;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.util.geometry.Geometry;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * For player at ball calculation
 * @author Klaus Dorer
 */
class DistanceToBallComparator implements Comparator<IPlayer>
{
	/**  */
	private final IGlobalMap worldModel;

	/**
	 * @param worldModel The world model.
	 */
	DistanceToBallComparator(IGlobalMap worldModel)
	{
		this.worldModel = worldModel;
	}

	@Override
	public int compare(IPlayer player1, IPlayer player2)
	{
		IMoveableObject ball = worldModel.getBall();
		Vector3D ballPosition = ball.getPosition();
		Vector3D otherGoalPos = worldModel.getOtherGoalPosition();
		double distance1;
		double distance2;
		// No predictions if ball speed is too low
		if (ball.isMoving()) {
			ballPosition = ball.getFuturePosition(150);
		}
		distance1 = player1.getDistanceToXY(ballPosition);
		distance2 = player2.getDistanceToXY(ballPosition);

		distance1 = checkLayingPenalty(player1, ballPosition, distance1);
		distance2 = checkLayingPenalty(player2, ballPosition, distance2);

		distance1 = checkBackwardRunning(player1, ball, otherGoalPos, distance1);
		distance2 = checkBackwardRunning(player2, ball, otherGoalPos, distance2);

		distance1 = checkMyselfBonus(player1, distance1);
		distance2 = checkMyselfBonus(player2, distance2);

		if (distance1 < distance2) {
			return -1;
		} else if (distance1 > distance2) {
			return 1;
		} else {
			return 0;
		}
	}

	/**
	 * @param player1
	 * @param ballPosition
	 * @param otherGoalPos
	 * @return
	 */
	private double checkBackwardRunning(IPlayer player, IMoveableObject ball,
			Vector3D otherGoalPos, double distance)
	{
		final double DISTANCE_PENALTY_BEHINDBALL = 2.0;
		double distance1 = player.getDistanceToXY(otherGoalPos);
		double distance2 = ball.getDistanceToXY(otherGoalPos);
		double delta = distance2 - distance1;
		if (distance1 < distance2) {
			return distance + Geometry.getLinearFuzzyValue(0, 1, true, delta)
					* DISTANCE_PENALTY_BEHINDBALL;
		}
		return distance;
	}

	/**
	 * @param player1
	 * @param ballPosition
	 * @param distance1
	 * @return
	 */
	private double checkLayingPenalty(IPlayer player, Vector3D ballPosition,
			double distance)
	{
		if (player.isLaying()) {
			distance += IMagmaConstants.DISTANCE_PENALTY_LYING;
		}
		return distance;
	}

	private double checkMyselfBonus(IPlayer player, double distance)
	{
		final double MYSELF_BONUS = 0.1;

		if (player.getID() == worldModel.getThisPlayer().getID()) {
			return distance - MYSELF_BONUS;
		}
		return distance;
	}
}