/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.thoughtmodel.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import magma.agent.IMagmaConstants;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.ILandmark;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.agent.worldmodel.impl.Landmark;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Calculates indexical functional objects (Agre and Chapman) using filters and
 * comparators
 * @author Klaus Dorer
 */
public class IFOCalculator
{
	/** order criteria is distance to ball */
	private final Comparator<IVisibleObject> distanceToBallComparator;

	/** order criteria is distance to this player */
	private final Comparator<IVisibleObject> distanceToMeComparator;

	/** accepts only teammates and not the goalie (+base checks) */
	private final IFilter teammateNotGoalieFilter;

	/** accepts only teammates (+base checks) */
	private final IFilter teammateFilter;

	/** accepts only opponents (+base checks) */
	private final IFilter opponentFilter;

	/** checking for age of player and that it is not this player */
	private final IFilter baseFilter;

	/** list of the four goal posts for collision avoidance */
	private final List<IVisibleObject> goalPosts;

	/**
	 * Single instance created once at startup
	 * @param worldModel reference to the global map
	 */
	public IFOCalculator(IGlobalMap worldModel)
	{
		distanceToBallComparator = new DistanceToBallComparator(worldModel);
		distanceToMeComparator = new DistanceToMeComparator(worldModel);
		teammateNotGoalieFilter = new TeammateFilter(worldModel,
				new NotGoalieFilter(worldModel));
		teammateFilter = new TeammateFilter(worldModel);
		opponentFilter = new OpponentFilter(worldModel);
		baseFilter = new BaseFilter(worldModel);
		goalPosts = new ArrayList<IVisibleObject>(4);

		initGoalPosts(worldModel);
	}

	private void initGoalPosts(IGlobalMap worldModel)
	{
		// the positions of the goal posts are slightly outside the field to allow
		// a bigger radius for avoidance
		Vector3D knownPos;

		for (ILandmark lm : worldModel.getLandmarks()) {
			double shiftOutside = 0.5;
			if (lm.getName().contains("G")) {
				knownPos = lm.getKnownPosition();

				if (knownPos.getX() < 0) {
					shiftOutside = -shiftOutside;
				}
				goalPosts.add(new Landmark(lm.getName() + "Obstacle", new Vector3D(
						knownPos.getX() + shiftOutside, knownPos.getY(), 0)));
			}
		}
	}

	/**
	 * @param sourceList the list containing the players to sort
	 * @return a filtered list sorted ascending by the distance of the own team's
	 *         players not including goalie to the ball
	 */
	public List<IPlayer> getTeammatesAtBall(List<IPlayer> sourceList)
	{
		return getFilteredAndSortedList(sourceList, teammateNotGoalieFilter,
				distanceToBallComparator);
	}

	/**
	 * @param sourceList the list containing the players to sort
	 * @return a filtered list sorted ascending by the distance of the opponent
	 *         team's players to the ball
	 */
	public List<IPlayer> getOpponentsAtBall(List<IPlayer> sourceList)
	{
		return getFilteredAndSortedList(sourceList, opponentFilter,
				distanceToBallComparator);
	}

	/**
	 * Filter all players currently near the ball out of the given list
	 * 
	 * @param sourceList Source player list
	 * @return Filtered list
	 */
	public List<IPlayer> getPlayersAtBall(List<IPlayer> sourceList)
	{
		return getFilteredAndSortedList(sourceList, baseFilter,
				distanceToBallComparator);
	}

	/**
	 * @param sourceList the list containing the players to sort
	 * @return a filtered list sorted ascending by the distance of the own team's
	 *         players including goalie to me
	 */
	public List<IPlayer> getPlayersAtMe(List<IPlayer> sourceList)
	{
		return getFilteredAndSortedList(sourceList, baseFilter,
				distanceToMeComparator);
	}

	/**
	 * @param sourceList the list containing the players to sort
	 * @return a filtered list sorted ascending by the distance of the own team's
	 *         players including goalie to me
	 */
	public List<IPlayer> getTeammatesAtMe(List<IPlayer> sourceList)
	{
		return getFilteredAndSortedList(sourceList, teammateFilter,
				distanceToMeComparator);
	}

	/**
	 * @param sourceList the list containing the players to sort
	 * @return a filtered list sorted ascending by the distance of the other
	 *         team's players including goalie to me
	 */
	public List<IPlayer> getOpponentsAtMe(List<IPlayer> sourceList)
	{
		return getFilteredAndSortedList(sourceList, opponentFilter,
				distanceToMeComparator);
	}

	/**
	 * Retrieve a lit of obstacles
	 * 
	 * @param sourceList the list containing the players to sort
	 * @param ball Reference to ball object
	 * @return a filtered list sorted ascending by the distance of the other
	 *         team's players including goalie to me
	 */
	public List<IVisibleObject> getObstacles(List<IPlayer> sourceList,
			IMoveableObject ball)
	{
		List<IPlayer> result = filter(sourceList, baseFilter);
		List<IVisibleObject> obstacles = new ArrayList<IVisibleObject>(result);

		// mergeObstacles(obstacles);

		obstacles.addAll(goalPosts);
		if (!ball.isMoving()) {
			// ball is only obstacle if not moving
			obstacles.add(ball);
		}
		Collections.sort(obstacles, distanceToMeComparator);
		return obstacles;
	}

	/**
	 * Merges obstacles that are close to each other into one.
	 * @param obstacles the list of obstacles to change
	 */
	protected void mergeObstacles(List<IVisibleObject> obstacles)
	{
		for (int i = 0; i < obstacles.size() - 1; i++) {
			IVisibleObject obstacle = obstacles.get(i);
			for (int j = i + 1; j < obstacles.size(); j++) {
				IVisibleObject obstacleToCompare = obstacles.get(j);
				double criticalDistance = obstacle.getCollisionDistance()
						+ obstacleToCompare.getCollisionDistance();

				double distance = obstacle.getDistanceToXY(obstacleToCompare);
				if (distance < criticalDistance) {
					Vector3D newMiddle = obstacle.getPosition().add(
							obstacleToCompare.getPosition()
									.subtract(obstacle.getPosition())
									.scalarMultiply(0.5));
					obstacle = new MergedObstacle("obstacle" + i, newMiddle,
							distance + criticalDistance * 0.5);
					obstacles.set(i, obstacle);
					obstacles.remove(j);
					j--;
				}
			}
		}
	}

	/**
	 * Returns a new list containing only the elements of the source list that
	 * were accepted by the filter and are sorted using the passed comparator.
	 * Filter or comparator may be null, but it does not make sense to pass both
	 * as null since the list will then be unchanged.
	 * @param sourceList the list with the data to filter and sort
	 * @param filter the filter to use for each element of the source list, null
	 *        if no filter should be used
	 * @param comparator for sorting, null if order should not be changed
	 * @return the filtered and sorted list
	 */
	public List<IPlayer> getFilteredAndSortedList(List<IPlayer> sourceList,
			IFilter filter, Comparator<? super IPlayer> comparator)
	{
		List<IPlayer> result = filter(sourceList, filter);
		if (comparator != null) {
			Collections.sort(result, comparator);
		}
		return result;
	}

	/**
	 * @param sourceList the list of players to filter
	 * @param filter the filter to use
	 * @return a new list that only contains the elements of the passed list that
	 *         passed the filter
	 */
	private List<IPlayer> filter(List<IPlayer> sourceList, IFilter filter)
	{
		List<IPlayer> result = new ArrayList<IPlayer>(sourceList.size());
		for (IPlayer player : sourceList) {
			if (filter == null || filter.accept(player)) {
				result.add(player);
			}
		}
		return result;
	}

	/**
	 * Instances allow to filter players from lists
	 * @author Klaus Dorer
	 */
	private interface IFilter
	{
		/**
		 * @param player the player to check
		 * @return true if the player is accepted by the filter, false if rejected
		 */
		boolean accept(IPlayer player);
	}

	private class BaseFilter implements IFilter
	{
		/** link to the global map of objects */
		protected final IGlobalMap worldModel;

		/** decorated filter, null if none is decorated */
		protected final IFilter decoratedFilter;

		public BaseFilter(IGlobalMap worldModel)
		{
			this.worldModel = worldModel;
			this.decoratedFilter = null;
		}

		public BaseFilter(IGlobalMap worldModel, IFilter decoratee)
		{
			this.worldModel = worldModel;
			this.decoratedFilter = decoratee;
		}

		@Override
		public boolean accept(IPlayer player)
		{
			if (decoratedFilter != null) {
				boolean result = decoratedFilter.accept(player);
				if (!result) {
					return false;
				}
			}

			// move this into its own filter, once this player should not be
			// filtered in one case
			if (player == null
					|| (player.isOwnTeam() && player.getID() == worldModel
							.getThisPlayer().getID())) {
				// never take ourselves into account
				return false;
			}

			if (player.getAge(worldModel.getGlobalTime()) > 4.0) {
				// only take into account recently seen players
				return false;
			}

			return localAccept(player);
		}

		/**
		 * Overwrite this method in subclasses to add filter criteria
		 * @param player the player to check for acceptance
		 * @return true if the player should remain in the list
		 */
		protected boolean localAccept(IPlayer player)
		{
			return true;
		}
	}

	/**
	 * Accepts only players of own team
	 */
	private class TeammateFilter extends BaseFilter
	{
		public TeammateFilter(IGlobalMap worldModel)
		{
			super(worldModel);
		}

		public TeammateFilter(IGlobalMap worldModel, IFilter decoratee)
		{
			super(worldModel, decoratee);
		}

		@Override
		public boolean localAccept(IPlayer player)
		{
			return !(player == null || !player.isOwnTeam());

		}
	}

	/**
	 * Accepts only players of other team
	 */
	private class OpponentFilter extends BaseFilter
	{
		public OpponentFilter(IGlobalMap worldModel)
		{
			super(worldModel);
		}

		public OpponentFilter(IGlobalMap worldModel, IFilter decoratee)
		{
			super(worldModel, decoratee);
		}

		@Override
		public boolean localAccept(IPlayer player)
		{
			return !(player == null || player.isOwnTeam());
		}
	}

	/**
	 * Accepts only players that are not goalies
	 */
	private class NotGoalieFilter extends BaseFilter
	{
		public NotGoalieFilter(IGlobalMap worldModel)
		{
			super(worldModel);
		}

		public NotGoalieFilter(IGlobalMap worldModel, IFilter decoratee)
		{
			super(worldModel, decoratee);
		}

		@Override
		public boolean localAccept(IPlayer player)
		{
			return !player.isGoalie();
		}
	}

	private abstract class IFOComparator implements Comparator<IVisibleObject>
	{
		protected final IGlobalMap worldModel;

		public IFOComparator(IGlobalMap worldModel)
		{
			this.worldModel = worldModel;
		}
	}

	/**
	 * For player at ball calculation
	 * @author Klaus Dorer
	 */
	private class DistanceToBallComparator extends IFOComparator
	{
		public DistanceToBallComparator(IGlobalMap worldModel)
		{
			super(worldModel);
		}

		@Override
		public int compare(IVisibleObject player1, IVisibleObject player2)
		{
			Vector3D ballPosition = worldModel.getBall().getPosition();
			double distance1 = player1.getDistanceToXY(ballPosition);
			if (player1.isLaying()) {
				distance1 += IMagmaConstants.DISTANCE_PENALTY_LYING;
			}
			double distance2 = player2.getDistanceToXY(ballPosition);
			if (player2.isLaying()) {
				distance2 += IMagmaConstants.DISTANCE_PENALTY_LYING;
			}

			if (distance1 < distance2) {
				return -1;
			} else if (distance1 > distance2) {
				return 1;
			} else {
				return 0;
			}
		}
	}

	/**
	 * Criteria is the distance to this player
	 */
	private class DistanceToMeComparator extends IFOComparator
	{
		public DistanceToMeComparator(IGlobalMap worldModel)
		{
			super(worldModel);
		}

		@Override
		public int compare(IVisibleObject player1, IVisibleObject player2)
		{
			IThisPlayer thisPlayer = worldModel.getThisPlayer();
			double distance1 = thisPlayer.getDistanceToXY(player1.getPosition());
			double distance2 = thisPlayer.getDistanceToXY(player2.getPosition());

			if (distance1 < distance2) {
				return -1;
			} else if (distance1 > distance2) {
				return 1;
			} else {
				return 0;
			}
		}
	}
}
