/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.flags.IFlagModel;
import magma.agent.perception.ICameraTiltPerceptor;
import magma.agent.perception.ICommandPerceptor;
import magma.agent.perception.IPerception;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.thoughtmodel.strategy.IRole;
import magma.agent.thoughtmodel.strategy.IRoleManager;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.util.geometry.Angle;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * @author Stefan Glaser
 */
public class ThoughtModel implements IThoughtModel, Serializable
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** the indexical functional object calculator. */
	private transient IFOCalculator ifoCalculator;

	/** list of all team-mates sorted by the distance to the ball. */
	private List<IPlayer> teamMateAtBallList;

	/** list of all opponents sorted by the distance to the ball. */
	private List<IPlayer> opponentsAtBallList;

	/** list of all players sorted by the distance to the ball. */
	private List<IPlayer> playersAtBallList;

	/** list of all players sorted by the distance to me. */
	private List<IPlayer> playersAtMeList;

	/** list of all opponents sorted by the distance to me. */
	private List<IPlayer> opponentsAtMeList;

	/** list of all obstacles to avoid sorted by the distance to me. */
	private List<IVisibleObject> obstacles;

	private InfluenceMap influenceMap;

	/** The kick direction profiler. */
	protected transient KickDirectionProfiler kickDirector;

	/** The role manager. */
	protected transient IRoleManager roleManager;

	/** The current role of the agent. */
	protected transient IRole role;

	/** The agent model. */
	protected final IAgentModel agentModel;

	/** The world model. */
	protected final IGlobalMap worldModel;

	/** The flags model. */
	protected transient final IFlagModel flags;

	/** Set to be used for remote control the agent */
	protected String remoteCommand;

	/** Indicator for camera tilt */
	protected String cameraTilt;

	public ThoughtModel(IAgentModel agentModel, IGlobalMap worldModel,
			IFlagModel flags)
	{
		this.agentModel = agentModel;
		this.worldModel = worldModel;
		this.flags = flags;

		ifoCalculator = new IFOCalculator(worldModel);
		influenceMap = new InfluenceMap(this);
		kickDirector = new KickDirectionProfiler(180, 7f);

		roleManager = null;
		role = null;
		// role = roleManager.determineRole(this, thisPlayer.getID());

		// functional Object list
		teamMateAtBallList = new ArrayList<IPlayer>(0);
		opponentsAtBallList = new ArrayList<IPlayer>(0);
		playersAtBallList = new ArrayList<IPlayer>(0);
		playersAtMeList = new ArrayList<IPlayer>(0);
		opponentsAtMeList = new ArrayList<IPlayer>(0);
		obstacles = new ArrayList<IVisibleObject>(0);
	}

	@Override
	public IAgentModel getAgentModel()
	{
		return agentModel;
	}

	@Override
	public IGlobalMap getWorldModel()
	{
		return worldModel;
	}

	@Override
	public IFlagModel getFlags()
	{
		return flags;
	}

	@Override
	public boolean update(IPerception perception)
	{
		boolean result = false;
		if (perception != null) {
			// Trigger flags model to process flag information
			flags.update(perception);

			// Trigger agent model to update all internal sensor values
			result = agentModel.update(perception);

			// Trigger world model to process vision information
			result = worldModel.update(perception) || result;

			ICommandPerceptor commandPerceptor = perception.getComandPerceptor();
			if (commandPerceptor != null) {
				remoteCommand = commandPerceptor.getCommand();
				// System.out.println("Remote Command: " + remoteCommand);
			}

			ICameraTiltPerceptor cameraTiltPerceptor = perception
					.getCameraTiltPerceptor();
			if (cameraTiltPerceptor != null) {
				cameraTilt = cameraTiltPerceptor.getCameraTilt();
				// System.out.println("Camera tilt: " + cameraTilt);
			}
		}

		// Recalculate indexical functional objects
		calculateIndexicalFunctionalObjects();

		// recalculate kick directions
		kickDirector.resetProfile();

		// determine new states based on actualized visual information
		if (perception == null || perception.containsVision()) {
			role = roleManager.determineRole(getClosestOwnPlayerAtBall(),
					worldModel.getThisPlayer().getID());
		}
		return result;
	}

	/**
	 * Calculates the lists of players sorted by different criteria.
	 */
	protected void calculateIndexicalFunctionalObjects()
	{
		List<IPlayer> visiblePlayers = worldModel.getVisiblePlayers();
		teamMateAtBallList = ifoCalculator.getTeammatesAtBall(visiblePlayers);
		opponentsAtBallList = ifoCalculator.getOpponentsAtBall(visiblePlayers);
		playersAtBallList = ifoCalculator.getPlayersAtBall(visiblePlayers);
		playersAtMeList = ifoCalculator.getPlayersAtMe(visiblePlayers);
		opponentsAtMeList = ifoCalculator.getOpponentsAtMe(playersAtMeList);
		obstacles = ifoCalculator.getObstacles(visiblePlayers,
				worldModel.getBall());
	}

	// --------------------------------------------------
	// IFOs
	@Override
	public List<IVisibleObject> getObstacles()
	{
		return obstacles;
	}

	@Override
	public List<IPlayer> getOpponentsAtMeList()
	{
		return opponentsAtMeList;
	}

	@Override
	public List<IPlayer> getPlayersAtMeList()
	{
		return playersAtMeList;
	}

	@Override
	public List<IPlayer> getTeammatesAtBall()
	{
		return teamMateAtBallList;
	}

	@Override
	public IPlayer getTeammateAtBall()
	{
		if (teamMateAtBallList.isEmpty()) {
			return null;
		}
		return teamMateAtBallList.get(0);
	}

	@Override
	public List<IPlayer> getOpponentsAtBallList()
	{
		return opponentsAtBallList;
	}

	@Override
	public IPlayer getOpponentAtBall()
	{
		if (opponentsAtBallList.isEmpty()) {
			return null;
		}
		return opponentsAtBallList.get(0);
	}

	@Override
	public List<IPlayer> getPlayersAtBallList()
	{
		return playersAtBallList;
	}

	@Override
	public IPlayer getClosestOwnPlayerAtBall()
	{
		List<IPlayer> playersAtBall = getTeammatesAtBall();
		List<IPlayer> newSort = new ArrayList<IPlayer>(playersAtBall);
		// we add ourselves to have the same calculation for all
		newSort.add(worldModel.getThisPlayer());

		return Collections.min(newSort, new DistanceToBallComparator(worldModel));
	}

	// --------------------------------------------------
	// Influence Map
	@Override
	public Vector3D getPointToKickBall()
	{
		return influenceMap.getTargetKickPositionForBall();
	}

	@Override
	public List<Vector3D[]> getInfluencePolys()
	{
		return influenceMap.getInfluenceArea(worldModel.getBall());
	}

	// --------------------------------------------------
	// Kick direction profile and estimation
	@Override
	public float[][] getKickDirectionProfile()
	{
		return kickDirector.getKickProfile(this);
	}

	@Override
	public Angle getPreferredKickDirection()
	{
		// Vector3D dirVec = getOtherGoalPosition().subtract(
		// thisPlayer.getPosition());
		// return Angle.rad(dirVec.getAlpha());
		// return Angle.deg(-80);

		return kickDirector.getBestKickDirection(this);
	}

	// --------------------------------------------------
	// Strategy and Role management
	@Override
	public IRoleManager getRoleManager()
	{
		return roleManager;
	}

	@Override
	public IRole getRole()
	{
		return role;
	}

	public void setRoleManager(IRoleManager manager, IRole role)
	{
		roleManager = manager;
		this.role = role;
	}

	// --------------------------------------------------
	// Other stuff
	@Override
	public boolean ownTeammateHasBall(double delta)
	{
		IPlayer teammateAtBall = getTeammateAtBall();
		if (teammateAtBall == null) {
			return false;
		}

		IPlayer opponentAtBall = getOpponentAtBall();
		if (opponentAtBall == null) {
			return true;
		}

		Vector3D ballPos = worldModel.getBall().getPosition();
		return opponentAtBall.getDistanceToXY(ballPos) + delta >= teammateAtBall
				.getDistanceToXY(ballPos);

	}

	/**
	 * @return the remoteCommand
	 */
	@Override
	public String getRemoteCommand()
	{
		// return "verbeugung";
		// return "kneebend";
		return remoteCommand;
	}

	@Override
	public String getCameraTilt()
	{
		return cameraTilt;
	}
}
