/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel.strategy.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import magma.agent.thoughtmodel.strategy.IRole;
import magma.agent.thoughtmodel.strategy.IRoleManager;
import magma.agent.thoughtmodel.strategy.ITeamStrategy;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * The RolePositionMangager sets up a strategy for the whole team. Here several
 * strategies could be set up.
 * 
 * @author Stephan Kammerer
 */
public class RoleManager implements IRoleManager
{
	private final RolePriorityComperator comperator = new RolePriorityComperator();

	private ITeamStrategy strategy;

	private IGlobalMap worldModel;

	/**
	 * This unit cares about the different roles and how they work together or
	 * with each other. Calculation of priorities.
	 */
	public RoleManager(IGlobalMap worldModel, ITeamStrategy strategy)
	{
		this.worldModel = worldModel;
		this.strategy = strategy;
	}

	@Override
	public IRole determineRole(IPlayer closestPlayer, int playerID)
	{

		HashMap<IPlayer, IRole> roleAssignments = new HashMap<IPlayer, IRole>();
		Vector3D ballPos = worldModel.getBall().getPosition();
		IThisPlayer thisPlayer = worldModel.getThisPlayer();

		// Am I goal keeper or closest player we can immediately return.
		if (thisPlayer.getID() == 1) {
			return IRoleConstants.DUMMY_ROLE; // strategy.getRoleForPlayerID(1);
		} else if (closestPlayer == thisPlayer) {
			return IRoleConstants.DUMMY_ROLE;//
			// strategy.getRoleForPlayerID(1);
		}

		List<IPlayer> visiblePlayers = worldModel.getVisiblePlayers();
		List<IPlayer> remainingPlayers = new ArrayList<IPlayer>();

		for (IPlayer player : visiblePlayers) {
			if (player.isOwnTeam() && player.getID() != 1
					&& player != closestPlayer) {
				remainingPlayers.add(player);
			}
		}
		// We have to add thisPlayer because he is not in remainingPlayers yet
		if (!remainingPlayers.contains(thisPlayer)) {
			remainingPlayers.add(thisPlayer);
		}

		// Calculating priorities
		List<IRole> availableRoles = new ArrayList<IRole>(
				strategy.getAvailableRoles());
		for (int i = 0; i < availableRoles.size(); i++) {
			availableRoles.get(i).update(worldModel);
		}
		Collections.sort(availableRoles, comperator); // Sorting priorities

		int roleIndex = 0;
		IRole currentRole;
		Vector3D currentTargetPose;
		Vector3D playerPos;
		float currentDistance;
		float closestDistance;

		while (remainingPlayers.size() > 0) {
			closestPlayer = null;
			closestDistance = 0;
			currentRole = availableRoles.get(roleIndex);
			currentTargetPose = currentRole.getTargetPosition().getPosition();
			currentTargetPose = new Vector3D(currentTargetPose.getX(),
					currentTargetPose.getY(), 0); // we just need X and Y

			for (int i = 0; i < remainingPlayers.size(); i++) {
				playerPos = remainingPlayers.get(i).getPosition();
				playerPos = new Vector3D(playerPos.getX(), playerPos.getY(), 0);
				// Calculate distances
				currentDistance = (float) currentTargetPose.subtract(playerPos)
						.getNorm();

				if (closestPlayer == null || currentDistance < closestDistance) {
					closestPlayer = remainingPlayers.get(i);
					closestDistance = currentDistance;
				}
			}

			roleAssignments.put(closestPlayer, currentRole);
			remainingPlayers.remove(closestPlayer);
			roleIndex++;
		}

		// Dynamic strategy change
		double actualBallX = ballPos.getX();
		double startField = -worldModel.fieldHalfLength();
		double def2_limit = -worldModel.fieldHalfLength() / 2;
		double def1_limit = 0;
		double off1_limit = worldModel.fieldHalfLength() / 2;
		double off2_limit = worldModel.fieldHalfLength();

		// Check where the ball is. It's important for choosing our strategy
		// TODO switch strategies
		if (actualBallX == startField || actualBallX <= def2_limit) {
			// System.out.println("ball is in def2");
		} else if (actualBallX > def2_limit && actualBallX <= def1_limit) {
			// System.out.println("ball is in def1");
		} else if (actualBallX > def1_limit && actualBallX <= off1_limit) {
			// System.out.println("ball is in off1");
		} else if (actualBallX > off1_limit && actualBallX <= off2_limit) {
			// System.out.println("ball is in off2");
		}

		IRole result = roleAssignments.get(thisPlayer);
		if (result == null) {
			System.out.println("Unexpected null role: " + roleAssignments);
			return IRoleConstants.DUMMY_ROLE;
		}
		return result;
	}

	private class RolePriorityComperator implements Comparator<IRole>
	{
		@Override
		public int compare(IRole o1, IRole o2)
		{
			if (o1.getPriority() < o2.getPriority()) {
				return 1;
			} else if (o1.getPriority() > o2.getPriority()) {
				return -1;
			}

			return 0;
		}
	}
}
