/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel.strategy.impl.roles;

import magma.agent.meta.server.impl.RelativeHomePosition;
import magma.agent.worldmodel.IGlobalMap;
import magma.util.ValueUtil;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * This role primarily guards the ball, trying to be between the ball and our
 * goal. Each of our guards has its own offsets to the ball.
 * 
 * @author Stephan Kammerer
 */
public class Ballguard extends Role
{
	/** Offset in x direction behind the ball */
	private final float xBallOffset;

	public Ballguard(String name, IGlobalMap worldModel, float xBallOffset,
			RelativeHomePosition relativeHomePosition, float priority)
	{
		super(name, relativeHomePosition, priority, //
				-worldModel.fieldHalfLength(), // minX
				worldModel.fieldHalfLength() - 3); // maxX

		this.xBallOffset = xBallOffset;
	}

	@Override
	protected Pose2D determinePosition(IGlobalMap worldModel)
	{
		Vector3D ballPos = worldModel.getBall().getPosition();
		Vector3D ownGoalPos = worldModel.getOwnGoalPosition();

		Vector3D homePos = homePosition.getHomePosition(
				worldModel.fieldHalfLength(), worldModel.fieldHalfLength());
		float absYWeight = Math.abs(homePosition.getYWeight());
		int sideFactor = homePos.getY() > 0 ? 1 : -1;

		// Comparison if our y-position is within a specific range. With this
		// query we can choose where the player is located (e.g. left, center,
		// right).

		// offensive center
		Vector3D pos;
		float xOffset = xBallOffset;
		// if (xBallOffset > 3.5
		// && worldModel.getBall().getDistanceToXY(ownGoalPos) < 11) {
		// // if we are in own back, we move back ballguards forward
		// xOffset -= 2;
		// }
		if (absYWeight < 0.1) {
			pos = Geometry.getPointOnLineAbsoluteEnd(ownGoalPos, ballPos, xOffset);
		} else {
			// left or right Ballguard
			if (sideFactor < 0) {
				pos = Geometry.getPointOnOrthogonal2D(ownGoalPos, ballPos, xOffset,
						-2.0);
			} else {
				pos = Geometry.getPointOnOrthogonal2D(ownGoalPos, ballPos, xOffset,
						1.3);
			}
		}

		// keep positions inside field
		double targetX = pos.getX();
		double targetY = pos.getY();
		targetY = ValueUtil.limitAbs(targetY, worldModel.fieldHalfWith() - 0.5);
		targetX = keepXLimits(targetX);

		Vector3D target = new Vector3D(targetX, targetY, 0);

		// check if we ended where some other player is currently located
		// temporarily disabled this feature, because more than one is in our way
		// target = checkOwnPlayerInWay(target, worldModel);
		Angle ballDirection = Angle.rad(ballPos.subtract(target).getAlpha());

		return new Pose2D(target, ballDirection);
	}

	// /**
	// * Checks if an own player is already at that position
	// * @param target the position to check
	// * @param worldModel
	// * @return the possibly new position if one is in way
	// */
	// private Vector3D checkOwnPlayerInWay(Vector3D target, IGlobalMap
	// worldModel)
	// {
	// final double KEEP_OFF_DISTANCE = 0.7;
	// List<IPlayer> teammates = worldModel.getTeammatesAtBall();
	// for (IPlayer teammate : teammates) {
	//
	// double distance = teammate.getDistanceToXY(target);
	// if (distance < KEEP_OFF_DISTANCE) {
	// System.out.println("A teammate is in our way" + teammate.getID());
	// // a teammate is close to where we run, so avoid that position by 1m
	// // we do not check again for the new position
	// Vector3D myPos = worldModel.getThisPlayer().getPosition();
	// Vector3D newPos = Geometry.getPointOnLineAbsoluteEnd(myPos,
	// teammate.getPosition(), KEEP_OFF_DISTANCE);
	// return newPos;
	// }
	// }
	//
	// return target;
	// }
}
