/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel.strategy.impl.roles;

import magma.agent.meta.server.impl.RelativeHomePosition;
import magma.agent.worldmodel.IGlobalMap;
import magma.util.ValueUtil;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * This role is like our normal defenders, but if the ball is really far away,
 * means in the enemy territory they are allowed to move further a bit.
 * 
 * But since our guards also go back to our penalty area, we move our defenders
 * a little bit into this area to avoid collision.
 * 
 * @author Stephan Kammerer
 */
public class OffensiveDefender extends Role
{

	public OffensiveDefender(String name, IGlobalMap worldModel,
			RelativeHomePosition relativeHomePosition, float priority)
	{
		super(name, relativeHomePosition, priority, //
				-worldModel.fieldHalfLength() + 0.3, // minX
				-worldModel.fieldHalfLength() + 5); // maxX
	}

	@Override
	protected Pose2D determinePosition(IGlobalMap worldModel)
	{
		Vector3D ownGoalPos = worldModel.getOwnGoalPosition();
		Vector3D ballPos = worldModel.getBall().getPosition();
		Vector3D homePos = homePosition.getHomePosition(
				worldModel.fieldHalfLength(), worldModel.fieldHalfWith());

		double deltaBallGoalX = Math.abs(ownGoalPos.getX() - ballPos.getX());
		double offensiveOffset = 0.0;
		double targetY, targetX = 0.0;

		// If ball is far in the enemy territory, our defenders should be more
		// offensive
		if (ballPos.getX() > worldModel.fieldHalfLength() / 2) {
			offensiveOffset = 1.5;
		}

		// Don't go further than 5m + offset
		targetX = minX + 0.3 * deltaBallGoalX;

		targetY = homePos.getY() + ballPos.getY() / 3;
		targetY = ValueUtil.limitAbs(targetY, 3.4);

		targetX = keepLimits(targetX, minX, maxX + offensiveOffset);

		return new Pose2D(targetX, targetY, Angle.ZERO);
	}
}
