/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel.strategy.impl.roles;

import magma.agent.meta.server.impl.RelativeHomePosition;
import magma.agent.thoughtmodel.strategy.IRole;
import magma.agent.worldmodel.IGlobalMap;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * This is the base class for roles. Each role knows its relative home-position
 * (x-, and y-coordinate) on the field. The determine Position-method returns
 * the target point each role has.
 * 
 * @author Stephan Kammerer
 */
public abstract class Role implements IRole
{
	private static final float SAFETY_DISTANCE = 0.3f;

	/** The name of the Role */
	protected final String name;

	/** Specific basic priority a role has in the team */
	protected final float basePriority;

	/** The dynamically calculated priority. */
	protected float priority;

	/** The target pose. */
	protected Pose2D targetPose;

	/** Minimal x-position where this role moves */
	protected final double minX;

	/** Maximal x-position where this role moves */
	protected final double maxX;

	/**
	 * Position of my home position. The offset is meant (in x-position) for
	 * offensive or defensive play.
	 */
	protected final RelativeHomePosition homePosition;

	/**
	 * @param name name of role
	 * @param relativeHomePosition the relative home position
	 * @param priority Our priority of role
	 * @param minX Minimal x-position where this agent moves
	 * @param maxX Maximal x-position where this agent moves
	 * @param positioning Indicates the order (in y-direction)
	 */
	public Role(String name, RelativeHomePosition relativeHomePosition,
			float priority, double minX, double maxX)
	{
		this.name = name;
		this.homePosition = relativeHomePosition;
		this.basePriority = priority;
		this.targetPose = new Pose2D();
		this.minX = minX;
		this.maxX = maxX;
	}

	@Override
	public String getName()
	{
		return name;
	}

	@Override
	public float getPriority()
	{
		return priority;
	}

	protected float getExtraPriority(IGlobalMap worldModel)
	{
		float fieldHalfLength = worldModel.fieldHalfLength();
		Vector3D goalVector = new Vector3D(fieldHalfLength, 0, 0);

		Vector3D goalBallPos = worldModel.getBall().getPosition().add(goalVector);
		Pose2D trans = new Pose2D(-fieldHalfLength, 0, Angle.rad(goalBallPos
				.getAlpha()));
		Vector3D tempPos = trans.applyInverseTo(targetPose.getPosition().add(
				goalVector));

		return (float) (1 - (Math.abs(tempPos.getY()) / worldModel
				.fieldHalfWith()));

		// return (float) (1 - Math.abs(targetPos.getY())
		// / worldModel.fieldHalfWith());
	}

	@Override
	public void update(IGlobalMap worldModel)
	{
		// update target pose
		switch (worldModel.getGameState()) {
		case OWN_GOAL:
		case OPPONENT_GOAL:
		case BEFORE_KICK_OFF:
			Vector3D homepos = homePosition.getHomePosition(
					worldModel.fieldHalfLength(), worldModel.fieldHalfWith());
			targetPose = new Pose2D(homepos.getX(), homepos.getY(), Angle.ZERO);
			break;
		default:
			targetPose = determinePosition(worldModel);
		}

		targetPose = avoidPenaltyArea(worldModel, targetPose);

		// update priority
		priority = basePriority + getExtraPriority(worldModel);
	}

	@Override
	public Pose2D getTargetPosition()
	{
		return targetPose;
	}

	/**
	 * @param worldModel
	 * @return
	 */
	protected abstract Pose2D determinePosition(IGlobalMap worldModel);

	/**
	 * @param target
	 * @param min
	 * @param max
	 * @return
	 */
	protected double keepLimits(double target, double min, double max)
	{
		if (target < min) {
			target = min;
		}

		if (target > max) {
			target = max;
		}
		return target;
	}

	/**
	 * @param targetX the target x-position
	 * @return the limited target x-position
	 */
	protected double keepXLimits(double targetX)
	{
		if (targetX < minX) {
			targetX = minX;
		}

		if (targetX > maxX) {
			targetX = maxX;
		}
		return targetX;
	}

	protected Pose2D avoidPenaltyArea(IGlobalMap worldModel, Pose2D target)
	{
		float penaltyLineY = worldModel.penaltyHalfLength()
				+ worldModel.goalHalfWith() + SAFETY_DISTANCE;
		float penaltyLineX = -worldModel.fieldHalfLength()
				+ worldModel.penaltyWith() + SAFETY_DISTANCE;

		if (Math.abs(target.getY()) < penaltyLineY
				&& target.getX() < penaltyLineX) {
			float newX = penaltyLineX;
			return new Pose2D(newX, target.getY(), target.getAngle());
		}

		return target;
	}

	@Override
	public String toString()
	{
		return name + " " + priority;
	}
}
