/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.thoughtmodel.strategy.impl.roles;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import magma.agent.meta.server.impl.RelativeHomePosition;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IPlayer;
import magma.util.ValueUtil;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * This quite normal defender scouts at the enemy team. His maximal x-position
 * is relative to the front enemy player.
 * 
 * @author Stephan Kammerer and Maximilian Kroeg
 */
public class ScoutingDefender extends Role
{
	private final XPositionComperator comperator = new XPositionComperator();

	public ScoutingDefender(String name, IGlobalMap worldModel,
			RelativeHomePosition relativeHomePosition, float priority)
	{
		super(name, relativeHomePosition, priority, //
				0, // minX
				0); // maxX
	}

	@Override
	protected Pose2D determinePosition(IGlobalMap worldModel)
	{
		double minX, maxX = 0.0;
		Vector3D ballPos = worldModel.getBall().getPosition();
		Vector3D ownGoalPos = worldModel.getOwnGoalPosition();

		List<IPlayer> visiblePlayers = worldModel.getVisiblePlayers();
		List<IPlayer> opponents = new ArrayList<IPlayer>();

		for (IPlayer player : visiblePlayers) {
			if (!player.isOwnTeam()) {
				opponents.add(player);
			}
		}

		// Sort opponents by x-position
		Collections.sort(opponents, comperator);

		// Now we know the opponent-player who is in front of the opponent team
		if (opponents.size() > 0) {
			IPlayer frontEnemyPlayer = opponents.get(0);
			maxX = frontEnemyPlayer.getPosition().getX();
		}

		minX = -worldModel.fieldHalfLength() + 0.3;

		Vector3D homePos = homePosition.getHomePosition(
				worldModel.fieldHalfLength(), worldModel.fieldHalfWith());

		double targetY = homePos.getY() + ballPos.getY() / 3;
		targetY = ValueUtil.limitAbs(targetY, 3.4);

		double deltaBallGoalX = Math.abs(ownGoalPos.getX() - ballPos.getX());

		double targetX = minX + 0.3 * deltaBallGoalX;
		if (Math.abs(targetY) < worldModel.penaltyHalfLength()
				+ worldModel.goalHalfWith() + 0.3) {
			// make sure not to run into penalty area
			minX = -worldModel.fieldHalfLength() + worldModel.penaltyWith() + 0.3;
			maxX = -worldModel.fieldHalfLength() + 7;
		}

		targetX = keepLimits(targetX, minX, maxX);

		return new Pose2D(targetX, targetY, Angle.ZERO);

	}

	private class XPositionComperator implements Comparator<IPlayer>
	{
		@Override
		public int compare(IPlayer p1, IPlayer p2)
		{
			if (p1.getPosition().getX() < p2.getPosition().getX()) {
				return 1;
			} else if (p1.getPosition().getX() > p2.getPosition().getX()) {
				return -1;
			}

			return 0;
		}
	}
}
