/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel;

import java.util.Collection;
import java.util.List;

import magma.agent.perception.IPerception;
import magma.util.observer.IObserver;
import magma.util.observer.ISubscribe;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Represents all environment data available to the agent
 */
public interface IGlobalMap extends ISoccerPitchDescription,
		ISubscribe<IGlobalMap>
{
	/**
	 * Called to trigger a update of the AgentModel based on the given perception
	 * object.
	 * 
	 * @param perception the Perception
	 */
	boolean update(IPerception perception);

	/**
	 * Get a reference to the ball object
	 * 
	 * @return The ball object
	 */
	IMoveableObject getBall();

	/**
	 * Get a list of all visible landmarks
	 * 
	 * @return An unmodifiable collection of the landmarks
	 */
	Collection<ILandmark> getLandmarks();

	/**
	 * @param name the name of the landmark
	 * @return the landmark specified by name
	 */
	ILandmark getLandmark(String name);

	/**
	 * Calculate the average error between known and calculated position
	 * 
	 * @return the average error of known position and calculated position,
	 *         Double.MAX_VALUE if we do not see any flags
	 */
	double getLandmarkError();

	/**
	 * Get a list of all visible field lines
	 * 
	 * @return An unmodifiable collection of the field lines
	 */
	Collection<IFieldLine> getFieldLines();

	/**
	 * @param name the name of the field line
	 * @return the field line specified by name
	 */
	IFieldLine getFieldLine(String name);

	/**
	 * Get a list of all visible players
	 * 
	 * @return An unmodifiable list of the currently visible players
	 */
	List<IPlayer> getVisiblePlayers();

	/**
	 * Get a reference to the player object of this player
	 * 
	 * @return This players information of position etc
	 */
	IThisPlayer getThisPlayer();

	/**
	 * Get the current global time
	 * 
	 * @return the time on the global clock
	 */
	float getGlobalTime();

	/**
	 * Get the running time of the current game
	 * 
	 * @return the time the game is running now
	 */
	float getGameTime();

	/**
	 * Get the current play mode
	 * @return the current mode of the game as String
	 */
	GameState getGameState();

	/**
	 * Get home position for this agent in conjunction with the current game
	 * state
	 * 
	 * @return vector to home position
	 */
	Vector3D getHomePosition();

	/**
	 * Get the number of goals our own team scored
	 * 
	 * @return number of goals scored by us
	 */
	int getGoalsWeScored();

	/**
	 * Get the number of goals the opponent team scored
	 * 
	 * @return number of goals scored by opponent
	 */
	int getGoalsTheyScored();

	/**
	 * @return the serverVersion used for playing
	 */
	int getServerVersion();

	/**
	 * Detaches the passed observer to this connection
	 * 
	 * @param observer the observer that is no longer interested in new messages
	 * @return true if the observer was in the list and detached
	 */
	boolean detach(IObserver<IGlobalMap> observer);

	/**
	 * @param id the player's number (1-11)
	 * @param ownTeam true if we want the player of our own team
	 * @return the referenced player, null if not in list
	 */
	IPlayer getVisiblePlayer(int id, boolean ownTeam);

	/**
	 * Writes out current situation information. Used for learning situations and
	 * implementation will change.
	 * @param which allows to indicate different situations
	 */
	void logSituation(int which);
}