/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel;

import java.io.Serializable;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

public interface ISoccerPitchDescription extends Serializable
{
	/**
	 * @return the half length of the soccer field
	 */
	public float fieldHalfLength();

	/**
	 * @return the half width of the soccer field
	 */
	public float fieldHalfWith();

	/**
	 * @return the half width of the goal
	 */
	public float goalHalfWith();

	/**
	 * @return the height of the goal
	 */
	public float goalHeight();

	/**
	 * @return the half length of the penalty area
	 */
	public float penaltyHalfLength();

	/**
	 * @return the width of the penalty area
	 */
	public float penaltyWith();

	/**
	 * @return the radius of the center circle
	 */
	public float centerCircleRadius();

	/**
	 * Retrieve the position of the own goal (position between the goal posts).
	 * 
	 * @return the position in the middle of the own goal posts
	 */
	Vector3D getOwnGoalPosition();

	/**
	 * Retrieve the position of the other goal (position between the goal posts).
	 * 
	 * @return the position in the middle of the other goal posts
	 */
	Vector3D getOtherGoalPosition();

	/**
	 * Checks if the given position is inside the own penalty area and the
	 * specified margin.
	 * 
	 * @param position the position to test
	 * @param margin the margin about which the own penalty area is enlarged
	 *        (positive values) or narrowed (negative values)
	 * @return true if the position is inside the own penalty area plus margin
	 */
	boolean isInOwnPenaltyArea(Vector3D position, float margin);

	/**
	 * Checks if the given position is inside the own half and the specified
	 * margin.
	 * 
	 * @param position the position to test
	 * @param margin the margin about which the area is enlarged (positive
	 *        values) or narrowed (negative values)
	 * @return true if the position is inside the own half plus margin
	 */
	boolean isInOwnHalf(Vector3D position, float margin);

	/**
	 * Checks if the given position is inside the other penalty area and the
	 * specified margin.
	 * 
	 * @param position the position to test
	 * @param margin the margin about which the own penalty area is enlarged
	 *        (positive values) or narrowed (negative values)
	 * @return true if the position is inside the other penalty area plus margin
	 */
	boolean isInOtherPenaltyArea(Vector3D position, float margin);

	/**
	 * Checks if the given position is inside the center circle area and the
	 * specified margin.
	 * 
	 * 
	 * @param position the position to test
	 * @param margin the margin about which the own penalty area is enlarged
	 *        (positive values) or narrowed (negative values)
	 * @return true if the position is inside the center circle area plus margin
	 */
	boolean isInCenterCircleArea(Vector3D position, float margin);

	/**
	 * Checks if the given position is inside the soccer pitch area and the
	 * specified margin.
	 * 
	 * @param position the position to test
	 * @param margin the margin about which the own penalty area is enlarged
	 *        (positive values) or narrowed (negative values)
	 * @return true if the position is inside the soccer pitch area plus margin
	 */
	boolean isInSoccerPitchArea(Vector3D position, float margin);
}
