/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.IMagmaConstants;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.ICompass;
import magma.agent.agentmodel.IGyroRate;
import magma.agent.meta.server.IRCServerMetaModel;
import magma.agent.perception.IGameStatePerceptor;
import magma.agent.perception.IGlobalPosePerceptor;
import magma.agent.perception.IHearPerceptor;
import magma.agent.perception.ILinePerceptor;
import magma.agent.perception.IPerception;
import magma.agent.perception.IPlayerPos;
import magma.agent.perception.ITimerPerceptor;
import magma.agent.perception.IVisibleObjectPerceptor;
import magma.agent.worldmodel.GameState;
import magma.agent.worldmodel.IFieldLine;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.ILandmark;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.agent.worldmodel.InformationSource;
import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.ILocalizer;
import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.common.spark.PlayMode;
import magma.common.spark.PlaySide;
import magma.common.spark.TeamColor;
import magma.util.SayCoder;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose3D;
import magma.util.geometry.positionFilter.IPositionFilter;
import magma.util.geometry.positionFilter.PositionFilter;
import magma.util.observer.IObserver;
import magma.util.observer.IPublishSubscribe;
import magma.util.observer.Subject;

import org.apache.commons.math3.geometry.euclidean.threed.Line;
import org.apache.commons.math3.geometry.euclidean.threed.Plane;
import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Container class for all visible objects on the field, for the time and game
 * state
 */
public class GlobalMap implements IGlobalMap, IEnvironmentModel, Serializable
{
	/**  */
	private static final float TIME_TO_TRUST_HEAR = 0.2f;

	/** the time we keep other players in mind if not seeing them (in seconds) */
	private static final int REMEMBERANCE_TIME = 30;

	// handle for logging
	private static final Logger logger = Logger.getLogger("");

	/** id of this class for serialization */
	private static final long serialVersionUID = 182243462936086603L;

	/** observers of world model */
	private final transient IPublishSubscribe<IGlobalMap> observer;

	private final transient IRCServerMetaModel serverMetaModel;

	/** the localizer that is able to calculate the absolute position */
	private final transient ILocalizer localizer;

	/** The soccer pitch description and helper class */
	protected SoccerPitchDescription soccerPitch;

	/** the ball of the game */
	protected Ball ball;

	/** the landmark visual objects (flags and goal posts) */
	private final Map<String, ILandmark> landmarks;

	/** the reference points used for localization (flags and goal posts) */
	private final transient List<IReferencePoint> referencePoints;

	/** the field line visual objects */
	private final Map<String, IFieldLine> fieldLines;

	/** the reference lines (field lines) used for localization */
	private final transient List<IReferenceLine> referenceLines;

	/**
	 * A map of all known players over time. As unique key, the team name
	 * appended by the player number is used.
	 */
	protected Map<String, IPlayer> knownPlayers;

	/** the list of all visible players */
	protected List<IPlayer> visiblePlayers;

	/** the time on the global clock (in cycles). Not the game time */
	private float globalTime;

	/** the time the game is running now (cycles since start) */
	private float gameTime;

	/** the time of the last change of the game state */
	private float lastGameStateChangeTime;

	/** The side which our team is playing. */
	// Note: It's important, that this value is initialized with the left side!
	private PlaySide playSide = PlaySide.LEFT_SIDE;

	/** color of team. Currently unknown in simspark! */
	private TeamColor teamColor = TeamColor.TeamBlue;

	/** the playmode the game is running now */
	private PlayMode playmode = PlayMode.BEFORE_KICK_OFF;

	/** the game state, the game is running now */
	protected GameState gameState = GameState.BEFORE_KICK_OFF;

	/** the game state, the game was running previously */
	protected GameState previousGameState;

	/** the global coordinates of the agent */
	protected ThisPlayer thisPlayer;

	/** goals scored by the opponent */
	protected int goalsTheyScored = 0;

	/** goals scored by us */
	protected int goalsWeScored = 0;

	/** reference to the agent model needed for gyro localization */
	private IAgentModel agentModel;

	/** The version of the soccer server used */
	protected int serverVersion;

	private transient IPositionFilter posFilter;

	/**
	 * The name of the other team.<br>
	 * <tt>null</tt> as long we haven't seen any other player.
	 */
	private String otherTeamName;

	/**
	 * The transformation from the camera body system to the root body system of
	 * the current cycle.
	 */
	private transient Pose3D cameraToRootVision;

	/** used to log some situations for learning */
	// private transient DataLogger dataLogger;

	// KlausTest
	// private transient AttitudeEstimator attitudeEstimator;
	//
	// public transient Rotation attitudeEstimation;

	/**
	 * Constructor
	 * 
	 * @param agentModel Reference to the agent model object
	 * @param localizer the module that calculates the agent's global position
	 * @param metaModel the meta model of the rc server
	 * @param teamname Team name
	 * @param playerNumber Player number
	 */
	public GlobalMap(IAgentModel agentModel, ILocalizer localizer,
			IRCServerMetaModel metaModel, String teamname, int playerNumber)
	{
		assert playerNumber < IMagmaConstants.homePositionsOwnKickoff.length : "No homePositionLeft for player id: "
				+ playerNumber;

		this.agentModel = agentModel;
		this.localizer = localizer;
		this.serverMetaModel = metaModel;
		this.serverVersion = metaModel.getVersion();
		this.soccerPitch = new SoccerPitchDescription(metaModel);

		// create objects
		landmarks = new HashMap<String, ILandmark>();
		fieldLines = new HashMap<String, IFieldLine>();
		referencePoints = new ArrayList<IReferencePoint>();
		referenceLines = new ArrayList<IReferenceLine>();
		createObjects(metaModel);

		thisPlayer = new ThisPlayer(teamname, playerNumber);
		thisPlayer.getPositionManager().setDesiredPosition(getHomePosition(),
				Angle.ZERO, false);
		observer = new Subject<IGlobalMap>();
		posFilter = new PositionFilter();
		// posFilter = new PositionKalmanFilter();
		// dataLogger = new DataLogger(this, agentModel);

		// KlausTest
		// attitudeEstimator = new AttitudeEstimator(true);
		// attitudeEstimator.setMagCalib(-1.0, 0.0, 0.0);
	}

	/**
	 * Creates all objects. We do not create new objects each time on update, but
	 * update the existing objects
	 */
	private void createObjects(IRCServerMetaModel metaModel)
	{
		ball = new Ball(metaModel.getBallSize(), metaModel.getBallDecay());

		// Landmarks
		Map<String, Vector3D> landmarkConfig = metaModel.getLandmarks();
		if (landmarkConfig != null) {
			Landmark landmark;

			for (String key : landmarkConfig.keySet()) {
				landmark = new Landmark(key, landmarkConfig.get(key));
				landmarks.put(key, landmark);
				referencePoints.add(landmark);
			}
		}

		// FieldLines
		Map<String, Vector3D[]> fieldLineConfig = metaModel.getFieldLines();
		if (fieldLineConfig != null) {
			Vector3D[] posArray;
			FieldLine fieldLine;

			for (String key : fieldLineConfig.keySet()) {
				posArray = fieldLineConfig.get(key);
				fieldLine = new FieldLine(key, posArray[0], posArray[1]);
				fieldLines.put(key, fieldLine);
				referenceLines.add(fieldLine);
			}
		}

		// Players
		int numberOfPlayersPerTeam = IMagmaConstants.NUMBER_OF_PLAYERS_PER_TEAM;
		knownPlayers = new HashMap<String, IPlayer>(numberOfPlayersPerTeam * 2);
		visiblePlayers = new ArrayList<IPlayer>(numberOfPlayersPerTeam * 2);
	}

	/**
	 * Called once perception is finished parsing a new incoming message
	 * @param perception the object containing the result from parsing
	 */
	@Override
	public boolean update(IPerception perception)
	{
		// update global time
		ITimerPerceptor timerPerceptor = perception.getTime();
		if (timerPerceptor != null) {
			globalTime = timerPerceptor.getTime();
		}

		// update game time, play side and play mode, as well as scored goals
		processGameState(perception);

		//
		// if (gyro.getGyro().getY() > 0.1) {
		// if (previousGameState == GameState.BEFORE_KICK_OFF) {
		// ICompass compass = agentModel.getTorso().getSensor(
		// ICompass.class);
		// if (compass != null) {
		// compass.reset();
		// }
		// }
		// previousGameState = gameState;
		// playmode = PlayMode.PLAY_ON;
		// gameState = GameState.PLAY_ON;
		// } else {
		// playmode = PlayMode.BEFORE_KICK_OFF;
		// gameState = GameState.BEFORE_KICK_OFF;
		// }
		// }
		// }

		// camera coordinate system is updated to current pose
		updateCameraCoordinateSystem();

		// calculate the new position of the agent
		calculatePosition(perception);

		// KlausTest
		// useAttitudeEstimator(perception);

		// update visible objects
		updateBall(perception);
		updateLandmarks(perception);
		updateFieldLines();
		updatePlayers(perception);

		processHear(perception);

		logger.log(Level.FINE, "time: {0} gameTime: {1}", new Object[] {
				globalTime, gameTime });

		// now inform observers about changes
		observer.onStateChange(this);
		return false;
	}

	// KlausTest
	// private void useAttitudeEstimator(IPerception perception)
	// {
	// IGyroRate gyroSensor = agentModel.getTorso().getSensor(IGyroRate.class);
	// if (gyroSensor == null) {
	// return;
	// }
	// Vector3D gyro = gyroSensor.getGyro();
	// // System.out.print("gyro: " + gyro);
	//
	// IAccelerometer accelerometer = agentModel.getTorso().getSensor(
	// IAccelerometer.class);
	// if (accelerometer == null) {
	// return;
	// }
	// Vector3D acceleration = accelerometer.getAcceleration();
	// // System.out.print(" acc: " + acceleration);
	//
	// attitudeEstimator.update(0.020, gyro.getX(), gyro.getY(), gyro.getZ(),
	// acceleration.getX(), acceleration.getY(), acceleration.getZ(), 0.0,
	// 0.0, 0.0 /* magnetometer */);
	// double[] attitude = attitudeEstimator.getAttitude();
	//
	// attitudeEstimation = new Rotation(attitude[0], attitude[1], attitude[2],
	// attitude[3], true);
	//
	// // Rotation oldEstimation = thisPlayer.getOrientation();
	// // System.out.print(" old: "
	// // + Arrays.toString(oldEstimation.getAngles(RotationOrder.ZYX)));
	//
	// // double distance = Rotation.distance(attitudeEstimation, oldEstimation);
	// // System.out.println(" distance: " + distance);
	// }

	private void processGameState(IPerception perception)
	{
		// update game time
		IGameStatePerceptor perceptor = perception.getGameState();
		if (perceptor == null) {
			return;
		}
		gameTime = perceptor.getTime();

		// update playSide
		String playSideString = perceptor.getTeamSide();
		PlaySide newPlaySide = PlaySide.parsePlaySide(playSideString);

		switch (newPlaySide) {
		case LEFT_SIDE:
			// System.out.println("PlaySide left");
			if (playSide != PlaySide.LEFT_SIDE) {
				// We switched from right to left side, so remember the new side and
				// mirror the flags
				playSide = PlaySide.LEFT_SIDE;
				mirrorLandmarks();
			}
			break;
		case RIGHT_SIDE:
			// System.out.println("PlaySide right");
			if (playSide != PlaySide.RIGHT_SIDE) {
				// We switched from left to right side, so remember the new side and
				// mirror the flags
				playSide = PlaySide.RIGHT_SIDE;
				mirrorLandmarks();
			}
			break;
		case UNKNOWN:
		default:
			// No playSide?!? Maybe I'm the referee?
			break;
		}

		// update team color
		teamColor = perceptor.getTeamColor();

		// update playMode
		String playModeString = perceptor.getPlaymode();
		previousGameState = gameState;

		playmode = PlayMode.parsePlayMode(playModeString);
		gameState = GameState.determineGameState(playmode, playSide);

		// If game state has changed, remember the current time
		if (gameState != previousGameState) {
			lastGameStateChangeTime = globalTime;
		}

		if (playmode == PlayMode.NONE) {
			logger.log(Level.SEVERE, "play mode not known: <{0}>",
					new Object[] { playModeString });
		}

		// update scored goals
		updateGoalsScored();
	}

	private void mirrorLandmarks()
	{
		Map<String, ILandmark> newLandmarks = new HashMap<String, ILandmark>();
		Vector3D pos;
		Landmark landmark;

		// Clear list of reference points
		referencePoints.clear();

		for (String key : landmarks.keySet()) {
			pos = landmarks.get(key).getKnownPosition();
			landmark = new Landmark(key, new Vector3D(-pos.getX(), -pos.getY(),
					pos.getZ()));
			newLandmarks.put(key, landmark);
			referencePoints.add(landmark);
		}

		landmarks.clear();
		landmarks.putAll(newLandmarks);
	}

	/**
	 * Update the number of scored goals
	 */
	private void updateGoalsScored()
	{
		if (gameState == GameState.OWN_GOAL
				&& previousGameState != GameState.OWN_GOAL) {
			goalsWeScored++;
		} else if (gameState == GameState.OPPONENT_GOAL
				&& previousGameState != GameState.OPPONENT_GOAL) {
			goalsTheyScored++;
		}
	}

	private void processHear(IPerception perception)
	{
		try {
			List<IHearPerceptor> hearList = perception.getHearPerceptors();
			for (IHearPerceptor hear : hearList) {
				String target = hear.getTarget();

				if ("self".equalsIgnoreCase(target)) {
					// message from ourselves
					break;
				}

				String message = hear.getMessage();
				if (message.length() != 4 * SayCoder.getPositionLength() + 2) {
					// our messages are fixed length
					break;
				}

				int posStart = 1;
				int posLength = SayCoder.getPositionLength();
				int playerID = SayCoder.decodeID(message.substring(0, posStart));
				Vector3D playerPos = SayCoder.decodePosition(message.substring(
						posStart, posStart + posLength));
				posStart += posLength;
				Vector3D ballPos = SayCoder.decodePosition(message.substring(
						posStart, posStart + posLength));

				posStart += posLength;
				Vector3D opponentPos = SayCoder.decodePosition(message.substring(
						posStart, posStart + posLength));

				posStart += posLength;
				Vector3D opponentPos2 = SayCoder.decodePosition(message.substring(
						posStart, posStart + posLength));

				posStart += posLength;
				String teamCode = message.substring(posStart, posStart + 1);

				if (!SayCoder.encodeName(getTeamname(true)).equals(teamCode)) {
					// message is not from our team
					break;
				}

				if (playerPos != null) {
					upatePlayerFromHear(playerID, true, playerPos);
				}

				if (ballPos != null) {
					updateBallFromHear(ballPos);
				}

				if (opponentPos != null) {
					upatePlayerFromHear(playerID, false, opponentPos);
				}

				if (opponentPos2 != null) {
					upatePlayerFromHear(
							((playerID + 5) % IMagmaConstants.NUMBER_OF_PLAYERS_PER_TEAM) + 1,
							false, opponentPos2);
				}
			}

		} catch (Exception e) {
			// might happen with opponent messages
		}
	}

	/**
	 * @param ballPos
	 */
	public void updateBallFromHear(Vector3D ballPos)
	{
		if (ball.getAge(globalTime) < TIME_TO_TRUST_HEAR
				&& ball.getInformationSource() == InformationSource.VISION) {
			// We have recently seen the ball so no update
			return;
		}

		Vector3D localPos = thisPlayer.calculateLocalPosition(ballPos);
		ball.updateFromAudio(localPos, ballPos, globalTime);
		logger.log(Level.FINER, "hear ball: ({0}, {1})",
				new Object[] { ballPos.getX(), ballPos.getY() });
	}

	/**
	 * Updates the local and global position of the corresponding player instance
	 * according to the given global position.
	 * 
	 * @param playerID the id of the player
	 * @param ownTeam true if the player to update is of the own team
	 * @param playerPos the new global position
	 */
	private void upatePlayerFromHear(int playerID, boolean ownTeam,
			Vector3D playerPos)
	{
		String teamName = getTeamname(ownTeam);
		if (teamName == null) {
			// As long we don't know the others team name, we aren't able to create
			// proper player instances
			return;
		}

		Player player = (Player) knownPlayers.get(teamName + playerID);
		if (player == null) {
			// We hear from a player that we have never seen before
			player = new Player(playerID, teamName, ownTeam);
			knownPlayers.put(teamName + playerID, player);
		}

		if (player.getAge(globalTime) < TIME_TO_TRUST_HEAR
				&& player.getInformationSource() == InformationSource.VISION) {
			// We have recently seen the player so no update
			return;
		}

		// Update player information
		Vector3D localPos = thisPlayer.calculateLocalPosition(playerPos);
		player.updateFromAudio(localPos, playerPos, globalTime);
	}

	/**
	 * If ownTeam is true we return the team name of thisPlayer. If false we
	 * return the name of the first seen other player ever. If we haven't seen
	 * any other player yet, this method returns <tt>null</tt>.
	 * @param ownTeam true if from own team, false if from other team
	 * @return the name of the team, if known, null if not
	 */
	public String getTeamname(boolean ownTeam)
	{
		if (ownTeam) {
			return thisPlayer.getTeamname();
		} else {
			return otherTeamName;
		}
	}

	/**
	 * @param perception the object containing the result from parsing
	 * 
	 */
	private void calculatePosition(IPerception perception)
	{
		IGlobalPosePerceptor globalPose = perception.getGlobalPose();
		if (globalPose != null) {
			calculatePositionFromGlobalPose(globalPose);
			return;
		}

		// Extract visible lines of the current cycle
		List<ILinePerceptor> visibleLines = perception.getVisibleLines();
		List<ILocalizationLine> visibleFieldLines = new ArrayList<ILocalizationLine>(
				visibleLines.size());
		for (ILinePerceptor pl : visibleLines) {
			visibleFieldLines.add(new LocalizationLine(toRootVisionSystem(pl
					.getPosition()), toRootVisionSystem(pl.getPosition2())));
		}

		// Reset reference lines (and implicitly the field lines as well)
		for (IReferenceLine line : referenceLines) {
			line.setVisible(false);
		}

		// Extract visible markers of the current cycle
		for (String name : landmarks.keySet()) {
			Landmark landmark = (Landmark) landmarks.get(name);
			IVisibleObjectPerceptor landmarkSeen = perception
					.getVisibleObject(name);
			if (landmarkSeen != null) {
				// this landmark is currently visible, so update it's local position
				// and store it to the visible flags for localization
				landmark.updateLocalPosition(toRootVisionSystem(landmarkSeen
						.getPosition()));
			} else {
				landmark.setVisible(false);
			}
		}

		// Calculate an estimation of the current orientation, based on the gyro
		// sensor in the root body (if existing)
		Rotation orientationEstimation = null;
		IGyroRate gyro = agentModel.getTorso().getSensor(IGyroRate.class);

		// If we have a gyro sensor in the root body, use it to estimate the
		// current orientation
		if (gyro != null) {
			Rotation globalGyro = Geometry.zTransformRotation(
					gyro.getOrientationChange(), -Math.PI / 2);
			orientationEstimation = thisPlayer.getGlobalOrientation().applyTo(
					globalGyro);
		} else {
			ICompass compass = agentModel.getTorso().getSensor(ICompass.class);

			// If we have a compass sensor in the root body, use it to estimate the
			// current orientation
			if (compass != null) {
				orientationEstimation = new Rotation(Vector3D.PLUS_K, compass
						.getAngle().radians());
			}
		}

		// Perform localization
		Pose3D localizeInfo = null;
		try {
			localizer.assignReferenceLines(this, visibleFieldLines,
					orientationEstimation);
			localizeInfo = localizer.localize(this, visibleFieldLines,
					orientationEstimation);
		} catch (RuntimeException e) {
			logger.log(Level.SEVERE, "Localization problem {0}", e);
		}

		if (localizeInfo != null) {
			thisPlayer
					.updateFromVision(Vector3D.ZERO,
							posFilter.filterPosition(localizeInfo.getPosition()),
							globalTime);
			thisPlayer.setGlobalOrientation(localizeInfo.getOrientation());
		} else {
			thisPlayer.setGlobalOrientation(orientationEstimation);
		}
	}

	/**
	 * 
	 */
	private void updateCameraCoordinateSystem()
	{
		// Update current camera-to-root-vision transformation, if possible
		IBodyPart cameraBodyPart = agentModel.getBodyPartContainingCamera();
		if (cameraBodyPart != null) {
			cameraToRootVision = cameraBodyPart.getPose();

			// Transform local body position and orientation into global field
			// system
			Vector3D pos = cameraToRootVision.getPosition();
			cameraToRootVision.position = new Vector3D(pos.getY(), -pos.getX(),
					pos.getZ());
			cameraToRootVision.orientation = Geometry.zTransformRotation(
					cameraToRootVision.getOrientation(), -Math.PI / 2);
		}
	}

	/**
	 * @param globalPose
	 */
	private void calculatePositionFromGlobalPose(IGlobalPosePerceptor globalPose)
	{
		Pose3D localizeInfo = globalPose.getGlobalPose();
		thisPlayer.updateFromVision(Vector3D.ZERO,
				posFilter.filterPosition(localizeInfo.getPosition()), globalTime);
		thisPlayer.setGlobalOrientation(localizeInfo.getOrientation());
	}

	/**
	 * @param perception the object containing the result from parsing
	 */
	private void updateBall(IPerception perception)
	{
		IVisibleObjectPerceptor ballVision = perception
				.getVisibleObject(IPerception.BALL);
		if (ballVision == null) {
			// the ball is currently not visible. This will be true every 2 out of
			// 3 cycles without camera image
			ball.setVisible(false);

			// we extrapolate the ball position in cycles we do not see it
			ball.updateNoVision(globalTime);
			return;
		}

		Vector3D ballPosition = ballVision.getPosition();
		if (!ballVision.hasDepth()) {
			IBodyPart cameraBodyPart = agentModel.getBodyPartContainingCamera();
			if (cameraBodyPart == null) {
				ball.setVisible(false);
				return;
			}
			Pose3D cameraPose = cameraBodyPart.getPose();
			ballPosition = estimateDepth(thisPlayer.getPose(), cameraPose,
					ballPosition, ball.getRadius());
			if (ballPosition == null) {
				ball.setVisible(false);
				// System.out.println("unplausible depth");
				return;
			}
		}
		Vector3D localPos = toRootVisionSystem(ballPosition);
		Vector3D globalPos = thisPlayer.calculateGlobalPosition(localPos);

		// Vector3D result = globalPos;
		// System.out.println(result.getX() + ";" + result.getY() + ";"
		// + result.getZ());

		ball.updateFromVision(localPos, globalPos, globalTime);
		// System.out.println("0.1;" + ball.getSpeed().getNorm() + ";"
		// + ball.getPosition().getX());
	}

	/**
	 * @param ballPosition
	 * @param radius
	 * @return
	 */
	Vector3D estimateDepth(Pose3D thisPlayerPose, Pose3D cameraPose,
			Vector3D position, float heightAboveGround)
	{

		// transform camera to global coordinate system
		Pose3D rotationVision = new Pose3D(Vector3D.ZERO, new Rotation(
				Vector3D.PLUS_K, Math.PI / 2));
		Pose3D rotationTorso = new Pose3D(Vector3D.ZERO, new Rotation(
				Vector3D.PLUS_K, -Math.PI / 2));
		Pose3D globalPose = thisPlayerPose.applyTo(rotationTorso)
				.applyTo(cameraPose).applyTo(rotationVision);

		// intersect ray to ball with field plane
		Plane objectPlane = new Plane(new Vector3D(0, 0, heightAboveGround),
				Vector3D.PLUS_K, 0.001);
		Vector3D p1 = globalPose.getPosition();
		Vector3D p2 = globalPose.applyTo(position);
		Line line = new Line(p1, p2, 0.001);
		Vector3D intersection = objectPlane.intersection(line);
		if (intersection == null) {
			return null;
		}

		// if the intersection is behind us, we do not accept that it is the ball
		// if (intersection.getX() < 0) {
		// // wrong ball or not on ground
		// // System.out.println("wrong");
		// return null;
		// }

		// convert back into camera coordinate system
		double distance = Vector3D.distance(p1, intersection);
		if (position.getNorm() < 0.00001) {
			position = Vector3D.PLUS_I;
		}
		Vector3D result = position.normalize().scalarMultiply(distance);
		return result;
	}

	/**
	 * Updates the local and global position of the visible landmarks.
	 * 
	 * @param perception the object containing the result from parsing
	 */
	private void updateLandmarks(IPerception perception)
	{
		for (String name : landmarks.keySet()) {
			Landmark landmark = (Landmark) landmarks.get(name);
			IVisibleObjectPerceptor landmarkSeen = perception
					.getVisibleObject(name);
			if (landmarkSeen == null) {
				// this landmark is currently not visible
				landmark.setVisible(false);
			} else {
				Vector3D localPos = toRootVisionSystem(landmarkSeen.getPosition());
				Vector3D globalPos = thisPlayer.calculateGlobalPosition(localPos);

				landmark.updateFromVision(localPos, globalPos, globalTime);
			}
		}
	}

	/**
	 * Updates the global position of the as visible marked field lines.
	 */
	private void updateFieldLines()
	{
		for (IFieldLine fieldLine : fieldLines.values()) {
			if (fieldLine.isVisible()) {
				((FieldLine) fieldLine).updatePositions(thisPlayer
						.calculateGlobalPosition(fieldLine.getLocalPosition1()),
						thisPlayer.calculateGlobalPosition(fieldLine
								.getLocalPosition2()), globalTime);
			}
		}
	}

	/**
	 * Updates the player information according to the new visual perception.
	 * 
	 * @param perception the object containing the result from vision parsing
	 */
	void updatePlayers(IPerception perception)
	{
		// We only act when vision information should be present
		if (!perception.containsVision()) {
			return;
		}

		// Clear visible players
		visiblePlayers.clear();

		// Process player visions
		List<IPlayerPos> playersVision = perception.getVisiblePlayers();
		Player currentPlayer;
		boolean isOwnTeam;

		for (IPlayerPos playerVision : playersVision) {
			isOwnTeam = thisPlayer.getTeamname()
					.equals(playerVision.getTeamname());

			// Don't process myself as a teammate
			if (isOwnTeam && playerVision.getId() == thisPlayer.getID()) {
				continue;
			}

			// Fetch a player instance
			if (playerVision.getTeamname() == null
					|| IMagmaConstants.UNKNOWN_PLAYER_TEAMNAME.equals(playerVision
							.getTeamname())
					|| playerVision.getId() == IMagmaConstants.UNKNOWN_PLAYER_NUMBER) {
				// If the player vision information is not complete, create just a
				// temporary player instance to hold the perceived information and
				// add it to the visiblePlayers array
				// TODO: Instead of simply adding an additional player by default,
				// one can perform some kind of lookup in the known players map and
				// try to match the player to an previously known player near by
				currentPlayer = new Player(playerVision.getId(),
						playerVision.getTeamname(), isOwnTeam);
				visiblePlayers.add(currentPlayer);

				logger.log(
						Level.FINE,
						"Saw player with incomplete information: ({0}, {1})",
						new Object[] { playerVision.getTeamname(),
								playerVision.getId() });
			} else {
				// If the player vision information is complete, fetch or create a
				// permanent instance to him
				currentPlayer = (Player) knownPlayers.get(playerVision
						.getTeamname() + playerVision.getId());
				if (currentPlayer == null) {
					currentPlayer = new Player(playerVision.getId(),
							playerVision.getTeamname(), isOwnTeam);
					knownPlayers.put(
							currentPlayer.getTeamname() + currentPlayer.getID(),
							currentPlayer);

					// Remember the name of the other team once we see the first
					// other player
					if (otherTeamName == null && !isOwnTeam) {
						otherTeamName = playerVision.getTeamname();
					}
				}
			}

			// Update the player information
			updatePlayer(currentPlayer, playerVision);
		}

		// Extract all recently seen players from known players map and mark the
		// others as not visible
		for (IPlayer player : knownPlayers.values()) {
			if (player.getAge(globalTime) < REMEMBERANCE_TIME) {

				// TODO: check if the player is in the visible cone and should have
				// been seen therefore

				visiblePlayers.add(player);
			} else {
				((Player) player).setVisible(false);
			}
		}
	}

	/**
	 * Updates a player's body parts as well as its local and global position
	 * 
	 * @param player the player to update
	 * @param playerVision the update information
	 */
	private void updatePlayer(Player player, IPlayerPos playerVision)
	{
		Vector3D localPos;
		Vector3D globalPos;

		// Transform body parts of player
		Map<String, Vector3D> bodyParts = new HashMap<String, Vector3D>();
		Map<String, Vector3D> allBodyParts = playerVision.getAllBodyParts();
		for (String name : allBodyParts.keySet()) {
			localPos = toRootVisionSystem(allBodyParts.get(name));
			globalPos = thisPlayer.calculateGlobalPosition(localPos);

			bodyParts.put(name, globalPos);
		}

		// Transform player positions
		localPos = toRootVisionSystem(playerVision.getPosition());
		globalPos = thisPlayer.calculateGlobalPosition(localPos);

		// Update positions and body parts of player
		player.update(localPos, globalPos, bodyParts, globalTime);
	}

	/**
	 * @param id the player's number (1-11)
	 * @param ownTeam true if we want the player of our own team
	 * @return the referenced player, null if not in list
	 */
	@Override
	public IPlayer getVisiblePlayer(int id, boolean ownTeam)
	{
		if (id < 1) {
			return null;
		}
		for (int i = 0; i < visiblePlayers.size(); i++) {
			IPlayer player = visiblePlayers.get(i);
			if (id == player.getID() && ownTeam == player.isOwnTeam()) {
				return player;
			}
		}
		return null;
	}

	private Vector3D toRootVisionSystem(Vector3D visionVector)
	{
		if (visionVector == null) {
			return null;
		}

		if (cameraToRootVision == null) {
			// There exists no transformation, so no system conversion
			return visionVector;
		}

		return cameraToRootVision.getPosition().add(
				cameraToRootVision.getOrientation().applyTo(visionVector));
	}

	// --------------------------------------------------
	// General environment and game information
	@Override
	public float getGlobalTime()
	{
		return globalTime;
	}

	/**
	 * @return the time the game is running now
	 */
	@Override
	public float getGameTime()
	{
		return gameTime;
	}

	@Override
	public int getGoalsTheyScored()
	{
		return goalsTheyScored;
	}

	@Override
	public int getGoalsWeScored()
	{
		return goalsWeScored;
	}

	/**
	 * Currently only known for Humanoid League!
	 * @return the color of our team
	 */
	public TeamColor getTeamColor()
	{
		return teamColor;
	}

	/**
	 * @return the current mode of the game as String
	 */
	public PlayMode getPlaymode()
	{
		return playmode;
	}

	/**
	 * Get play side
	 * 
	 * @return Play side
	 */
	public PlaySide getPlaySide()
	{
		return playSide;
	}

	@Override
	public GameState getGameState()
	{
		return gameState;
	}

	/**
	 * @return the time when the current game state was entered
	 */
	public float getTimeOfLastGameStateChange()
	{
		return lastGameStateChangeTime;
	}

	@Override
	public Vector3D getHomePosition()
	{
		Vector3D startPos;

		int id = thisPlayer.getID();
		if (id < 0 || id >= IMagmaConstants.homePositionsOwnKickoff.length) {
			return Vector3D.ZERO;
		}

		if (// gameState == GameState.OWN_GOAL
		gameState == GameState.BEFORE_KICK_OFF
				|| gameState == GameState.OPPONENT_GOAL
				|| gameState == GameState.OWN_KICK_OFF) {
			startPos = IMagmaConstants.homePositionsOwnKickoff[id]
					.getHomePosition(serverMetaModel);
		} else {
			startPos = IMagmaConstants.homePositionsOtherKickoff[id]
					.getHomePosition(serverMetaModel);
		}

		logger.log(Level.FINE, "playmode: {0}", new Object[] { gameState });
		return startPos;
	}

	// --------------------------------------------------
	// Visual object getters
	@Override
	public Ball getBall()
	{
		return ball;
	}

	@Override
	public Collection<ILandmark> getLandmarks()
	{
		return Collections.unmodifiableCollection(landmarks.values());
	}

	@Override
	public ILandmark getLandmark(String name)
	{
		return landmarks.get(name);
	}

	/**
	 * @return the average error of known position and calculated position,
	 *         Double.MAX_VALUE if we do not see any flags
	 */
	@Override
	public double getLandmarkError()
	{
		double result = 0.0;
		int count = 0;
		for (String name : landmarks.keySet()) {
			Landmark landmark = (Landmark) landmarks.get(name);
			// this landmark is not visible currently
			if (landmark.isVisible()) {
				result += landmark.getPosition()
						.subtract(landmark.getKnownPosition()).getNorm();
				count++;
			}
		}
		if (count < 1) {
			return Double.MAX_VALUE;
		}

		return result / count;
	}

	@Override
	public Collection<IFieldLine> getFieldLines()
	{
		return Collections.unmodifiableCollection(fieldLines.values());
	}

	@Override
	public IFieldLine getFieldLine(String name)
	{
		return fieldLines.get(name);
	}

	@Override
	public List<IPlayer> getVisiblePlayers()
	{
		return Collections.unmodifiableList(visiblePlayers);
	}

	@Override
	public IThisPlayer getThisPlayer()
	{
		return thisPlayer;
	}

	// --------------------------------------------------
	// EnvironmentModel Methods
	@Override
	public List<IReferenceLine> getReferenceLines()
	{
		return referenceLines;
	}

	@Override
	public List<IReferencePoint> getReferencePoints()
	{
		return referencePoints;
	}

	// --------------------------------------------------
	// Server version and soccer pitch description delegates
	@Override
	public int getServerVersion()
	{
		return serverVersion;
	}

	@Override
	public float fieldHalfLength()
	{
		return soccerPitch.fieldHalfLength;
	}

	@Override
	public float fieldHalfWith()
	{
		return soccerPitch.fieldHalfWidth;
	}

	@Override
	public float penaltyHalfLength()
	{
		return soccerPitch.penaltyHalfLength;
	}

	@Override
	public float penaltyWith()
	{
		return soccerPitch.penaltyWidth;
	}

	@Override
	public float goalHalfWith()
	{
		return soccerPitch.goalHalfWidth;
	}

	@Override
	public float goalHeight()
	{
		return soccerPitch.goalHeight;
	}

	@Override
	public float centerCircleRadius()
	{
		return soccerPitch.centerCircleRadius;
	}

	@Override
	public Vector3D getOwnGoalPosition()
	{
		return soccerPitch.ownGoalPosition;
	}

	@Override
	public Vector3D getOtherGoalPosition()
	{
		return soccerPitch.otherGoalPosition;
	}

	@Override
	public boolean isInOwnPenaltyArea(Vector3D position, float margin)
	{
		return soccerPitch.isInOwnPenaltyArea(position, margin);
	}

	@Override
	public boolean isInOtherPenaltyArea(Vector3D position, float margin)
	{
		return soccerPitch.isInOtherPenaltyArea(position, margin);
	}

	@Override
	public boolean isInOwnHalf(Vector3D position, float margin)
	{
		return soccerPitch.isInOwnHalf(position, margin);
	}

	@Override
	public boolean isInCenterCircleArea(Vector3D position, float margin)
	{
		return soccerPitch.isInCenterCircleArea(position, margin);
	}

	@Override
	public boolean isInSoccerPitchArea(Vector3D position, float margin)
	{
		return soccerPitch.isInSoccerPitchArea(position, margin);
	}

	// --------------------------------------------------
	// Observer methods
	@Override
	public void attach(IObserver<IGlobalMap> newObserver)
	{
		observer.attach(newObserver);
	}

	@Override
	public boolean detach(IObserver<IGlobalMap> oldObserver)
	{
		return observer.detach(oldObserver);
	}

	// --------------------------------------------------
	// Basic object methods
	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof GlobalMap)) {
			return false;
		}
		GlobalMap other = (GlobalMap) o;

		if (!ball.equals(other.ball)) {
			return false;
		}

		if (!ball.equals(other.ball)) {
			return false;
		}

		if (!landmarks.equals(other.landmarks)) {
			return false;
		}

		return visiblePlayers.equals(other.visiblePlayers);
	}

	/**
	 * Writes out current situation information. Used for learning situations and
	 * implementation will change.
	 * @param which allows to indicate different situations
	 */
	@Override
	public void logSituation(int which)
	{
		// dataLogger.logSituation(which);
	}
}