/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import magma.agent.worldmodel.ILandmark;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.util.FuzzyCompare;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Non movable orientation point on the field
 * 
 * @author Klaus Dorer
 */
public class Landmark extends VisibleObject implements ILandmark,
		IReferencePoint
{
	/** id of this class for serialization */
	private static final long serialVersionUID = -5016797299200496064L;

	/** the position as defined in robocup configuration files */
	private final Vector3D knownPosition;

	/**
	 * Constructor
	 * 
	 * @param name Landmark name
	 * @param knownPosition Landmark position
	 */
	public Landmark(String name, Vector3D knownPosition)
	{
		super(name);
		this.knownPosition = knownPosition;
	}

	@Override
	public Vector3D getKnownPosition()
	{
		return knownPosition;
	}

	/**
	 * @param localPosition the position as observed in the root body system
	 */
	public void updateLocalPosition(Vector3D localPosition)
	{
		this.localPosition = localPosition;
		setVisible(true);
	}

	@Override
	public Vector3D getPosition()
	{
		// TODO: a delegate to getKnownPosition is not necessarily good here (was
		// introduced because of IFOs). We avoid obstacles, even if we see them
		// somewhere else - in case of coming from the outside to a goal when the
		// ball lies somewhere very close to a goal post, this could lead to
		// unnecessary obstacle avoidance, even if the ball is between us and the
		// goal post. maybe:
		// if(isVisible()){ return position; } else {
		return getKnownPosition();
		// }
	}

	@Override
	public int compareTo(IReferencePoint other)
	{

		if (localPosition.getAlpha() < other.getLocalPosition().getAlpha()) {
			return 1;
		} else if (localPosition.getAlpha() > other.getLocalPosition().getAlpha()) {
			return -1;
		}
		return 0;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof Landmark)) {
			return false;
		}
		Landmark other = (Landmark) o;
		if (!super.equals(other)) {
			return false;
		}
		return FuzzyCompare.eq(knownPosition, other.knownPosition, 0.00001f);
	}
}