/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import magma.agent.worldmodel.IMoveableObject;
import magma.util.FuzzyCompare;
import magma.util.geometry.Geometry;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Base class for all movable objects on the field
 * 
 * @author Klaus Dorer
 */
public abstract class MovableObject extends VisibleObject implements
		IMoveableObject
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** the speed calculated from previous and current positions in m/cycle */
	protected Vector3D speed;

	/** the speed as has been before last update */
	protected Vector3D oldSpeed;

	/** caching future positions */
	protected transient Vector3D[] futurePositions;

	/**
	 * Constructor
	 * 
	 * @param name Object name
	 */
	MovableObject(String name)
	{
		super(name);
		speed = Vector3D.ZERO;
		oldSpeed = Vector3D.ZERO;
	}

	@Override
	public Vector3D[] getFuturePositions(int howMany)
	{
		if (futurePositions == null || futurePositions.length < howMany) {
			futurePositions = calculateFuturePositions(howMany);
		}

		return futurePositions;
	}

	@Override
	public Vector3D getFuturePosition(int when)
	{
		if (futurePositions == null || futurePositions.length < when) {
			futurePositions = calculateFuturePositions(when);
		}

		return futurePositions[when - 1];
	}

	protected Vector3D[] calculateFuturePositions(int howMany)
	{
		return Geometry.getFuturePositions(position, speed, howMany);
	}

	/**
	 * @return the speed of this object in m/cycle
	 */
	@Override
	public Vector3D getSpeed()
	{
		return speed;
	}

	/**
	 * @return the previous speed of this object
	 */
	@Override
	public Vector3D getOldSpeed()
	{
		return oldSpeed;
	}

	@Override
	public void updateFromVision(Vector3D localPosition,
			Vector3D globalPosition, float time)
	{
		// WARNING: getLastSeenTime() MUST be called before super.update(),
		// because
		// super.update() updates the lastSeenTime with the current time
		float lastSeenTime = getLastSeenTime();
		super.updateFromVision(localPosition, globalPosition, time);
		calculateSpeed(lastSeenTime, time);

		futurePositions = null;
	}

	@Override
	public void updateFromAudio(Vector3D localPosition, Vector3D globalPosition,
			float time)
	{
		super.updateFromAudio(localPosition, globalPosition, time);

		futurePositions = null;
	}

	/**
	 * Calculate the maximum possible speed this object can move at in m/cycle
	 * 
	 * @return Possible speed
	 */
	double getPossibleSpeed()
	{
		return 0.0;
	}

	/**
	 * Calculate the speed this object is moving at
	 * 
	 * @param lastSeenTime Timestamp this object was last seen
	 * @param time Current timestamp
	 */
	void calculateSpeed(float lastSeenTime, float time)
	{
		oldSpeed = speed;
		speed = calculateSpeed(getPosition(), time, getPreviousPosition(),
				lastSeenTime);
	}

	/**
	 * Calculate the speed this object is moving at
	 * 
	 * @param lastSeenTime Timestamp this object was last seen
	 * @param time Current timestamp
	 */
	Vector3D calculateSpeed(Vector3D newPosition, float newTime,
			Vector3D oldPosition, float oldTime)
	{
		// need previousPosition, hence speed is calculated after update
		Vector3D checkSpeed = oldSpeed;
		if (oldPosition != null) {
			checkSpeed = newPosition.subtract(oldPosition);
			// 0.02 is the cycle time
			float deltaCycles = Math.round((newTime - oldTime) / 0.02f);
			if (deltaCycles > 0.01f) {
				checkSpeed = checkSpeed.scalarMultiply(1 / deltaCycles);
			}

			// we don't set speed for impossible values
			if (checkSpeed.getNorm() > getPossibleSpeed()) {
				return oldSpeed;
			}
			// we cut small speed values (why, filter should care!)
			// we did not do this for Sweaty
			if (checkSpeed.getNorm() < 0.005) {
				return Vector3D.ZERO;
			}
		}
		return checkSpeed;
	}

	@Override
	public boolean isMoving()
	{
		return getSpeed().getNorm() > 0.017;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof MovableObject)) {
			return false;
		}
		MovableObject other = (MovableObject) o;
		if (!super.equals(other)) {
			return false;
		}
		return FuzzyCompare.eq(speed, other.speed, 0.00001f);
	}

	/**
	 * @param globalTime
	 */
	public void updateNoVision(float globalTime)
	{
		super.updateNoVision(position.add(speed), globalTime);

		futurePositions = null;
	}
}