/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.agent.worldmodel.impl;

import magma.agent.meta.server.IRCServerMetaModel;
import magma.agent.worldmodel.ISoccerPitchDescription;
import magma.util.ValueUtil;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * This class describes a soccer pitch and some geometric calculations to it.
 * 
 * @author Stefan Glaser
 */
public class SoccerPitchDescription implements ISoccerPitchDescription
{
	/**  */
	private static final long serialVersionUID = -41059962451694494L;

	/** The half length of the soccer pitch (global x) */
	public final float fieldHalfLength;

	/** The half width of the soccer pitch (global y) */
	public final float fieldHalfWidth;

	/** The half width of the goals (global y) */
	public final float goalHalfWidth;

	/** The height of the goals (global z) */
	public final float goalHeight;

	/** The width of the penalty areas (global x) */
	public final float penaltyWidth;

	/** The half length of the penalty areas (global y) */
	public final float penaltyHalfLength;

	/** The radius of the center circle */
	public final float centerCircleRadius;

	/** The position of the own goal (between the goal posts) */
	public final Vector3D ownGoalPosition;

	/** The position of the other goal (between the goal posts) */
	public final Vector3D otherGoalPosition;

	public SoccerPitchDescription(float fieldHalfLength, float fieldHalfWidth,
			float goalHalfWidth, float goalHeight, float penaltyWidth,
			float penaltyHalfLength, float centerCircleRadius)
	{
		this.fieldHalfLength = fieldHalfLength;
		this.fieldHalfWidth = fieldHalfWidth;
		this.goalHalfWidth = goalHalfWidth;
		this.goalHeight = goalHeight;
		this.penaltyWidth = penaltyWidth;
		this.penaltyHalfLength = penaltyHalfLength;
		this.centerCircleRadius = centerCircleRadius;

		ownGoalPosition = new Vector3D(-fieldHalfLength, 0, 0);
		otherGoalPosition = new Vector3D(fieldHalfLength, 0, 0);
	}

	public SoccerPitchDescription(IRCServerMetaModel metaModel)
	{
		this(metaModel.getFieldDimensions()[0] / 2, //
				metaModel.getFieldDimensions()[1] / 2, //
				metaModel.getGoalDimensions()[0] / 2, //
				metaModel.getGoalDimensions()[1], //
				metaModel.getPenaltyAreaDimensions()[1], //
				metaModel.getPenaltyAreaDimensions()[0] / 2, //
				metaModel.getMiddleCircleRadii());
	}

	@Override
	public float fieldHalfLength()
	{
		return fieldHalfLength;
	}

	@Override
	public float fieldHalfWith()
	{
		return fieldHalfWidth;
	}

	@Override
	public float goalHalfWith()
	{
		return goalHalfWidth;
	}

	@Override
	public float goalHeight()
	{
		return goalHeight;
	}

	@Override
	public float penaltyWith()
	{
		return penaltyWidth;
	}

	@Override
	public float penaltyHalfLength()
	{
		return penaltyHalfLength;
	}

	@Override
	public float centerCircleRadius()
	{
		return centerCircleRadius;
	}

	@Override
	public Vector3D getOwnGoalPosition()
	{
		return ownGoalPosition;
	}

	@Override
	public Vector3D getOtherGoalPosition()
	{
		return otherGoalPosition;
	}

	@Override
	public boolean isInOwnPenaltyArea(Vector3D position, float margin)
	{
		return ValueUtil.isValueInRange(position.getX(), //
				-fieldHalfLength, //
				-fieldHalfLength + penaltyWidth, //
				margin) //
				&& ValueUtil.isValueInRange(position.getY(), //
						-penaltyHalfLength - goalHalfWidth, //
						penaltyHalfLength + goalHalfWidth, //
						margin);
	}

	@Override
	public boolean isInOtherPenaltyArea(Vector3D position, float margin)
	{
		return ValueUtil.isValueInRange(position.getX(), //
				fieldHalfLength - penaltyWidth, //
				fieldHalfLength, //
				margin) //
				&& ValueUtil.isValueInRange(position.getY(), //
						-penaltyHalfLength - goalHalfWidth, //
						penaltyHalfLength + goalHalfWidth, //
						margin);
	}

	@Override
	public boolean isInCenterCircleArea(Vector3D position, float margin)
	{
		float norm = (float) Math.sqrt(position.getX() * position.getX()
				+ position.getY() * position.getY());

		return norm < centerCircleRadius + margin;
	}

	@Override
	public boolean isInSoccerPitchArea(Vector3D position, float margin)
	{
		return ValueUtil.isValueInRange(position.getX(), //
				-fieldHalfLength, //
				fieldHalfLength, //
				margin) //
				&& ValueUtil.isValueInRange(position.getY(), //
						-fieldHalfWidth, //
						fieldHalfWidth, //
						margin);
	}

	@Override
	public boolean isInOwnHalf(Vector3D position, float margin)
	{
		return ValueUtil.isValueInRange(position.getX(), //
				-fieldHalfLength, 0, //
				margin) //
				&& ValueUtil.isValueInRange(position.getY(), //
						-fieldHalfWidth, //
						fieldHalfWidth, //
						margin);
	}

}
