/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.util.Random;

import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;
import magma.util.geometry.Pose3D;

import org.apache.commons.math3.geometry.euclidean.threed.CardanEulerSingularityException;
import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.RotationOrder;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Container for this agent's information that is unique to the player this
 * agent is representing
 * 
 * @author Klaus Dorer, Stefan Glaser
 */
public class ThisPlayer extends Player implements IThisPlayer
{
	/** id of this class for serialization */
	private static final long serialVersionUID = 3307599892830771538L;

	/** manager for desired positions */
	private final PositionManager positionManager;

	/** the orientation of the player relative to the local coordinate system */
	private transient Rotation localOrientation;

	/** only used for experiments with artificial gyro noise */
	private double gyroNoise;

	/**
	 * Constructor
	 * 
	 * @param teamname Team name
	 * @param id Player number
	 */
	public ThisPlayer(String teamname, int id)
	{
		super(id, teamname, true);
		setGyroNoise(0.000);

		positionManager = new PositionManager();
	}

	@Override
	public void setGlobalOrientation(Rotation orientation)
	{
		// Rotation noisyOrientation = addNoiseToGyro(orientation);
		// super.setGlobalOrientation(noisyOrientation);
		super.setGlobalOrientation(orientation);
		localOrientation = Geometry.zTransformRotation(globalOrientation,
				Math.PI / 2);
	}

	/**
	 * Used for evaluating dependency on noise on gyro
	 * @param orientation
	 * @return
	 */
	@SuppressWarnings("unused")
	private Rotation addNoiseToGyro(Rotation orientation)
	{
		try {
			double[] anglesBefore = orientation.getAngles(RotationOrder.XYZ);
			Random rand = new Random();
			double q0 = orientation.getQ0();
			double q1 = orientation.getQ1();
			double q2 = orientation.getQ2();
			double q3 = orientation.getQ3();

			// q0 += rand.nextGaussian() * noise;
			q1 += rand.nextGaussian() * gyroNoise;
			q2 += rand.nextGaussian() * gyroNoise;
			q3 += rand.nextGaussian() * gyroNoise;
			// System.out.println("q0:" + q0 + " q1:" + q1 + " q2:" + q2 + " q3:" +
			// q3);
			Rotation noisyOrientation = new Rotation(q0, q1, q2, q3, true);

			double[] angles = noisyOrientation.getAngles(RotationOrder.XYZ);
			angles[0] = Math.toDegrees(Math.abs(anglesBefore[0] - angles[0]));
			angles[1] = Math.toDegrees(Math.abs(anglesBefore[1] - angles[1]));
			angles[2] = Math.toDegrees(Math.abs(anglesBefore[2] - angles[2]));
			// System.out.println(Arrays.toString(angles));
			return noisyOrientation;

		} catch (CardanEulerSingularityException e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public Rotation getOrientation()
	{
		// Handle not serialized localOrientation
		if (localOrientation == null) {
			return Geometry.zTransformRotation(globalOrientation, Math.PI / 2);
		}
		return localOrientation;
	}

	@Override
	public Vector3D calculateGlobalPosition(Vector3D localPos)
	{
		return position.add(globalOrientation.applyTo(localPos));
	}

	@Override
	public Vector3D calculateGlobal2DPosition(Vector3D localPos)
	{
		return position.add(Geometry.createZRotation(
				getHorizontalAngle().radians()).applyTo(localPos));
	}

	@Override
	public Pose2D calculateGlobalBodyPose2D(Pose3D poseToTranslate)
	{
		Pose2D result = new Pose2D(//
				calculateGlobalPosition(new Vector3D(poseToTranslate.getY(),
						-poseToTranslate.getX(), poseToTranslate.getZ())), //
				Angle.rad(-Geometry.getTopViewZAngle(getOrientation().applyTo(
						poseToTranslate.getOrientation()))));
		return result;
	}

	@Override
	public Vector3D calculateLocalPosition(Vector3D globalPosition)
	{
		Vector3D localPos = globalPosition.subtract(position);
		return globalOrientation.applyInverseTo(localPos);
	}

	/**
	 * Returns the relative angle (rad) this player's torso has to the specified
	 * position
	 * @param position the position to which to calculate the body angle
	 * @return the relative angle (rad) this player's torso has to the specified
	 *         position
	 */
	@Override
	public Angle getBodyDirectionTo(Vector3D position)
	{
		return getDirectionTo(position).subtract(getHorizontalAngle());
	}

	/**
	 * Checks if the passed 2D xy-point (absolute coordinate system) is inside
	 * the passed polygon (xy, relative coordinate system)
	 * @param absolutePosition 2D vector (x,y) of the point to check in the
	 *        global coordinate system
	 * @param polygon corner points of the polygon (x, y), in this player's
	 *        coordinate system and in clockwise order. The coordinates should
	 *        NOT take neck angles into account, i.e. assume head is straight.
	 * @return true if the point is inside the polygon
	 */
	@Override
	public boolean isInsidePolygonXY(Vector3D absolutePosition,
			Vector3D[] polygon)
	{
		Vector3D[] polygonGlobal = new Vector3D[polygon.length];
		for (int i = 0; i < polygon.length; i++) {
			polygonGlobal[i] = calculateGlobalPosition(polygon[i]);
		}

		return Geometry.isInsidePolygon(absolutePosition, polygonGlobal);
	}

	/**
	 * @param position the position to check
	 * @return true if the specified position is left of this player with respect
	 *         to the root body orientation
	 */
	@Override
	public boolean positionIsLeft(Vector3D absolutePosition)
	{
		return calculateLocalPosition(absolutePosition).getY() >= 0;
	}

	/**
	 * @param position the position to check
	 * @return true if the specified position is right of this player with
	 *         respect to the root body orientation
	 */
	@Override
	public boolean positionIsRight(Vector3D absolutePosition)
	{
		return calculateLocalPosition(absolutePosition).getY() <= 0;
	}

	/**
	 * @param position the position to check
	 * @return true if the specified position is behind this player with respect
	 *         to the root body orientation
	 */
	@Override
	public boolean positionIsBehind(Vector3D absolutePosition)
	{
		return calculateLocalPosition(absolutePosition).getX() <= 0;
	}

	/**
	 * @return the positionManager
	 */
	@Override
	public PositionManager getPositionManager()
	{
		return positionManager;
	}

	/**
	 * @param gyroNoise the gyroNoise to set
	 */
	@Override
	public void setGyroNoise(double gyroNoise)
	{
		this.gyroNoise = gyroNoise;
	}

	@Override
	public double getMaxSpeed()
	{
		return 0.85;
	}

	public Pose3D getPose()
	{
		return new Pose3D(position, globalOrientation);
	}

	@Override
	public void setGlobalPosition(Vector3D position, float time)
	{
		updateFromAudio(Vector3D.ZERO, position, time);
	}

}
