/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.localizer.impl;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.ILocalizer;
import magma.util.geometry.Pose3D;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;

/**
 * Composite localizer, calling a set of other localizer sequentially until a
 * localization / line assignment was successful.
 * 
 * @author Stefan Glaser
 * 
 */
public class CompositeLocalizer extends LocalizerBase
{
	/** handle for logging */
	private static final Logger logger = Logger.getLogger("");

	private ILocalizer[] localizer;

	public CompositeLocalizer(ILocalizer[] localizer)
	{
		this.localizer = localizer;
	}

	@Override
	public Pose3D localize(IEnvironmentModel environment,
			List<ILocalizationLine> lines, Rotation estimatedOrientation)
	{
		if (!containsVisibleReferencePoints(environment)) {
			// there is no visible flag
			logger.log(Level.FINER,
					"Localization not possible, there is no visible reference point!");
			return null;
		}

		Pose3D result = null;

		for (int i = 0; i < localizer.length; i++) {
			if (result == null) {
				try {
					result = localizer[i].localize(environment, lines,
							estimatedOrientation);
				} catch (Exception e) {
					// in case a localizer fails we want to be able to continue with
					// the others
					logger.log(Level.SEVERE,
							"Localization caused exception " + e.toString());
				}
			}
		}

		return result;
	}

	@Override
	public boolean assignReferenceLines(IEnvironmentModel environment,
			List<ILocalizationLine> lines, Rotation estimatedOrientation)
	{
		if (!containsVisibleReferencePoints(environment)) {
			// there is no visible flag
			logger.log(Level.FINER,
					"Line assignment not possible, there is no visible reference point!");
			return false;
		}

		boolean result = false;

		for (int i = 0; i < localizer.length; i++) {
			if (!result) {
				result = localizer[i].assignReferenceLines(environment, lines,
						estimatedOrientation);
			}
		}

		return result;
	}
}
