/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.localizer.impl;

import java.util.ArrayList;
import java.util.List;

import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.ILocalizer;
import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Base class for Localizer components
 * @author Stefan Glaser
 */
public abstract class LocalizerBase implements ILocalizer
{
	@Override
	public boolean assignReferenceLines(IEnvironmentModel environment,
			List<ILocalizationLine> lines, Rotation estimatedOrientation)
	{
		// 1. idea
		// Create an utility system, in which we assign an utility (or likelyhood)
		// to each visible line and each reference line. After that we check, if
		// all reference lines are just preferred once. If this is the case, we
		// have a collision free assignment and hope everything is fine. If we
		// have collisions, we have to take further checks on the
		// second-most-preferred reference lines, etc.

		// 2. idea
		// Try to build a set of relations between all known reference lines
		// (like: "x.dir == y.dir && x.p1 == y.p1-(0,14) && x.p2 == y.p2-(0,14)"
		// or "x.p1 == refPoint.p", etc.) and then try to match the set of visible
		// lines in such a way, that no relation is violated.
		// Example:
		// 1. Normalize all vision information according to the normal of the
		// least square fitting plane through all line points (--> remove x- and
		// y-rotation).
		// 2. Take the longest unassigned line you see and try to find a unique
		// assignment with respect to all the other visible lines
		// 3. If an unique assignment was possible or not, remove the current line
		// from the set of lines to assign. Continue with Step 2 as long as the
		// set of lines to assign is not empty.
		// 4. If one run finished without assigning all lines, take the remaining
		// unassigned lines and restart with step 2. Stop if no single line could
		// be assigned during one run.

		// 3. idea
		// Try to start with an known reference point and search for lines which
		// end at the reference point and then start to assign lines from that
		// reference point on. Since all lines are interconnected with some other
		// lines, we can follow up the graph, until we don't see no matching
		// lines any more.

		return false;
	}

	/**
	 * Assigns the given locLine to the refLine.<br>
	 * If <tt>crossAssignemnt == true</tt> the two points of the lines are
	 * assigned reverse (refLine.p1 = locLine.p1 && refLine.p2 = locLine.p2).<br>
	 * The parameter <tt>firstPointMatched</tt> is used to indicate which point
	 * of the line was matching it's known position. This information is then
	 * used to adjust the known position of the other point, if the line is not
	 * fully visible (shorter than expected). If none of the points match their
	 * known position, the line assignment makes no sense, since we can't
	 * estimate, which part of the line we are actually seeing.
	 * 
	 * @param refLine - the reference line to update
	 * @param locLine - the unlabeled line, which should be assigned to the
	 *        refLine
	 * @param crossAssignment - cross assignment of points<br>
	 *        (<tt>false</tt> => refLine.p1 = locLine.p1 && refLine.p2 =
	 *        locLine.p2;<br>
	 *        <tt>true</tt> => refLine.p1 = locLine.p2 && refLine.p2 =
	 *        locLine.p1)
	 * @param firstPointMatched - indicator for adjusting the known position of
	 *        one line point, as a consequence of the restricted vision.<br>
	 *        (<tt>true</tt> => locLine.p1 is matching the known position,
	 *        locLine.p2 may not <br>
	 *        <tt>false</tt> => locLine.p2 is matching the known position,
	 *        locLine.p1 may not)
	 */
	protected void assignReferenceLine(IReferenceLine refLine,
			ILocalizationLine locLine, boolean crossAssignment,
			boolean firstPointMatched)
	{
		double refLineLength, locLineLength;
		Vector3D refLineKnownPos1 = refLine.getKnownPosition1();
		Vector3D refLineKnownPos2 = refLine.getKnownPosition2();
		Vector3D locLinePos1 = locLine.getLocalPosition1();
		Vector3D locLinePos2 = locLine.getLocalPosition2();

		if (crossAssignment) {
			refLine.updateLocalPositions(locLinePos2, locLinePos1);

			// Check if we have to adjust the known position because of restricted
			// vision
			if (locLinePos1 != null && locLinePos2 != null) {
				refLineLength = refLineKnownPos1.subtract(refLineKnownPos2)
						.getNorm();
				locLineLength = locLinePos1.subtract(locLinePos2).getNorm();

				if (locLineLength < refLineLength * 0.97) {
					if (firstPointMatched) {
						refLineKnownPos1 = refLineKnownPos2.add(refLineKnownPos1
								.subtract(refLineKnownPos2).normalize()
								.scalarMultiply(locLineLength));
					} else {
						refLineKnownPos2 = refLineKnownPos1.add(refLineKnownPos2
								.subtract(refLineKnownPos1).normalize()
								.scalarMultiply(locLineLength));
					}
				}
			}
		} else {
			refLine.updateLocalPositions(locLinePos1, locLinePos2);

			// Check if we have to adjust the known position because of restricted
			// vision
			if (locLinePos1 != null && locLinePos2 != null) {
				refLineLength = refLineKnownPos1.subtract(refLineKnownPos2)
						.getNorm();
				locLineLength = locLinePos1.subtract(locLinePos2).getNorm();

				if (locLineLength < refLineLength * 0.97) {
					if (firstPointMatched) {
						refLineKnownPos2 = refLineKnownPos1.add(refLineKnownPos2
								.subtract(refLineKnownPos1).normalize()
								.scalarMultiply(locLineLength));
					} else {
						refLineKnownPos1 = refLineKnownPos2.add(refLineKnownPos1
								.subtract(refLineKnownPos2).normalize()
								.scalarMultiply(locLineLength));
					}
				}
			}
		}

		refLine.adjustKnownPosition1(refLineKnownPos1);
		refLine.adjustKnownPosition2(refLineKnownPos2);
	}

	/**
	 * Extracts the visible reference points of the environment into a new list.
	 * If no reference point is visible, an empty list is returned.
	 * 
	 * @param environment - the environment containing the reference points
	 * @return a new list containing just the visible reference points of the
	 *         environment
	 */
	protected List<IReferencePoint> getVisibleReferencePoints(
			IEnvironmentModel environment)
	{
		List<IReferencePoint> refPoints = new ArrayList<IReferencePoint>();
		if (environment.getReferencePoints() != null) {
			for (IReferencePoint refPoint : environment.getReferencePoints()) {
				if (refPoint.isVisible()) {
					refPoints.add(refPoint);
				}
			}
		}

		return refPoints;
	}

	/**
	 * Checks if the environment contains any visible reference points.
	 * 
	 * @param environment - the environment containing the reference points
	 * @return <tt>true</tt> if the environment contains any visible reference
	 *         point, else <tt>false</tt>
	 */
	protected boolean containsVisibleReferencePoints(
			IEnvironmentModel environment)
	{
		if (environment.getReferencePoints() != null) {
			for (IReferencePoint refPoint : environment.getReferencePoints()) {
				if (refPoint.isVisible()) {
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Extracts the visible reference lines of the environment into a new list.
	 * If no reference line is visible, an empty list is returned.
	 * 
	 * @param environment - the environment containing the reference lines
	 * @return a new list containing just the visible reference lines of the
	 *         environment
	 */
	protected List<IReferenceLine> getVisibleReferenceLines(
			IEnvironmentModel environment)
	{
		List<IReferenceLine> refLines = new ArrayList<IReferenceLine>();

		if (environment.getReferenceLines() != null) {
			for (IReferenceLine refLine : environment.getReferenceLines()) {
				if (refLine.isVisible()) {
					refLines.add(refLine);
				}
			}
		}

		return refLines;
	}
}
