/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots;

import java.util.logging.Level;

import magma.agent.IMagmaConstants;
import magma.agent.IServerConfigFilesConstants;
import magma.agent.general.ComponentFactory;
import magma.agent.general.PlayerParameters;
import magma.robots.nao.general.NAOComponentFactory;
import magma.robots.nao1.general.NAO1ComponentFactory;
import magma.robots.nao2.general.NAO2ComponentFactory;
import magma.robots.nao3.general.NAO3ComponentFactory;
import magma.robots.naotoe.general.NAOToeComponentFactory;
import magma.util.connection.IServerConnection;

/**
 * This is the main class of the application used to start the magma Robocup
 * agent.
 * 
 * @author Klaus Dorer
 */
public class CommandLineParser
{
	/**
	 * Instantiates and starts an agent.
	 * 
	 * @param args Command line arguments <br>
	 *        <table>
	 *        <tr>
	 *        <td>--teamname=</td>
	 *        <td>Name-of-the-Team</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--playerid=</td>
	 *        <td>PlayerID of the Agent</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--server=</td>
	 *        <td>Server IP</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--port=</td>
	 *        <td>Server Port</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--loglevel=</td>
	 *        <td>Different Logging levels, default: SEVERE<br>
	 *        ('fine','finer','finest','warning','info')</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--serverversion=</td>
	 *        <td>Simspark Server Version<br>
	 *        ('60','61','62','63',...)</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--decisionmaker=</td>
	 *        <td>DecisionMaker of the Agent<br>
	 *        (0 (default) if we decide based on player number, <br>
	 *        1 for goalie, <br>
	 *        2 for field player, <br>
	 *        3 for simple decision maker, <br>
	 *        4 for Penalty decision maker, <br>
	 *        5 for do nothing decision maker, <br>
	 *        6 for training decision maker, <br>
	 *        7 for goalie penalty decision maker, <br>
	 *        8 for Walk Demo decision maker, <br>
	 *        9 for training decision maker 2<br>
	 *        10 for run challenge dm</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--factory=</td>
	 *        <td>Factory (robot model)<br>
	 *        Nao (default) for RoboCup simulation 3D Nao robot, Naoi with i the
	 *        heterogeneous type of Nao [0 (default and same as Nao) - no of
	 *        types], NaoWebots for the Webots Nao over Naoqi and real Nao robot,
	 *        Nimbro for RoboCup simulation 3D Nimbro robot, Sweaty for simspark
	 *        Sweaty simulation, SweatyReal for real sweaty robot.</td>
	 *        </tr>
	 *        <tr>
	 *        <td>--serialPort=</td>
	 *        <td>Serial Port Name<br>
	 *        Check with ls /dev/tty* or ls /dev/ttyACM* which port are
	 *        available.</td>
	 *        </tr>
	 *        </table>
	 */
	public static PlayerParameters parse(String[] args)
	{
		String teamname = IMagmaConstants.DEFAULT_TEAMNAME;
		byte teamID = 0;
		int playerID = 0;
		String host = IServerConnection.SERVER_IP;
		int port = IServerConnection.SERVER_PORT;
		Level level = Level.SEVERE;
		int serverVersion = IServerConfigFilesConstants.DEFAULT_SERVER_VERSION;
		int decisionMakerID = 0;
		String factory = "Nao";
		String serialPort = "/dev/ttyACM0";

		boolean fallback = false;
		for (int i = 0; i < args.length && !fallback; i++) {
			String arg = args[i];
			if (arg.startsWith("--teamname=")) {
				teamname = arg.replaceFirst("--teamname=", "");

			} else if (arg.startsWith("--teamid=")) {
				teamID = Byte.valueOf(arg.replaceFirst("--teamid=", ""));
				if (teamID < 0 || teamID >= 127) {
					System.out.println("Invalid team ID: " + teamID
							+ "! Maximum is: 127");
					fallback = true;
				}

			} else if (arg.startsWith("--playerid=")) {
				playerID = Integer.valueOf(arg.replaceFirst("--playerid=", ""));
				if (playerID < 0
						|| playerID >= IMagmaConstants.homePositionsOwnKickoff.length) {
					System.out.println("Invalid player ID: " + playerID
							+ "! Maximum is: "
							+ (IMagmaConstants.homePositionsOwnKickoff.length - 1));
					fallback = true;
				}

			} else if (arg.startsWith("--server=")) {
				host = arg.replaceFirst("--server=", "");

			} else if (arg.startsWith("--port=")) {
				port = Integer.valueOf(arg.replaceFirst("--port=", ""));
				if (port < 0) {
					System.out.println("Invalid port: " + port);
					fallback = true;
				}

			} else if (arg.startsWith("--loglevel=")) {
				String loglevel = arg.replaceFirst("--loglevel=", "");
				level = getLoglevel(loglevel);

			} else if (arg.startsWith("--serverversion=")) {
				serverVersion = Integer.valueOf(arg.replaceFirst(
						"--serverversion=", ""));
				if (serverVersion < 60) {
					System.out.println("Invalid server version: " + serverVersion);
					fallback = true;
				}

			} else if (arg.startsWith("--decisionmaker=")) {
				decisionMakerID = Integer.valueOf(arg.replaceFirst(
						"--decisionmaker=", ""));
				if (decisionMakerID < 0 || decisionMakerID > 10) {
					System.out.println("Invalid decision maker id: "
							+ decisionMakerID);
					fallback = true;
				}

			} else if (arg.startsWith("--factory=")) {
				factory = arg.replaceFirst("--factory=", "");
				if (factory.equals("")) {
					System.out.println("Invalid factory: " + factory);
					fallback = true;
				}

			} else if (arg.startsWith("--serialPort=")) {
				serialPort = arg.replaceFirst("--serialPort=", "");
				if (serialPort.equals("")) {
					System.out.println("Invalid serialPort: " + serialPort);
					fallback = true;
				}

			} else {
				// Activate again when all scripts support new commandline
				// arguments.
				// System.out.println("Invalid argument: " + arg);
				fallback = true;
			}
		}

		if (!fallback) {
			ComponentFactory componentFactory = getComponentFactory(factory);
			return new PlayerParameters(teamname, teamID, playerID, host, port,
					null, level, serverVersion, componentFactory, decisionMakerID);

		} else {
			return fallbackStart(args);
		}
	}

	/**
	 * @param level
	 * @param loglevel
	 * @return
	 */
	protected static Level getLoglevel(String loglevel)
	{
		if (loglevel.equalsIgnoreCase("fine")) {
			return Level.FINE;
		} else if (loglevel.equalsIgnoreCase("finer")) {
			return Level.FINER;
		} else if (loglevel.equalsIgnoreCase("finest")) {
			return Level.FINEST;
		} else if (loglevel.equalsIgnoreCase("warning")) {
			return Level.WARNING;
		} else if (loglevel.equalsIgnoreCase("info")) {
			return Level.INFO;
		}
		return Level.SEVERE;
	}

	/**
	 * @return
	 */
	protected static ComponentFactory getComponentFactory(String name)
	{
		ComponentFactory componentFactory;
		if ("Nao1".equals(name)) {
			componentFactory = new NAO1ComponentFactory();
		} else if ("Nao2".equals(name)) {
			componentFactory = new NAO2ComponentFactory();
		} else if ("Nao3".equals(name)) {
			componentFactory = new NAO3ComponentFactory();
		} else if ("NaoToe".equals(name)) {
			componentFactory = new NAOToeComponentFactory();
		} else {
			componentFactory = new NAOComponentFactory();
		}
		return componentFactory;
	}

	/**
	 * @param args command line args (@see main)
	 */
	public static PlayerParameters fallbackStart(String[] args)
	{
		String teamname = IMagmaConstants.DEFAULT_TEAMNAME;
		byte teamID = 0;
		int playerID = 0;
		String host = IServerConnection.SERVER_IP;
		int port = IServerConnection.SERVER_PORT;
		Level level = Level.SEVERE;
		int serverVersion = IServerConfigFilesConstants.DEFAULT_SERVER_VERSION;
		int decisionMakerID = 0;
		String factory = "Nao";

		if (args.length > 0) {
			teamname = args[0];
		}

		if (args.length > 1) {
			playerID = Integer.valueOf(args[1]);
			if (playerID < 0
					|| playerID >= IMagmaConstants.homePositionsOwnKickoff.length) {
				System.out.println("Invalid player ID: " + playerID
						+ "! Maximum is: "
						+ (IMagmaConstants.homePositionsOwnKickoff.length - 1));
				System.exit(-1);
			}
		}

		if (args.length > 2) {
			host = args[2];
		}

		if (args.length > 3) {
			port = Integer.valueOf(args[3]);
			if (port < 0) {
				System.out.println("Invalid port: " + port);
				System.exit(-1);
			}
		}

		if (args.length > 4) {
			String loglevel = args[4];
			level = getLoglevel(loglevel);
		}

		if (args.length > 5) {
			serverVersion = Integer.valueOf(args[5]);
			if (serverVersion < 60) {
				System.out.println("Invalid server version: " + serverVersion);
				System.exit(-1);
			}
		}

		if (args.length > 6) {
			decisionMakerID = Integer.valueOf(args[6]);
			if (decisionMakerID < 0 || decisionMakerID > 8) {
				System.out.println("Invalid decision maker id: " + decisionMakerID);
				System.exit(-1);
			}
		}

		if (args.length > 7) {
			factory = args[7];
			if (factory.equals("")) {
				System.out.println("Invalid factory: " + factory);
				System.exit(-1);
			}
		}

		ComponentFactory componentFactory = getComponentFactory(factory);
		return new PlayerParameters(teamname, teamID, playerID, host, port, null,
				level, serverVersion, componentFactory, decisionMakerID);
	}
}
