/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
/**
 * 
 */
package magma.robots.nao.behavior.complex;

import java.util.Map;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.IKick;
import magma.agent.behavior.SupportFoot;
import magma.agent.behavior.complex.SingleComplexBehavior;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.util.geometry.Angle;
import magma.util.geometry.IPose2D;
import magma.util.parameter.ParameterMap;

/**
 * Implements a behavior which performs kicking by dynamically balancing on one
 * leg and then statically kicking
 * 
 * @author Klaus Dorer
 */
public class DynamicStaticKick extends SingleComplexBehavior implements IKick
{
	/**  */
	private static final long serialVersionUID = 1L;

	private IKick balanceBehavior;

	private IKick kickBehavior;

	private boolean isFinished;

	/**
	 * Constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param params Behavior parameters
	 * @param behaviors List of available behaviors
	 */
	public DynamicStaticKick(String name, IThoughtModel thoughtModel,
			ParameterMap params, Map<String, IBehavior> behaviors,
			IBehavior defaultBehavior, IKick balance, IKick kick,
			boolean isForwardKick)
	{
		super(name, thoughtModel, behaviors, defaultBehavior);
		balanceBehavior = balance;
		kickBehavior = kick;
		isFinished = false;
	}

	public static DynamicStaticKick getSloMoKickLeft(IThoughtModel thoughtModel,
			ParameterMap params, Map<String, IBehavior> behaviors)
	{
		IKick balanceBehavior = (IKick) behaviors
				.get(IBehaviorConstants.DYNAMIC_SLOMO_KICK_LEFT);
		IKick kickBehavior = (IKick) behaviors
				.get(IBehaviorConstants.SHOOT_SLOMO_LEFT);

		return new DynamicStaticKick(IBehaviorConstants.SLOMO_KICK_LEFT,
				thoughtModel, params, behaviors, balanceBehavior, balanceBehavior,
				kickBehavior, true);
	}

	public static DynamicStaticKick getSloMoKickRight(
			IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors)
	{
		IKick balanceBehavior = (IKick) behaviors
				.get(IBehaviorConstants.DYNAMIC_SLOMO_KICK_RIGHT);
		IKick kickBehavior = (IKick) behaviors
				.get(IBehaviorConstants.SHOOT_SLOMO_RIGHT);

		return new DynamicStaticKick(IBehaviorConstants.SLOMO_KICK_RIGHT,
				thoughtModel, params, behaviors, balanceBehavior, balanceBehavior,
				kickBehavior, true);
	}

	public static DynamicStaticKick getForwardKickLeft(
			IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors)
	{
		IKick balanceBehavior = (IKick) behaviors
				.get(IBehaviorConstants.DYNAMIC_KICK_LEFT);
		IKick kickBehavior = (IKick) behaviors
				.get(IBehaviorConstants.SHOOT_STRAIGHT_ONE_LEG_LEFT);
		// IKick kickBehavior = (IKick) behaviors
		// .get(IBehaviorConstants.KICK_CHIP_LEFT);

		return new DynamicStaticKick(IBehaviorConstants.DYNAMIC_STATIC_KICK_LEFT,
				thoughtModel, params, behaviors, balanceBehavior, balanceBehavior,
				kickBehavior, true);
	}

	public static DynamicStaticKick getForwardKickRight(
			IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors)
	{
		IKick balanceBehavior = (IKick) behaviors
				.get(IBehaviorConstants.DYNAMIC_KICK_RIGHT);
		IKick kickBehavior = (IKick) behaviors
				.get(IBehaviorConstants.SHOOT_STRAIGHT_ONE_LEG_RIGHT);
		// IKick kickBehavior = (IKick) behaviors
		// .get(IBehaviorConstants.KICK_CHIP_RIGHT);

		return new DynamicStaticKick(
				IBehaviorConstants.DYNAMIC_STATIC_KICK_RIGHT, thoughtModel, params,
				behaviors, balanceBehavior, balanceBehavior, kickBehavior, true);
	}

	public static DynamicStaticKick getSidewardKickLeft(
			IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors)
	{
		IKick balanceBehavior = (IKick) behaviors
				.get(IBehaviorConstants.DYNAMIC_SIDEKICK_LEFT);
		IKick kickBehavior = (IKick) behaviors
				.get(IBehaviorConstants.SHOOT_LEFT_ONE_LEG);

		return new DynamicStaticKick(
				IBehaviorConstants.DYNAMIC_STATIC_SIDEKICK_LEFT, thoughtModel,
				params, behaviors, balanceBehavior, balanceBehavior, kickBehavior,
				false);
	}

	public static DynamicStaticKick getSidewardKickRight(
			IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors)
	{
		IKick balanceBehavior = (IKick) behaviors
				.get(IBehaviorConstants.DYNAMIC_SIDEKICK_RIGHT);
		IKick kickBehavior = (IKick) behaviors
				.get(IBehaviorConstants.SHOOT_RIGHT_ONE_LEG);

		return new DynamicStaticKick(
				IBehaviorConstants.DYNAMIC_STATIC_SIDEKICK_RIGHT, thoughtModel,
				params, behaviors, balanceBehavior, balanceBehavior, kickBehavior,
				false);
	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		IBehavior current = getCurrentBehavior();
		if (current == balanceBehavior) {
			if (current.isFinished()) {
				// if (isForwardKick && willFail()) {
				// isFinished = true;
				// return current;
				// }
				current = kickBehavior;
			} else {
				// still balancing, write out gyro
				worldModel.logSituation(1);
			}
		} else {
			if (current.isFinished()) {
				current = balanceBehavior;
			}
		}
		return current;
	}

	// private boolean willFail()
	// {
	// // check if the kick is doomed to fail
	// Vector3D ballPos = worldModel.getBall().getLocalPosition();
	//
	// // values from 400 kicks weka file
	// // left kick x < 0.06 > 0.18 y < -0.06 > 0.13
	// // right kick x < 0.06 > 0.17 y < -0.13 > 0.04
	// boolean bad = false;
	// boolean isLeftKick = balanceBehavior.getKickingFoot() == SupportFoot.LEFT;
	// if (ballPos.getX() < 0.06 || ballPos.getX() > 0.18) {
	// bad = true;
	// } else if (isLeftKick) {
	// if (ballPos.getY() < -0.06 || ballPos.getY() > 0.13) {
	// bad = true;
	// }
	// } else if (!isLeftKick) {
	// if (ballPos.getY() < -0.13 || ballPos.getY() > 0.06) {
	// bad = true;
	// }
	// }
	//
	// if (bad) {
	// // System.out.print("badPos");
	// return true;
	// } else {
	// // System.out.print("goodPos");
	// }
	//
	// bad = false;
	// double[][] orientation = worldModel.getThisPlayer().getOrientation()
	// .getMatrix();
	// double gyroX = orientation[2][0];
	// double gyroY = orientation[2][1];
	//
	// // left kick gx < -0.45 > -0.12 gy < -0.02 > 0.25
	// // right kick gx < 0.11 > 0.47 gy < -0.02 > 0.22
	// if (gyroY < -0.02 || gyroY > 0.25) {
	// bad = true;
	// } else if (isLeftKick) {
	// if (gyroX < -0.47 || gyroX > -0.11) {
	// bad = true;
	// }
	// } else if (!isLeftKick) {
	// if (gyroX < 0.11 || gyroX > 0.47) {
	// bad = true;
	// }
	// }
	//
	// if (bad) {
	// // System.out.print("badGyro");
	// return true;
	// } else {
	// // System.out.print("goodGyro");
	// }
	// return false;
	// }

	@Override
	public void setIntendedKickDirection(Angle intendedKickDirection)
	{
		balanceBehavior.setIntendedKickDirection(intendedKickDirection);
	}

	@Override
	public SupportFoot getKickingFoot()
	{
		return balanceBehavior.getKickingFoot();
	}

	@Override
	public float isBallKickable()
	{
		return balanceBehavior.isBallKickable();
	}

	@Override
	public IPose2D getRelativeRunToPose()
	{
		return balanceBehavior.getRelativeRunToPose();
	}

	@Override
	public double getMaxKickDistance()
	{
		return kickBehavior.getMaxKickDistance();
	}

	@Override
	public void setKickPower(float kickPower)
	{
		kickBehavior.setKickPower(kickPower);
	}

	@Override
	public void init()
	{
		isFinished = false;
		super.init();
	}

	@Override
	public boolean isFinished()
	{
		IBehavior currentBehavior = getCurrentBehavior();
		boolean localFinished = (currentBehavior == kickBehavior && currentBehavior
				.isFinished()) || isFinished;
		return localFinished;
	}

	/**
	 * To prevent switching from lower behaviors, we return this behavior here.
	 */
	@Override
	public IBehavior getRootBehavior()
	{
		return this;
	}
}