/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
/**
 * 
 */
package magma.robots.nao.behavior.complex;

import java.util.Map;
import java.util.logging.Level;

import magma.agent.behavior.IBaseWalk;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.IWalk;
import magma.agent.behavior.complex.SingleComplexBehavior;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.util.ValueUtil;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;
import magma.util.parameter.ParameterMap;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Implements a behavior which performs bi-pedal walking based on the "Balance"
 * behavior.
 * 
 * @author Ingo Schindler
 */
public class Walk extends SingleComplexBehavior implements IWalk
{
	/**  */
	private static final double SLOW_DOWN_DISTANCE = 0.49;

	/**  */
	private static final long serialVersionUID = 1L;

	/**
	 * Constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param params Behavior parameters
	 * @param behaviors List of available behaviors
	 */
	public Walk(String name, IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors, IBaseWalk baseWalk)
	{
		super(name, thoughtModel, behaviors, baseWalk);
	}

	/**
	 * Constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param params Behavior parameters
	 * @param behaviors List of available behaviors
	 */
	public Walk(IThoughtModel thoughtModel, ParameterMap params,
			Map<String, IBehavior> behaviors, IBaseWalk baseWalk)
	{
		this(IBehaviorConstants.WALK, thoughtModel, params, behaviors, baseWalk);
	}

	/**
	 * Set parameters for Walk. It is possible to combine all different
	 * parameters, e.g. forwards and sidesteps.
	 * 
	 * @param forwards_backwards positive = forwards; negative = backwards (-100
	 *        .. 100)
	 * @param step_left_right positive = right; negative = left (-100 .. 100)
	 * @param turn_left_right positive = right; negative = left (-30 .. 30)
	 */
	@Override
	public void walk(double forwards_backwards, double step_left_right,
			double turn_left_right)
	{
		walk(forwards_backwards, step_left_right, turn_left_right, 100, true);
	}

	/**
	 * Set parameters for Walk. It is possible to combine all different
	 * parameters, e.g. forwards and sidesteps.
	 * 
	 * @param forwards_backwards positive = forwards; negative = backwards (-100
	 *        .. 100)
	 * @param step_left_right positive = right; negative = left (-100 .. 100)
	 * @param turn_left_right positive = right; negative = left (-30 .. 30)
	 */
	@Override
	public void walk(double forwards_backwards, double step_left_right,
			double turn_left_right, double maxSpeedForward, boolean hurry)
	{
		logger.log(Level.FINE, "walk (x/y/turn): {0}/{1}/{2}", new Object[] {
				forwards_backwards, step_left_right, turn_left_right });

		IBaseWalk base = (IBaseWalk) getCurrentBehavior();
		base.setMovement(forwards_backwards, step_left_right, turn_left_right,
				maxSpeedForward, hurry);
	}

	/**
	 * Set parameters to walk in a global direction
	 * 
	 * @param left_right negative = own goal; positive = opponent goal
	 * @param up_down negative = left field side; positive = right field side
	 * @param direction local angle where to go
	 */
	@Override
	public void globalWalk(Pose2D walkTo, Vector3D speedThere,
			double minSpeedForward, double minSpeedSideward,
			double maxSpeedForward, boolean hurry)
	{
		Angle goalDirection = worldModel.getThisPlayer().getHorizontalAngle();

		Vector3D fieldVector = walkTo.getPosition();

		Vector3D playerVector = Geometry.createZRotation(
				-(float) goalDirection.radians()).applyTo(fieldVector);

		double forwardsBackwards = playerVector.getX();
		double stepLeftRight = playerVector.getY();
		double speed = 0.0;
		double sideSpeed = 0.0;
		double desiredTurn = walkTo.angle.degrees();

		if (Math.abs(forwardsBackwards) > Math.abs(stepLeftRight)
				|| Math.abs(desiredTurn) > 40) {
			// mainly forward/backward move
			speed = speedThere.getX() * 100 + Math.abs(forwardsBackwards) * 100
					/ SLOW_DOWN_DISTANCE;
			speed = Math.min(speed, maxSpeedForward)
					* Math.signum(forwardsBackwards);

			// System.out.println("ForwardBackward: " + forwardsBackwards
			// + " speed: " + speed + " there: " + speedThere.getX());

			double ratio = 0.0;
			if (Math.abs(forwardsBackwards) > 0.01) {
				ratio = Math.abs(stepLeftRight / forwardsBackwards);
			}
			// if mainly forward we limit sideward walk considerably
			sideSpeed = Math.abs(speed) * ratio * Math.signum(stepLeftRight) * 0.2;

			if (speed > 0) {
				speed = Math.max(minSpeedForward, speed);
			} else {
				speed = Math.min(-minSpeedForward, speed);
			}

		} else {
			// mainly sideward move
			sideSpeed = speedThere.getX() * 100 + Math.abs(stepLeftRight) * 100
					/ SLOW_DOWN_DISTANCE;
			sideSpeed = Math.min(sideSpeed, maxSpeedForward)
					* Math.signum(stepLeftRight);
			//
			// System.out.println("LeftRight: " + stepLeftRight + " sideSpeed: "
			// + sideSpeed);

			double ratio = 0.0;
			if (Math.abs(stepLeftRight) > 0.01) {
				ratio = Math.abs(forwardsBackwards / stepLeftRight);
			}
			speed = Math.abs(sideSpeed) * ratio * Math.signum(forwardsBackwards);
		}

		logger.log(
				Level.FINE,
				"globalWalk (x/y/direction): {0}/{1}/{2} goalDir: {3} (fb/lr): ({4}/{5})",
				new Object[] { fieldVector.getX(), fieldVector.getY(),
						walkTo.angle.degrees(), goalDirection.degrees(),
						forwardsBackwards, stepLeftRight });

		speed = ValueUtil.limitAbs(speed, maxSpeedForward);
		walk(speed, sideSpeed, desiredTurn, maxSpeedForward, hurry);
	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		return getCurrentBehavior();
	}
}