/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.robots.nao.behavior.dynamic;

import java.util.Collection;
import java.util.Map;

import magma.agent.agentmodel.IBodyPart;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.Behavior;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.ILandmark;
import magma.robots.nao.INaoConstants;
import magma.util.FuzzyCompare;
import magma.util.geometry.Pose3D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Implements a behavior which lets the robot head focus on the ball
 * 
 * @author Ingo Schindler
 */
public class FocusBall extends Behavior
{

	/**  */
	private static final long serialVersionUID = 1L;

	private static final float TIME_TO_LOOK_AROUND = 2;

	private static final float LOOK_ARROUND_DELAY = 8;

	private final IBehavior turnHead;

	private float timeToTurnHead;

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param behaviors List of applicable behaviors
	 */
	public FocusBall(IThoughtModel thoughtModel, Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.FOCUS_BALL, thoughtModel);
		turnHead = behaviors.get(IBehaviorConstants.TURN_HEAD);
		timeToTurnHead = worldModel.getGlobalTime();
	}

	@Override
	public void perform()
	{
		boolean seeLandMark = false;
		Collection<ILandmark> landmarks = worldModel.getLandmarks();
		for (ILandmark landmark : landmarks) {
			float time = landmark.getAge(worldModel.getGlobalTime());
			if (time < 0.1) {
				seeLandMark = true;
				break;
			}
		}

		Vector3D ballPos = worldModel.getBall().getLocalPosition();

		float timeDelta = worldModel.getGlobalTime() - timeToTurnHead;

		if (worldModel.getBall().getAge(worldModel.getGlobalTime()) > 0.1
				|| !seeLandMark
				|| (ballPos.getNorm() > 1.5 && FuzzyCompare
						.gte(timeDelta, LOOK_ARROUND_DELAY + TIME_TO_LOOK_AROUND,
								TIME_TO_LOOK_AROUND))) {
			turnHead.perform();
			if (timeDelta > LOOK_ARROUND_DELAY + TIME_TO_LOOK_AROUND)
				timeToTurnHead = worldModel.getGlobalTime();

		} else {
			// Calculate position of ball relative to camera body
			IBodyPart head = agentModel.getBodyPart(INaoConstants.Head);
			Pose3D headPose = head.getPose();

			Vector3D ballVec = new Vector3D(-ballPos.getY(), ballPos.getX(),
					ballPos.getZ());

			ballVec = headPose.getPosition().negate()
					.add(headPose.getOrientation().applyInverseTo(ballVec));
			ballPos = new Vector3D(ballVec.getY(), -ballVec.getX(), ballVec.getZ());

			// Adjust head joints to focus the ball
			agentModel.getWriteableHJ(INaoConstants.NeckYaw).adjustAxisPosition(
					Math.toDegrees(ballPos.getAlpha() / 3));
			agentModel.getWriteableHJ(INaoConstants.NeckPitch).adjustAxisPosition(
					Math.toDegrees(ballPos.getDelta() / 3));
		}
	}
}
