/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
/**
 * 
 */
package magma.robots.nao.behavior.dynamic.walk;

import magma.util.function.SplineFunction;

/**
 * Represents an ankle in the robot body.
 * 
 * @author Ingo Schindler
 */
public class Ankle
{
	private SplineFunction thetaSpline, xValuesSpline, zValuesSpline,
			gammaValuesSpline;

	private double len;

	/**
	 * Retrieve ankle length
	 * 
	 * @return Length
	 */
	public double getLen()
	{
		return len;
	}

	/**
	 * Set walk behavior parameters
	 * 
	 * @param p Parameters
	 */
	void setParam(WalkBehaviorParameter p)
	{
		double[] yValuesTheta = { p.getQgs(), p.getQb(), p.getQf(), p.getQge() };
		double[] xValuesTheta = { p.getK() * p.getTc(),
				p.getK() * p.getTc() + p.getTd(), (p.getK() + 1) * p.getTc(),
				(p.getK() + 1) * p.getTc() + p.getTd() };

		thetaSpline = new SplineFunction(xValuesTheta, yValuesTheta, 1);

		double[] yValuesX = {
				(p.getK()) * p.getDs(),
				(p.getK() - 1) * p.getDs() + p.getL_an()
						* Math.sin(Math.toRadians(p.getQb())) + p.getL_af()
						* (1 - Math.cos(Math.toRadians(p.getQb()))),
				p.getK() * p.getDs() + p.getLao(),
				(p.getK() + 2) * p.getDs() - p.getL_an()
						* Math.sin(Math.toRadians(p.getQf())) - p.getL_ab()
						* (1 - Math.cos(Math.toRadians(p.getQf()))),
				(p.getK()) * p.getDs() };
		double[] xValuesX = { p.getK() * p.getTc(),
				p.getK() * p.getTc() + p.getTd(), p.getK() * p.getTc() + p.getTm(),
				(p.getK() + 1) * p.getTc(), (p.getK() + 1) * p.getTc() + p.getTd() };

		xValuesSpline = new SplineFunction(xValuesX, yValuesX, 1);

		double[] yValuesZ = {
				p.getH_gs() + p.getL_an(),
				p.getH_gs() + p.getL_af() * Math.sin(Math.toRadians(p.getQb()))
						+ p.getL_an() * Math.cos(Math.toRadians(p.getQb())),
				p.getHao(),
				p.getH_ge() + p.getL_ab() * Math.sin(Math.toRadians(p.getQf()))
						+ p.getL_an() * Math.cos(Math.toRadians(p.getQf())),
				p.getH_ge() + p.getL_an() };
		double[] xValuesZ = { p.getK() * p.getTc(),
				p.getK() * p.getTc() + p.getTd(), p.getK() * p.getTc() + p.getTm(),
				(p.getK() + 1) * p.getTc(), (p.getK() + 1) * p.getTc() + p.getTd() };

		zValuesSpline = new SplineFunction(xValuesZ, yValuesZ, 1);

		double[] yValuesGamma = { 0, 0, 0, 0 };
		double[] xValuesGamma = { p.getK() * p.getTc(),
				p.getK() * p.getTc() + p.getTd(), (p.getK() + 1) * p.getTc(),
				(p.getK() + 1) * p.getTc() + p.getTd() };

		gammaValuesSpline = new SplineFunction(xValuesGamma, yValuesGamma, 1);

		len = ((p.getK() + 1) * p.getTc() + p.getTd());
	}

	/**
	 * Constructor
	 * 
	 * @param p Walking parameters
	 */
	public Ankle(WalkBehaviorParameter p)
	{
		setParam(p);
	}

	/**
	 * Calculate the theta value based on the current cycle
	 * 
	 * @param t Cycle number (>0)
	 * @return Calculated value
	 */
	public double getTheta(double t)
	{
		if (t < 0) {
			throw new IllegalArgumentException("t should be > 0");
		}

		return thetaSpline.value(t % len);
	}

	/**
	 * Calculate the gamma value based on the current cycle
	 * 
	 * @param t Cycle number (>0)
	 * @return Calculated value
	 */
	public double getGamma(double t)
	{
		if (t < 0) {
			throw new IllegalArgumentException("t should be > 0");
		}

		return gammaValuesSpline.value(t % len);
	}

	/**
	 * Calculate the x value based on the current cycle
	 * 
	 * @param t Cycle number (>0)
	 * @return Calculated value
	 */
	public double getXValue(double t)
	{
		if (t < 0) {
			throw new IllegalArgumentException("t should be > 0");
		}

		return xValuesSpline.value(t % len);
	}

	/**
	 * Calculate the z value based on the current cycle
	 * 
	 * @param t Cycle number (>0)
	 * @return Calculated value
	 */
	public double getZValue(double t)
	{
		if (t < 0) {
			throw new IllegalArgumentException("t should be > 0");
		}

		return zValuesSpline.value(t % len);
	}
}
