/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.robots.nao.behavior.dynamic.walk;

import static java.lang.Math.asin;
import static java.lang.Math.cos;
import static java.lang.Math.sin;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;

import magma.agent.agentmodel.IBodyModel;
import magma.agent.behavior.IBaseWalk;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.Behavior;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.robots.nao.INaoConstants;
import magma.util.FuzzyCompare;
import magma.util.ValueUtil;
import magma.util.geometry.Geometry;
import magma.util.parameter.IParameterList;
import magma.util.parameter.ParameterMap;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * @author Ingo Schindler
 * 
 */
public class Balance extends Behavior implements IBaseWalk
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** which side of the body is meant */
	private enum BodySide {
		LEFT, RIGHT
	}

	/** Constant that defines the maximal turn angle */
	private static final double MAX_TURN_ANGLE = 40;

	/** defines the maximal allowed angle of the HipRoll for a SideStep */
	private static final double MAX_ALLOWD_SIDESTEP_ANGLE = 50;

	private static final double SPEED_FACTOR = 1.0;

	/** counter for the movement (currentStep) */
	private double movementCycle = 0;

	/** Speed you want to reach (forward - backward) */
	private double desiredSpeed = 0;

	/** Actual Speed (forward - backward) */
	private double currentSpeed = 0;

	/** side step speed */
	private double desiredDirection = 0;

	/** Direction you actually walking to (sidesteps) */
	private double currentDirection = 0;

	/** Orientation you want to reach (turning) */
	private double desiredAngle = 0;

	/** Current angle to turn for the next step */
	private double currentAngle = 0;

	/** forward-backward speed delta calculated from gyro for not falling */
	private double compensationSpeed = 0;

	/** sidewise speed delta calculated from gyro for not falling */
	private final double compensationDirection = 0;

	/** true if the next call to movement method should be ignored */
	private boolean ignoreNextCallToMovement;

	/** the foot on which we are standing */
	private BodySide supportFoot = BodySide.RIGHT;

	/** the foot on which we were standing last cycle */
	private BodySide oldSupportFoot = BodySide.LEFT;

	/** average value of x axis gyro when walking */
	private double averageGyroXLongTerm;

	/** average value of y axis gyro when walking */
	private double averageGyroYLongTerm;

	/** max speed forwards */
	private double maxSpeedForward;

	/**  */
	private transient Map<WalkMode, WalkSituation> walkParams;

	private transient WalkSituation situation;

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param thoughtModel Reference to the thought model object
	 * @param params Behavior parameters
	 */
	public Balance(IThoughtModel thoughtModel, ParameterMap params)
	{
		this(IBehaviorConstants.BALANCE, thoughtModel, params);
	}

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param name name of the Behavior
	 * @param thoughtModel Reference to the thought model object
	 * @param params Behavior parameters
	 */
	public Balance(String name, IThoughtModel thoughtModel, ParameterMap params)
	{
		super(name, thoughtModel);
		averageGyroXLongTerm = 0;
		averageGyroYLongTerm = 0;
		init();
		IParameterList param = params.get(name);
		setParams(param);
	}

	/**
	 * Set WalkBehaviorParameter
	 * 
	 * @param params Behavior parameter
	 */
	private void setParams(IParameterList params)
	{
		walkParams = new HashMap<WalkMode, WalkSituation>();
		for (WalkMode mode : WalkMode.values()) {
			WalkSituation s = new WalkSituation(params, mode);
			walkParams.put(mode, s);
		}
	}

	/**
	 * This method returns an Ankle's and Hip's-Roll angle to perform a SideStep
	 * 
	 * @param right true if the angle is thought for the right leg
	 * @return an Ankle's and Hip's-Roll angle to perform a SideStep (degrees)
	 */
	protected double getSideStepRatio(BodySide side)
	{
		double localDirection = getDirection() / 100 * MAX_ALLOWD_SIDESTEP_ANGLE;

		if (localDirection > 0 && supportFoot == BodySide.RIGHT) {
			// nothing to do
		} else if (localDirection < 0 && supportFoot == BodySide.LEFT) {
			localDirection = -localDirection;
		} else {
			localDirection = 0;
		}
		if (side == BodySide.RIGHT) {
			localDirection = -localDirection;
		}
		return localDirection;
	}

	/**
	 * Retrieve the current walking speed
	 * 
	 * @return Current walking speed
	 */
	protected double getSpeed()
	{
		return currentSpeed + compensationSpeed;
	}

	/**
	 * Retrieve the current side speed
	 * 
	 * @return Current side speed
	 */
	protected double getDirection()
	{
		return currentDirection + compensationDirection;
	}

	/**
	 * Retrieve the current turn angle
	 * 
	 * @return Current turn angle
	 */
	protected double getAngle()
	{
		return currentAngle;
	}

	/**
	 * Retrieve the YawPitch angle for the left leg, to perform a turn-Movement
	 * 
	 * @return YawPitch angle in degrees
	 */
	protected double getYawPitch()
	{
		if (currentAngle > 0 && supportFoot == BodySide.RIGHT) {
			return -currentAngle;
		}

		if (currentAngle < 0 && supportFoot == BodySide.LEFT) {
			return currentAngle;
		}

		return 0;
	}

	@Override
	public void init()
	{
		super.init();
		currentSpeed = 0;
		currentDirection = 0;
		currentAngle = 0;
		desiredSpeed = 0;
		movementCycle = 0;
		ignoreNextCallToMovement = false;
		compensationSpeed = 0;
		setSupportFoot();
	}

	@Override
	public boolean isFinished()
	{
		// checks if both knees and hip yawpitchs have roughly the same angle
		float axis = agentModel.getHJExpected(INaoConstants.LKneePitch).getAxis();
		float axis2 = agentModel.getHJExpected(INaoConstants.RKneePitch)
				.getAxis();
		float yawpitchl = agentModel.getHJExpected(INaoConstants.LHipYawPitch)
				.getAxis();
		float yawpitchr = agentModel.getHJExpected(INaoConstants.RHipYawPitch)
				.getAxis();

		double precision = 6.0;

		boolean result = FuzzyCompare.eq(axis, axis2, precision)
				&& FuzzyCompare.eq(yawpitchl, yawpitchr, 1.0);

		logger.log(Level.FINE, "axis: {0} axis2: {1} yawL: {2} yawR: {3}",
				new Object[] { axis, axis2, yawpitchl, yawpitchr });

		return result;
	}

	/**
	 * Support foot is that, which have more force resistance
	 */
	protected void setSupportFoot()
	{
		oldSupportFoot = supportFoot;
		boolean result = agentModel.getForceResistance(INaoConstants.LFootForce)
				.getForce().getZ() < agentModel
				.getForceResistance(INaoConstants.RFootForce).getForce().getZ();

		if (result) {
			supportFoot = BodySide.RIGHT;
		} else {
			supportFoot = BodySide.LEFT;
		}
	}

	/**
	 * Main method called to perform the behavior
	 */
	@Override
	public void perform()
	{
		WalkMode mode = detectMode();
		situation = walkParams.get(mode);
		Ankle ankle = situation.getAnkle();
		double stepOffset = situation.getStepOffset();
		movementCycle += 1;
		averageOutGyroDrift();
		setSupportFoot();
		accelerate(0, 0);
		compensateGyro();

		logger.log(
				Level.FINE,
				"balance(desired/cur/comp/used) speed: ({0}/{1}/{2}/{3})"
						+ " direction: ({4}/{5}/{6}/{7}) angle: ({8}/{9}) support (new/old): ({10}/{11})",
				new Object[] { desiredSpeed, currentSpeed, compensationSpeed,
						getSpeed(), desiredDirection, currentDirection,
						compensationDirection, getDirection(), desiredAngle,
						currentAngle, supportFoot, oldSupportFoot });

		// in order to use floor compensation values in accelerate we have to use
		// old speed values to calculate speedfactor
		double speedfactor = getSpeed() / 100.0 * SPEED_FACTOR;
		double floorX = calculateFloorX();
		double floorY = calculateFloorY(speedfactor);
		// System.out.println("floor: " + floorX + "/" + floorY + " speedfactor: "
		// + speedfactor);

		double turnR = getYawPitch();
		double turnRrad = Math.toRadians(turnR);
		double floorXcompensationR = cos(turnRrad) * floorY + sin(turnRrad)
				* floorX;
		double floorYcompensationR = (sin(turnRrad) * floorY + cos(turnRrad)
				* floorX);

		double turnLrad = Math.toRadians(turnR);
		double floorXcompensationL = cos(turnLrad) * floorY + sin(turnLrad)
				* floorX;
		double floorYcompensationL = (sin(turnLrad) * floorY + cos(turnLrad)
				* floorX);

		// double floorXcompensationR = 0;
		// double floorXcompensationL = 0;
		// double floorYcompensationR = 0;
		// double floorYcompensationL = 0;

		double desiredAnklePitchR = ankle.getTheta(movementCycle);
		double desiredAnklePitchL = ankle.getTheta(movementCycle + stepOffset);

		// desired position of the feet, currently y coordinate is neglected
		Vector3D anklePositionR = new Vector3D(ankle.getXValue(movementCycle)
				* speedfactor, 0, ankle.getZValue(movementCycle));
		Vector3D anklePositionL = new Vector3D(ankle.getXValue(movementCycle
				+ stepOffset)
				* speedfactor, 0, ankle.getZValue(movementCycle + stepOffset));

		// calculate the point where the knee is desired to be
		Vector3D kneePositionR = intersectionPoint(anklePositionR);
		Vector3D kneePositionL = intersectionPoint(anklePositionL);

		// calculate the angles of hip and knee pitch to reach the positions
		double[] angleR = Geometry.calculateAngles(kneePositionR, anklePositionR);
		angleR[0] += floorXcompensationR;
		double[] angleL = Geometry.calculateAngles(kneePositionL, anklePositionL);
		angleL[0] += floorXcompensationL;

		// adjust roll of hip and ankle for side step
		double sideStepRatioRight = getSideStepRatio(BodySide.RIGHT);
		double sideStepRatioLeft = getSideStepRatio(BodySide.LEFT);
		double sideStepRollRight = sideStepRatioRight + sideStepRatioLeft / 2;
		double sideStepRollLeft = sideStepRatioLeft + sideStepRatioRight / 2;

		// calculate final joint values
		double rHipRoll = -floorYcompensationR + sideStepRollRight + turnR / 6;
		double lHipRoll = -floorYcompensationL + sideStepRollLeft - turnR / 6;
		double rHipPitch = angleR[0] + turnR / 2;
		double lHipPitch = angleL[0] + turnR / 2;

		double rKneePitch = angleR[1] + turnR / 10;
		double lKneePitch = angleL[1] + turnR / 10;

		double rFootPitch = -(angleR[0] + angleR[1]) + desiredAnklePitchR
				* speedfactor;
		double lFootPitch = -(angleL[0] + angleL[1]) + desiredAnklePitchL
				* speedfactor;
		double rFootRoll = floorYcompensationR - sideStepRollRight;
		double lFootRoll = floorYcompensationL - sideStepRollLeft;

		// set values
		IBodyModel bodyModelCopy = moveLegs(turnR, turnR, rHipRoll, lHipRoll,
				rHipPitch, lHipPitch, rKneePitch, lKneePitch, rFootPitch,
				lFootPitch, rFootRoll, lFootRoll);

		// adjust feet to ground
		// adjustFeetToGround(bodyModelCopy);
		swingArms(bodyModelCopy);
	}

	/**
	 * Detect mode from parameters
	 * @return the mode
	 */
	private WalkMode detectMode()
	{
		// we have to use current values not desired since the acceleration has to
		// fit to current movement
		double turningSpeed = Math.abs(currentAngle / 30.0 * 100);
		double forwardSpeed = Math.abs(currentSpeed);
		double sidewardspeed = Math.abs(currentDirection);
		if (forwardSpeed < sidewardspeed) {
			if (turningSpeed < sidewardspeed) {
				return WalkMode.SIDEWARD;
			}
			return WalkMode.TURNING;
		} else {
			if (forwardSpeed < turningSpeed) {
				return WalkMode.TURNING;
			}
			if (currentSpeed > 0) {
				return WalkMode.FORWARD;
			} else {
				return WalkMode.BACKWARD;
			}
		}
	}

	/**
	 * once the gyro drifts, we average out the problem
	 */
	private void averageOutGyroDrift()
	{
		double[][] orientation = worldModel.getThisPlayer().getOrientation()
				.getMatrix();
		double gyroX = orientation[2][0];
		averageGyroXLongTerm = (500 * averageGyroXLongTerm + gyroX) / 501;

		double gyroY = orientation[2][1];
		averageGyroYLongTerm = (500 * averageGyroYLongTerm + gyroY) / 501;

		logger.log(Level.FINE, "gyroX/Y: {0}/{1} average: {2}/{3}", new Object[] {
				gyroX, gyroY, averageGyroXLongTerm, averageGyroYLongTerm });
	}

	private double getCompensatedGyroX()
	{
		return worldModel.getThisPlayer().getOrientation().getMatrix()[2][0]
				- averageGyroXLongTerm;
	}

	private double getCompensatedGyroY()
	{
		return worldModel.getThisPlayer().getOrientation().getMatrix()[2][1];
		// we subtract 0.05 since when walking that is the average hang we have
		// switched off
		// - averageGyroYLongTerm;
	}

	/**
	 * @param speedfactor
	 * @return the angle (in degrees) to compensate in x direction
	 */
	private double calculateFloorX()
	{
		double floorX = Math.toDegrees(asin(getCompensatedGyroX())) * 2;
		// limiting the floor compensation
		floorX = ValueUtil.limitAbs(floorX, 10);
		return floorX;
	}

	/**
	 * @return the angle (in degrees) to compensate in y direction
	 */
	private double calculateFloorY(double speedfactor)
	{

		double deviation = situation.getHip().getXValue(movementCycle)
				* speedfactor;
		double floorY = Math.toDegrees(asin(getCompensatedGyroY())) + deviation;
		// limiting the floor compensation
		floorY = ValueUtil.limitAbs(floorY, 20);
		return floorY;
	}

	private IBodyModel moveLegs(double rHipYawPitch, double lHipYawPitch,
			double rHipRoll, double lHipRoll, double rHipPitch, double lHipPitch,
			double rKneePitch, double lKneePitch, double rFootPitch,
			double lFootPitch, double rFootRoll, double lFootRoll)
	{
		IBodyModel copy = agentModel.getFutureBodyModel();
		copy.setHingeJointPosition(INaoConstants.RHipYawPitch, rHipYawPitch);
		copy.setHingeJointPosition(INaoConstants.LHipYawPitch, lHipYawPitch);
		copy.setHingeJointPosition(INaoConstants.RHipRoll, rHipRoll);
		copy.setHingeJointPosition(INaoConstants.LHipRoll, lHipRoll);
		copy.setHingeJointPosition(INaoConstants.RHipPitch, rHipPitch);
		copy.setHingeJointPosition(INaoConstants.LHipPitch, lHipPitch);
		copy.setHingeJointPosition(INaoConstants.RKneePitch, rKneePitch);
		copy.setHingeJointPosition(INaoConstants.LKneePitch, lKneePitch);
		copy.setHingeJointPosition(INaoConstants.RFootPitch, rFootPitch);
		copy.setHingeJointPosition(INaoConstants.LFootPitch, lFootPitch);
		copy.setHingeJointPosition(INaoConstants.RFootRoll, rFootRoll);
		copy.setHingeJointPosition(INaoConstants.LFootRoll, lFootRoll);

		return copy;
	}

	/**
	 * Accelerate all speeds. This is basically the morphing between different
	 * walk-Behaviors
	 * @param floorY
	 * @param floorX
	 */
	protected void accelerate(double floorX, double floorY)
	{
		// Accelerate by n% per cycle
		double speedChange;
		WalkBehaviorParameter parameter = situation.getParameter();
		double deceleration = parameter
				.getValue(WalkBehaviorParameter.Param.DECELERATION);
		double acceleration = parameter
				.getValue(WalkBehaviorParameter.Param.ACCELERATION);

		if (desiredSpeed > 0) {
			// want forward
			if (currentSpeed < desiredSpeed) {
				speedChange = acceleration;
			} else {
				speedChange = deceleration;
			}
		} else {
			// want backward or stop
			if (currentSpeed > desiredSpeed) {
				speedChange = acceleration;
			} else {
				speedChange = deceleration;
			}
		}

		if (maxSpeedForward < 99) {
			// limit acceleration if we do not need 100% speed
			speedChange *= 0.4;
		}

		if (compensationSpeed < 0) {
			// limit acceleration if hanging backwards
			speedChange *= (Geometry.getLinearFuzzyValue(-20, -5, true,
					compensationSpeed) * 0.8 + 0.2);
		} else if (compensationSpeed > 0) {
			// limit acceleration if hanging forwards
			speedChange *= (Geometry.getLinearFuzzyValue(5, 20, false,
					compensationSpeed) * 0.8 + 0.2);
		}

		double accelerationSide = 1;
		if (Math.abs(currentAngle) > 15
				&& Math.abs(currentSpeed) < Math.abs(desiredSpeed)) {
			// we have to wait with acceleration until turning is finished
			speedChange = 0.1;
			accelerationSide = 0.1;
		}

		currentSpeed = ValueUtil.adjustValue(currentSpeed, desiredSpeed,
				speedChange);

		currentDirection = ValueUtil.adjustValue(currentDirection,
				desiredDirection, accelerationSide);

		// Calculate how much the the robot is allowed to turn
		double maxSpeed = Math
				.max(Math.abs(getSpeed()), Math.abs(getDirection()));
		double percentOfTurn = Geometry.getLinearFuzzyValue(10, 50, false,
				maxSpeed) * 0.8 + 0.2;

		// if desired angle has to be restricted, do it
		double localAngle = desiredAngle;
		if (Math.abs(desiredAngle) > (MAX_TURN_ANGLE * percentOfTurn)) {
			if (desiredAngle < 0) {
				localAngle = -(MAX_TURN_ANGLE * percentOfTurn);
			} else {
				localAngle = (MAX_TURN_ANGLE * percentOfTurn);
			}
		}

		double accelerationTurn = 4;
		if (maxSpeed > 50) {
			// when too fast we may not turn immediately
			accelerationTurn = 1;
		}
		// accelerationTurn = (Geometry.getLinearFuzzyValue(3, 6, false, floorX) +
		// 0.1)
		// * accelerationTurn;
		// accelerationTurn = (Geometry.getLinearFuzzyValue(2, 5, false, floorY) +
		// 0.1)
		// * accelerationTurn;
		currentAngle = ValueUtil.adjustValue(currentAngle, localAngle,
				accelerationTurn);
	}

	/**
	 * Changes the speed and direction according to our gyro to work against
	 * falling down
	 * 
	 * TODO: This should base on the values of the ACC not on GYR
	 * 
	 * TODO: I, think I know what you are trying with this, but you have also to
	 * take care, which is the support foot at this moment
	 */
	private void compensateGyro()
	{
		double gyroY = getCompensatedGyroY();

		// WalkBehaviorParameter newParams = parameter;
		// newParams.setQgs(gyroY
		// * agentModel.getBodyPart(INaoConstants.RFoot).getPosition().getY());
		// newParams.setQge(gyroY);
		// setParams(newParams);

		double factor = -Geometry.getLinearFuzzyValue(0, 20, true, desiredSpeed)
				* 200 * Math.signum(desiredSpeed);
		double newSpeed = gyroY * factor;
		newSpeed = ValueUtil.limitAbs(newSpeed, 90);
		// FEATURE: make compensation speed inert
		// compensationSpeed = (22 * compensationSpeed + newSpeed) / 23;
		compensationSpeed = newSpeed;

		// double gyroX = agentModel.getGyroRate(INaoConstants.TorsoGyro)
		// .getOrientation().m20;
		// double factorSide = factor;
		// compensationDirection = gyroX * factorSide;
		// logger.log(Level.FINE, "gyroX/Y: {0}/{1}", new Object[] { gyroX, gyroY
		// });
	}

	/**
	 * Set angel for Orientation (max: 30; min: -30)
	 * 
	 * @param turnLeftRight values >0 means right turn; <0 left turn
	 * 
	 */
	public void setAngle(double turnLeftRight)
	{
		this.desiredAngle = ValueUtil.limitAbs(turnLeftRight, MAX_TURN_ANGLE);
	}

	/**
	 * Set forwards or backwards speed in percent. (-100% .. -100%)
	 * 
	 * @param speed values (forward > 0 > backward)
	 */
	public void setDesiredSpeed(double speed)
	{
		this.desiredSpeed = ValueUtil.limitAbs(speed, 100);
	}

	/**
	 * Set Direction in percent (-100 .. 100)
	 * 
	 * @param direction values (right > 0 > left)
	 */
	public void setDirection(double direction)
	{
		this.desiredDirection = ValueUtil.limitAbs(direction, 100);
	}

	/**
	 * Swinging arms in the opposite way to the legs
	 * 
	 * TODO: Make Arm swing dynamic
	 */
	private void swingArms(IBodyModel bodyModel)
	{
		// Arms
		if (supportFoot == BodySide.RIGHT) {
			bodyModel.setHingeJointPosition(INaoConstants.RShoulderPitch, -120);
			bodyModel.setHingeJointPosition(INaoConstants.RShoulderYaw, -15);
			bodyModel.setHingeJointPosition(INaoConstants.RArmRoll, 68);
			bodyModel.setHingeJointPosition(INaoConstants.RArmYaw, 81);
			bodyModel.setHingeJointPosition(INaoConstants.LShoulderPitch, -60);
			bodyModel.setHingeJointPosition(INaoConstants.LShoulderYaw, 20);
			bodyModel.setHingeJointPosition(INaoConstants.LArmRoll, -25);
			bodyModel.setHingeJointPosition(INaoConstants.LArmYaw, -90);

		} else {
			bodyModel.setHingeJointPosition(INaoConstants.RShoulderPitch, -60);
			bodyModel.setHingeJointPosition(INaoConstants.RShoulderYaw, -20);
			bodyModel.setHingeJointPosition(INaoConstants.RArmRoll, 25);
			bodyModel.setHingeJointPosition(INaoConstants.RArmYaw, 90);
			bodyModel.setHingeJointPosition(INaoConstants.LShoulderPitch, -120);
			bodyModel.setHingeJointPosition(INaoConstants.LShoulderYaw, 15);
			bodyModel.setHingeJointPosition(INaoConstants.LArmRoll, -68);
			bodyModel.setHingeJointPosition(INaoConstants.LArmYaw, -81);
		}
	}

	/**
	 * Intersect two circles. First circle is always at (0|0) with radius 0.14.
	 * Second is always (x|z) with radius 0.11.
	 * 
	 * Basically this is the inverse kinematics for the leg.
	 * 
	 * @param x X-Coordinate for the second circle
	 * @param z Z-Coordinate for the second circle
	 * 
	 * @return two Points where the two circle intersect
	 */
	private Vector3D intersectionPoint(Vector3D point)
	{
		double l1 = 0.14;
		double l2 = 0.11;
		double z_ = -((l1 + l2) - point.getZ());

		Vector3D[] intersection = Geometry.intersectionPoint(0, 0, l1,
				point.getX(), z_, l2);
		return intersection[0];
	}

	@Override
	public IBehavior switchFrom(IBehavior actualBehavior)
	{
		IBehavior realBehavior = actualBehavior.getRootBehavior();

		if (realBehavior instanceof Balance) {
			this.morphing((Balance) realBehavior);
			actualBehavior.onLeavingBehavior(this);
			return this;
		}

		return super.switchFrom(actualBehavior);
	}

	/**
	 * @param speed the desired forward speed
	 * @param direction the desired side wise speed
	 * @param angle the desired turn angle
	 * @param ignoreNextCall true if the next call to this method should be
	 *        ignored
	 */
	@Override
	public void setMovement(double speed, double direction, double angle,
			double maxSpeedForward, boolean hurry)
	{
		setMovement(speed, direction, angle, false, maxSpeedForward);
	}

	public void setMovement(double speed, double direction, double angle,
			boolean ignoreNextCall, double maxSpeedForward)
	{
		if (!ignoreNextCall && ignoreNextCallToMovement) {
			ignoreNextCallToMovement = false;
			return;
		}

		ignoreNextCallToMovement = ignoreNextCall;
		setDesiredSpeed(speed);
		setDirection(direction);
		setAngle(angle);
		this.maxSpeedForward = maxSpeedForward;
	}

	/**
	 * Morphes from one Balance-Behavior to another
	 * 
	 * @param source Behavior from which to take the values for the morphing
	 */
	public void morphing(Balance source)
	{
		this.init();
		this.walkParams = source.walkParams;
		this.currentSpeed = source.currentSpeed;
		this.currentAngle = source.currentAngle;
		this.currentDirection = source.currentDirection;
		this.movementCycle = source.movementCycle;
	}
}
