/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.dynamic.walk;

import magma.util.parameter.Parameter;
import magma.util.parameter.ParameterList;

/**
 * Represents the parameter set necessary for walking
 * 
 * @author Ingo Schindler
 * 
 *         What exactly this parameters are is documented in my Bachelor-Thesis
 *         (Bipedal-Walking in the simulated RoboCup-3D environment)
 */
public class WalkBehaviorParameter extends ParameterList
{
	// extremely good and stable
	// Tc 18.585773
	// Td 10.824421
	// Tm 3.0998294
	// Ds 0.061073594
	// Hao 0.07191266
	// Lao 0.011017295
	// Qb -2.19759
	// Qf 3.104993
	// x_ed -4.1913013
	// x_sd 6.942412
	// HipHeight 0.03596818
	// acceleration_forwards 3.9780335

	// learned in Istanbul
	// Tc 17.772072
	// Td 10.288188
	// Tm 3.0008128
	// Ds 0.0602371
	// Hao 0.07223502
	// Lao 0.010303149
	// Qb -1.8156232
	// Qf 3.2881079
	// x_ed -4.7516885
	// x_sd 7.3903484
	// HipHeight 0.02998544
	// acceleration_forwards 3.933716

	private static final long serialVersionUID = -2324636896847512436L;

	// k = step.
	// Tc = duration of the whole motion (Td < Tm < Tc)
	// Tm = time when the foot should touch the ground again
	// Td = middle of the motion
	// h_gs = Height of Ground when leaving the Ground (hip height)
	// h_ge = Height of Ground when touching the Ground again (hip height)
	// Ds * 2 = Step-Length
	// Hao = Highest Point of the Foot during a step
	// Lao = Distance of Hao relative to the Point the foot leaves the Ground
	// qgs = Angle of the Ground when leaving the Ground
	// qb = Foot-Angle when leaving the Ground
	// qf = Foot-Angle when touching the Ground again
	// qge = Angle of the Ground when touching the Ground again
	// x_ed = Hip to Ankle distance at the start of the single support phase
	// should be a value 0 < x_ed <= 0.5Ds
	// x_sd = Hip to Ankle distance at the end of the single support phase should
	// be a value 0 < x_sd <= 0.5Ds
	// l_ab = Distance Ankle->Heel
	// l_af = Distance Ankle->Tiptoe
	// l_an = Distance Ankle->Sole
	public enum Param {
		HIP_HEIGHT, K, TC, TM, TD, DS, HAO, LAO, //
		QGS, QB, QF, QGE, XED, XSD, LAB, LAF, LAN, ACCELERATION, DECELERATION,
	};

	private WalkMode mode;

	public WalkBehaviorParameter(Parameter[] params)
	{
		super(params);

		// create the parameters, order DOES matter
		// addParameter(Param.HIP_HEIGHT, 0, true, 0.1f);
		// addParameter(Param.K, 0f, false, 0f);
		// addParameter(Param.TC, 0, true, 0.5f);
		// addParameter(Param.TM, 0, true, 0.5f);
		// addParameter(Param.TD, 0, true, 0.5f);
		// addParameter(Param.DS, 0, true, 0.004f);
		// addParameter(Param.HAO, 0, true, 0.005f);
		// addParameter(Param.LAO, 0, true, 0.004f);
		// addParameter(Param.QGS, 0f, false, 0f);
		// addParameter(Param.QB, 0, true, 0.1f);
		// addParameter(Param.QF, 0, true, 0.1f);
		// addParameter(Param.QGE, 0f, false, 0f);
		// addParameter(Param.XED, 0, true, 0.4f);
		// addParameter(Param.XSD, 0, true, 0.4f);
		// addParameter(Param.LAB, 0, true, 0.01f);
		// addParameter(Param.LAF, 0, true, 0.01f);
		// addParameter(Param.LAN, 0, true, 0.005f);
		// addParameter(Param.ACCELERATION, 0, true, 0.3f);
		// addParameter(Param.DECELERATION, 0, true, 0.3f);
	}

	public static WalkBehaviorParameter getInstance()
	{
		WalkBehaviorParameter result = new WalkBehaviorParameter(createParams());
		result.setValues();
		return result;
	}

	/**
	 * @return
	 */
	protected static Parameter[] createParams()
	{
		Parameter[] params = new Parameter[Param.values().length];
		int i = 0;
		for (Param param : Param.values()) {
			params[i++] = new Parameter(param.name());
		}
		return params;
	}

	protected void setValues()
	{
		// add the values in a separate step
		// this is done to simplify the copy paste of values from console
		setValue(Param.HIP_HEIGHT, 0.03596818f);
		setValue(Param.K, 0);
		setValue(Param.TC, 18.585773f);
		setValue(Param.TM, 10.824421f);
		setValue(Param.TD, 3.0998294f);
		setValue(Param.DS, 0.061073594f);
		setValue(Param.HAO, 0.07191266f);
		setValue(Param.LAO, 0.011017295f);
		setValue(Param.QGS, 0);
		setValue(Param.QB, -2.19759f);
		setValue(Param.QF, 3.104993f);
		setValue(Param.QGE, 0);
		setValue(Param.XED, -4.1913013f);
		setValue(Param.XSD, 6.942412f);
		setValue(Param.LAB, 0.05f);
		setValue(Param.LAF, 0.11f);
		setValue(Param.LAN, 0.015f);
		setValue(Param.ACCELERATION, 3.9780335f);
		setValue(Param.DECELERATION, 3);
	}

	/**
	 * @param accelerationForwards
	 * @param value
	 */
	protected void setValue(Param which, float value)
	{
		setValue(which.ordinal(), value);
	}

	/**
	 * @param index the 0 based index of the parameter.
	 * @return the value of the parameter specified by index
	 */
	final public float getValue(Param which)
	{
		return getValue(which.ordinal());
	}

	public WalkMode getMode()
	{
		return mode;
	}

	public double getHipHeight()
	{
		return getValue(Param.HIP_HEIGHT);
	}

	public double getK()
	{
		return getValue(Param.K);
	}

	public double getTd()
	{
		return getValue(Param.TD);
	}

	public double getH_gs()
	{
		return getValue(Param.HIP_HEIGHT);
	}

	public double getH_ge()
	{
		return getValue(Param.HIP_HEIGHT);
	}

	public double getDs()
	{
		return getValue(Param.DS);
	}

	public double getHao()
	{
		return getValue(Param.HAO);
	}

	public double getLao()
	{
		return getValue(Param.LAO);
	}

	public double getX_ed()
	{
		return getValue(Param.XED);
	}

	public double getX_sd()
	{
		return getValue(Param.XSD);
	}

	public double getL_ab()
	{
		return getValue(Param.LAB);
	}

	public double getL_af()
	{
		return getValue(Param.LAF);
	}

	public double getL_an()
	{
		return getValue(Param.LAN);
	}

	public double getTm()
	{
		return getValue(Param.TM);
	}

	public double getQge()
	{
		return getValue(Param.QGE);
	}

	public double getQf()
	{
		return getValue(Param.QF);
	}

	public double getTc()
	{
		return getValue(Param.TC);
	}

	public double getQb()
	{
		return getValue(Param.QB);
	}

	public double getQgs()
	{
		return getValue(Param.QGS);
	}
}
