/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.behavior.IKick;
import magma.agent.behavior.SupportFoot;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementBehavior;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IThisPlayer;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * @author Stefan Glaser
 * 
 */
public abstract class Kick extends MovementBehavior implements IKick
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** The intended kick direction in the global system */
	protected Angle intendedKickDirection;

	/**
	 * The pose relative to the ball and intended kick direction, to which we
	 * should navigate in order to be able to perform this kick.<br>
	 * (The pose is interpreted in the global system)
	 */
	protected final Pose2D relativeRunToPose;

	/** the maximum kick distance this kick is able to kick */
	private double maxKickDistance;

	/** the area in which the ball has to be to be kickable */
	private final Vector3D[] shootPoly;

	/** get stopped on both legs */
	protected Movement standStill;

	/** move to stand on left leg only */
	protected Movement getOnLeftLeg;

	/** move to stand on right leg only */
	protected Movement getOnRightLeg;

	/** slow down power for managing kick speed */
	protected float kickPower;

	/**
	 * @param name the name of the kick
	 * @param thoughtModel Reference to the thought model object
	 * @param kickDirection angle in which the shoot will go relative to the body
	 *        direction
	 * @param ballFacingDirection direction we have to face relative to the ball
	 * @param maxBallDistance max distance to the ball to hit the ball
	 * @param maxKickDistance the maximum kick distance this kick is able to kick
	 * @param shootPoly polygon inside which the ball has to be to be kickable
	 * @param initCount the value to start counting with cnt
	 */
	public Kick(String name, IThoughtModel thoughtModel,
			Pose2D relativeRunToPose, double maxKickDistance, Vector3D[] shootPoly)
	{
		super(name, thoughtModel);
		this.relativeRunToPose = relativeRunToPose;
		this.maxKickDistance = maxKickDistance;
		this.shootPoly = shootPoly;

		standStill = new Movement("standStill");
		standStill.add(new MovementPhase("standStill", 50) //
				.add(INaoConstants.LShoulderPitch, -80, 3f) //
				.add(INaoConstants.RShoulderPitch, -80, 3f) //
				.add(INaoConstants.LHipYawPitch, 0, 0.5f) //
				.add(INaoConstants.RHipYawPitch, 0, 0.5f) //
				.add(INaoConstants.LHipRoll, 0, 0.5f) //
				.add(INaoConstants.LFootRoll, 0, 0.5f) //
				.add(INaoConstants.RHipRoll, 0, 0.5f) //
				.add(INaoConstants.RFootRoll, 0, 0.5f) //
				.add(INaoConstants.LHipPitch, 15, 0.5f) //
				.add(INaoConstants.LKneePitch, -30, 1f) //
				.add(INaoConstants.LFootPitch, 15, 0.5f) //
				.add(INaoConstants.RHipPitch, 15, 0.5f) //
				.add(INaoConstants.RKneePitch, -30, 1f) //
				.add(INaoConstants.RFootPitch, 15, 0.5f) //
				);

		getOnRightLeg = new Movement("getOnRightLeg");
		getOnRightLeg.add(new MovementPhase("getOnRightLeg", 10)
				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -40, 4f) //
				.add(INaoConstants.RFootPitch, 20, 2f) //
				.add(INaoConstants.LHipPitch, 0, 2f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //
				);

		getOnRightLeg.add(new MovementPhase("getOnRightLegSlowDown", 10) //
				.add(INaoConstants.RHipPitch, 20, 1f) //
				.add(INaoConstants.RKneePitch, -40, 2f) //
				.add(INaoConstants.RFootPitch, 20, 1f) //
				.add(INaoConstants.RFootRoll, -8, 1f) //
				.add(INaoConstants.RHipRoll, -8, 1f) //
				);
		getOnLeftLeg = getOnRightLeg.getLeftVersion();

		currentMovement = standStill;
	}

	/**
	 * @param kickPower the kickPower to set (factor 2 means half long)
	 */
	@Override
	public void setKickPower(float kickPower)
	{
		this.kickPower = kickPower;
	}

	@Override
	protected Movement getInitialMovement()
	{
		return standStill;
	}

	@Override
	public void setIntendedKickDirection(Angle intendedKickDirection)
	{
		this.intendedKickDirection = intendedKickDirection;
	}

	@Override
	public SupportFoot getKickingFoot()
	{
		return null;
	}

	@Override
	public Pose2D getRelativeRunToPose()
	{
		return relativeRunToPose;
	}

	/**
	 * @return the maximum kick distance this kick is able to kick
	 */
	@Override
	public double getMaxKickDistance()
	{
		return maxKickDistance;
	}

	@Override
	public float isBallKickable()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		Vector3D ballPos = worldModel.getBall().getPosition();

		if (thisPlayer.isInsidePolygonXY(ballPos, shootPoly)) {
			float absDevianveAngle = (float) Math.abs(intendedKickDirection
					.subtract(thisPlayer.getHorizontalAngle()).degrees());

			if (absDevianveAngle < 30) {
				return absDevianveAngle / 10;
			} else {
				return -1;
			}
		}

		return -1;
	}
}
