/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;
import magma.util.parameter.ParameterMap;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Kick ball straight forward. This Behavior assumes that stabilizing on one leg
 * was done already
 * 
 * @author Stefan Glaser
 */
public class KickSlomo extends Kick
{
	/**  */
	private static final long serialVersionUID = 1L;

	private static final double MAX_KICK_DISTANCE = 6;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.20;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.15;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/** used to check if the ball is still kickable */
	Vector3D[] internalShootPoly;

	/** movement to fast kick with right leg */
	protected Movement kickFast;

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickSlomo(String name, IThoughtModel thoughtModel,
			ParameterMap params, Pose2D relativeRunToPose, double maxKickDistance,
			Vector3D[] shootPoly)
	{
		super(name, thoughtModel, relativeRunToPose, maxKickDistance, shootPoly);

		internalShootPoly = Geometry
				.getPolygon(new double[][] { { 0, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, -MAX_Y_KICK - 0.04 },
						{ 0, -MAX_Y_KICK - 0.04 } });

		KickStraightOneLegBehaviorParameter param = (KickStraightOneLegBehaviorParameter) params
				.get(name);

		kickFast = new Movement("kickRightStraightFast");

		// // Move right leg back
		float speed = param.getRHYP2();
		kickFast.add(new MovementPhase("MoveLegBack", param.getTime1())
				.add(INaoConstants.LHipYawPitch, 0, speed) //
				.add(INaoConstants.LHipPitch, 21, speed) // lean in the front
				.add(INaoConstants.LKneePitch, 0, 2 * speed) //
				.add(INaoConstants.LFootPitch, 3, speed) //
				// .add(INaoConstants.LFootPitch, -12, 0.5f) //
				.add(INaoConstants.RHipYawPitch, 0, speed) //
				.add(INaoConstants.RHipPitch, -14, speed) //
				.add(INaoConstants.RKneePitch, -35, speed) //
				.add(INaoConstants.RFootPitch, 18, speed) //

				.add(INaoConstants.LHipRoll, 12.0f, speed) //
				.add(INaoConstants.LFootRoll, 12.0f * 0.7, speed) //

				.add(INaoConstants.RHipRoll, 10.0f, speed) //
				.add(INaoConstants.RFootRoll, 10.0f, speed) //
				);

		// Move right leg to the front as soon as possible, but not with full
		// speed at beginning
		kickFast.add(new MovementPhase("KickAcceleration0", 2)
				.add(INaoConstants.LHipPitch, -14, 1.0f) //
				.add(INaoConstants.RHipPitch, 21, 1.0f) //
				.add(INaoConstants.RKneePitch, 0, 1.0f) //
				.add(INaoConstants.RFootPitch, -14, 1.0f) //
				);

		kickFast.add(new MovementPhase("KickAcceleration1", 1)
				.add(INaoConstants.LHipPitch, -14, 2.0f) //
				.add(INaoConstants.RHipPitch, 21, 2.0f) //
				.add(INaoConstants.RKneePitch, 0, 2.0f) //
				.add(INaoConstants.RFootPitch, -14, 2.0f) //
				);

		kickFast.add(new MovementPhase("KickAcceleration2", 1)
				.add(INaoConstants.LHipPitch, -14, 4.0f) //
				// .add(INaoConstants.LKneePitch, -21, 2.0f) //
				// .add(INaoConstants.LFootPitch, 10, 1.0f) //
				.add(INaoConstants.RHipPitch, 21, 4.0f) //
				.add(INaoConstants.RKneePitch, 0, 4.0f) //
				.add(INaoConstants.RFootPitch, -14, 4.0f) //
				);
		kickFast.add(new MovementPhase("KickAcceleration3", 1)
				.add(INaoConstants.LHipPitch, -14, 6.0f) //
				.add(INaoConstants.LKneePitch, -21, 4.0f) //
				// .add(INaoConstants.LFootPitch, 10, 2.0f) //
				.add(INaoConstants.RHipPitch, 21, 6.0f) //
				.add(INaoConstants.RKneePitch, 0, 6.0f) //
				.add(INaoConstants.RFootPitch, -14, 6.0f) //

				// .add(INaoConstants.LHipYawPitch, -14, 2.0f) //
				// .add(INaoConstants.RHipYawPitch, -14, 2.0f) //
				);

		// Now full speed
		kickFast.add(new MovementPhase("KickTargetPointwithFullSpeed", 4)
				.add(INaoConstants.LHipPitch, -14, 7.0f) //
				.add(INaoConstants.LKneePitch, -14, 4.0f) //
				// .add(INaoConstants.LFootPitch, 10, 7.0f) //
				.add(INaoConstants.RHipPitch, 42, 7.0f) //
				.add(INaoConstants.RKneePitch, 0, 7.0f) //
				.add(INaoConstants.RFootPitch, -35, 7.0f) //

				.add(INaoConstants.LHipYawPitch, -14, 7.0f) //
				.add(INaoConstants.RHipYawPitch, -14, 7.0f) //
				);

		kickFast.add(new MovementPhase("KickDeceleration0", 1)
				.add(INaoConstants.LHipPitch, -14, 5.0f) //
				.add(INaoConstants.LKneePitch, -14, 2.0f) //
				// .add(INaoConstants.LFootPitch, 9, 3.0f) //
				.add(INaoConstants.RHipPitch, 35, 5.0f) //
				.add(INaoConstants.RKneePitch, 0, 5.0f) //
				.add(INaoConstants.RFootPitch, -28, 5.0f) //
				);
		kickFast.add(new MovementPhase("KickDeceleration1", 1)
				.add(INaoConstants.LHipPitch, -14, 3.0f) //
				.add(INaoConstants.LKneePitch, -14, 1.0f) //
				// .add(INaoConstants.LFootPitch, 9, 3.0f) //
				.add(INaoConstants.RHipPitch, 35, 3.0f) //
				.add(INaoConstants.RKneePitch, 0, 3.0f) //
				.add(INaoConstants.RFootPitch, -28, 3.0f) //
				);

		kickFast.add(new MovementPhase("moveToNormal1", 10)
				.add(INaoConstants.LHipPitch, 5, 2f) //
				.add(INaoConstants.LKneePitch, -14, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //

				.add(INaoConstants.RHipYawPitch, 0, 6) //
				.add(INaoConstants.RHipRoll, -20, 6) //

				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -48, 4f) //
				.add(INaoConstants.RFootPitch, 20, 4f) //
				);
		kickFast.add(new MovementPhase("moveToNormal2", 20)
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				);

		currentMovement = kickFast;
	}

	@Override
	protected Movement getInitialMovement()
	{
		return kickFast;
	}

	@Override
	public void perform()
	{
		MovementPhase currentPhase = kickFast.getCurrentPhase();
		if (!currentMovement.isFinished()) {
			if (currentPhase.getName().equals("KickTargetPointwithFullSpeed")) {
				currentPhase.setRelativeSpeed(kickPower);
			} else if (currentPhase.getName().equals("MoveLegBack")) {
			}
		}
		super.perform();
	}

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public static KickSlomo getRightKick(IThoughtModel thoughtModel,
			ParameterMap params)
	{
		KickSlomo result = new KickSlomo(IBehaviorConstants.SHOOT_SLOMO_RIGHT,
				thoughtModel, params, new Pose2D(-0.12, 0, Angle.ZERO),
				MAX_KICK_DISTANCE, SHOOT_POLY);
		return result;
	}

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public static KickSlomo getLeftKick(IThoughtModel thoughtModel,
			ParameterMap params)
	{
		KickSlomo result = new KickSlomo(IBehaviorConstants.SHOOT_SLOMO_LEFT,
				thoughtModel, params, new Pose2D(-0.12, 0, Angle.ZERO),
				MAX_KICK_DISTANCE, SHOOT_POLY);
		result.kickFast = result.kickFast.getLeftVersion();
		result.currentMovement = result.kickFast;
		return result;
	}
}
