/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.agentmodel.IBodyModel;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;
import magma.util.parameter.ParameterMap;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Kick ball straight forward. This Behavior first stabilizes and then decides
 * to kick either with the left or the right. It also decides if we should do a
 * fast-kick, if an opponent is close by.
 * 
 * @author Stefan Glaser
 */
public class KickStraight extends Kick
{
	/**  */
	private static final long serialVersionUID = 1L;

	/**  */
	private static final int CYCLES_FOR_KICK = 8;

	private static final double MAX_KICK_DISTANCE = 4.7;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.20;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.15;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/** used to check if the ball is still kickable */
	Vector3D[] internalShootPoly;

	/** movement to kick with right leg */
	protected Movement kickRight;

	/** movement to fast kick with right leg */
	protected Movement kickRightFast;

	/** movement to kick with left leg */
	protected Movement kickLeft;

	/** movement to fast kick with left leg */
	protected Movement kickLeftFast;

	private int cyclesLeft;

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickStraight(String name, IThoughtModel thoughtModel,
			ParameterMap params, Pose2D relativeRunToPose, double maxKickDistance,
			Vector3D[] shootPoly)
	{
		super(name, thoughtModel, relativeRunToPose, maxKickDistance, shootPoly);

		KickFastBehaviorParameter param = (KickFastBehaviorParameter) params
				.get(name);

		internalShootPoly = Geometry
				.getPolygon(new double[][] { { 0, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, -MAX_Y_KICK - 0.04 },
						{ 0, -MAX_Y_KICK - 0.04 } });

		kickRight = new Movement("kickRightStraight");
		kickRight.add(new MovementPhase("moveLegBackwards", 15) //
				.add(INaoConstants.LHipPitch, 10, 0.5f) //
				.add(INaoConstants.LKneePitch, -20, 1f) //
				.add(INaoConstants.LFootPitch, 10, 0.5f) //
				.add(INaoConstants.RHipPitch, -5, 2f) //
				.add(INaoConstants.RKneePitch, -10, 2f) //
				.add(INaoConstants.RFootPitch, 10, 4f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 16, 2f) //
				);
		kickRight.add(new MovementPhase("moveLegBackwards2", 20) //
				.add(INaoConstants.LHipPitch, 30, 1f) //
				.add(INaoConstants.LKneePitch, -20, 1f) //
				.add(INaoConstants.LFootPitch, 10, 0.5f) //
				.add(INaoConstants.RHipPitch, -5, 2f) //
				.add(INaoConstants.RKneePitch, -40, 2f) //
				.add(INaoConstants.RFootPitch, 20, 2f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 16, 2f) //
				);

		kickRight.add(new MovementPhase("kickRight", CYCLES_FOR_KICK) //
				.add(INaoConstants.LHipPitch, -20, 7f) //
				.add(INaoConstants.LFootPitch, 20, 5f) //
				// .add(INaoConstants.RHipPitch, 35, 7f) //
				// .add(INaoConstants.RKneePitch, 0, 7f) //
				// .add(INaoConstants.RFootPitch, -50, 7f) //
				);

		kickRight.add(new MovementPhase("kickRight2", 30) //
				.add(INaoConstants.LHipPitch, 0, 1f) //
				.add(INaoConstants.RHipPitch, 0, 2f) //
				.add(INaoConstants.RFootPitch, 0, 2f) //
				);
		kickRight.add(new MovementPhase("kickRight3", 20) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 0, 1f) //
				);

		kickLeft = kickRight.getLeftVersion();

		kickRightFast = new Movement("kickRightStraightFast");
		kickRightFast.add(new MovementPhase("getOnOneLegFast1", 3)
				.add(INaoConstants.RFootRoll, 8, 2f) //
				.add(INaoConstants.RHipRoll, -8, 2f) //
				.add(INaoConstants.LFootRoll, 8, 2f) //
				.add(INaoConstants.LHipRoll, -8, 2f) //

				.add(INaoConstants.LHipPitch, 0, 4.5f) //
				.add(INaoConstants.LKneePitch, 0, 7f) //
				.add(INaoConstants.LFootPitch, 0, 2.5f) //

				.add(INaoConstants.RHipPitch, 0, 3.5f) //
				.add(INaoConstants.RKneePitch, 0, 7f) //
				.add(INaoConstants.RFootPitch, 0, 3.5f) //
				);
		kickRightFast.add(new MovementPhase("getOnOneLegFast2", 4)
				.add(INaoConstants.RFootRoll, param.getRRoll(), 1.5f) //
				.add(INaoConstants.RHipRoll, -param.getRRoll(), 1.5f) //
				.add(INaoConstants.LFootRoll, param.getLRoll(), 1.5f) //
				.add(INaoConstants.LHipRoll, -param.getLRoll(), 1.5f) //

				.add(INaoConstants.LHipPitch, 0, 3.5f) //
				.add(INaoConstants.LKneePitch, 0, 7f) //
				.add(INaoConstants.LFootPitch, 0, 3.5f) //
				);
		kickRightFast.add(new MovementPhase("moveLegBackward1", param.getTime1()) //
				.add(INaoConstants.LHipPitch, param.getLHP1(), 1f) //
				.add(INaoConstants.RHipPitch, param.getRHP1(), 2f) //
				.add(INaoConstants.RKneePitch, param.getRKP1(), param.getRKPS1()) //
				.add(INaoConstants.RFootPitch, param.getRFP1(), 4f) //
				);
		kickRightFast.add(new MovementPhase("kickFast", param.getTime2()) //
				.add(INaoConstants.LHipPitch, param.getLHP2(), 7f) //
				.add(INaoConstants.LKneePitch, 0, 2f) //
				.add(INaoConstants.LFootPitch, 5, 2f) //

				.add(INaoConstants.RHipYawPitch, param.getRHYP2(), 7) //
				.add(INaoConstants.RHipRoll, param.getRHYP2() * -0.6, 3) //

				.add(INaoConstants.RHipPitch, param.getRHP2(), param.getRHPS2()) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 7f) //
				.add(INaoConstants.RFootPitch, param.getRFP2(), param.getRFPS2()) //
				);
		kickRightFast.add(new MovementPhase("moveToNormal1", 10)
				.add(INaoConstants.LHipPitch, 5, 2f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //

				.add(INaoConstants.RHipYawPitch, 0, 6) //
				.add(INaoConstants.RHipRoll, -20, 6) //

				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -48, 4f) //
				.add(INaoConstants.RFootPitch, 20, 4f) //
				);
		kickRightFast.add(new MovementPhase("moveToNormal2", 20)
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				);

		kickLeftFast = kickRightFast.getLeftVersion();
	}

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickStraight(IThoughtModel thoughtModel, ParameterMap params)
	{
		this(IBehaviorConstants.SHOOT_STRAIGHT, thoughtModel, params, new Pose2D(
				-0.12, 0, Angle.ZERO), MAX_KICK_DISTANCE, SHOOT_POLY);
	}

	@Override
	public void perform()
	{
		if (currentMovement == kickLeft) {
			if (currentMovement.getCurrentPhase().getName().equals("kickLeft")) {

				if (cyclesLeft > 0) {

					Vector3D leftFootPosition = agentModel.getBodyPart(
							INaoConstants.LFoot).getPosition();
					Vector3D leftFootStepPosition = getNextFootPosition(leftFootPosition);

					// double anglex = Geometry.getLinearFuzzyValue(0,
					// CYCLES_FOR_KICK,
					// true, cyclesLeft) * 15;
					Vector3D leftFootAngles = new Vector3D(0, -8, 0);

					IBodyModel bodyModel = agentModel.getFutureBodyModel();
					bodyModel.moveBodyPartToPose(INaoConstants.LFoot,
							leftFootStepPosition, leftFootAngles);
					cyclesLeft--;
					// return;
				} else {
					// go to the next movement phase
					currentMovement.proceed();
				}
			}
		}

		if (currentMovement == kickRight) {
			if (currentMovement.getCurrentPhase().getName().equals("kickRight")) {

				if (cyclesLeft > 0) {
					Vector3D rightFootPosition = agentModel.getBodyPart(
							INaoConstants.RFoot).getPosition();
					Vector3D rightFootStepPosition = getNextFootPosition(rightFootPosition);
					Vector3D rightFootAngles = new Vector3D(0, 8, 0);

					IBodyModel bodyModel = agentModel.getFutureBodyModel();
					bodyModel.moveBodyPartToPose(INaoConstants.RFoot,
							rightFootStepPosition, rightFootAngles);
					cyclesLeft--;
					// return;
				} else {
					// go to the next movement phase
					currentMovement.proceed();
				}
			}
		}

		super.perform();
	}

	@Override
	protected Movement getNextMovement()
	{
		if (currentMovement == standStill) {
			IThisPlayer thisPlayer = worldModel.getThisPlayer();
			Vector3D ballPos = worldModel.getBall().getPosition();
			if (thisPlayer.isInsidePolygonXY(ballPos, internalShootPoly)) {
				// ball is kickable by both feet
				boolean kickRight = true;
				if (thisPlayer.positionIsLeft(ballPos)) {
					kickRight = false;
				}
				IPlayer opponentAtBall = thoughtModel.getOpponentAtBall();
				boolean shootFast = false;
				if (opponentAtBall != null) {
					shootFast = opponentAtBall.getDistanceToXY(worldModel.getBall()) < 1
							&& Math.abs(worldModel.getThisPlayer()
									.getBodyDirectionTo(opponentAtBall.getPosition())
									.degrees()) < 100;
				}

				shootFast = true;
				if (shootFast) {
					if (kickRight) {
						return kickRightFast;
					} else {
						return kickLeftFast;
					}

				} else {
					if (kickRight) {
						return getOnLeftLeg;
					} else {
						return getOnRightLeg;
					}
				}

			} else {
				return null;
			}
		} else if (currentMovement == getOnLeftLeg) {
			cyclesLeft = CYCLES_FOR_KICK;
			return kickRight;
		} else if (currentMovement == getOnRightLeg) {
			cyclesLeft = CYCLES_FOR_KICK;
			return kickLeft;
		} else if (currentMovement == kickRight || currentMovement == kickLeft
				|| currentMovement == kickRightFast
				|| currentMovement == kickLeftFast) {
			return getReady;
		} else {
			return null;
		}
	}

	/**
	 * @param rightFootPosition
	 * @return
	 */
	private Vector3D getNextFootPosition(Vector3D rightFootPosition)
	{
		Vector3D ballPos = worldModel.getBall().getLocalPosition();
		Vector3D ballPosition = new Vector3D(-ballPos.getY(), ballPos.getX(),
				-0.35);

		Vector3D delta = ballPosition.subtract(rightFootPosition);
		delta = delta.add(delta.normalize().scalarMultiply(0.10));
		Vector3D rightFootStepPosition = rightFootPosition.add(delta
				.scalarMultiply(1.0f / cyclesLeft));
		// Vector3D leftFootStepPosition = leftPosition;
		// System.out.print("" + cyclesLeft);
		// System.out.print(" ball: " + ballPosition);
		// System.out.print(" delta: " + delta);
		// System.out.println(" foot: " + rightFootPosition);
		return rightFootStepPosition;
	}

}
