/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IThisPlayer;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;
import magma.util.parameter.ParameterMap;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Kick ball straight forward. This Behavior first stabilizes and then decides
 * to kick either with the left or the right.
 * 
 * @author Stefan Glaser
 */
public class KickStraight2 extends Kick
{
	/**  */
	private static final long serialVersionUID = 1L;

	private static final double MAX_KICK_DISTANCE = 8;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.23;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.15;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/** used to check if the ball is still kickable */
	Vector3D[] internalShootPoly;

	/** movement to fast kick with right leg */
	protected Movement kickRightFast;

	/** movement to fast kick with left leg */
	protected Movement kickLeftFast;

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickStraight2(String name, IThoughtModel thoughtModel,
			ParameterMap params, Pose2D relativeRunToPose, double maxKickDistance,
			Vector3D[] shootPoly)
	{
		super(name, thoughtModel, relativeRunToPose, maxKickDistance, shootPoly);

		KickFastBehaviorParameter param = (KickFastBehaviorParameter) params
				.get(name);

		internalShootPoly = Geometry
				.getPolygon(new double[][] { { 0, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, -MAX_Y_KICK - 0.04 },
						{ 0, -MAX_Y_KICK - 0.04 } });

		kickRightFast = new Movement("kickRightStraightFast");
		kickRightFast.add(new MovementPhase("getOnOneLegFast1", 3)
				.add(INaoConstants.RFootRoll, 8, 2f) //
				.add(INaoConstants.RHipRoll, -8, 2f) //
				.add(INaoConstants.LFootRoll, 8, 2f) //
				.add(INaoConstants.LHipRoll, -8, 2f) //

				.add(INaoConstants.LHipPitch, 0, 4.5f) //
				.add(INaoConstants.LKneePitch, 0, 7f) //
				.add(INaoConstants.LFootPitch, 0, 2.5f) //

				.add(INaoConstants.RHipPitch, 0, 3.5f) //
				.add(INaoConstants.RKneePitch, 0, 7f) //
				.add(INaoConstants.RFootPitch, 0, 3.5f) //
				);
		kickRightFast.add(new MovementPhase("getOnOneLegFast2", 9)
				.add(INaoConstants.RFootRoll, param.getRRoll(), 1.5f) //
				.add(INaoConstants.RHipRoll, -param.getRRoll(), 1.5f) //
				.add(INaoConstants.LFootRoll, param.getLRoll() * 0.7, 1.5f) //
				.add(INaoConstants.LHipRoll, param.getLRoll(), 1.5f) //
				.add(INaoConstants.LHipPitch, 0, 3.5f) //
				.add(INaoConstants.LKneePitch, 0, 7f) //
				.add(INaoConstants.LFootPitch, 0, 3.5f) //
				);
		kickRightFast.add(new MovementPhase("moveLegBackward0", 10)
				.add(INaoConstants.RHipPitch, -10, 1f) //
				.add(INaoConstants.RKneePitch, -20, 2) //
				.add(INaoConstants.RFootPitch, 30, 3f) //
				);
		kickRightFast.add(new MovementPhase("moveLegBackward1", 20) //
				.add(INaoConstants.RHipPitch, param.getRHP1(), 4f) //
				.add(INaoConstants.RKneePitch, param.getRKP1(), param.getRKPS1()) //
				.add(INaoConstants.RFootPitch, param.getRFP1(), 4f) //
				);
		kickRightFast.add(new MovementPhase("kickFast", 2) //
				.add(INaoConstants.RHipPitch, param.getRHP2(), 3) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 5f) //
				);
		kickRightFast.add(new MovementPhase("kickFast2", param.getTime1()) //
				.add(INaoConstants.RHipPitch, param.getRHP2(), 7.03f) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 7.03f) //
				.add(INaoConstants.RFootPitch, param.getRFP2(), param.getRFPS2()));

		kickRightFast.add(new MovementPhase("ballHit", param.getTime2()) //
				.add(INaoConstants.LHipPitch, param.getLHP2(), 7.03f) //
				.add(INaoConstants.LFootPitch, 5, 2f) //
				.add(INaoConstants.RHipPitch, param.getRHP2(), 7.03f) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 7.03f) //
				.add(INaoConstants.RFootPitch, param.getRFP2(), 7.03f) //
				.add(INaoConstants.RHipYawPitch, param.getRHYP2(), 4) //
				);

		kickRightFast.add(new MovementPhase("moveToNormal1", 10)
				.add(INaoConstants.LHipPitch, 5, 2f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //

				.add(INaoConstants.RHipYawPitch, 0, 6) //
				.add(INaoConstants.RHipRoll, -20, 6) //

				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -48, 4f) //
				.add(INaoConstants.RFootPitch, 20, 4f) //
				);
		kickRightFast.add(new MovementPhase("moveToNormal2", 20)
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				);

		kickLeftFast = kickRightFast.getLeftVersion();
	}

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickStraight2(IThoughtModel thoughtModel, ParameterMap params)
	{
		this(IBehaviorConstants.SHOOT_STRAIGHT, thoughtModel, params, new Pose2D(
				-0.12, 0, Angle.ZERO), MAX_KICK_DISTANCE, SHOOT_POLY);
	}

	@Override
	protected Movement getNextMovement()
	{
		IMoveableObject ball = worldModel.getBall();
		if (currentMovement == standStill) {
			IThisPlayer thisPlayer = worldModel.getThisPlayer();
			Vector3D ballPos = ball.getPosition();
			if (thisPlayer.isInsidePolygonXY(ballPos, internalShootPoly)) {
				// ball is kickable by both feet
				boolean kickRight = true;
				if (thisPlayer.positionIsLeft(ballPos)) {
					kickRight = false;
				}
				if (kickRight) {
					return kickRightFast;
				} else {
					return kickLeftFast;
				}

				// System.out.println("see ball at: " + ball.getLocalPosition());

			} else {
				return null;
			}
		} else if (currentMovement == kickRightFast
				|| currentMovement == kickLeftFast) {
			return getReady;
		} else {
			return null;
		}
	}
}
