/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.Pose2D;
import magma.util.parameter.ParameterMap;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Kick ball straight forward. This Behavior assumes that stabilizing on one leg
 * was done already
 * 
 * @author Stefan Glaser
 */
public class KickStraightOneLegRight extends Kick
{
	/**  */
	private static final long serialVersionUID = 1L;

	protected static final double MAX_KICK_DISTANCE = 6;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.20;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.15;

	protected static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/** used to check if the ball is still kickable */
	Vector3D[] internalShootPoly;

	/** movement to fast kick with right leg */
	protected Movement kickFast;

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickStraightOneLegRight(String name, IThoughtModel thoughtModel,
			ParameterMap params, Pose2D relativeRunToPose, double maxKickDistance,
			Vector3D[] shootPoly)
	{
		super(name, thoughtModel, relativeRunToPose, maxKickDistance, shootPoly);

		KickStraightOneLegBehaviorParameter param = (KickStraightOneLegBehaviorParameter) params
				.get(name);

		internalShootPoly = Geometry
				.getPolygon(new double[][] { { 0, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, -MAX_Y_KICK - 0.04 },
						{ 0, -MAX_Y_KICK - 0.04 } });

		kickFast = new Movement("kickRightStraightFast");

		kickFast.add(new MovementPhase("getSupportLegStraight", 10)
				.add(INaoConstants.RFootRoll, param.getRRoll(), 0.8f) //
				.add(INaoConstants.RHipRoll, -param.getRRoll(), 0.8f) //
				.add(INaoConstants.RHipPitch, param.getRHP1(), 2f) //
				.add(INaoConstants.RKneePitch, param.getRKP1(), 3) //
				.add(INaoConstants.RFootPitch, param.getRFP1(), 1.5f) //
				.add(INaoConstants.RHipYawPitch, 0, 1) //

				.add(INaoConstants.LFootRoll, param.getLRoll() * 0.7, 0.5f) //
				.add(INaoConstants.LHipRoll, param.getLRoll(), 0.5f) //
				.add(INaoConstants.LHipPitch, 0, 1.5f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2.4f) //
				.add(INaoConstants.LHipYawPitch, 0, 1) //
				);

		kickFast.add(new MovementPhase("kickFast", param.getTime1()) //
				.add(INaoConstants.RHipPitch, param.getRHP2(), 3) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 5f) //
				);
		kickFast.add(new MovementPhase("kickFast2", param.getTime2()) //
				.add(INaoConstants.RHipPitch, param.getRHP2(), 7.03f) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 7.03f) //
				.add(INaoConstants.RFootPitch, param.getRFP2(), param.getRFPS2()));

		kickFast.add(new MovementPhase("ballHit", param.getTime3()) //
				.add(INaoConstants.LHipPitch, param.getLHP2(), 7.03f) //
				.add(INaoConstants.LFootPitch, 5, 2f) //
				.add(INaoConstants.RHipPitch, param.getRHP2(), 7.03f) //
				.add(INaoConstants.RKneePitch, param.getRKP2(), 7.03f) //
				.add(INaoConstants.RFootPitch, param.getRFP2(), 7.03f) //
				.add(INaoConstants.RHipYawPitch, param.getRHYP2(),
						param.getRHYPS2()) //
				);

		kickFast.add(new MovementPhase("moveToNormal1", 10)
				.add(INaoConstants.LHipPitch, 5, 2f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //

				.add(INaoConstants.RHipYawPitch, 0, 6) //
				.add(INaoConstants.RHipRoll, -20, 6) //

				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -48, 4f) //
				.add(INaoConstants.RFootPitch, 20, 4f) //
				);
		kickFast.add(new MovementPhase("moveToNormal2", 20)
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				);

		currentMovement = kickFast;
	}

	@Override
	protected Movement getInitialMovement()
	{
		return kickFast;
	}

	/**
	 * Default constructor
	 * 
	 * @param thoughtModel Reference to the thought model object
	 */
	public KickStraightOneLegRight(IThoughtModel thoughtModel,
			ParameterMap params)
	{
		this(IBehaviorConstants.SHOOT_STRAIGHT_ONE_LEG_RIGHT, thoughtModel,
				params, new Pose2D(-0.12, 0, Angle.ZERO), MAX_KICK_DISTANCE,
				SHOOT_POLY);
	}

	@Override
	public void perform()
	{
		// System.out.println(agentModel.getHJ(INaoConstants.LHipPitch).getAxis()
		// + ";" + agentModel.getHJ(INaoConstants.LKneePitch).getAxis() + ";"
		// + agentModel.getHJ(INaoConstants.LFootPitch).getAxis());
		if (!currentMovement.isFinished()) {
			MovementPhase currentPhase = kickFast.getCurrentPhase();
			if (currentPhase.getName().equals("ballHit")) {
				currentPhase.setRelativeSpeed(kickPower);
			}
		}
		super.perform();
	}
}
