/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.general;

import java.util.Map;

import magma.agent.agentmodel.ik.IAgentIKSolver;
import magma.agent.behavior.IBaseWalk;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.complex.GetInScorePosition;
import magma.agent.behavior.complex.GoaliePositioning;
import magma.agent.behavior.complex.Keep;
import magma.agent.behavior.complex.MoveWithBall;
import magma.agent.behavior.complex.RunToPosition;
import magma.agent.behavior.complex.SearchBall;
import magma.agent.behavior.complex.ShootToGoal;
import magma.agent.behavior.complex.WalkForward;
import magma.agent.behavior.complex.WalkSideward;
import magma.agent.behavior.complex.WalkTurning;
import magma.agent.belief.IBelief;
import magma.agent.belief.impl.BallInScorableArea;
import magma.agent.belief.impl.BallIsKeepable;
import magma.agent.belief.impl.ResourceBelieve;
import magma.agent.communication.IChannel;
import magma.agent.communication.IChannelManager;
import magma.agent.communication.IInputChannel;
import magma.agent.communication.IOutputChannel;
import magma.agent.communication.base.ChannelManager;
import magma.agent.communication.base.ChannelParameters;
import magma.agent.general.ComponentFactory;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.thoughtmodel.IThoughtModel;
import magma.robots.nao.INaoConstants;
import magma.robots.nao.agent.communication.base.SimsparkChannel;
import magma.robots.nao.agentmodel.ik.impl.NAOLegCalculator;
import magma.robots.nao.behavior.basic.Cry;
import magma.robots.nao.behavior.basic.Dance;
import magma.robots.nao.behavior.complex.DynamicStaticKick;
import magma.robots.nao.behavior.complex.Walk;
import magma.robots.nao.behavior.dynamic.FocusBall;
import magma.robots.nao.behavior.dynamic.walk.Balance;
import magma.robots.nao.behavior.dynamic.walk.ForwardWalkBehaviorParameter;
import magma.robots.nao.behavior.movement.FallBack;
import magma.robots.nao.behavior.movement.FallForward;
import magma.robots.nao.behavior.movement.KeeperLeft;
import magma.robots.nao.behavior.movement.KeeperRight;
import magma.robots.nao.behavior.movement.KickBack;
import magma.robots.nao.behavior.movement.KickBackPull;
import magma.robots.nao.behavior.movement.KickChipLeft;
import magma.robots.nao.behavior.movement.KickChipRight;
import magma.robots.nao.behavior.movement.KickChipRightBehaviorParameters;
import magma.robots.nao.behavior.movement.KickFastBehaviorParameter;
import magma.robots.nao.behavior.movement.KickSlomo;
import magma.robots.nao.behavior.movement.KickSlomoBehaviorParameter;
import magma.robots.nao.behavior.movement.KickStraight2;
import magma.robots.nao.behavior.movement.KickStraightOneLegBehaviorParameter;
import magma.robots.nao.behavior.movement.KickStraightOneLegLeft;
import magma.robots.nao.behavior.movement.KickStraightOneLegRight;
import magma.robots.nao.behavior.movement.MoveArmsToFall;
import magma.robots.nao.behavior.movement.Push;
import magma.robots.nao.behavior.movement.SideKickLeft;
import magma.robots.nao.behavior.movement.SideKickLeftOneLeg;
import magma.robots.nao.behavior.movement.SideKickRight;
import magma.robots.nao.behavior.movement.SideKickRightOneLeg;
import magma.robots.nao.belief.impl.SoccerPosition;
import magma.robots.nao.meta.agent.NaoAgentMetaModel;
import magma.util.parameter.ParameterMap;

public class NAOComponentFactory extends ComponentFactory
{
	private static final int GOALPREDICTION_TIME = 60;

	/**
	 * Create a agent meta model
	 * 
	 * @return the agent meta model
	 */
	@Override
	public IAgentMetaModel createAgentMetaModel()
	{
		return NaoAgentMetaModel.INSTANCE;
	}

	@Override
	public IChannelManager createChannelManager(ChannelParameters info)
	{
		IChannelManager channelManager = new ChannelManager();

		// This can be used to record perceptions or to run on recorded
		// File filePercept = new File("log/perceptions.log");
		// File fileAct = new File("log/actions1.log");
		// IChannel channel = new SimsparkLogfileWriterChannel(channelManager,
		// info,
		// filePercept, fileAct);
		// IChannel channel = new SimsparkLogfileReaderChannel(channelManager,
		// info,
		// filePercept, fileAct);

		IChannel channel = new SimsparkChannel(channelManager, info);
		channel.init(createAgentMetaModel());
		channelManager.addInputChannel((IInputChannel) channel, true);
		channelManager.addOutputChannel((IOutputChannel) channel);
		return channelManager;
	}

	@Override
	public IAgentIKSolver createAgentIKSolver()
	{
		return new NAOLegCalculator();
	}

	@Override
	protected void createSpecificBeliefs(IThoughtModel thoughtModel,
			Map<String, IBelief> beliefs)
	{
		IBelief belief = new ResourceBelieve("resourceBody", 1);
		beliefs.put(belief.getName(), belief);

		belief = new BallIsKeepable(thoughtModel, INaoConstants.PLAYER_HEIGHT,
				GOALPREDICTION_TIME);
		beliefs.put(belief.getName(), belief);

		belief = new SoccerPosition(thoughtModel, (float) Math.toDegrees(-1.05),
				-1000);
		beliefs.put(belief.getName(), belief);

		belief = new BallInScorableArea(thoughtModel,
				INaoConstants.KICKABLE_DISTANCE);
		beliefs.put(belief.getName(), belief);
	}

	@Override
	protected String getBehaviorFilesBasePath()
	{
		return "behaviors/nao/";
	}

	@Override
	protected void createSpecificBehaviors(IThoughtModel thoughtModel,
			ParameterMap params, Map<String, IBehavior> behaviors)
	{
		// General behaviors
		IBehavior behavior = new FocusBall(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new MoveArmsToFall(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new FallBack(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new FallForward(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		// Movement behaviors
		behavior = new RunToPosition(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		IBaseWalk walkBehavior = createWalk(thoughtModel, params, behaviors);
		behavior = new Walk(thoughtModel, params, behaviors, walkBehavior);
		behaviors.put(behavior.getName(), behavior);

		behavior = new WalkForward(thoughtModel, params, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new WalkSideward(thoughtModel, params, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new WalkTurning(thoughtModel, params, behaviors);
		behaviors.put(behavior.getName(), behavior);

		// Positioning behaviors
		behavior = new GetInScorePosition(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new MoveWithBall(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new GoaliePositioning(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		// Kick behaviors
		behavior = new ShootToGoal(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickStraight2(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickStraightOneLegRight(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickStraightOneLegLeft(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SideKickRightOneLeg(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SideKickLeftOneLeg(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = KickSlomo.getRightKick(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = KickSlomo.getLeftKick(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickBack(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickBackPull(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SideKickLeft(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SideKickRight(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickChipRight(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickChipLeft(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		// Keeper behavior
		behavior = new KeeperLeft(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KeeperRight(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new Keep(thoughtModel, behaviors, INaoConstants.PLAYER_HEIGHT,
				GOALPREDICTION_TIME, behaviors.get(IBehaviorConstants.KEEP_LEFT),
				behaviors.get(IBehaviorConstants.KEEP_RIGHT));
		behaviors.put(behavior.getName(), behavior);

		// Game result reflection behaviors
		behavior = new Cry(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new Dance(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = DynamicStaticKick.getForwardKickLeft(thoughtModel, params,
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = DynamicStaticKick.getForwardKickRight(thoughtModel, params,
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = DynamicStaticKick.getSloMoKickLeft(thoughtModel, params,
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = DynamicStaticKick.getSloMoKickRight(thoughtModel, params,
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = DynamicStaticKick.getSidewardKickLeft(thoughtModel, params,
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = DynamicStaticKick.getSidewardKickRight(thoughtModel, params,
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		// Experimental behaviors
		behavior = new Push(thoughtModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SearchBall(thoughtModel, behaviors);
		behaviors.put(behavior.getName(), behavior);
	}

	protected IBaseWalk createWalk(IThoughtModel thoughtModel,
			ParameterMap params, Map<String, IBehavior> behaviors)
	{
		// 2011 walk
		IBaseWalk behavior = new Balance(thoughtModel, params);
		behaviors.put(behavior.getName(), behavior);

		return behavior;
	}

	@Override
	protected ParameterMap createSpecificParameters()
	{
		ParameterMap result = new ParameterMap();

		result.put(IBehaviorConstants.WALK,
				ForwardWalkBehaviorParameter.getInstance());

		result.put(IBehaviorConstants.SHOOT_STRAIGHT,
				KickFastBehaviorParameter.getInstance());
		result.put(IBehaviorConstants.SHOOT_BACK,
				KickFastBehaviorParameter.getInstance());
		result.put(IBehaviorConstants.SHOOT_BACK_PULL,
				KickFastBehaviorParameter.getInstance());

		result.put(IBehaviorConstants.SHOOT_STRAIGHT_ONE_LEG_LEFT,
				KickStraightOneLegBehaviorParameter.getInstance());
		result.put(IBehaviorConstants.SHOOT_STRAIGHT_ONE_LEG_RIGHT,
				KickStraightOneLegBehaviorParameter.getInstance());

		result.put(IBehaviorConstants.SHOOT_SLOMO_LEFT,
				KickSlomoBehaviorParameter.getInstance());
		result.put(IBehaviorConstants.SHOOT_SLOMO_RIGHT,
				KickSlomoBehaviorParameter.getInstance());

		result.put(IBehaviorConstants.KICK_CHIP_RIGHT,
				KickChipRightBehaviorParameters.getInstance());
		result.put(IBehaviorConstants.KICK_CHIP_LEFT,
				KickChipRightBehaviorParameters.getInstance());

		return result;
	}
}
