/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.common.spark;

/**
 * Play modes enumeration. Every enum stores the corresponding message used in
 * the server protocol.
 * 
 * @author Stefan Glaser
 */
public enum PlayMode {
	/** Humanoid Initial State before the game */
	INITIAL("Initial"),

	/** The game has not started yet, Set in humanoid games */
	BEFORE_KICK_OFF("BeforeKickOff"),

	/**
	 * The ball is in the center, and the team on the left side is allowed to
	 * kick it first
	 */
	KICK_OFF_LEFT("KickOff_Left"),

	/**
	 * The ball is in the center, and the team on the right side is allowed to
	 * kick it first
	 */
	KICK_OFF_RIGHT("KickOff_Right"),

	/**
	 * The game is procesing normally, no special rules are in place
	 */
	PLAY_ON("PlayOn"),

	/**
	 * The ball left the playfield, the team on the left is allowed to kick it
	 * back into the field
	 */
	KICK_IN_LEFT("KickIn_Left"),

	/**
	 * The ball left the playfield, the team on the right is allowed to kick it
	 * back into the field
	 */
	KICK_IN_RIGHT("KickIn_Right"),

	/**
	 * The ball left the playfield, the team on the left is allowed to kick it
	 * back into the field from a corner
	 */
	CORNER_KICK_LEFT("CORNER_KICK_LEFT"),

	/**
	 * The ball left the playfield, the team on the right is allowed to kick it
	 * back into the field from a corner
	 */
	CORNER_KICK_RIGHT("CORNER_KICK_RIGHT"),

	/**
	 * The ball left the playfield while it was near the left goal, and the left
	 * team is allowed to kick it first
	 */
	GOAL_KICK_LEFT("goal_kick_left"),

	/**
	 * The ball left the playfield while it was near the right goal, and the
	 * right team is allowed to kick it first
	 */
	GOAL_KICK_RIGHT("goal_kick_right"),

	/**
	 * The team on the right violated the offside rule and the team on the left
	 * is allowed to kick the ball first
	 */
	OFFSIDE_LEFT("OFFSIDE_LEFT"),

	/**
	 * The team on the left violated the offside rule and the team on the right
	 * is allowed to kick the ball first
	 */
	OFFSIDE_RIGHT("OFFSIDE_RIGHT"),

	/**
	 * The game is over
	 */
	GAME_OVER("GameOver"),

	/**
	 * A goal was counted for the team on the left
	 */
	GOAL_LEFT("Goal_Left"),

	/**
	 * A goal was counted for the team on the right
	 */
	GOAL_RIGHT("Goal_Right"),

	/**
	 * The left team got a free kick (probably because the referee decided to do
	 * so) and may kick the ball first
	 */
	FREE_KICK_LEFT("FREE_KICK_LEFT"),

	/**
	 * The right team got a free kick (probably because the referee decided to do
	 * so) and may kick the ball first
	 */
	FREE_KICK_RIGHT("FREE_KICK_RIGHT"),

	/**
	 * No playmode or an unknown one (should not be used!)
	 */
	NONE("NONE");

	private final String serverMsg;

	private PlayMode(String serverMsg)
	{
		this.serverMsg = serverMsg;
	}

	/**
	 * Convert a play mode into an enum
	 * 
	 * @param playModeString Play mode string
	 * @return Resulting enum
	 */
	public static PlayMode parsePlayMode(String playModeString)
	{
		for (PlayMode mode : values()) {
			if (playModeString.equalsIgnoreCase(mode.serverMsg))
				return mode;
		}
		return NONE;
	}

	/**
	 * Convert a PlayMode to a string.
	 * 
	 * @return The playmode string that will be accepted by the server.
	 */
	public String getServerString()
	{
		return serverMsg;
	}
}
