/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package magma.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Field;

import magma.util.file.StreamConsumer;

/**
 * A collection of command line utilities for direct unix commandxs.
 * @author kdorer
 */
public class UnixCommandUtil
{
	/**
	 * Retrieves the pid of a unix process. (Possibly in Java 1.8 this will be
	 * added to the API of process??).
	 * @param process the process object to get the pid from
	 * @return the pid, -1 if not able to get it
	 * @throws NoSuchFieldException
	 * @throws SecurityException
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 */
	public static int getPID(Process process) throws NoSuchFieldException,
			SecurityException, IllegalArgumentException, IllegalAccessException
	{
		if (process.getClass().getName().equals("java.lang.UNIXProcess")) {
			/* get the PID on unix/linux systems */
			Field f = process.getClass().getDeclaredField("pid");
			f.setAccessible(true);
			return f.getInt(process);
		} else {
			return -1;
		}
	}

	/**
	 * Checks if the process with pid passed contains the passed name when run
	 * with ps. If so it is killed.
	 * @param pid the pid of the process to kill.
	 * @param name the name that has to appear when calling ps
	 * @return true if the process was killed.
	 */
	public static boolean killProcessConditional(String pid, String name)
	{
		String command = "ps " + pid;
		Process ps;
		try {
			ps = Runtime.getRuntime().exec(command);
			if (UnixCommandUtil.checkForOutput(ps, name)) {
				Runtime.getRuntime().exec("kill -9 " + pid);
				return true;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		return false;
	}

	/**
	 * Kills all processes with a given name. Use with care!
	 * @param name the name that has to appear when calling ps
	 */
	public static boolean killAll(String name)
	{
		try {
			Runtime.getRuntime().exec("killall -9 " + name);
			return true;

		} catch (IOException e) {
			e.printStackTrace();
		}
		return false;
	}

	/**
	 * Checks if the passed string is part of the output of the passed process.
	 * @param ps the process to get output from
	 * @param text the text that we expect to be part of the output
	 */
	public static boolean checkForOutput(Process ps, String text)
	{
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				ps.getInputStream()));
		String line;
		boolean found = false;
		try {
			line = reader.readLine();
			while (line != null) {
				if (line.contains(text)) {
					found = true;
					break;
				}
				line = reader.readLine();
			}
			reader.close();
			return found;

		} catch (IOException e) {
			e.printStackTrace();
		}
		return false;
	}

	public static Process launch(String command, String[] params, File path)
	{
		try {
			Process ps = Runtime.getRuntime().exec(command, params, path);
			return ps;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	public static Process launchAndConsume(String command, String[] params,
			File path, boolean print)
	{
		try {
			Process ps = Runtime.getRuntime().exec(command, params, path);
			new StreamConsumer(ps.getInputStream(), print);
			new StreamConsumer(ps.getErrorStream(), print);
			return ps;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	public static void launchAndConsume(String command, boolean print)
	{
		try {
			Process ps = Runtime.getRuntime().exec(command);
			new StreamConsumer(ps.getInputStream(), print);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
