/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.connection.impl;

import java.io.BufferedInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.ConnectException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.logging.Level;

import magma.util.connection.ConnectionException;

/**
 * Server connection implementation
 * 
 * @author Klaus Dorer
 */
public class ServerConnection extends SocketConnection
{
	// Network socket
	private Socket socket;

	// Outgoing data stream
	private DataOutputStream out;

	// Incoming data stream
	private InputStream in;

	/** the time when receive started (in ns) */
	private long startTime;

	private long endTime;

	/**
	 * Default constructor
	 */
	public ServerConnection()
	{
		this(SERVER_IP, SERVER_PORT);
	}

	/**
	 * Assignment constructor
	 * 
	 * @param host Host address
	 * @param port Port number
	 */
	public ServerConnection(String host, int port)
	{
		super(host, port);
	}

	/**
	 * Actually connect to the server
	 */
	@Override
	public void establishConnection() throws ConnectionException
	{
		try {
			socket = new Socket(host, port);
			socket.setTcpNoDelay(true);

			in = new BufferedInputStream(socket.getInputStream());
			out = new DataOutputStream(socket.getOutputStream());

			// System.out.println("Connection to: " + host + ":" + port);

			connected = true;
		} catch (UnknownHostException e) {
			throw new ConnectionException(e.getMessage(), e);
		} catch (ConnectException e) {
			String message = host
					+ ":"
					+ port
					+ " refused the connection. Is rcssserver3d running? Are you using an IPv6-enabled"
					+ " system and the host name translates to an IPv6 address?";

			throw new ConnectionException(message, e);
		} catch (IOException e) {
			e.printStackTrace();
			throw new ConnectionException("IO problems", e);
		}
	}

	/**
	 * Send a message to the server
	 * 
	 * @param msg Message in ASCII form
	 */
	@Override
	public void sendMessage(byte[] msg) throws ConnectionException
	{
		logger.log(Level.FINEST, "sending: {0}", msg);
		byte[] body = msg;

		// System.out.println(new String(msg));

		// creation of the messages header (4 bytes)
		int len = body.length;
		int byte0 = (len >> 24) & 0xFF;
		int byte1 = (len >> 16) & 0xFF;
		int byte2 = (len >> 8) & 0xFF;
		int byte3 = len & 0xFF;

		try {
			out.writeByte((byte) byte0);
			out.writeByte((byte) byte1);
			out.writeByte((byte) byte2);
			out.writeByte((byte) byte3);
			out.write(body);
			out.flush();
		} catch (IOException e) {
			shutDown = true;
			throw new ConnectionException(
					"Error writing to socket, shuting down...", e);
		}

		// System.out.println("Sent: " + msg);
	}

	/**
	 * Starts the endless loop of receiving messages and notifying observers
	 */
	@Override
	public void startReceiveLoop() throws ConnectionException
	{
		try {
			while (!shutDown) {
				byte[] currentMessage = receiveMessage();

				// debug information
				// long delta = startTime - endTime;
				// if (delta > 25000000) {
				// System.out.println("late Server Message: " + delta);
				// }
				// if (delta < 5000000) {
				// System.out.println("early Server Message: " + delta);
				// }

				logger.log(Level.FINER, "Received message from server");
				logger.log(Level.FINEST, "message: {0}", currentMessage);

				// trigger main update cycle
				observer.onStateChange(currentMessage);

				// debug information
				endTime = System.nanoTime();
				// if (endTime - startTime > 20000000) {
				// System.out.println("slow processing: " + (endTime - startTime));
				// }
				logger.log(Level.FINE, "Message processed in (ns): "
						+ (endTime - startTime));
			}
		} finally {
			closeConnection();
			shutDown = false;
		}
	}

	/*
	 * Receive server message
	 */
	protected byte[] receiveMessage() throws ConnectionException
	{
		byte[] result;
		int length;

		try {
			int byte0 = in.read();

			startTime = System.nanoTime();

			int byte1 = in.read();
			int byte2 = in.read();
			int byte3 = in.read();
			length = byte0 << 24 | byte1 << 16 | byte2 << 8 | byte3; // analyzes
			// the header
			int total = 0;

			if (length < 0) {
				// server was shutdown
				shutDown = true;
				throw new ConnectionException("Server shut down");
			}

			result = new byte[length];
			int count = 0;
			while (total < length) {
				total += in.read(result, total, length - total);
				count++;
			}
			if (count > 1) {
				logger.log(Level.FINE, "Received in multiple pieces: " + count);
			}

			// System.out.println(new String(result));

		} catch (IOException e) {
			shutDown = true;
			throw new ConnectionException(
					"Error when reading from socket, closing down...", e);
		}
		return result;
	}

	/*
	 * Close server connection
	 */
	public void closeConnection()
	{
		try {
			connected = false;
			in.close();
			out.close();
			socket.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		// System.out.println("Connection disconnected");
	}
}
