/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.connection.impl;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.util.logging.Level;

import magma.util.connection.ConnectionException;

/**
 * UDP socket connection implementation
 * 
 * @author Klaus Dorer
 */
public class UDPConnection extends SocketConnection
{
	/** the socket we listen to */
	private DatagramSocket datagramSocket;

	private int bufferSize;

	private IByteBufferReceiver receiver;

	private InetAddress address;

	/**
	 * Default
	 */
	public UDPConnection(String host, int port, int bufferSize,
			IByteBufferReceiver receiver)
	{
		super(host, port);
		this.bufferSize = bufferSize;
		this.receiver = receiver;
	}

	public void establishSending() throws ConnectionException
	{
		try {
			datagramSocket = new DatagramSocket(null);
			address = InetAddress.getByName(host);
			connected = true;
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Error on start port for sending " + port);
		}
	}

	@Override
	public void establishConnection() throws ConnectionException
	{
		try {
			datagramSocket = new DatagramSocket(null);
			datagramSocket.setReuseAddress(true);
			datagramSocket.setBroadcast(true);
			datagramSocket.setSoTimeout(500);
			datagramSocket.bind(new InetSocketAddress(port));
			connected = true;
		} catch (SocketException e) {
			logger.log(Level.SEVERE, "Error on start listening to port " + port);
		}
	}

	@Override
	public void sendMessage(byte[] msg) throws ConnectionException
	{
		try {
			DatagramPacket packet = new DatagramPacket(msg, msg.length, address,
					port);
			datagramSocket.send(packet);
		} catch (IOException e) {
			System.out.println("Error while sending");
			e.printStackTrace();
		}
	}

	@Override
	public void startReceiveLoop() throws ConnectionException
	{
		try {
			while (!shutDown) {
				final ByteBuffer buffer = ByteBuffer.wrap(new byte[bufferSize]);

				final DatagramPacket packet = new DatagramPacket(buffer.array(),
						buffer.array().length);

				try {
					datagramSocket.receive(packet);
					buffer.rewind();
					receiver.update(buffer);

				} catch (SocketTimeoutException e) {
					// ignore, because we set a timeout
				} catch (IOException e) {
					logger.log(Level.SEVERE, "Error while listening to port " + port);
				}
			}
		} finally {
			closeConnection();
			shutDown = false;
		}
	}

	public void setReceiver(IByteBufferReceiver receiver)
	{
		this.receiver = receiver;
	}

	/*
	 * Close server connection
	 */
	public void closeConnection()
	{
		connected = false;
		datagramSocket.close();
	}
}
