/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.util.function;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Factory for creating motor file functions. <br>
 * <br>
 * A function can be stored as a line in a file, in the following format:<br>
 * <code>&lt;joint-name&gt;,&lt;function-type&gt;,&lt;support-point-count&gt;,
 * &lt;phase&gt;,[&lt;support- point&gt;]{ 2...*}</code><br>
 * where:<br>
 * <code>&lt;joint-name&gt;</code>: String: Name of the joint<br>
 * <code>&lt;function-type&gt;</code>: String: Name of the function (sinus,
 * linear, bezier, spline)<br>
 * <code>&lt;support-point-count&gt;</code>: Integer: The number of support
 * points in this line<br>
 * <code>&lt;support-point-count&gt;</code>: Integer: The number of support
 * points in this line<br>
 * <code>&lt;support-point&gt;</code>: format v1: two separate float values,
 * format v2: 2 or 6 in brackets encapsulated, space separated float values<br>
 * 
 * 
 * @author Stefan Glaser
 */
public class FunctionLoader
{
	public static final FunctionLoader INSTANCE = new FunctionLoader();

	private static final String[] KNOWN_FUNCTIONS = new String[] {
			SinFunction.SIN_NAME, PiecewiseLinearFunction.LINEAR_NAME,
			PiecewiseBezierFunction.BEZIER_NAME, SplineFunction.SPLINE_NAME,
			PiecewiseSinusSquare.SIN_SQUARE_NAME };

	/** Constant for format version 1 */
	public static final int VERSION_1 = 1;

	/** Constant for format version 2 */
	public static final int VERSION_2 = 2;

	public String[] getKnownFunctionTypes()
	{
		return KNOWN_FUNCTIONS;
	}

	public String toCSVString(IFunction function, int version)
	{
		DecimalFormatSymbols symbols = new DecimalFormatSymbols();
		symbols.setDecimalSeparator('.');
		DecimalFormat df = new DecimalFormat("##0.0#", symbols);
		StringBuilder result = new StringBuilder();

		// Append Name
		result.append(function.getName());

		if (version != VERSION_2 && function instanceof SinFunction) {
			float[] params = ((SinFunction) function).getSinParameters();
			result.append(",").append(df.format(params[0]));
			result.append(",").append(df.format(params[1]));
			result.append(",").append(df.format(params[2]));
			result.append(",").append(df.format(params[3]));

			return result.toString();
		}

		// Append number of support points
		result.append(",").append(function.getSupportPointSize());

		// Append a zero float value (previously the phase)
		result.append(",").append("0.0");

		// Append SupportPoints
		for (int i = 0; i < function.getSupportPointSize(); i++) {
			ISupportPoint point = function.getSupportPoint(i);

			if (version == VERSION_2) {
				boolean addTangentPoints = point.getTangentPointAfter() != null
						&& point.getTangentPointBefore() != null;

				result.append(",(");
				if (addTangentPoints) {
					result.append(df.format(point.getTangentPointBefore().getX()));
					result.append(" ");
					result.append(df.format(point.getTangentPointBefore().getY()));
					result.append(" ");
				}

				result.append(df.format(point.getX()));
				result.append(" ").append(df.format(point.getY()));

				if (addTangentPoints) {
					result.append(" ");
					result.append(df.format(point.getTangentPointAfter().getX()));
					result.append(" ");
					result.append(df.format(point.getTangentPointAfter().getY()));
				}
				result.append(")");
			} else {
				result.append(",").append(df.format(point.getX()));
				result.append(",").append(df.format(point.getY()));
			}
		}

		return result.toString();
	}

	public IFunction fromCSVString(String csvLine, int version)
	{
		String[] csvContents = csvLine.split(",");

		String functionName = csvContents[0];
		float stiffness = 1;

		// Workaround to parse old SinFunction format
		if (version != VERSION_2 && functionName.equals(SinFunction.SIN_NAME)) {
			return new SinFunction(Float.parseFloat(csvContents[1]),
					Float.parseFloat(csvContents[2]),
					Float.parseFloat(csvContents[3]),
					Float.parseFloat(csvContents[4]), 1);
		} else {
			stiffness = Float.parseFloat(csvContents[2]);
		}

		// int supportPointSize = (int) Float.parseFloat(csvContents[1]);
		// float phase = Float.parseFloat(csvContents[2]); // Not used any more

		ArrayList<SupportPoint> supportPoints = parseSupportPoints(csvContents,
				version);

		return createFunction(functionName, supportPoints, stiffness);
	}

	protected ArrayList<SupportPoint> parseSupportPoints(String[] csvLine,
			int version)
	{
		ArrayList<SupportPoint> supportPoints = new ArrayList<SupportPoint>();
		int startIndex = 3;
		int minSize = version == VERSION_2 ? 2 : 4;

		if (csvLine.length < startIndex + minSize) {
			throw new IllegalArgumentException("Expected at least "
					+ (startIndex + minSize) + " values in "
					+ Arrays.toString(csvLine));
		}

		if (version == VERSION_2) {
			for (int i = startIndex; i < csvLine.length; i++) {
				if (!csvLine[i].startsWith("(") || !csvLine[i].endsWith(")")) {
					throw new MalformatedSupportPointException(
							"Expected brackets! v" + version + " value: " + csvLine[i]);
				}

				String[] supportPointValues = csvLine[i].substring(1,
						csvLine[i].length() - 1).split(" ");

				if (supportPointValues.length == 2) {
					float x = Float.parseFloat(supportPointValues[0]);
					float y = Float.parseFloat(supportPointValues[1]);
					supportPoints.add(new SupportPoint(x, y));
				} else if (supportPointValues.length == 6) {
					float xBefore = Float.parseFloat(supportPointValues[0]);
					float yBefore = Float.parseFloat(supportPointValues[1]);
					float x = Float.parseFloat(supportPointValues[2]);
					float y = Float.parseFloat(supportPointValues[3]);
					float xAfter = Float.parseFloat(supportPointValues[4]);
					float yAfter = Float.parseFloat(supportPointValues[5]);
					supportPoints.add(new SupportPoint(xBefore, yBefore, x, y,
							xAfter, yAfter));
				} else {
					throw new MalformatedSupportPointException(
							"Expected 2 or 6 values but got: "
									+ supportPointValues.length + " in v" + version
									+ " value: " + csvLine[i]);
				}
			}
		} else {
			if ((csvLine.length - startIndex) % 2 != 0) {
				throw new MalformatedSupportPointException(
						"Expected an even number of x,y pairs but got: "
								+ (csvLine.length - startIndex) + " in v" + version
								+ " value: " + csvLine);
			}

			for (int i = startIndex; i < csvLine.length; i += 2) {
				float x = Float.parseFloat(csvLine[i]);
				float y = Float.parseFloat(csvLine[i + 1]);
				supportPoints.add(new SupportPoint(x, y));
			}
		}

		return supportPoints;
	}

	/**
	 * @param csvLine - the line in the behavior-file defining a function
	 * @param version - the file format version
	 * @return the function defined by the csv-line
	 */
	public IFunction createFunction(String name,
			ArrayList<SupportPoint> supportPoints, float stiffness)
	{
		if (name.equals(SinFunction.SIN_NAME)) {
			return new SinFunction(supportPoints, stiffness);
		} else if (name.equals(PiecewiseSinusSquare.SIN_SQUARE_NAME)) {
			return new PiecewiseSinusSquare(supportPoints, stiffness);
		} else if (name.equals(SplineFunction.SPLINE_NAME)) {
			return new SplineFunction(supportPoints, stiffness);
		} else if (name.equals(PiecewiseBezierFunction.BEZIER_NAME)) {
			return new PiecewiseBezierFunction(supportPoints, stiffness);
		} else if (name.equals(PiecewiseLinearFunction.LINEAR_NAME)) {
			return new PiecewiseLinearFunction(supportPoints, stiffness);
		} else {
			throw new IllegalArgumentException("Unknown fuction " + name + "!");
		}
	}
}
