/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import java.util.ArrayList;

import magma.util.geometry.Geometry;

import org.apache.commons.math3.geometry.euclidean.twod.Vector2D;

/**
 * Represents a piecewise linear function with any number of pieces
 * 
 * @author Klaus Dorer
 */
public class PiecewiseLinearFunction extends SupportPointFunction
{
	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** name identifying a this kind of function */
	public static final String LINEAR_NAME = "linear";

	/**
	 * Default Constructor. Instantiates a new PiecewiseLinearFunction object
	 * with no initial SupportPoints.
	 */
	public PiecewiseLinearFunction()
	{
		super(1);
	}

	/**
	 * Instantiates a new PiecewiseLinearFunction object
	 */
	public PiecewiseLinearFunction(ArrayList<SupportPoint> points,
			float stiffness)
	{
		super(points, stiffness);
	}

	// /**
	// * Instantiates a new PiecewiseLinearFunction object
	// *
	// * @param xValues Array containing x values
	// * @param yValues Array containing y values
	// */
	// public PiecewiseLinearFunction(double[] xValues, double[] yValues)
	// {
	// super(xValues, yValues, 0.0);
	// }

	// /**
	// * Creates the function from a sin line in a motor function file
	// *
	// * @param csvLineContents the values that were comma separated in the file
	// */
	// public PiecewiseLinearFunction(String[] csvLineContents)
	// {
	// super(csvLineContents);
	// }

	/**
	 * Copy Constructor
	 * 
	 * @param source Source object to copy
	 */
	private PiecewiseLinearFunction(PiecewiseLinearFunction source)
	{
		super(source);
	}

	@Override
	protected void setDefaultSupportPoints()
	{
	}

	@Override
	protected int getMinimumSupportPointSize()
	{
		return 0;
	}

	@Override
	protected int getMaximumSupportPointSize()
	{
		return Integer.MAX_VALUE;
	}

	// /**
	// * Instantiates the object from a maximum x value and a phase
	// *
	// * @param maxX maximum x value
	// * @param phase sinus phase
	// */
	// public PiecewiseLinearFunction(int maxX, double phase)
	// {
	// super(maxX, phase);
	// }

	@Override
	public String getName()
	{
		return LINEAR_NAME;
	}

	/**
	 * Returns the sin value at the passed x coordinate. Parameters are used as
	 * written in constructor.
	 * @param t the coordinate at which to return the functions value
	 * @return the sin value at the passed x coordinate
	 */
	@Override
	public double value(double t)
	{
		// Special condition for logging use
		if (supportPoints.size() == 0) {
			return 0;
		}

		// Catch border conditions
		if (t <= getMinX()) {
			return supportPoints.get(0).y;
		} else if (t >= getMaxX()) {
			return supportPoints.get(supportPoints.size() - 1).y;
		}

		int index = 1;

		// Find index after the SupportPoint that is closest, but smaller or equal
		// to t
		while (index < supportPoints.size() && t >= supportPoints.get(index).x) {
			index++;
		}

		SupportPoint sPointBefore = supportPoints.get(index - 1);
		SupportPoint sPointAfter = supportPoints.get(index);

		// linear interpolation
		return Geometry.linearInterpolation(sPointBefore.x, sPointBefore.y,
				sPointAfter.x, sPointAfter.y, t);
	}

	/**
	 * @see magma.util.function.IFunction#copy()
	 */
	@Override
	public IFunction copy()
	{
		return new PiecewiseLinearFunction(this);
	}

	@Override
	public Vector2D moveTangentPointBeforeTo(int index, float x, float y)
	{
		return null;
	}

	@Override
	public Vector2D moveTangentPointAfterTo(int index, float x, float y)
	{
		return null;
	}
}