/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import java.util.ArrayList;

import org.apache.commons.math3.geometry.euclidean.twod.Vector2D;

/**
 * Represents a sin function with amplitude, period, phaseShift and offset by
 * using three SupportPoints:<br>
 * 1: (phaseShift, offset)<br>
 * 2: (phaseShift + phase/4, offset+amplitude)<br>
 * 3: (phaseShift + phase/2, offset)<br>
 * 
 * @author Stefan Glaser
 */
public class SinFunction extends SupportPointFunction
{
	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** name identifying a sine function */
	public static final String SIN_NAME = "sinus";

	/** The phaseShift of the sin-function */
	float phaseShift;

	/** The yOffset of the sin-function */
	float yOffset;

	/** The period of the sin-function */
	float period;

	/** The amplitude of the sin-function */
	float amplitude;

	/**
	 * Default Constructor. Creates a new SinFunction with a period of 20 and an
	 * amplitude of 1.
	 */
	public SinFunction()
	{
		super(1);
	}

	/**
	 * Construct a new SinFunction from 3 special SupportPoints.
	 * 
	 * @param points - a list of 3 special SupportPoints representing a sinus
	 *        function
	 */
	public SinFunction(ArrayList<SupportPoint> points, float stiffness)
	{
		super(points, stiffness);
	}

	/**
	 * Construct a new SinFunction with the given properties.
	 * 
	 * @param period - the period
	 * @param amplitude - the amplitude
	 * @param phaseShift - the phase shift
	 * @param yOffset - the offset in Y-direction
	 */
	public SinFunction(float period, float amplitude, float phaseShift,
			float offset, float stiffness)
	{
		super(stiffness);

		setSinusSupportPoints(period, amplitude, phaseShift, offset);
	}

	/**
	 * Copy constructor.
	 * 
	 * @param source The source object to copy
	 */
	public SinFunction(SinFunction source)
	{
		this(source.supportPoints, source.stiffness);
	}

	@Override
	protected void setDefaultSupportPoints()
	{
		setSinusSupportPoints(20, 1, 0, 0);
	}

	@Override
	protected int getMaximumSupportPointSize()
	{
		return 3;
	}

	@Override
	protected int getMinimumSupportPointSize()
	{
		return 3;
	}

	/**
	 * Set the properties of this sinus function to the given values and create
	 * the corresponding SupportPoints.
	 * 
	 * @param period
	 * @param amplitude
	 * @param phaseShift
	 * @param yOffset
	 */
	private void setSinusSupportPoints(float period, float amplitude,
			float phaseShift, float yOffset)
	{
		if (supportPoints.size() != 3) {
			if (supportPoints.size() > 3) {
				for (int i = supportPoints.size() - 1; i > 2; i--) {
					supportPoints.remove(i);
				}
			} else {
				for (int i = supportPoints.size(); i < 3; i++) {
					supportPoints.add(new SupportPoint(0, 0));
				}
			}
		}

		SupportPoint s0 = supportPoints.get(0);
		SupportPoint s1 = supportPoints.get(1);
		SupportPoint s2 = supportPoints.get(2);

		s0.x = phaseShift;
		s0.y = yOffset;
		s1.x = phaseShift + period / 4;
		s1.y = yOffset + amplitude;
		s2.x = phaseShift + period / 2;
		s2.y = yOffset;

		initialize();
	}

	/**
	 * Get Function name
	 * 
	 * @return String
	 */
	@Override
	public String getName()
	{
		return SIN_NAME;
	}

	@Override
	public double value(double t)
	{
		return amplitude * Math.sin((t - phaseShift) / period * 2 * Math.PI)
				+ yOffset;
	}

	@Override
	public Vector2D moveSupportPointTo(int index, float x, float y)
	{
		if (index < 0 || index > 2) {
			return null;
		}

		Vector2D previousPos = null;

		// Truncate x value of SupportPoint
		x = truncateX(x);

		if (index == 0) {
			// shift whole function
			previousPos = supportPoints.get(0).asVector2D();
			setSinusSupportPoints(period, amplitude, x, y);
		} else if (index == 1) {
			// shift amplitude
			previousPos = supportPoints.get(1).asVector2D();
			setSinusSupportPoints(period, y - yOffset, phaseShift, yOffset);
		} else {
			// shift period
			previousPos = supportPoints.get(2).asVector2D();
			float newPeriod = (x - phaseShift) * 2;
			if (newPeriod < 0) {
				newPeriod *= -1;
			}
			if (newPeriod < 0.2) {
				newPeriod = 0.2f;
			}
			setSinusSupportPoints(newPeriod, amplitude, phaseShift, yOffset);
		}

		initialize();

		return previousPos;
	}

	@Override
	public void initialize()
	{
		SupportPoint sp0 = supportPoints.get(0);
		SupportPoint sp1 = supportPoints.get(1);
		SupportPoint sp2 = supportPoints.get(2);

		phaseShift = sp0.x;
		yOffset = sp0.y;
		period = (sp2.x - phaseShift) * 2;
		amplitude = sp1.y - yOffset;

		sp1.x = sp0.x + period / 4;
		sp2.y = sp0.y;

		if (period < 0) {
			minXValue = phaseShift + period / 2;
			maxXValue = phaseShift;
		} else {
			minXValue = phaseShift;
			maxXValue = phaseShift + period / 2;
		}
		minYValue = yOffset - Math.abs(amplitude);
		maxYValue = yOffset + Math.abs(amplitude);
	}

	@Override
	public IFunction copy()
	{
		return new SinFunction(this);
	}

	public float[] getSinParameters()
	{
		return new float[] { period, amplitude, phaseShift, yOffset };
	}

	@Override
	public Vector2D moveTangentPointBeforeTo(int index, float x, float y)
	{
		return null;
	}

	@Override
	public Vector2D moveTangentPointAfterTo(int index, float x, float y)
	{
		return null;
	}
}