/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import java.text.FieldPosition;

import magma.util.FuzzyCompare;

import org.apache.commons.math3.geometry.euclidean.threed.CardanEulerSingularityException;
import org.apache.commons.math3.geometry.euclidean.threed.NotARotationMatrixException;
import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.RotationOrder;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3DFormat;
import org.apache.commons.math3.linear.RealMatrix;

/**
 * Utilities for geometric calculations
 * 
 * @author Klaus Dorer
 */
public class Geometry
{
	/**
	 * The formatter to format {@link Vector3D} objects.
	 */
	private static final Vector3DFormat VECTOR_FORMAT;

	static {
		VECTOR_FORMAT = Vector3DFormat.getInstance();
		VECTOR_FORMAT.getFormat().setMaximumFractionDigits(5);
	}

	/**
	 * Formats the given {@link Vector3D} object to produce a String.
	 * 
	 * @param vector the vector to format
	 * @return the formated vector
	 */
	public static String format(Vector3D vector)
	{
		return VECTOR_FORMAT.format(vector);
	}

	/**
	 * Formats the given {@link Vector3D} object to produce a String, which is
	 * appended to the given StringBuilder at the given position.
	 * 
	 * @param vector the vector to format
	 * @param buffer the StringBuffer to which the formated vector should be
	 *        appended
	 * @param pos the position in the buffer at which the vector should be
	 *        appended
	 * @return the value passed as buffer
	 */
	public static StringBuffer format(Vector3D vector, StringBuffer buffer,
			FieldPosition pos)
	{
		return VECTOR_FORMAT.format(vector, buffer, pos);
	}

	/**
	 * Checks if the passed 2D point is inside the passed polygon
	 * @param position the point to check (x,y) (it has to be in the same
	 *        coordinate system as the points of the polygon
	 * @param polygon corner points 2D of the polygon in clockwise direction
	 * @return true if the passed point is inside the polygon
	 */
	public static boolean isInsidePolygon(Vector3D position, Vector3D[] polygon)
	{
		int index0 = polygon.length - 1;
		int index1 = 0;
		double x2 = position.getX();
		double y2 = position.getY();
		for (index1 = 0; index1 < polygon.length;) {
			double x0 = polygon[index0].getX();
			double y0 = polygon[index0].getY();
			double x1 = polygon[index1].getX();
			double y1 = polygon[index1].getY();
			// if points are clockwise then the determinant has to be positive
			// ----1 | x0 y0 1 |
			// A = - | x1 y1 1 |
			// ----2 | x2 y2 1 |
			double result = 0.5 * (x1 * y2 - y1 * x2 - x0 * y2 + y0 * x2 + x0 * y1 - y0
					* x1);
			if (result > 0) {
				return false;
			}

			index0 = index1;
			index1++;
		}

		return true;
	}

	/**
	 * Converts the passed points to a Vector3D array representing a polygon
	 * @param points the corner points (x,y)
	 * @return a Vector3D array representing a polygon
	 */
	public static Vector3D[] getPolygon(double[][] points)
	{
		Vector3D[] polygon = new Vector3D[points.length];
		for (int i = 0; i < points.length; i++) {
			double[] point = points[i];
			polygon[i] = new Vector3D(point[0], point[1], 0);
		}
		return polygon;
	}

	/**
	 * calculates with the given coordinates the desired angles for Hip and Knee
	 * the Hip-Coordinates are always (0|0). When Leg is straight Foot is
	 * (0|-0.25) and Knee is (0|-0.14)
	 * 
	 * @param xKnee - x Position of the Knee
	 * @param zKnee - z Position of the Knee
	 * @param xFoot - x Position of the Foot (Ankle)
	 * @param zFoot - y Position of the Foot (Ankle)
	 * @return two-Dimensional Array first Value is Hip-Angle Second is
	 *         Knee-Angle
	 */
	public static double[] calculateAngles(Vector3D kneePosition,
			Vector3D anklePosition)
	{
		double xKnee = kneePosition.getX();
		double zKnee = kneePosition.getZ();
		double xFoot = anklePosition.getX();
		double zFoot = anklePosition.getZ();

		double[] ret = { 0, 0 };

		double l1 = 0.14, l2 = 0.11;

		zFoot = -((l1 + l2) - zFoot);

		double hip = Math.toDegrees(Math.atan(zKnee / xKnee));

		if (xKnee < 0) {
			hip = hip - 180;
		}
		hip = 90 + hip;

		double l = -Math.sqrt((xFoot * xFoot) + (zFoot * zFoot));

		double calc = ((l2 * l2 + l1 * l1 - l * l) / (2 * l2 * l1));

		double knee = 0;

		if (!FuzzyCompare.eq(zKnee / xKnee, zFoot / xFoot, 0.01)) {
			if (calc > 1) {
				knee = Math.toDegrees(Math.acos(calc - 2));
			} else if (calc < -1) {
				knee = Math.toDegrees(Math.acos(calc + 2));
			} else {
				knee = Math.toDegrees(Math.acos(calc));
			}
			knee = -(180 - knee);
		}

		ret[0] = hip;
		ret[1] = knee;

		return ret;
	}

	/**
	 * Calculates intersection Point(s) of two circles. Center of the first
	 * circle is (x1|z1) and its radius is r1. The center of the second circle is
	 * given by the parameters x2 and z2. The radius of the second circle is r2.
	 * If there is no intersection between the two circles a straight line from
	 * center to center is calculated and the intersection of this line and the
	 * first circle is returned. <br>
	 * Note: Since we intersect the two circles in just 2 dimensions (the
	 * x-z-plane), the y value is always forced to 0 in the resulting 3D vector.
	 * 
	 * @param x1 X coordinate of the first circle
	 * @param z1 Z coordinate of the first circle
	 * @param r1 radius of the first circle
	 * @param x2 X coordinate of the second circle
	 * @param z2 Z coordinate of the second circle
	 * @param r2 radius of the second circle
	 * @return Array with two elements for the coordinate of the two
	 *         intersection-points - the y-value is set to 0 by default
	 */
	public static Vector3D[] intersectionPoint(double x1, double z1, double r1,
			double x2, double z2, double r2)
	{
		Vector3D[] ret = new Vector3D[2];
		// double[] ret = { 0, 0, 0, 0 };

		double xc1, xc2, yc1, yc2, a1, a2, b, c, d, e, f, g;
		double ret_y0, ret_y1, ret_x0, ret_x1;

		xc1 = x1;
		yc1 = z1;

		xc2 = x2;
		yc2 = z2;

		if (xc2 != 0) {
			a1 = r1 * r1 - xc1 * xc1 - yc1 * yc1;
			a2 = r2 * r2 - xc2 * xc2 - yc2 * yc2;
			b = (a1 - a2) / 2;
			c = b / (xc2 - xc1);
			d = (yc2 - yc1) / (xc2 - xc1);
			e = d * d + 1;
			f = 2 * (d * xc1 - yc1 - c * d);
			g = c * c - 2 * c * xc1 - a1;
			if (f * f - 4 * e * g < 0) {
				double dx = xc1 - xc2;
				double dy = yc1 - yc2;

				double dr = Math.sqrt(dx * dx + dy * dy);

				double D = xc1 * yc2 - xc2 * yc1;

				double det = r1 * r1 * dr * dr - D * D;

				ret_x1 = (D * dy - (dy < 0 ? -1 : 1) * dx * Math.sqrt(det))
						/ (dr * dr);

				ret_y1 = (-D * dx - Math.abs(dy) * Math.sqrt(det)) / (dr * dr);

				ret[0] = new Vector3D(ret_x1, 0, ret_y1);
				ret[1] = new Vector3D(ret_x1, 0, ret_y1);

			} else {

				ret_y0 = ((-f - Math.sqrt(f * f - 4 * e * g)) / (2 * e));
				ret_x0 = (c - d * ret_y0);

				ret_y1 = ((-f + Math.sqrt(f * f - 4 * e * g)) / (2 * e));
				ret_x1 = (c - d * ret_y1);

				ret[0] = new Vector3D(ret_x1 > 0 ? ret_x1 : ret_x0, 0,
						ret_x1 > 0 ? ret_y1 : ret_y0);
				ret[1] = new Vector3D(ret_x1 > 0 ? ret_x0 : ret_x1, 0,
						ret_x1 > 0 ? ret_y0 : ret_y1);
			}

		} else {
			c = -yc2;
			double a = 0.11;
			b = 0.14;

			double alpha = Math.toDegrees(Math.acos((b * b + c * c - a * a)
					/ (2 * b * c)));

			double h = b * Math.sin(Math.toRadians(alpha));

			ret[0] = new Vector3D(h, 0, -Math.sqrt(h * h + b * b - 2 * h * b
					* Math.cos(Math.toRadians(180 - 90 - alpha))));
			ret[1] = new Vector3D(-h, 0, -Math.sqrt(h * h + b * b - 2 * h * b
					* Math.cos(Math.toRadians(180 - 90 - alpha))));
		}

		return ret;
	}

	/**
	 * Returns a linear interpolation between the two passed points at the passed
	 * x coordinate. Depending on the ascending parameter the function grows
	 * (true) from 0 to 1 otherwise it decreases from 1 to 0. Outside the range
	 * specified by x0 and x1 the values are 0 (left, true or right, false) or 1
	 * respectively (left, false or right, true.
	 * @param x0 lower x value
	 * @param x1 higher x value
	 * @param ascending true if the lower x value represents 0 false if it
	 *        represents 1
	 * @param x the x value for which to get the fuzzy value
	 * @return a fuzzy value that is linearly interpolated between the two passed
	 *         values
	 */
	public static double getLinearFuzzyValue(double x0, double x1,
			boolean ascending, double x)
	{
		if (x0 > x1) {
			double temp = x0;
			x0 = x1;
			x1 = temp;
		}

		if (ascending) {
			if (x <= x0) {
				return 0.0;
			}
			if (x >= x1) {
				return 1.0;
			}
			return linearInterpolation(x0, 0, x1, 1, x);

		}

		if (x <= x0) {
			return 1.0;
		}

		if (x >= x1) {
			return 0.0;
		}

		return linearInterpolation(x0, 1, x1, 0, x);
	}

	/**
	 * Returns the linear interpolation (y value) of the passed points (x0,y0)
	 * and (x1,y1) at the passed coordinate x
	 * @param x0 x coordinate of the first point
	 * @param y0 y coordinate of the first point
	 * @param x1 x coordinate of the second point
	 * @param y1 y coordinate of the second point
	 * @param x x coordinate of the point of interest
	 * @return the y coordinate of position x as linear interpolation of the two
	 *         passed points
	 */
	public static double linearInterpolation(double x0, double y0, double x1,
			double y1, double x)
	{
		return y0 + (y1 - y0) / (x1 - x0) * (x - x0);
	}

	/**
	 * Method to rotate a vector in 2D by an angle about the origin.
	 * <p>
	 * <b><i>Note:</i> Rotation displaces the polygon</b>
	 * @param vectorToRotate - the vector that has to be rotated
	 * @param angleInDegrees - the angle of rotation in degrees
	 * @param clockwise - true if the vector has to be rotated clockwise, false
	 *        if anti-clockwise
	 * 
	 * @return Rotated vector
	 */
	public static Vector3D rotateVectorIn2D(Vector3D vectorToRotate,
			double angleInDegrees, boolean clockwise)
	{
		double originalX = vectorToRotate.getX();
		double originalY = vectorToRotate.getY();
		double newX, newY;

		if (clockwise) {
			newX = originalX * Math.cos(angleInDegrees) - originalY
					* Math.sin(angleInDegrees);
			newY = originalX * Math.sin(angleInDegrees) + originalY
					* Math.cos(angleInDegrees);
		} else {
			newX = originalX * Math.cos(angleInDegrees) + originalY
					* Math.sin(angleInDegrees);
			newY = -originalX * Math.sin(angleInDegrees) + originalY
					* Math.cos(angleInDegrees);
		}
		return new Vector3D(newX, newY);
	}

	/**
	 * Method to calculate the vector at a <code>distance</code> from
	 * <code>vector</code> at an <code>angle</code> to the norm of the passed
	 * vector
	 * @param vector - the vector from which to find the displaced and rotated
	 *        vector
	 * @param distance - the distance from the <code>vector</code>
	 * @param angle - Angle object
	 * @return Vector3D (the vector from Apache)
	 */
	public static Vector3D getVectorAtDistanceAndAngle(Vector3D vector,
			double distance, Angle angle)
	{
		double newX, newY;
		newX = vector.getX() + (distance * Math.cos(angle.radians()));
		newY = vector.getY() + (distance * Math.sin(angle.radians()));

		return new Vector3D(newX, newY, 0.0);
	}

	/**
	 * Calculates the area of the triangle passed. Uses co-ordinate geometry.
	 * @param t - triangle
	 * @return double
	 */
	public static double areaOfTriangle(Vector3D[] t)
	{
		Vector3D p1 = t[0];
		Vector3D p2 = t[1];
		Vector3D p3 = t[2];

		// area of a triangle by co-ordinate geometry
		double area = (p1.getX() * (p2.getY() - p3.getY()))
				+ (p2.getX() * (p3.getY() - p1.getY()))
				+ (p3.getX() * (p1.getY() - p2.getY()));

		return Math.abs(area);
	}

	/**
	 * @param source source point
	 * @param destination destination point
	 * @param distanceFromSource distance from source point
	 * @return a point on the line between source and destination with distance
	 *         from source as specified
	 */
	public static Pose2D getPointInterpolation(Vector3D source,
			Vector3D destination, double distanceFromSource)
	{
		Vector3D srcDest = destination.subtract(source);
		if (srcDest.getNorm() < 0.001) {
			// to avoid normalization problems
			return new Pose2D(source, Angle.deg(0));
		}
		srcDest = srcDest.normalize().scalarMultiply(distanceFromSource);
		Angle directionSrcDest = Angle.rad(srcDest.getAlpha());
		srcDest = srcDest.add(source);
		return new Pose2D(srcDest, directionSrcDest);
	}

	/**
	 * Create a Rotation representing a rotation around the x-axis.
	 * 
	 * @param angle - the turn angle (rad)
	 * @return a rotation around the x-axis about angle
	 */
	public static Rotation createXRotation(double angle)
	{
		return new Rotation(Vector3D.PLUS_I, angle);
	}

	/**
	 * Create a Rotation representing a rotation around the y-axis.
	 * 
	 * @param angle - the turn angle (rad)
	 * @return a rotation around the y-axis about angle
	 */
	public static Rotation createYRotation(double angle)
	{
		return new Rotation(Vector3D.PLUS_J, angle);
	}

	/**
	 * Create a Rotation representing a rotation around the z-axis.
	 * 
	 * @param angle - the turn angle (rad)
	 * @return a rotation around the z-axis about angle
	 */
	public static Rotation createZRotation(double angle)
	{
		return new Rotation(Vector3D.PLUS_K, angle);
	}

	/**
	 * Extract the rotation about the z-axis from a given rotation.
	 * 
	 * @param rotation the rotation
	 * @return the z-axis rotation angle
	 */
	public static Angle getHorizontalAngle(Rotation rotation)
	{
		try {
			double[] angles = rotation.getAngles(RotationOrder.ZXY);
			return Angle.rad(angles[0]);
		} catch (CardanEulerSingularityException e) {
			e.printStackTrace();
		}

		return Angle.ZERO;
	}

	/**
	 * Transform a given matrix to a rotation.
	 * 
	 * @param matrix the matrix containing a rotation
	 * @return a Rotation representing the same rotation as the given matrix
	 */
	public static Rotation toRotation(RealMatrix matrix)
	{
		try {
			return new Rotation(matrix.getData(), 1.0e-8);
		} catch (NotARotationMatrixException e) {
			e.printStackTrace();
		}

		return Rotation.IDENTITY;
	}

	/**
	 * Transforms the given rotation relative to a system that is rotated around
	 * the x-axis by xAngle.
	 * 
	 * @param rotation - the rotation to transform
	 * @param xAngle - difference x-angle in rad
	 * 
	 * @return the given rotation relative to a system that is rotated by xAngle
	 *         around the x-axis
	 */
	public static Rotation xTransformRotation(Rotation rotation, double xAngle)
	{
		Rotation diffRot = createXRotation(xAngle);
		return diffRot.applyTo(rotation).applyTo(diffRot.revert());
	}

	/**
	 * Transforms the given rotation relative to a system that is rotated around
	 * the y-axis by yAngle.
	 * 
	 * @param rotation - the rotation to transform
	 * @param yAngle - difference y-angle in rad
	 * 
	 * @return the given rotation relative to a system that is rotated by yAngle
	 *         around the y-axis
	 */
	public static Rotation yTransformRotation(Rotation rotation, double yAngle)
	{
		Rotation diffRot = createYRotation(yAngle);
		return diffRot.applyTo(rotation).applyTo(diffRot.revert());
	}

	/**
	 * Transforms the given rotation relative to a system that is rotated around
	 * the z-axis by zAngle.
	 * 
	 * @param rotation - the rotation to transform
	 * @param zAngle - difference z-angle in rad
	 * 
	 * @return the given rotation relative to a system that is rotated by zAngle
	 *         around the z-axis
	 */
	public static Rotation zTransformRotation(Rotation rotation, double zAngle)
	{
		Rotation diffRot = createZRotation(zAngle);
		return diffRot.applyTo(rotation).applyTo(diffRot.revert());
	}

	/**
	 * Processes the top view z-normalized orientation.
	 * 
	 * @param orientation - global orientation
	 * @return the given orientation without z-rotation
	 */
	public static Rotation getTopViewOrientation(Rotation orientation)
	{
		// Process the top view z-normalized orientation
		double angle = getTopViewZAngle(orientation);
		Rotation ret = new Rotation(Vector3D.PLUS_K, angle).applyTo(orientation);

		// Geometry.printRotationMatrix(ret);

		return ret;
	}

	/**
	 * Processes the top view z-rotation.
	 * 
	 * @param orientation - global orientation
	 * @return the given top view z-rotation in rad
	 */
	public static double getTopViewZAngle(Rotation orientation)
	{
		// Process the top view z-normalized orientation
		double[][] tempMat = orientation.getMatrix();
		return Math.atan2(tempMat[0][1], tempMat[1][1]);
	}

	/**
	 * Print a rotation matrix.
	 * @param rotation - the rotation to print.
	 */
	public static void printRotationMatrix(Rotation rot)
	{
		printMatrix(rot.getMatrix());
	}

	/**
	 * Print a matrix.
	 * @param matrix the matrix to print.
	 */
	public static void printMatrix(double[][] matrix)
	{
		for (int row = 0; row < matrix.length; row++) {
			for (int col = 0; col < matrix[0].length; col++) {
				if (col > 0) {
					System.out.print(" ");
				}
				System.out.print(matrix[row][col]);
			}
			System.out.println();
		}
	}

	/**
	 * Calculates a point on the line from start to end that is distance away
	 * from the end point towards the start point.
	 * @param start the start point of the line
	 * @param end the end point of the line
	 * @param distance the distance from end (absolute)
	 * @return the point
	 */
	public static Vector3D getPointOnLineAbsoluteEnd(Vector3D start,
			Vector3D end, final double distance)
	{
		Vector3D endStart = start.subtract(end);
		if (endStart.getNorm() < 0.001) {
			// avoid vector normalization exception
			return end;
		}
		return end.add(endStart.normalize().scalarMultiply(distance));
	}

	/**
	 * Calculates a point that is orthogonal to the start - end line (in 2D) at
	 * the point that is distanceEnd away from end and distanceLeft away from
	 * start - end
	 * @param start the start point of the line
	 * @param end the end point
	 * @param distanceEnd the distance from end (absolute) towards start
	 * @param distanceLeft the distance of the point from the line
	 * @return the point as calculated above
	 */
	public static Vector3D getPointOnOrthogonal2D(Vector3D start, Vector3D end,
			double distanceEnd, double distanceLeft)
	{
		Vector3D start2D = new Vector3D(start.getX(), start.getY(), 0);
		Vector3D end2D = new Vector3D(end.getX(), end.getY(), 0);
		Vector3D onLine2D = getPointOnLineAbsoluteEnd(start2D, end2D, distanceEnd);
		Vector3D endStart2D = start2D.subtract(end2D);
		if (endStart2D.getNorm() < 0.001) {
			// avoid vector normalization exception
			return end2D;
		}
		Vector3D endStartNorm = endStart2D.normalize();
		Vector3D orthogonal2D = new Vector3D(endStartNorm.getY(),
				-endStartNorm.getX(), 0);
		Vector3D result = onLine2D.add(orthogonal2D.scalarMultiply(distanceLeft));
		return result;
	}

	/**
	 * Returns an array of future positions as predicted by the speed.
	 * 
	 * @param position The current position of the object.
	 * @param speedVector The speed vector of the object.
	 * @param howMany The number of future positions that should be returned
	 * @return an array of future positions
	 */
	public static Vector3D[] getFuturePositions(Vector3D position,
			Vector3D speedVector, int howMany)
	{
		Vector3D[] futurePositions = new Vector3D[howMany];

		futurePositions[0] = position.add(speedVector);
		for (int i = 1; i < howMany; i++) {
			futurePositions[i] = futurePositions[i - 1].add(speedVector);
		}
		return futurePositions;
	}

	/**
	 * Returns an array of future positions as predicted by the speed and
	 * speedDecay.
	 * 
	 * @param position The current position of the object.
	 * @param speedVector The speed vector of the object.
	 * @param speedDecay The decay of the speed vector (0 &lt; speedDecay &lt;=
	 *        1).
	 * @param howMany The number of future positions that should be returned.
	 * @return an array of future positions
	 */
	public static Vector3D[] getFuturePositions(Vector3D position,
			Vector3D speedVector, float speedDecay, int howMany)
	{
		Vector3D[] futurePositions = new Vector3D[howMany];
		Vector3D currentSpeed = speedVector;

		futurePositions[0] = position.add(currentSpeed);
		for (int i = 1; i < howMany; i++) {
			currentSpeed = currentSpeed.scalarMultiply(speedDecay);
			futurePositions[i] = futurePositions[i - 1].add(currentSpeed);
		}

		return futurePositions;
	}

	/**
	 * @param startPosition
	 * @param targetPositions
	 * @param ourSpeed
	 */
	public static Vector3D getInterceptionPoint(Vector3D startPosition,
			Vector3D[] targetPositions, double ourSpeed)
	{
		double cycleSpeed = ourSpeed * 3.0 / 50;
		double possibleDistance = 0;
		int left = 0;
		int right = targetPositions.length - 1;
		while (left < right) {
			int i = (left + right) / 2;
			double ballDistance = startPosition.subtract(targetPositions[i])
					.getNorm();
			possibleDistance = cycleSpeed * i;
			if (possibleDistance < ballDistance) {
				left = i + 1;
			} else {
				right = i;
			}
		}
		return targetPositions[right];
	}

}
