/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Stores a 2D position and orientation.
 */
public class Pose2D implements IPose2D
{
	/** */
	private static final long serialVersionUID = 3511039404384477686L;

	/** The x-position value */
	public double x;

	/** The y-position value */
	public double y;

	/** The horizontal angle. */
	public Angle angle;

	/**
	 * Default constructor initiating all values with zero.
	 */
	public Pose2D()
	{
		this(0, 0, Angle.ZERO);
	}

	/**
	 * @param x the x-position value
	 * @param y the y-position value
	 */
	public Pose2D(double x, double y)
	{
		this(x, y, Angle.ZERO);
	}

	/**
	 * @param x the x-position value
	 * @param y the y-position value
	 * @param angle the horizontal angle
	 */
	public Pose2D(double x, double y, Angle angle)
	{
		this.x = x;
		this.y = y;
		this.angle = angle;
	}

	/**
	 * @param position the position
	 * @param angle the horizontal angle
	 */
	public Pose2D(Vector3D position, Angle angle)
	{
		this(position.getX(), position.getY(), angle);
	}

	/**
	 * The copy constructor.
	 * 
	 * @param other the other pose
	 */
	public Pose2D(Pose2D other)
	{
		this.x = other.x;
		this.y = other.y;
		this.angle = other.angle;
	}

	public void set(Pose2D other)
	{
		this.x = other.x;
		this.y = other.y;
		this.angle = other.angle;
	}

	@Override
	public double getX()
	{
		return x;
	}

	@Override
	public double getY()
	{
		return y;
	}

	@Override
	public Vector3D getPosition()
	{
		return new Vector3D(x, y, 0);
	}

	@Override
	public Angle getAngle()
	{
		return angle;
	}

	/**
	 * @param position: the position to set
	 */
	public void setPosition(Vector3D position)
	{
		this.x = position.getX();
		this.y = position.getY();
	}

	/**
	 * Calculate the average of a list of Pose2D.
	 * 
	 * @param input: array of input values
	 * @return calculated average
	 */
	public static Pose2D average(Pose2D[] input)
	{
		// float angle = averageAngle(angles);
		double avgX = 0.0;
		double avgY = 0.0;

		// average position
		for (Pose2D value : input) {
			avgX += value.x;
			avgY += value.y;
		}
		if (input.length > 1) {
			avgX /= input.length;
			avgY /= input.length;
		}

		// average angles
		Angle[] angles = new Angle[input.length];
		for (int i = 0; i < input.length; i++) {
			angles[i] = input[i].angle;
		}

		return new Pose2D(avgX, avgY, Angle.average(angles));
	}

	@Override
	public Pose2D applyTo(IPose2D other)
	{
		Vector3D rotPos = angle.applyTo(other.getX(), other.getY());

		return new Pose2D(x + rotPos.getX(), y + rotPos.getY(), angle.add(other
				.getAngle()));
	}

	@Override
	public Pose2D applyInverseTo(IPose2D other)
	{
		Vector3D rotPos = angle
				.applyInverseTo(other.getX() - x, other.getY() - y);

		return new Pose2D(rotPos, other.getAngle().subtract(angle));
	}

	@Override
	public Vector3D applyTo(Vector3D position)
	{
		Vector3D rotPos = angle.applyTo(position);

		return new Vector3D(x + rotPos.getX(), y + rotPos.getY(), position.getZ());
	}

	@Override
	public Vector3D applyInverseTo(Vector3D position)
	{
		return angle.applyInverseTo(position.getX() - x, position.getY() - y);
	}

	@Override
	public double getDistanceTo(IPose2D other)
	{
		double deltax = x - other.getX();
		double deltay = y - other.getY();
		return Math.sqrt(deltax * deltax + deltay * deltay);
	}

	@Override
	public Angle getDeltaAngle(IPose2D other)
	{
		return other.getAngle().subtract(angle);
	}

	@Override
	public Angle getAngleTo(IPose2D other)
	{
		Pose2D relative = applyInverseTo(other);
		return Angle.rad(relative.getPosition().getAlpha());
	}

	@Override
	public Vector3D getUnitVector()
	{
		return new Vector3D(angle.radians(), 0);
	}

	@Override
	public String toString()
	{
		StringBuilder result = new StringBuilder();
		result.append("(").append(x).append(",");
		result.append(y).append("/");
		result.append(angle.degrees()).append(")");
		return result.toString();
	}

	@Override
	public boolean equals(Object other)
	{
		if (!(other instanceof Pose2D))
			return false;
		Pose2D p = (Pose2D) other;
		return (this.x == p.x && this.y == p.y && this.angle.equals(p.angle));
	}
}
