/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import org.apache.commons.math3.geometry.euclidean.threed.Rotation;
import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Stores a 3D position and orientation
 * 
 * @author Klaus Dorer
 */
public class Pose3D implements IPose3D
{
	/** */
	private static final long serialVersionUID = 1969426925498896576L;

	/** The 3D position. */
	public Vector3D position;

	/** The 3D orientation. */
	public Rotation orientation;

	/**
	 * Instantiate a new Pose3D object. The position is initialized with the
	 * {@link Vector3D#ZERO} vector and the {@link Rotation#IDENTITY} rotation.
	 */
	public Pose3D()
	{
		this(Vector3D.ZERO, Rotation.IDENTITY);
	}

	/**
	 * Instantiate a new Pose3D object and initialize position and orientation.
	 * 
	 * @param position: the position
	 * @param orientation: the orientation
	 */
	public Pose3D(Vector3D position, Rotation orientation)
	{
		this.position = position;
		this.orientation = orientation;
	}

	@Override
	public double getX()
	{
		return position.getX();
	}

	@Override
	public double getY()
	{
		return position.getY();
	}

	@Override
	public double getZ()
	{
		return position.getZ();
	}

	@Override
	public Vector3D getPosition()
	{
		return position;
	}

	@Override
	public Rotation getOrientation()
	{
		return orientation;
	}

	@Override
	public Pose3D applyTo(IPose3D other)
	{
		return new Pose3D(//
				position.add(orientation.applyTo(other.getPosition())), //
				orientation.applyTo(other.getOrientation()));
	}

	@Override
	public Pose3D applyInverseTo(IPose3D other)
	{
		return new Pose3D(//
				orientation.applyInverseTo(other.getPosition().subtract(position)), //
				orientation.applyInverseTo(other.getOrientation()));
	}

	@Override
	public Vector3D applyTo(Vector3D pos)
	{
		return position.add(orientation.applyTo(pos));
	}

	@Override
	public Vector3D applyInverseTo(Vector3D pos)
	{
		return orientation.applyInverseTo(pos.subtract(position));
	}

	@Override
	public boolean equals(Object obj)
	{
		if (obj instanceof IPose3D) {
			IPose3D other = (IPose3D) obj;
			return position.equals(other.getPosition())
					&& Rotation.distance(orientation, other.getOrientation()) < 0.0001;
		}

		return super.equals(obj);
	}

	@Override
	public String toString()
	{
		double[][] mat = orientation.getMatrix();

		StringBuilder result = new StringBuilder();
		result.append("(").append(position.getX()).append(",");
		result.append(position.getY()).append(",");
		result.append(position.getZ()).append(")\n");
		result.append(String.format(//
				"%2.3f %2.3f %2.3f\n" //
						+ "%2.3f %2.3f %2.3f\n" //
						+ "%2.3f %2.3f %2.3f", //
				mat[0][0], mat[0][1], mat[0][2], //
				mat[1][0], mat[1][1], mat[1][2], //
				mat[2][0], mat[2][1], mat[2][2]));
		return result.toString();
	}
}
