/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry.positionFilter;

import java.util.Iterator;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Interpolate a more reliable position value for an object from a list of past,
 * less reliable values, but accept high position changes unfiltered
 * 
 * @author Klaus Dorer
 */
public class LowFrequencyPositionFilter extends PositionFilter
{
	/**
	 * Default constructor
	 */
	public LowFrequencyPositionFilter()
	{
		super();
	}

	/**
	 * Constructor
	 * 
	 * @param i Internal buffer size for past position values
	 */
	public LowFrequencyPositionFilter(int i)
	{
		super(i);
	}

	/**
	 * Execute a filter pass: takes the current position of an object, adds it to
	 * the internal buffer, and interpolates a (hopefully) more reliable
	 * estimate.
	 * 
	 * @param position Current position
	 * @return Filtered position
	 */
	@Override
	public Vector3D filterPosition(Vector3D position)
	{
		if (position == null)
			return null;

		Vector3D speed = getAverageSpeed();
		if (filterbuffer.size() > 1) {
			Vector3D oldPos = filterbuffer.peek();
			if (position.subtract(oldPos.add(speed)).getNorm() > 0.2) {
				// we accept very high changes
				// System.out.println("clear filterbuffer: " + oldPos.add(speed)
				// + " new: " + position);
				reset();
				filterbuffer.push(oldPos);
			}
		}
		return super.filterPosition(position);
	}

	/**
	 * @return
	 */
	private Vector3D getAverageSpeed()
	{
		Vector3D speed = new Vector3D(0, 0, 0);
		if (filterbuffer.size() < 2) {
			return speed;
		}

		Iterator<Vector3D> filterIterator = filterbuffer.iterator();
		Vector3D newPosition = filterIterator.next();
		while (filterIterator.hasNext()) {
			Vector3D oldPosition = filterIterator.next();
			speed = speed.add(newPosition.subtract(oldPosition));
			newPosition = oldPosition;
		}
		return speed.scalarMultiply(1f / (filterbuffer.size() - 1));
	}

}
