/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.parameter;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;

/**
 * Base class for sets of parameters used to modify Behaviors
 * @author Klaus Dorer
 */
public class ParameterList implements Serializable, IParameterList
{
	/**  */
	private static final long serialVersionUID = 7491159595772287927L;

	/** parameters used by the underlying behavior */
	private Parameter[] parameters;

	/**
	 * Default constructor
	 */
	public ParameterList(Parameter[] params)
	{
		parameters = params;
	}

	/**
	 * @return the a new array containing all parameter values
	 */
	@Override
	public float[] getParameterValues()
	{
		float[] result = new float[parameters.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = parameters[i].getValue();
		}
		return result;
	}

	/**
	 * @return the overall number of parameters
	 */
	public int getNoOfParameters()
	{
		return parameters.length;
	}

	/**
	 * @param index the 0 based index of the parameter.
	 * @return the value of the parameter specified by index
	 */
	@Override
	final public float getValue(int index)
	{
		return parameters[index].getValue();
	}

	/**
	 * @param index the 0 based index of the parameter
	 * @param value the new value of the parameter
	 */
	@Override
	final public void setValue(int index, float value)
	{
		parameters[index].setValue(value);
	}

	/**
	 * Writes the values of each parameter to a file
	 * @param filename the name to use as filename
	 */
	@Override
	public void writeToFile(String filename)
	{
		PrintWriter out = null;
		try {
			out = new PrintWriter(new FileWriter(filename));

			// write all data for each joint
			for (Parameter param : parameters) {
				String line = param.getValue() + "\n";
				out.println(line);
			}

		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (out != null) {
				out.close();
			}
		}
	}

	@Override
	public void readFromFile(String filepath)
	{
		try {

			BufferedReader in = new BufferedReader(new FileReader(filepath));

			String line;
			int i = 0;
			while ((line = in.readLine()) != null) {
				float value = Float.valueOf(line);
				setValue(i, value);
				i++;
			}
			in.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	@Override
	public String toString()
	{
		StringBuilder builder = new StringBuilder(300);
		for (Parameter param : parameters) {
			builder.append("\n").append(param.toString());
		}
		return builder.toString();
	}

	@Override
	public String getParamsString()
	{
		StringBuilder builder = new StringBuilder(100);
		for (Parameter param : parameters) {
			builder.append("\nsetValue(Param.").append(param.getName());
			builder.append(", ").append(param.getValue()).append("f);");
		}
		return builder.toString();
	}

	@Override
	public Parameter[] getParameters()
	{
		return parameters;
	}
}
