/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.statemachine;

import java.util.ArrayList;
import java.util.List;

/**
 * Base class for all states implementing a state machine
 * 
 * @author Klaus Dorer
 * @param <K> the enum used to identify states
 * @param <E> the enum used to identify events
 */
public abstract class ObjectState<K, E>
{
	/**
	 * reference to the state factory that contains this state and the
	 * transitions
	 */
	private final IStateFactory<K, E> stateFactory;

	/** map containing all allowed state transitions from this state */
	private final List<StateTransition<K, E>> transitions;

	/**
	 * Default constructor
	 * @param factory the state factory that contains this state
	 */
	protected ObjectState(IStateFactory<K, E> factory)
	{
		stateFactory = factory;
		transitions = new ArrayList<StateTransition<K, E>>(2);
	}

	/**
	 * Adds a transition to the list of transitions of this state
	 * @param transition the new transition to add
	 */
	public void addTransition(StateTransition<K, E> transition)
	{
		transitions.add(transition);
	}

	/**
	 * Returns the id of this state
	 * @return the id of this state
	 */
	protected abstract K getStateID();

	/**
	 * A new event occurred and is now processed
	 * @param eventID the event that happened
	 * @param context the object on which the event happened
	 * @return the possibly new state we are in. Make sure to assign it to the
	 *         context.
	 */
	public ObjectState<K, E> tiggerEvent(E eventID, Object context)
	{
		for (StateTransition<K, E> transition : transitions) {
			if (transition.canTrigger(eventID, context)) {
				// we have a state change
				transition.transitionAction(context);

				return switchTo(transition.getDestination(), context);
			}
		}
		// event did not cause state change
		return this;
	}

	/**
	 * Switches from this state to the specified state. If the specified state is
	 * identical to this it triggers a self transition. It is assumed that it was
	 * checked before if the transition is valid.
	 * @param newStateID the state to which to switch.
	 * @param context the context object we are state from
	 * @return the possibly new state we are in. Make sure to assign it to the
	 *         context.
	 */
	protected ObjectState<K, E> switchTo(K newStateID, Object context)
	{
		if (newStateID == this.getStateID()) {
			// staying in the same state
			onStayInState(context);
			return this;
		}

		// change to different state
		ObjectState<K, E> newState = stateFactory.getState(newStateID);
		this.onLeavingState(newState, context);
		newState.onEnteringState(this, context);
		return newState;
	}

	/**
	 * Notification sent before this state is entered. The default implementation
	 * does nothing.
	 * 
	 * @param previousState State the machine was in before the transition
	 * @param context Context information object
	 */
	void onEnteringState(ObjectState<K, E> previousState, Object context)
	{
	}

	/**
	 * Notification sent when a state change is called that returns to this state
	 * again. The default implementation generally allows transitions from one
	 * state to itself and calls onLeavingState() and onEnteringState().
	 * 
	 * @param context Context information object
	 */
	void onStayInState(Object context)
	{
		onLeavingState(this, context);
		onEnteringState(this, context);
	}

	/**
	 * Notification sent before this state is entered. The default implementation
	 * does nothing.
	 * 
	 * @param nextState State the machine will be in after the transition
	 * @param context Context information object
	 */
	void onLeavingState(ObjectState<K, E> nextState, Object context)
	{
	}
}