/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IMotor;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.impl.AgentMetaModel;
import magma.agent.meta.agent.impl.BodyPartConfiguration;
import magma.agent.meta.agent.impl.HingeJointConfiguration;
import magma.agent.meta.agent.impl.SensorConfiguration;
import magma.agent.perception.IAccelerometerPerceptor;
import magma.agent.perception.IGyroPerceptor;
import magma.agent.perception.IHingeJointPerceptor;
import magma.agent.perception.IPerception;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;
import org.junit.Before;
import org.junit.Test;

import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.verify;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotSame;
import static org.junit.Assert.assertSame;

/**
 * 
 * @author dorer
 */
public class BodyModelTest
{
	private BodyModel testee;

	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception
	{
		AgentMetaModel metaModel2 = new AgentMetaModel("name", "sceneString",
				null, null);
		metaModel2.setBodyPartConfiguration(getTestBodyPartConfig());
		testee = new BodyModel(metaModel2, null);
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#BodyModel(magma.agent.meta.agent.IAgentMetaModel)}
	 * .
	 */
	@Test
	public void testBodyModelIAgentMetaModel()
	{

		assertEquals("torso", testee.getTorso().getName());
		assertEquals(1, testee.getTorso().getNoOfChildren());
		assertSame(testee.getTorso().getJoint(), testee.getJoint("torso"));
		assertEquals("TorsoGyro", testee.getSensor("TorsoGyro").getName());
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#BodyModel(magma.agent.agentmodel.impl.BodyModel)}
	 * .
	 */
	@Test
	public void testBodyModelBodyModel()
	{
		IBodyModel copy = new BodyModel(testee);
		assertNotSame(copy.getTorso(), testee.getTorso());
		assertEquals(1, copy.getTorso().getNoOfChildren());
		assertEquals(testee.getSensor("TorsoGyro"), copy.getSensor("TorsoGyro"));
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#getJoint(java.lang.String)} .
	 */
	@Test
	public void testGetJoint()
	{
		assertEquals("joint1", testee.getJoint("joint1").getName());
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#generateJointActions(magma.agent.action.IAction)}
	 * .
	 */
	@Test
	public void testGenerateJointActions()
	{
		Map<String, float[]> actions = new HashMap<String, float[]>();
		testee.generateJointActions(actions);
		assertEquals(1, actions.size());
		float[] values = actions.get("he1");
		assertEquals(1, actions.size());
		assertEquals(0, values[0], 0.0001);
		assertEquals(0, values[1], 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#updateFromPerception(magma.agent.perception.IPerception)}
	 * .
	 */
	@Test
	public void testUpdateFromPerception()
	{
		IPerception perception = createMock(IPerception.class);
		IHingeJointPerceptor hingePerceptor = createMock(IHingeJointPerceptor.class);
		expect(hingePerceptor.getAxis()).andStubReturn(20.0f);
		expect(hingePerceptor.getSpeed()).andReturn(1.0f);
		expect(hingePerceptor.getLoad()).andReturn(2.0f);
		expect(hingePerceptor.getVoltage()).andReturn(3.0f);
		expect(hingePerceptor.getTemperature()).andReturn(4.0f);
		expect(hingePerceptor.getCalculatedTemperature()).andReturn(5.0f);
		expect(hingePerceptor.getCalculatedTemperatureCoil()).andReturn(6.0f);
		expect(hingePerceptor.getError()).andReturn((byte) 7);

		IGyroPerceptor gyroPerceptor = createMock(IGyroPerceptor.class);
		expect(gyroPerceptor.getGyro()).andReturn(new Vector3D(1, 2, 3));

		IAccelerometerPerceptor accPerceptor = createMock(IAccelerometerPerceptor.class);
		expect(accPerceptor.getAcceleration()).andReturn(new Vector3D(1, 2, 3));

		expect(perception.getHingeJointPerceptor("hj1"))
				.andReturn(hingePerceptor);
		expect(perception.getGyroRatePerceptor("torsoGyro")).andReturn(
				gyroPerceptor);
		expect(perception.getAccelerationPerceptor("torsoAccel")).andReturn(
				accPerceptor);
		replay(perception, hingePerceptor, gyroPerceptor, accPerceptor);

		testee.updateFromPerception(perception);

		HingeJoint hingeJoint = (HingeJoint) testee.getJoint("joint1");
		assertEquals(20.0f, hingeJoint.getAxis(), 0.0001);
		assertEquals(0.0f, hingeJoint.getNextAxisSpeed(), 0.0001);
		IMotor motor = hingeJoint.getMotor();
		assertEquals(2.0f, motor.getLoad(), 0.0001);
		assertEquals(3.0f, motor.getVoltage(), 0.0001);
		assertEquals(4.0f, motor.getTemperature(), 0.0001);
		assertEquals(5.0f, motor.getCalculatedTemperature(), 0.0001);
		assertEquals(6.0f, motor.getCalculatedTemperatureCoil(), 0.0001);
		assertEquals(7, motor.getError());
		// TODO switch these three lines on once the miraculous factor 2 in
		// performAxisPosition is removed
		// hingeJoint.performAxisPosition(25);
		// assertEquals(25.0f, hingeJoint.getAxis(), 0.0001);
		// assertEquals(5.0f, hingeJoint.getNextAxisSpeed(), 0.0001);

		verify(perception);
	}

	private List<IBodyPartConfiguration> getTestBodyPartConfig()
	{
		List<IBodyPartConfiguration> configs = new ArrayList<IBodyPartConfiguration>();

		IBodyPartConfiguration currentConfig;

		currentConfig = new BodyPartConfiguration("torso", null, new Vector3D(0,
				0, 0), 1.2171f, new Vector3D(0.1, 0.1, 0.18), null, new Vector3D(0,
				0, 0), new SensorConfiguration("TorsoGyro", "torsoGyro"),
				new SensorConfiguration("TorsoAccel", "torsoAccel"), null, null);
		configs.add(currentConfig);

		currentConfig = new BodyPartConfiguration("neck", "torso", new Vector3D(
				0, 0, 0.09), 0.05f, new Vector3D(0.03, 0.03, 0.08),
				new HingeJointConfiguration("joint1", "hj1", "he1", new Vector3D(0,
						0, 1), -120, 120, 7), new Vector3D(0, 0, 0), null, null,
				null, null);
		configs.add(currentConfig);

		return configs;
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#BodyModel(magma.agent.agentmodel.impl.BodyModel)}
	 * .
	 */
	@Test
	public void testUpdateJointSpeed()
	{
		BodyModel copy = new BodyModel(testee);

		HingeJoint hj1 = (HingeJoint) copy.getJoint("joint1");
		hj1.performAxisSpeed(3.0f);
		testee.updateJointsSpeed(copy);

		HingeJoint hj1Source = (HingeJoint) testee.getJoint("joint1");
		assertEquals(3.0f, hj1Source.getNextAxisSpeed(), 0.00001);
	}

}
