/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package magma.agent.behavior.complex;

import magma.agent.behavior.complex.WalkEstimator.WalkMode;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Pose2D;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;
import org.junit.Before;
import org.junit.Test;

import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.junit.Assert.assertEquals;

/**
 * 
 * @author kdorer
 */
public class WalkEstimatorTest
{
	private WalkEstimator testee;

	private IThisPlayer thisPlayerMock;

	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception
	{
		thisPlayerMock = createMock(IThisPlayer.class);

		// we do not want to change unit tests if we are able to run faster
		float[] speeds = { 1.0f, 0.75f, 0.5f, 0.5f, 0.6f, 0.6f, 90 };
		testee = new WalkEstimator(speeds);
	}

	@Test
	public void testEstimateWalkTime()
	{
		// we stand at 0,0 looking to other goal (abs angle 0)
		Vector3D position = new Vector3D(2, 0, 0);
		expect(thisPlayerMock.getBodyDirectionTo(position)).andStubReturn(
				Angle.deg(90));
		expect(thisPlayerMock.getDistanceToXY(position)).andStubReturn(2.0);
		expect(thisPlayerMock.getDirectionTo(position)).andStubReturn(
				Angle.deg(90));
		replay(thisPlayerMock);

		Angle absFacingDirectionAtPosition = Angle.deg(45);
		Pose2D pose = new Pose2D(position, absFacingDirectionAtPosition);
		assertEquals(1 + 2 + 0.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.FORWARD),
				0.001);
		assertEquals(1 + 2 / 0.75 + 1.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.BACKWARD),
				0.001);
		assertEquals(0 + 4 + 0.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.LEFTSIDE),
				0.001);
		assertEquals(2 + 4 + 1.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.RIGHTSIDE),
				0.001);
	}

	@Test
	public void testEstimateWalkTimeLookDown()
	{
		// we stand at 0,0 looking to own goal (abs angle 180)
		Vector3D position = new Vector3D(2, 0, 0);
		expect(thisPlayerMock.getBodyDirectionTo(position)).andStubReturn(
				Angle.deg(-90));
		expect(thisPlayerMock.getDistanceToXY(position)).andStubReturn(2.0);
		expect(thisPlayerMock.getDirectionTo(position)).andStubReturn(
				Angle.deg(90));
		replay(thisPlayerMock);

		Angle absFacingDirectionAtPosition = Angle.deg(45);
		Pose2D pose = new Pose2D(position, absFacingDirectionAtPosition);

		assertEquals(1 + 2 + 0.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.FORWARD),
				0.001);
		assertEquals(1 + 2 / 0.75 + 1.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.BACKWARD),
				0.001);
		assertEquals(2 + 4 + 0.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.LEFTSIDE),
				0.001);
		assertEquals(0 + 4 + 1.5,
				testee.estimateWalkTime(thisPlayerMock, pose, WalkMode.RIGHTSIDE),
				0.001);
	}

	@Test
	public void testGetFastestWalkMode()
	{
		// we stand at 0,0 looking to own goal (abs angle 180)
		Vector3D position = new Vector3D(2, 0, 0);
		expect(thisPlayerMock.getBodyDirectionTo(position)).andStubReturn(
				Angle.deg(-90));
		expect(thisPlayerMock.getDistanceToXY(position)).andStubReturn(2.0);
		expect(thisPlayerMock.getDirectionTo(position)).andStubReturn(
				Angle.deg(90));
		expect(thisPlayerMock.getHorizontalAngle()).andStubReturn(Angle.deg(180));
		expect(thisPlayerMock.getPosition()).andStubReturn(new Vector3D(0, 0, 0));
		replay(thisPlayerMock);

		Angle absFacingDirectionAtPosition = Angle.deg(45);
		Pose2D pose = new Pose2D(position, absFacingDirectionAtPosition);

		assertEquals(WalkMode.FORWARD,
				testee.getFastestWalkMode(thisPlayerMock, pose));
	}
}
