/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.thoughtmodel.impl;

import java.util.ArrayList;
import java.util.List;

import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.agent.worldmodel.impl.WorldModelBaseTest;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;
import org.junit.Before;
import org.junit.Test;

import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.junit.Assert.assertEquals;

/**
 * Test for the {@link IFOCalculator} class
 * 
 * @author Klaus Dorer
 */
public class IFOCalculatorTest extends WorldModelBaseTest
{
	private IFOCalculator testee;

	private List<IPlayer> sourceList;

	/**
	 * Setup method
	 */
	@Override
	@Before
	public void setUp()
	{
		super.setUp();
		sourceList = new ArrayList<IPlayer>();
		sourceList.add(playerMock1);
		sourceList.add(playerMock2);
		sourceList.add(playerMock3);
		sourceList.add(playerMock4);

		replay(worldModelMock);
		testee = new IFOCalculator(worldModelMock);
	}

	/**
	 * Test for {@link IFOCalculator#getTeammatesAtBall(List)}
	 */
	@Test
	public void getTeammatesAtBall()
	{
		Vector3D ballPosition = new Vector3D(2, 0, 0);
		expect(ballMock.getPosition()).andStubReturn(ballPosition);
		expect(playerMock2.getDistanceToXY(ballPosition)).andStubReturn(1.0);
		expect(playerMock2.isLaying()).andStubReturn(false);
		expect(playerMock4.getDistanceToXY(ballPosition)).andStubReturn(2.0);
		expect(playerMock4.isLaying()).andStubReturn(false);

		replay(thisPlayerMock, playerMock1, playerMock2, playerMock3,
				playerMock4, ballMock);
		List<IPlayer> result = testee.getTeammatesAtBall(sourceList);
		assertEquals(2, result.size());
		assertEquals(playerMock2, result.get(0));
		assertEquals(playerMock4, result.get(1));
	}

	/**
	 * Test for {@link IFOCalculator#getTeammatesAtBall(List)}
	 */
	@Test
	public void getTeammatesAtBallLaying()
	{
		Vector3D ballPosition = new Vector3D(2, 0, 0);
		expect(ballMock.getPosition()).andStubReturn(ballPosition);
		expect(playerMock2.getDistanceToXY(ballPosition)).andStubReturn(1.0);
		expect(playerMock2.isLaying()).andStubReturn(true);
		expect(playerMock4.getDistanceToXY(ballPosition)).andStubReturn(2.0);
		expect(playerMock4.isLaying()).andStubReturn(false);

		replay(thisPlayerMock, playerMock1, playerMock2, playerMock3,
				playerMock4, ballMock);
		List<IPlayer> result = testee.getTeammatesAtBall(sourceList);
		assertEquals(2, result.size());
		assertEquals(playerMock4, result.get(0));
		assertEquals(playerMock2, result.get(1));
	}

	@Test
	public void testMergeObstacles()
	{
		List<IVisibleObject> obstacles = new ArrayList<IVisibleObject>();
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 0, 0), 1.0f));
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 0.5, 0), 1.0f));
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 4.0, 0), 1.0f));
		testee.mergeObstacles(obstacles);
		assertEquals(2, obstacles.size());
		assertEquals(1.5, obstacles.get(0).getCollisionDistance(), 0.0001f);
		assertEquals(0.25, obstacles.get(0).getPosition().getY(), 0.0001f);
	}

	@Test
	public void testMergeObstaclesNonConsecutive()
	{
		List<IVisibleObject> obstacles = new ArrayList<IVisibleObject>();
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 0, 0), 1.0f));
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 4.0, 0), 1.0f));
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 1.0, 0), 1.0f));
		testee.mergeObstacles(obstacles);
		assertEquals(2, obstacles.size());
		assertEquals(2, obstacles.get(0).getCollisionDistance(), 0.0001f);
		assertEquals(0.5, obstacles.get(0).getPosition().getY(), 0.0001f);
		assertEquals(4, obstacles.get(1).getPosition().getY(), 0.0001f);
	}

	@Test
	public void testMergeObstaclesMultiple()
	{
		List<IVisibleObject> obstacles = new ArrayList<IVisibleObject>();
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 0, 0), 1.0f));
		obstacles.add(new MergedObstacle("test", new Vector3D(0, -1.0, 0), 1.0f));
		obstacles.add(new MergedObstacle("test", new Vector3D(0, 1.0, 0), 1.0f));
		testee.mergeObstacles(obstacles);
		assertEquals(1, obstacles.size());
		assertEquals(3, obstacles.get(0).getCollisionDistance(), 0.0001f);
		assertEquals(0.25, obstacles.get(0).getPosition().getY(), 0.0001f);
	}
}
