/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import org.apache.commons.math3.geometry.euclidean.twod.Vector2D;
import org.junit.Before;
import org.junit.Test;

/**
 * @author Klaus Dorer
 * 
 */
public class PiecewiseLinearFunctionTest
{
	private PiecewiseLinearFunction testee;

	/**
	 * P0: &nbsp; 0, &nbsp;-2<br>
	 * P1: &nbsp; 4, &nbsp; 2<br>
	 * P2: &nbsp; 5, &nbsp; 2<br>
	 * P3: &nbsp; 6, &nbsp; 0
	 */
	private double[] supportPointArr;

	/**
	 * Setup method
	 */
	@Before
	public void setUp()
	{
		supportPointArr = new double[] { //
		0, -2, //
				4, 2, //
				5, 2, //
				6, 0 };
		testee = new PiecewiseLinearFunction(
				SupportPoint.fromArray(supportPointArr), 1);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#value(double)} .
	 */
	@Test
	public void testValueValid()
	{
		assertEquals(-2, testee.value(0), 0.0001);
		assertEquals(0, testee.value(6), 0.0001);
		assertEquals(1, testee.value(5.5), 0.0001);
		assertEquals(2, testee.value(4.5), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#getName()} .
	 */
	@Test
	public void testGetName()
	{
		assertEquals(PiecewiseLinearFunction.LINEAR_NAME, testee.getName());
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#addPoint(double, double)}
	 * .
	 */
	@Test
	public void testAddPointFirst()
	{
		testee.addSupportPoint(-1, 7);
		assertEquals(5, testee.getSupportPointSize(), 0.0001);
		assertEquals(-1, testee.getSupportPoint(0).x, 0.0001);
		assertEquals(7, testee.getSupportPoint(0).y, 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#addPoint(double, double)}
	 * .
	 */
	@Test
	public void testAddPointMiddle()
	{
		testee.addSupportPoint(4.5f, 7f);
		assertEquals(5, testee.getSupportPointSize(), 0.0001);
		assertEquals(4.5, testee.getSupportPoint(2).x, 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#addPoint(double, double)}
	 * .
	 */
	@Test
	public void testAddPointLast()
	{
		testee.addSupportPoint(8, 7);
		assertEquals(5, testee.getSupportPointSize(), 0.0001);
		assertEquals(8, testee.getSupportPoint(4).x, 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testHasSupportPointInArea()
	{
		assertEquals(false, testee.hasSupportPointInArea(1.0, 2.0));
		assertEquals(true, testee.hasSupportPointInArea(3.9, 4.1));
		assertEquals(true, testee.hasSupportPointInArea(3.0, 7));
		assertEquals(true, testee.hasSupportPointInArea(-1.0, 1.0));
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testRemovePointFirst()
	{
		testee.removeSupportPoint(0);
		assertEquals(3, testee.getSupportPointSize(), 0.0001);
		assertEquals(4, testee.getSupportPoint(0).x, 0.0001);
		assertEquals(2, testee.getSupportPoint(0).y, 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testRemovePointMiddle()
	{
		testee.removeSupportPoint(2);
		assertEquals(3, testee.getSupportPointSize(), 0.0001);
		assertEquals(6, testee.getSupportPoint(2).x, 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testRemovePointLast()
	{
		testee.removeSupportPoint(3);
		assertEquals(3, testee.getSupportPointSize(), 0.0001);
		assertEquals(5, testee.getSupportPoint(2).x, 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#changePointCoordinates(int, double, double)}
	 */
	@Test
	public void testChangePointCoordinates()
	{
		// Test Move P1 from (4, 2) to (3, -2.5)
		// Order: --> no change expected
		Vector2D old = testee.moveSupportPointTo(1, 3.0f, -2.5f);

		assertEquals(old.getX(), 4.0, 0.00001);
		assertEquals(old.getY(), 2.0, 0.00001);

		assertEquals(testee.getSupportPoint(0).x, supportPointArr[0], 0.000001);
		assertEquals(testee.getSupportPoint(0).y, supportPointArr[1], 0.000001);
		assertEquals(testee.getSupportPoint(1).x, 3.0, 0.000001);
		assertEquals(testee.getSupportPoint(1).y, -2.5, 0.000001);
		assertEquals(testee.getSupportPoint(2).x, supportPointArr[4], 0.000001);
		assertEquals(testee.getSupportPoint(2).y, supportPointArr[5], 0.000001);
		assertEquals(testee.getSupportPoint(3).x, supportPointArr[6], 0.000001);
		assertEquals(testee.getSupportPoint(3).y, supportPointArr[7], 0.000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#changePointCoordinates(int, double, double)}
	 */
	@Test
	public void testChangePointCoordinatesLessThanPrevious()
	{
		// Test Move P2 from (5, 2) to (3, -2.5)
		// Order: --> P1 and P2 should exchange
		Vector2D old = testee.moveSupportPointTo(2, 3.0f, -2.5f);

		assertEquals(old.getX(), 5.0, 0.00001);
		assertEquals(old.getY(), 2.0, 0.00001);

		assertEquals(testee.getSupportPoint(0).x, supportPointArr[0], 0.000001);
		assertEquals(testee.getSupportPoint(0).y, supportPointArr[1], 0.000001);
		assertEquals(testee.getSupportPoint(1).x, 3.0, 0.000001);
		assertEquals(testee.getSupportPoint(1).y, -2.5, 0.000001);
		assertEquals(testee.getSupportPoint(2).x, supportPointArr[2], 0.000001);
		assertEquals(testee.getSupportPoint(2).y, supportPointArr[3], 0.000001);
		assertEquals(testee.getSupportPoint(3).x, supportPointArr[6], 0.000001);
		assertEquals(testee.getSupportPoint(3).y, supportPointArr[7], 0.000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#changePointCoordinates(int, double, double)}
	 */
	@Test
	public void testChangePointCoordinatesMoreThanNext()
	{
		// Test Move P2 from (5, 2) to (7, 2.5)
		// Order: --> P3 and P2 should exchange
		Vector2D old = testee.moveSupportPointTo(2, 7.0f, 2.5f);

		assertEquals(old.getX(), 5.0, 0.00001);
		assertEquals(old.getY(), 2.0, 0.00001);

		assertEquals(testee.getSupportPoint(0).x, supportPointArr[0], 0.000001);
		assertEquals(testee.getSupportPoint(0).y, supportPointArr[1], 0.000001);
		assertEquals(testee.getSupportPoint(1).x, supportPointArr[2], 0.000001);
		assertEquals(testee.getSupportPoint(1).y, supportPointArr[3], 0.000001);
		assertEquals(testee.getSupportPoint(2).x, supportPointArr[6], 0.000001);
		assertEquals(testee.getSupportPoint(2).y, supportPointArr[7], 0.000001);
		assertEquals(testee.getSupportPoint(3).x, 7.0f, 0.000001);
		assertEquals(testee.getSupportPoint(3).y, 2.5f, 0.000001);
	}

	/**
	 * Test method for {@link magma.util.function.PiecewiseLinearFunction#copy()}
	 */
	@Test
	public void testCopy()
	{
		PiecewiseLinearFunction copy = (PiecewiseLinearFunction) testee.copy();

		assertEquals(testee.getMaxX(), copy.getMaxX(), 0.0000001);
		assertEquals(testee.getMinX(), copy.getMinX(), 0.0000001);
		assertEquals(testee.getMaxY(), copy.getMaxY(), 0.0000001);
		assertEquals(testee.getMinY(), copy.getMinY(), 0.0000001);
		assertEquals(testee.getName(), copy.getName());
		assertEquals(testee.getSupportPointSize(), copy.getSupportPointSize(),
				0.0000001);

		SupportPoint sp1, sp2;
		for (int i = 0; i < testee.getSupportPointSize(); i++) {
			sp1 = testee.getSupportPoint(i);
			sp2 = copy.getSupportPoint(i);
			if (sp1 == sp2) {
				fail("Copy of function is not deep!");
			}

			assertEquals(sp1.x, sp2.x, 0.0000001);
			assertEquals(sp1.y, sp2.y, 0.0000001);
		}
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#toCSVString()}
	 */
	@Test
	public void testToCSVString()
	{
		assertEquals("linear,4,0,0.0,-2.0,4.0,2.0,5.0,2.0,6.0,0.0",
				testee.toCSVString());
	}
}
