/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;
import org.junit.Test;

import static org.junit.Assert.assertEquals;

/**
 * @author Klaus Dorer
 * 
 */
public class GeometryTest
{
	/**
	 * Test for {@link Geometry#isInsidePolygon(Vector3D, Vector3D[])}
	 */
	@Test
	public void testIsInsidePolygon()
	{
		double[][] points = { { 0, 1 }, { 1, 0 }, { 0, -1 }, { -1, 0 } };
		Vector3D[] polygon = Geometry.getPolygon(points);

		// point outside
		assertEquals(false,
				Geometry.isInsidePolygon(new Vector3D(1, 1, 0), polygon));
		assertEquals(false,
				Geometry.isInsidePolygon(new Vector3D(-1, 1, 0), polygon));
		// point inside
		assertEquals(true,
				Geometry.isInsidePolygon(new Vector3D(0, 0, 0), polygon));
		// point on corner
		assertEquals(true,
				Geometry.isInsidePolygon(new Vector3D(0, 1, 0), polygon));
		// point on edge
		assertEquals(true,
				Geometry.isInsidePolygon(new Vector3D(0.5, 0.5, 0), polygon));
	}

	/**
	 * Test for
	 * {@link Geometry#getLinearFuzzyValue(double, double, boolean, double)}
	 */
	@Test
	public void testGetLinearFuzzyValue()
	{
		assertEquals(0.0, Geometry.getLinearFuzzyValue(10, 20, true, 0), 0.0001);
		assertEquals(0.1, Geometry.getLinearFuzzyValue(10, 20, true, 11), 0.0001);
		assertEquals(1.0, Geometry.getLinearFuzzyValue(10, 20, true, 25), 0.0001);

		assertEquals(1.0, Geometry.getLinearFuzzyValue(10, 20, false, 0), 0.0001);
		assertEquals(0.9, Geometry.getLinearFuzzyValue(10, 20, false, 11), 0.0001);
		assertEquals(0.0, Geometry.getLinearFuzzyValue(10, 20, false, 25), 0.0001);

		assertEquals(0.0, Geometry.getLinearFuzzyValue(20, 10, false, 25), 0.0001);
	}

	/**
	 * Test for
	 * {@link Geometry#linearInterpolation(double, double, double, double, double)}
	 */
	@Test
	public void testLinearInterpolation()
	{
		assertEquals(0.0, Geometry.linearInterpolation(10, 0, 20, 1, 10), 0.0001);
		assertEquals(-1.0, Geometry.linearInterpolation(10, 0, 20, 1, 0), 0.0001);
		assertEquals(0.5, Geometry.linearInterpolation(10, 0, 20, 1, 15), 0.0001);
		assertEquals(1.0, Geometry.linearInterpolation(10, 0, 20, 1, 20), 0.0001);
		assertEquals(1.5, Geometry.linearInterpolation(10, 0, 20, 1, 25), 0.0001);

		assertEquals(0.5, Geometry.linearInterpolation(20, 0, 10, 1, 15), 0.0001);

		assertEquals(-0.5, Geometry.linearInterpolation(10, 1, 20, 0, 25), 0.0001);
	}

	/**
	 * Test for
	 * {@link Geometry#getPointInterpolation(Vector3D, Vector3D, double)}
	 */
	@Test
	public void testGetPointInterpolation()
	{
		Pose2D result = Geometry.getPointInterpolation(new Vector3D(0, 0, 0),
				new Vector3D(1, 0, 0), 0.5);
		assertEquals(0.5, result.getPosition().getX(), 0.0001);
		assertEquals(0, result.getPosition().getY(), 0.0001);
		assertEquals(0, result.getAngle().degrees(), 0.0001);

		result = Geometry.getPointInterpolation(new Vector3D(2, 2, 0),
				new Vector3D(1, 1, 0), 0.5 * Math.sqrt(2));
		assertEquals(1.5, result.getPosition().getX(), 0.0001);
		assertEquals(1.5, result.getPosition().getY(), 0.0001);
		assertEquals(-135, result.getAngle().degrees(), 0.0001);
	}

	@Test
	public void testGetPointOnLineAbsoluteEnd()
	{
		Vector3D start = new Vector3D(0, 0, 0);
		Vector3D end = new Vector3D(1, 0, 0);
		Vector3D pos = Geometry.getPointOnLineAbsoluteEnd(start, end, 0.7);
		assertEquals(0.3, pos.getX(), 0.0001);
		assertEquals(0.0, pos.getY(), 0.0001);
		assertEquals(0.0, pos.getZ(), 0.0001);

		start = new Vector3D(0, 2, 0);
		end = new Vector3D(0, 4, 0);
		pos = Geometry.getPointOnLineAbsoluteEnd(start, end, 0.5);
		assertEquals(0.0, pos.getX(), 0.0001);
		assertEquals(3.5, pos.getY(), 0.0001);
	}

	@Test
	public void testGetPointOnOrthogonal2D()
	{
		Vector3D start = new Vector3D(0, 0, 0);
		Vector3D end = new Vector3D(1, 0, 1);
		Vector3D pos = Geometry.getPointOnOrthogonal2D(start, end, 0.4, 0.4);
		assertEquals(0.6, pos.getX(), 0.0001);
		assertEquals(0.4, pos.getY(), 0.0001);
		assertEquals(0.0, pos.getZ(), 0.0001);
	}

	@Test
	public void testGetInterceptionPoint() throws Exception
	{
		double ourSpeed = 0.7;
		Vector3D startPosition = Vector3D.ZERO;
		Vector3D[] targetPositions = new Vector3D[4];
		targetPositions[0] = new Vector3D(0.01, 0, 0);
		targetPositions[1] = new Vector3D(0.05, 0, 0);
		targetPositions[2] = new Vector3D(0.08, 0, 0);
		targetPositions[3] = new Vector3D(0.10, 0, 0);
		Vector3D interceptionPoint = Geometry.getInterceptionPoint(startPosition,
				targetPositions, ourSpeed);
		assertEquals(targetPositions[2], interceptionPoint);
	}
}