/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 * 
 * This file is part of magmaOffenburg.
 * 
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg.  If not, see <http://www.gnu.org/licenses/>.
 ********************************************************************************/
package magma.util.geometry;

import static org.junit.Assert.assertEquals;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;
import org.junit.Test;

public class Pose2DTest
{
	/**
	 * Test for {@link Pose2D#applyTo(IPose2D)}, {@link Pose2D#applyTo(Vector3D)}
	 * , {@link Pose2D#applyInverseTo(IPose2D)} and
	 * {@link Pose2D#applyInverseTo(Vector3D)}
	 */
	@Test
	public void testApplyTo()
	{
		Pose2D testee = new Pose2D(3, 1, Angle.deg(90));

		// Test applyTo other pose
		Pose2D otherPose = new Pose2D(1, 0, Angle.deg(20));
		Pose2D resPose = testee.applyTo(otherPose);

		assertEquals(3, resPose.getX(), 0.0001);
		assertEquals(2, resPose.getY(), 0.0001);
		assertEquals(110, resPose.getAngle().degrees(), 0.0001);

		// Test applyInverseTo other pose
		Pose2D resPoseInverse = testee.applyInverseTo(resPose);

		assertEquals(otherPose.getX(), resPoseInverse.getX(), 0.0001);
		assertEquals(otherPose.getY(), resPoseInverse.getY(), 0.0001);
		assertEquals(otherPose.getAngle().degrees(), resPoseInverse.getAngle()
				.degrees(), 0.0001);

		// Test applyTo other position
		Vector3D otherPosition = new Vector3D(2, 1, 0);
		Vector3D resPos = testee.applyTo(otherPosition);

		assertEquals(2, resPos.getX(), 0.0001);
		assertEquals(3, resPos.getY(), 0.0001);

		// Test applyInverseTo other position
		Vector3D resPosInverse = testee.applyInverseTo(resPos);

		assertEquals(otherPosition.getX(), resPosInverse.getX(), 0.0001);
		assertEquals(otherPosition.getY(), resPosInverse.getY(), 0.0001);
	}

	/**
	 * Test for {@link Pose2D#applyInverseTo(IPose2D)} and
	 * {@link Pose2D#applyInverseTo(Vector3D)}
	 */
	@Test
	public void testApplyInverseTo()
	{
		Pose2D testee = new Pose2D(3, 1, Angle.deg(90));

		// Test applyTo other pose
		Pose2D otherPose = new Pose2D(1, 0, Angle.deg(20));
		Pose2D resPose = testee.applyInverseTo(otherPose);

		assertEquals(-1, resPose.getX(), 0.0001);
		assertEquals(2, resPose.getY(), 0.0001);
		assertEquals(-70, resPose.getAngle().degrees(), 0.0001);

		// Test applyTo other position
		Vector3D otherPosition = new Vector3D(2, 1, 0);
		Vector3D resPos = testee.applyInverseTo(otherPosition);

		assertEquals(0, resPos.getX(), 0.0001);
		assertEquals(1, resPos.getY(), 0.0001);
	}

	@Test
	public void testGetDistanceTo()
	{
		Pose2D pose1 = new Pose2D(-2, -2);
		Pose2D pose2 = new Pose2D(1, 2);
		assertEquals(5.0, pose1.getDistanceTo(pose2), 0.001);

		pose1 = new Pose2D(2, 2);
		pose2 = new Pose2D(-1, -2);
		assertEquals(5.0, pose1.getDistanceTo(pose2), 0.001);
	}

	@Test
	public void testGetDeltaAngle()
	{
		Pose2D pose1 = new Pose2D(-2, -2, Angle.deg(90));
		Pose2D pose2 = new Pose2D(-2, -2, Angle.deg(45));
		assertEquals(-45.0, pose1.getDeltaAngle(pose2).degrees(), 0.001);
	}

	@Test
	public void testGetUnitVector()
	{
		Pose2D pose = new Pose2D(17, 95, Angle.deg(45));
		Vector3D result = pose.getUnitVector();
		assertEquals(result.getX(), 0.707, 0.001);
		assertEquals(result.getY(), 0.707, 0.001);
	}

	@Test
	public void testGetAngleTo()
	{
		Pose2D pose1 = new Pose2D(-2, -2, Angle.deg(90));
		Pose2D pose2 = new Pose2D(2, 2);
		assertEquals(-45, pose1.getAngleTo(pose2).degrees(), 0.001);

		pose1 = new Pose2D(2, 2, Angle.deg(91));
		pose2 = new Pose2D(2, 0);
		assertEquals(179, pose1.getAngleTo(pose2).degrees(), 0.001);
	}

}
