/********************************************************************************
 * Copyright 2008 - 2014 Hochschule Offenburg
 * All Authors are named in file Authors.txt
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util;

import org.apache.commons.math3.geometry.euclidean.threed.Vector3D;

/**
 * Implements a set of fuzzy compare operators for float and double data types.
 * 
 * @author Ingo Schindler
 */
public class FuzzyCompare
{
	/**
	 * Fuzzy "equal" ( == ) comparison
	 * 
	 * @param reference Reference value
	 * @param compare Value to be compared to the reference
	 * @param range The range within the two values are equal
	 * @return true if the both values are within the same range
	 */
	public static boolean eq(double reference, double compare, double range)
	{
		if (reference > compare) {
			return reference - range <= compare;
		}

		return compare - range <= reference;
	}

	/**
	 * Fuzzy "Greater Than or Equal" ( >= ) comparison
	 * 
	 * @param reference Reference value
	 * @param compare Value to be compared to the reference
	 * @param range The range within the two values are considered equal
	 * @return True if greater or equal, false if not
	 */
	public static boolean gte(double reference, double compare, double range)
	{
		return eq(reference, compare, range) || reference > compare;

	}

	/**
	 * Fuzzy "Lower Than or Equal" ( <= ) comparison
	 * 
	 * @param reference Reference value
	 * @param compare Value to be compared to the reference
	 * @param range The range within the two values are considered equal
	 * @return True if less than or equal, false if not
	 */
	public static boolean lte(double reference, double compare, double range)
	{
		return gte(compare, reference, range);
	}

	/**
	 * Fuzzy "Greater Than" ( > ) comparison
	 * 
	 * @param reference Reference value
	 * @param compare the lower of two values
	 * @param range The range within the two values are considered equal
	 * @return true if the first parameter is greater than the second one and the
	 *         second one is not within the range of the first value
	 */
	public static boolean gt(double reference, double compare, double range)
	{
		return gte(reference, compare, range);
	}

	/**
	 * Fuzzy "Lower Than" ( < ) comparison
	 * 
	 * @param reference Reference value
	 * @param compare Value to be compared to the reference
	 * @param range The range within the two values are considered equal
	 * @return True if lower, false if not
	 */
	public static boolean lt(double reference, double compare, double range)
	{
		return gte(compare, reference, range);
	}

	/**
	 * Fuzzy "Equal" ( == ) comparison for vectors
	 * 
	 * @param reference Reference value
	 * @param compare Value to be compared to the reference
	 * @param range The range within the vector values are considered to be equal
	 * @return True if both vectors are equal
	 */
	public static boolean eq(Vector3D reference, Vector3D compare, double range)
	{
		if (reference == null && compare == null) {
			return true;
		}
		if (reference == null || compare == null) {
			return false;
		}

		if (!eq((float) reference.getX(), (float) compare.getX(), range)) {
			return false;
		}
		if (!eq((float) reference.getY(), (float) compare.getY(), range)) {
			return false;
		}
		return eq((float) reference.getZ(), (float) compare.getZ(), range);
	}
}
